# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.1 `zeros` Chapter.

``zeros`` - Zeros of Polynomials

This module is concerned with computing the zeros of a polynomial with
real or complex coefficients.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/c02/c02intro.html

See Also
--------
``naginterfaces.library.examples.zeros`` :
    This subpackage contains examples for the ``zeros`` module.
    See also the :ref:`library_zeros_ex` subsection.
"""

# NAG Copyright 2017-2020.

import collections as _collections

from ..base import utils as _b_utils
from ..base import zeros as _b_zeros
from ..library import _utils

PolyComplexFpmlReturnData = _collections.namedtuple(
    'PolyComplexFpmlReturnData',
    ['z', 'berr', 'cond', 'conv'],
)

def poly_complex_fpml(a, itmax=30, polish=1):
    u"""
    Finds all the zeros of complex polynomial, fast modified Laguerre's
    method.

    ``poly_complex_fpml`` finds all the roots of a complex polynomial
    equation, using a fourth-order convergent modification of Laguerre's
    method.

    For full information please refer to the NAG Library document for
    c02aa

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/c02/c02aaf.html

    Parameters
    ----------
    a : complex, array-like, shape (n+1)
        `a`[i] must contain the coefficient of z^(n-i), for i =
        0,1,...,n.

    itmax : int, optional
        The maximum number of iterations to be performed.

    polish : int, optional
        Specifies the polishing technique used to refine root
        approximations.

        `polish` = 0

            No polishing.

        `polish` = 1

            Single iteration of Newton's method.

        `polish` = 2

            Iterative refinement using the compensated Horner's method.

        See also [Selecting a Polishing Process].

    Returns
    -------
    z : complex, ndarray, shape (n)
        `z`[j-1] holds the approximation of the root z_j, for j =
        1,2,...,n.

    berr : float, ndarray, shape (n)
        `berr`[j-1] holds the relative backward error, eta(z_j), for j =
        1,2,...,n.

    cond : float, ndarray, shape (n)
        `cond`[j-1] holds the condition number, kappa(z_j), for j =
        1,2,...,n.

    conv : int, ndarray, shape (n)
        `conv`[j-1] indicates the convergence status of the root
        approximation, z_j, for j = 1,2,...,n.

        `conv`[j-1] >= 0

            Successfully converged after `conv`[j-1] iterations.

        `conv`[j-1] = -1

            Failed to converge after `itmax` iterations.

        `conv`[j-1] = -2

            Overflow was encountered.

        Note: if `polish` = 2, `conv` refers to convergence in the
        compensated polishing process.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, the complex variable `a`[0] = (0.0,0.0).

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 3)
            On entry, `itmax` = *<value>*.

            Constraint: `itmax` >= 1.

        (`errno` 4)
            On entry, `polish` = *<value>*.

            Constraint: `polish` = 0, 1 or 2.

        (`errno` 5)
            Convergence has failed for at least one root approximation.

        (`errno` 6)
            ``poly_complex_fpml`` encountered overflow during at least
            one root approximation.

    Notes
    -----
    ``poly_complex_fpml`` attempts to find all the roots of the nth
    degree complex polynomial equation

        P(z) = a_0 z^n+a_1 z^(n-1)+a_2 z^(n-2)+...+a_{n-1}z+a_n = 0.

    The roots are located using a modified form of Laguerre's method, as
    implemented by Cameron (2018).
    An implicit deflation strategy is employed, which allows for high
    accuracy even when solving high degree polynomials.
    Linear (n = 1) and quadratic (n = 2) problems are obtained via the
    'standard' closed formulae.

    First, initial estimates of the roots are made using a method
    proposed by Bini (1996), which selects complex numbers along circles
    of suitable radii.
    Updates to each root approximation z_j, for j = 1,2,...,n, are then
    made using the iterative formula from Petković `et al.` (1997):

        z^_j = z_j-n/(G_j+/-sqrt((n-1)(nH_j-G_j^2))),

    where

        G_j = P'(z_j)/P(z_j)- sum(1/(z_j-z_i)) and H_j =
        -(P'(z_j)/P(z_j))'- sum(1/(z_j-z_i)^2).

    The nearest z^_j to the current approximation is used, by selecting
    the sign that maximizes the denominator of the correction term.
    The subtraction of the sum terms when computing G_j and H_j
    determines the implicit deflation strategy of the modified Laguerre
    method.

    The relative backward error of a root approximation z_j is given by

        eta(z_j) = |P(z_j)|/alpha(z_j),

    where alpha is the perturbed polynomial

        alpha(z) = sum((3.8*k+1)|a_{n-k}||z|^k).

    A root approximation is deemed to have converged if eta(z_j) <= 2 *
    machine precision, at which point updates of that root cease.
    If the stopping criterion holds, then the computed root is the exact
    root of a polynomial whose coefficients are no more perturbed than
    the floating-point computation of P(z_j).

    The condition number of each root is also computed, as a measure of
    sensitivity to changes in the coefficients of the polynomial.
    It is given by

        kappa(z_j) = alpha(z_j)/|z_j||P'(z_j)|.

    Root approximations can be further refined with optional 'polishing'
    processes.
    A simple polishing process is provided that carries out a single
    iteration of Newton's method, which proves quick yet often
    effective.
    Alternatively, a compensated polishing process from Cameron and
    O'Neill (2019) can be applied.
    This iterative method combines the implicit deflation of the
    modified Laguerre method, with the accuracy of evaluating
    polynomials and their derivatives using the compensated Horner's
    method from Graillat `et al.` (2005).
    Compensated polishing yields approximations with a limiting accuracy
    as if computed in twice the working precision.

    It is recommended that you read [Selecting a Polishing Process] for
    advice on selecting an appropriate polishing process.

    References
    ----------
    Bini, D A, 1996, `Numerical computation of polynomial zeros by means
    of Aberth's method`, Numerical Algorithms (13), 179--200, Springer
    US

    Cameron, T R, 2018, `An effective implementation of a modified
    Laguerre method for the roots of a polynomial`, Numerical
    Algorithms, Springer US, https://doi.org/10.1007/s11075-018-0641-9

    Cameron, T R and O'Neill, A, 2019, `On a compensated polishing
    technique for polynomial root solvers`, To be published,

    Graillat, S, Louvet, N, and Langlois, P, 2005, `Compensated Horner
    scheme`, Technical Report, Université de Perpignan Via Domitia

    Petković, M, Ilić, S, and Tričković, S, 1997, `A family of
    simultaneous zero-finding methods`, Computers & Mathematics with
    Applications (Volume 34) (10), 49--59,
    https://doi.org/10.1016/S0898-1221(97)00206-X

    Wilkinson, J H, 1959, `The evaluation of the zeros of
    ill-conditioned polynomials. Part I`, Numerische Mathematik (Volume
    1) (1), 150--166, Springer-Verlag

    See Also
    --------
    :meth:`naginterfaces.library.examples.zeros.poly_complex_fpml_ex.main`
    """

    fun_name = 'naginterfaces.library.zeros.poly_complex_fpml'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _z = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _cond = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _conv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    _b_zeros.poly_complex_fpml(
        _a,
        _n,
        itmax,
        polish,
        _z,
        _berr,
        _cond,
        _conv,
    )
    _z_py = _z
    _berr_py = _berr
    _cond_py = _cond
    _conv_py = _conv
    _return_data = PolyComplexFpmlReturnData(
        _z_py,
        _berr_py,
        _cond_py,
        _conv_py,
    )
    return _return_data

PolyRealFpmlReturnData = _collections.namedtuple(
    'PolyRealFpmlReturnData',
    ['z', 'berr', 'cond', 'conv'],
)

def poly_real_fpml(a, itmax=30, polish=1):
    u"""
    Finds all the zeros of real polynomial, fast modified Laguerre's
    method.

    ``poly_real_fpml`` finds all the roots of a real polynomial
    equation, using a fourth-order convergent modification of Laguerre's
    method.

    For full information please refer to the NAG Library document for
    c02ab

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/c02/c02abf.html

    Parameters
    ----------
    a : float, array-like, shape (n+1)
        `a`[i] must contain the coefficient of z^(n-i), for i =
        0,1,...,n.

    itmax : int, optional
        The maximum number of iterations to be performed.

    polish : int, optional
        Specifies the polishing technique used to refine root
        approximations.

        `polish` = 0

            No polishing.

        `polish` = 1

            Single iteration of Newton's method.

        `polish` = 2

            Iterative refinement using the compensated Horner's method.

        See also [Selecting a Polishing Process].

    Returns
    -------
    z : complex, ndarray, shape (n)
        `z`[j-1] holds the approximation of the root z_j, for j =
        1,2,...,n.

    berr : float, ndarray, shape (n)
        `berr`[j-1] holds the relative backward error, eta(z_j), for j =
        1,2,...,n.

    cond : float, ndarray, shape (n)
        `cond`[j-1] holds the condition number, kappa(z_j), for j =
        1,2,...,n.

    conv : int, ndarray, shape (n)
        `conv`[j-1] indicates the convergence status of the root
        approximation, z_j, for j = 1,2,...,n.

        `conv`[j-1] >= 0

            Successfully converged after `conv`[j-1] iterations.

        `conv`[j-1] = -1

            Failed to converge after `itmax` iterations.

        `conv`[j-1] = -2

            Overflow was encountered.

        Note: if `polish` = 2, `conv` refers to convergence in the
        compensated polishing process.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, the real variable `a`[0] = 0.0.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 3)
            On entry, `itmax` = *<value>*.

            Constraint: `itmax` >= 1.

        (`errno` 4)
            On entry, `polish` = *<value>*.

            Constraint: `polish` = 0, 1 or 2.

        (`errno` 5)
            Convergence has failed for at least one root approximation.

        (`errno` 6)
            ``poly_real_fpml`` encountered overflow during at least one
            root approximation.

    Notes
    -----
    ``poly_real_fpml`` attempts to find all the roots of the nth degree
    real polynomial equation

        P(z) = a_0 z^n+a_1 z^(n-1)+a_2 z^(n-2)+...+a_{n-1}z+a_n = 0.

    The roots are located using a modified form of Laguerre's method, as
    implemented by Cameron (2018).

    ``poly_real_fpml`` is a wrapper around the corresponding complex
    routine ``poly_complex_fpml``.

    The relative backward error of a root approximation z_j is given by

        eta(z_j) = |P(z_j)|/alpha(z_j),

    where alpha is the perturbed polynomial

        alpha(z) = sum((3.8*k+1)|a_{n-k}||z|^k).

    A root approximation is deemed to have converged if eta(z_j) <= 2 *
    machine precision, at which point updates of that root cease.
    If the stopping criterion holds, then the computed root is the exact
    root of a polynomial whose coefficients are no more perturbed than
    the floating-point computation of P(z_j).

    The condition number of each root is also computed, as a measure of
    sensitivity to changes in the coefficients of the polynomial.
    It is given by

        kappa(z_j) = alpha(z_j)/|z_j||P'(z_j)|.

    Root approximations can be further refined with optional 'polishing'
    processes.
    A simple polishing process is provided that carries out a single
    iteration of Newton's method, which proves quick yet often
    effective.
    Alternatively, a compensated polishing process from Cameron and
    O'Neill (2019) can be applied.
    This iterative method combines the implicit deflation of the
    modified Laguerre method, with the accuracy of evaluating
    polynomials and their derivatives using the compensated Horner's
    method from Graillat `et al.` (2005).
    Compensated polishing yields approximations with a limiting accuracy
    as if computed in twice the working precision.

    It is recommended that you read [Selecting a Polishing Process] for
    advice on selecting an appropriate polishing process.

    References
    ----------
    Bini, D A, 1996, `Numerical computation of polynomial zeros by means
    of Aberth's method`, Numerical Algorithms (13), 179--200, Springer
    US

    Cameron, T R, 2018, `An effective implementation of a modified
    Laguerre method for the roots of a polynomial`, Numerical
    Algorithms, Springer US, https://doi.org/10.1007/s11075-018-0641-9

    Cameron, T R and O'Neill, A, 2019, `On a compensated polishing
    technique for polynomial root solvers`, To be published,

    Graillat, S, Louvet, N, and Langlois, P, 2005, `Compensated Horner
    scheme`, Technical Report, Université de Perpignan Via Domitia

    Petković, M, Ilić, S, and Tričković, S, 1997, `A family of
    simultaneous zero-finding methods`, Computers & Mathematics with
    Applications (Volume 34) (10), 49--59,
    https://doi.org/10.1016/S0898-1221(97)00206-X

    Wilkinson, J H, 1959, `The evaluation of the zeros of
    ill-conditioned polynomials. Part I`, Numerische Mathematik (Volume
    1) (1), 150--166, Springer-Verlag
    """

    fun_name = 'naginterfaces.library.zeros.poly_real_fpml'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _z = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _cond = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _conv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    _b_zeros.poly_real_fpml(
        _a,
        _n,
        itmax,
        polish,
        _z,
        _berr,
        _cond,
        _conv,
    )
    _z_py = _z
    _berr_py = _berr
    _cond_py = _cond
    _conv_py = _conv
    _return_data = PolyRealFpmlReturnData(
        _z_py,
        _berr_py,
        _cond_py,
        _conv_py,
    )
    return _return_data

@_b_utils._withdrawal(
    'naginterfaces.library.zeros.poly_complex',
    mkwdrn=None,
    advice=[
        'Please use poly_complex_fpml instead.',
        'See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html',
    ]
)
def poly_complex(a, scal=True):
    u"""
    All zeros of complex polynomial, modified Laguerre's method.

    ``poly_complex`` finds all the roots of a complex polynomial
    equation, using a variant of Laguerre's method.

    Note: ``poly_complex`` is deprecated.
    Please use ``poly_complex_fpml`` instead.
    See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html

    For full information please refer to the NAG Library document for
    c02af

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/c02/c02aff.html

    Parameters
    ----------
    a : complex, array-like, shape (n+1)
        Re(`a`[i]) and Im(`a`[i]) must contain the real and imaginary
        parts of a_i (i.e., the coefficient of z^(n-i)), for i =
        0,1,...,n.

    scal : bool, optional
        Indicates whether or not the polynomial is to be scaled. See
        [Further Comments] for advice on when it may be preferable to
        set `scal` = False and for a description of the scaling
        strategy.

    Returns
    -------
    z : complex, ndarray, shape (n)
        The roots.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `a`[0] = 0.0.

            Constraint: `a`[0] != 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            The iterative procedure has failed to converge. This error
            is very unlikely to occur. If it does, please contact NAG
            immediately, as some basic assumption for the arithmetic has
            been violated.

        (`errno` 3)
            ``poly_complex`` cannot evaluate P(z) near some of its zeros
            without underflow. If this message occurs please contact
            NAG.

        (`errno` 3)
            ``poly_complex`` cannot evaluate P(z) near some of its zeros
            without overflow. If this message occurs please contact NAG.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``poly_complex`` attempts to find all the roots of the nth degree
    complex polynomial equation

        P(z) = a_0 z^n+a_1 z^(n-1)+a_2 z^(n-2)+...+a_{n-1}z+a_n = 0.

    The roots are located using a modified form of Laguerre's method,
    originally proposed by Smith (1967).

    The method of Laguerre (see Wilkinson (1965)) can be described by
    the iterative scheme

        L(z_k) = z_{k+1}-z_k = -nP(z_k)/(P'(z_k)+/-sqrt(H(z_k))),

    where H(z_k) = (n-1)[(n-1)(P'(z_k))^2-nP(z_k)P^('')(z_k)] and z_0 is
    specified.

    The sign in the denominator is chosen so that the modulus of the
    Laguerre step at z_k, viz. |L(z_k)|, is as small as possible.
    The method can be shown to be cubically convergent for isolated
    roots (real or complex) and linearly convergent for multiple roots.

    The function generates a sequence of iterates z_1,z_2,z_3,..., such
    that |P(z_{k+1})| < |P(z_k)| and ensures that z_{k+1}+L(z_{k+1})
    'roughly' lies inside a circular region of radius |F| about z_k
    known to contain a zero of P(z); that is, |L(z_{k+1})| <= |F|, where
    F denotes the Fejér bound (see Marden (1966)) at the point z_k.
    Following Smith (1967), F is taken to be min(B, 1.445nR), where B is
    an upper bound for the magnitude of the smallest zero given by

        B = 1.0001*min(sqrt(n)L(z_k), |r_1|, |a_n/a_0|^(1/n)),

    r_1 is the zero X of smaller magnitude of the quadratic equation

        P^('')(z_k)/(2n(n-1))X^2+P'(z_k)/nX+1/2P(z_k) = 0

    and the Cauchy lower bound R for the smallest zero is computed
    (using Newton's Method) as the positive root of the polynomial
    equation

        |a_0|z^n+|a_1|z^(n-1)+|a_2|z^(n-2)+...+|a_{n-1}|z-|a_n| = 0.

    Starting from the origin, successive iterates are generated
    according to the rule z_{k+1} = z_k+L(z_k), for k = 1,2,3,..., and
    L(z_k) is 'adjusted' so that |P(z_{k+1})| < |P(z_k)| and
    |L(z_{k+1})| <= |F|.
    The iterative procedure terminates if P(z_{k+1}) is smaller in
    absolute value than the bound on the rounding error in P(z_{k+1})
    and the current iterate z_p = z_{k+1} is taken to be a zero of P(z).
    The deflated polynomial P~(z) = P(z)/(z-z_p) of degree n-1 is then
    formed, and the above procedure is repeated on the deflated
    polynomial until n < 3, whereupon the remaining roots are obtained
    via the 'standard' closed formulae for a linear (n = 1) or quadratic
    (n = 2) equation.

    Note that ``quadratic_complex``, ``cubic_complex`` and
    ``quartic_complex`` can be used to obtain the roots of a quadratic,
    cubic (n = 3) and quartic (n = 4) polynomial, respectively.

    References
    ----------
    Marden, M, 1966, `Geometry of polynomials`, Mathematical Surveys
    (3), American Mathematical Society, Providence, RI

    Smith, B T, 1967, `ZERPOL: a zero finding algorithm for polynomials
    using Laguerre's method`, Technical Report, Department of Computer
    Science, University of Toronto, Canada

    Thompson, K W, 1991, `Error analysis for polynomial solvers`,
    Fortran Journal (Volume 3) (3), 10--13

    Wilkinson, J H, 1965, `The Algebraic Eigenvalue Problem`, Oxford
    University Press, Oxford
    """

    fun_name = 'naginterfaces.library.zeros.poly_complex'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _z = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py,),
    )
    _b_zeros.poly_complex(
        _a,
        _n,
        scal,
        _z,
    )
    _z_py = _z
    return _z_py

@_b_utils._withdrawal(
    'naginterfaces.library.zeros.poly_real',
    mkwdrn=None,
    advice=[
        'Please use poly_real_fpml instead.',
        'See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html',
    ]
)
def poly_real(a, scal=True):
    u"""
    All zeros of real polynomial, modified Laguerre's method.

    ``poly_real`` finds all the roots of a real polynomial equation,
    using a variant of Laguerre's method.

    Note: ``poly_real`` is deprecated.
    Please use ``poly_real_fpml`` instead.
    See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html

    For full information please refer to the NAG Library document for
    c02ag

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/c02/c02agf.html

    Parameters
    ----------
    a : float, array-like, shape (n+1)
        `a`[i] must contain a_i (i.e., the coefficient of z^(n-i)), for
        i = 0,1,...,n.

    scal : bool, optional
        Indicates whether or not the polynomial is to be scaled. See
        [Further Comments] for advice on when it may be preferable to
        set `scal` = False and for a description of the scaling
        strategy.

    Returns
    -------
    z : complex, ndarray, shape (n)
        The roots. Complex conjugate pairs of roots are stored in
        consecutive elements of `z`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `a`[0] = 0.0.

            Constraint: `a`[0] != 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            The iterative procedure has failed to converge. This error
            is very unlikely to occur. If it does, please contact NAG
            immediately, as some basic assumption for the arithmetic has
            been violated.

        (`errno` 3)
            ``poly_real`` cannot evaluate P(z) near some of its zeros
            without underflow. If this message occurs please contact
            NAG.

        (`errno` 3)
            ``poly_real`` cannot evaluate P(z) near some of its zeros
            without overflow. If this message occurs please contact NAG.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``poly_real`` attempts to find all the roots of the nth degree real
    polynomial equation

        P(z) = a_0 z^n+a_1 z^(n-1)+a_2 z^(n-2)+...+a_{n-1}z+a_n = 0.

    The roots are located using a modified form of Laguerre's method,
    originally proposed by Smith (1967).

    The method of Laguerre (see Wilkinson (1965)) can be described by
    the iterative scheme

        L(z_k) = z_{k+1}-z_k = -nP(z_k)/(P'(z_k)+/-sqrt(H(z_k))),

    where H(z_k) = (n-1)[(n-1)(P'(z_k))^2-nP(z_k)P^('')(z_k)] and z_0 is
    specified.

    The sign in the denominator is chosen so that the modulus of the
    Laguerre step at z_k, viz. |L(z_k)|, is as small as possible.
    The method can be shown to be cubically convergent for isolated
    roots (real or complex) and linearly convergent for multiple roots.

    The function generates a sequence of iterates z_1,z_2,z_3,..., such
    that |P(z_{k+1})| < |P(z_k)| and ensures that z_{k+1}+L(z_{k+1})
    'roughly' lies inside a circular region of radius |F| about z_k
    known to contain a zero of P(z); that is, |L(z_{k+1})| <= |F|, where
    F denotes the Fejér bound (see Marden (1966)) at the point z_k.
    Following Smith (1967), F is taken to be min(B, 1.445nR), where B is
    an upper bound for the magnitude of the smallest zero given by

        B = 1.0001*min(sqrt(n)L(z_k), |r_1|, |a_n/a_0|^(1/n)),

    r_1 is the zero X of smaller magnitude of the quadratic equation

        P^('')(z_k)/(2n(n-1))X^2+P'(z_k)/nX+1/2P(z_k) = 0

    and the Cauchy lower bound R for the smallest zero is computed
    (using Newton's Method) as the positive root of the polynomial
    equation

        |a_0|z^n+|a_1|z^(n-1)+|a_2|z^(n-2)+...+|a_{n-1}|z-|a_n| = 0.

    Starting from the origin, successive iterates are generated
    according to the rule z_{k+1} = z_k+L(z_k), for k = 1,2,3,..., and
    L(z_k) is 'adjusted' so that |P(z_{k+1})| < |P(z_k)| and
    |L(z_{k+1})| <= |F|.
    The iterative procedure terminates if P(z_{k+1}) is smaller in
    absolute value than the bound on the rounding error in P(z_{k+1})
    and the current iterate z_p = z_{k+1} is taken to be a zero of P(z)
    (as is its conjugate z__p if z_p is complex).
    The deflated polynomial P~(z) = P(z)/(z-z_p) of degree n-1 if z_p is
    real (P~(z) = P(z)/((z-z_p)(z-z__p)) of degree n-2 if z_p is
    complex) is then formed, and the above procedure is repeated on the
    deflated polynomial until n < 3, whereupon the remaining roots are
    obtained via the 'standard' closed formulae for a linear (n = 1) or
    quadratic (n = 2) equation.

    Note that ``quadratic_real``, ``cubic_real`` and ``quartic_real``
    can be used to obtain the roots of a quadratic, cubic (n = 3) and
    quartic (n = 4) polynomial, respectively.

    References
    ----------
    Marden, M, 1966, `Geometry of polynomials`, Mathematical Surveys
    (3), American Mathematical Society, Providence, RI

    Smith, B T, 1967, `ZERPOL: a zero finding algorithm for polynomials
    using Laguerre's method`, Technical Report, Department of Computer
    Science, University of Toronto, Canada

    Thompson, K W, 1991, `Error analysis for polynomial solvers`,
    Fortran Journal (Volume 3) (3), 10--13

    Wilkinson, J H, 1965, `The Algebraic Eigenvalue Problem`, Oxford
    University Press, Oxford
    """

    fun_name = 'naginterfaces.library.zeros.poly_real'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _z = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py,),
    )
    _b_zeros.poly_real(
        _a,
        _n,
        scal,
        _z,
    )
    _z_py = _z
    return _z_py

QuadraticComplexReturnData = _collections.namedtuple(
    'QuadraticComplexReturnData',
    ['z_small', 'z_large'],
)

def quadratic_complex(a, b, c):
    u"""
    All zeros of complex quadratic equation.

    ``quadratic_complex`` determines the roots of a quadratic equation
    with complex coefficients.

    For full information please refer to the NAG Library document for
    c02ah

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/c02/c02ahf.html

    Parameters
    ----------
    a : complex
        a, the coefficient of z^2.

    b : complex
        b, the coefficient of z.

    c : complex
        c, the constant coefficient.

    Returns
    -------
    z_small : complex
        The smallest root in magnitude.

    z_large : complex
        The largest root in magnitude.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `a` = 0.0.

        (`errno` 2)
            On entry, `a` = 0.0 and `b` = 0.0.

        (`errno` 3)
            On entry, `a` = 0.0 and the root -`c`/`b` overflows: Re(`a`)
            = *<value>*, Re(`c`) = *<value>*, Re(`b`) = *<value>*,
            Im(`a`) = *<value>*, Im(`c`) = *<value>*, Im(`b`) =
            *<value>*.

        (`errno` 4)
            On entry, `c` = 0.0 and the root -`b`/`a` overflows: Re(`c`)
            = *<value>*, Re(`b`) = *<value>*, Re(`a`) = *<value>*,
            Im(`c`) = *<value>*, Im(`b`) = *<value>*, Im(`a`) =
            *<value>*.

        (`errno` 5)
            On entry, B = max(abs(Re(`b`)), abs(Im(`b`))) = *<value>*, A
            = max(abs(Re(`a`)), abs(Im(`a`))) = *<value>* and C =
            max(abs(Re(`c`)), abs(Im(`c`))) = *<value>*. B is so large
            that B^2 is indistinguishable from (B^2-4*A*C) and the root
            -`b`/`a` overflows. Re(`b`) = *<value>*, Im(`b`) =
            *<value>*, Re(`a`) = *<value>* and Im(`a`) = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``quadratic_complex`` attempts to find the roots of the quadratic
    equation az^2+bz+c = 0 (where a, b and c are complex coefficients),
    by carefully evaluating the 'standard' closed formula

        z = (-b+/-sqrt(b^2-4ac))/2a.

    It is based on the function CQDRTC from Smith (1967).

    **Note:** it is not necessary to scale the coefficients prior to
    calling the function.

    References
    ----------
    Smith, B T, 1967, `ZERPOL: a zero finding algorithm for polynomials
    using Laguerre's method`, Technical Report, Department of Computer
    Science, University of Toronto, Canada
    """

    try:
        _return_scalars = _b_zeros.quadratic_complex(
            a,
            b,
            c,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _z_small_py = _return_scalars.z_small
    _z_large_py = _return_scalars.z_large
    _return_data = QuadraticComplexReturnData(
        _z_small_py,
        _z_large_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

QuadraticRealReturnData = _collections.namedtuple(
    'QuadraticRealReturnData',
    ['z_small', 'z_large'],
)

def quadratic_real(a, b, c):
    u"""
    All zeros of real quadratic equation.

    ``quadratic_real`` determines the roots of a quadratic equation with
    real coefficients.

    For full information please refer to the NAG Library document for
    c02aj

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/c02/c02ajf.html

    Parameters
    ----------
    a : float
        Must contain a, the coefficient of z^2.

    b : float
        Must contain b, the coefficient of z.

    c : float
        Must contain c, the constant coefficient.

    Returns
    -------
    z_small : complex
        The smallest root in magnitude.

    z_large : complex
        The largest root in magnitude.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `a` = 0.0.

        (`errno` 2)
            On entry, `a` = 0.0 and `b` = 0.0.

        (`errno` 3)
            On entry, `a` = 0.0 and the root -`c`/`b` overflows: `a` =
            *<value>*, `c` = *<value>* and `b` = *<value>*.

        (`errno` 4)
            On entry, `c` = 0.0 and the root -`b`/`a` overflows: `c` =
            *<value>*, `b` = *<value>* and `a` = *<value>*.

        (`errno` 5)
            On entry, `b` is so large that `b`^2 is indistinguishable
            from (`b`^2-4*`a`*`c`) and the root -`b`/`a` overflows: `b`
            = *<value>*, `a` = *<value>* and `c` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``quadratic_real`` attempts to find the roots of the quadratic
    equation az^2+bz+c = 0 (where a, b and c are real coefficients), by
    carefully evaluating the 'standard' closed formula

        z = (-b+/-sqrt(b^2-4ac))/2a.

    It is based on the function QDRTC from Smith (1967).

    **Note:** it is not necessary to scale the coefficients prior to
    calling the function.

    References
    ----------
    Smith, B T, 1967, `ZERPOL: a zero finding algorithm for polynomials
    using Laguerre's method`, Technical Report, Department of Computer
    Science, University of Toronto, Canada
    """

    try:
        _return_scalars = _b_zeros.quadratic_real(
            a,
            b,
            c,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _z_small_py = _return_scalars.z_small
    _z_large_py = _return_scalars.z_large
    _return_data = QuadraticRealReturnData(
        _z_small_py,
        _z_large_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

CubicRealReturnData = _collections.namedtuple(
    'CubicRealReturnData',
    ['zero', 'errest'],
)

def cubic_real(u, r, s, t):
    u"""
    All zeros of real cubic equation.

    ``cubic_real`` determines the roots of a cubic equation with real
    coefficients.

    For full information please refer to the NAG Library document for
    c02ak

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/c02/c02akf.html

    Parameters
    ----------
    u : float
        u, the coefficient of z^3.

    r : float
        r, the coefficient of z^2.

    s : float
        s, the coefficient of z.

    t : float
        t, the constant coefficient.

    Returns
    -------
    zero : complex, ndarray, shape (3)
        `zero`[i-1] contains the ith root.

    errest : float, ndarray, shape (3)
        `errest`[i-1] contains an approximate error estimate for the ith
        root.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `u` = 0.0.

            Constraint: `u` != 0.0.

        (`errno` 2)
            The companion matrix H cannot be formed without overflow.

        (`errno` 3)
            Failure to converge in ``lapackeig.dhseqr``.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``cubic_real`` attempts to find the roots of the cubic equation

        uz^3+rz^2+sz+t = 0,

    where u, r, s and t are real coefficients with u != 0.
    The roots are located by finding the eigenvalues of the associated
    3*3 (upper Hessenberg) companion matrix H given by

        H = [table omitted].

    The eigenvalues are obtained by a call to ``lapackeig.dhseqr``.
    Further details can be found in [Further Comments].

    To obtain the roots of a quartic equation, ``quartic_real`` can be
    used.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    _zero = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (3,),
    )
    _errest = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    _b_zeros.cubic_real(
        u,
        r,
        s,
        t,
        _zero,
        _errest,
    )
    _zero_py = _zero
    _errest_py = _errest
    _return_data = CubicRealReturnData(
        _zero_py,
        _errest_py,
    )
    return _return_data

QuarticRealReturnData = _collections.namedtuple(
    'QuarticRealReturnData',
    ['zero', 'errest'],
)

def quartic_real(e, a, b, c, d):
    u"""
    All zeros of real quartic equation.

    ``quartic_real`` determines the roots of a quartic equation with
    real coefficients.

    For full information please refer to the NAG Library document for
    c02al

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/c02/c02alf.html

    Parameters
    ----------
    e : float
        e, the coefficient of z^4.

    a : float
        a, the coefficient of z^3.

    b : float
        b, the coefficient of z^2.

    c : float
        c, the coefficient of z.

    d : float
        d, the constant coefficient.

    Returns
    -------
    zero : complex, ndarray, shape (4)
        `zero`[i-1] contains the ith root.

    errest : float, ndarray, shape (4)
        `errest`[i-1] contains an approximate error estimate for the ith
        root.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `e` = 0.0.

            Constraint: `e` != 0.0.

        (`errno` 2)
            The companion matrix H cannot be formed without overflow.

        (`errno` 3)
            Failure to converge in ``lapackeig.dhseqr``.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``quartic_real`` attempts to find the roots of the quartic equation

        ez^4+az^3+bz^2+cz+d = 0,

    where e, a, b, c and d are real coefficients with e != 0.
    The roots are located by finding the eigenvalues of the associated
    4*4 (upper Hessenberg) companion matrix H given by

        H = [table omitted].

    The eigenvalues are obtained by a call to ``lapackeig.dhseqr``.
    Further details can be found in [Further Comments].

    To obtain the roots of a cubic equation, ``cubic_real`` can be used.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    _zero = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (4,),
    )
    _errest = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    _b_zeros.quartic_real(
        e,
        a,
        b,
        c,
        d,
        _zero,
        _errest,
    )
    _zero_py = _zero
    _errest_py = _errest
    _return_data = QuarticRealReturnData(
        _zero_py,
        _errest_py,
    )
    return _return_data

CubicComplexReturnData = _collections.namedtuple(
    'CubicComplexReturnData',
    ['zero', 'errest'],
)

def cubic_complex(u, r, s, t):
    u"""
    All zeros of complex cubic equation.

    ``cubic_complex`` determines the roots of a cubic equation with
    complex coefficients.

    For full information please refer to the NAG Library document for
    c02am

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/c02/c02amf.html

    Parameters
    ----------
    u : complex
        u, the coefficient of z^3.

    r : complex
        r, the coefficient of z^2.

    s : complex
        s, the coefficient of z.

    t : complex
        t, the constant coefficient.

    Returns
    -------
    zero : complex, ndarray, shape (3)
        `zero`[i-1] contains the ith root.

    errest : float, ndarray, shape (3)
        `errest`[i-1] contains an approximate error estimate for the ith
        root.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `u` = (0.0,0.0).

            Constraint: `u` != (0.0,0.0)

        (`errno` 2)
            The companion matrix H cannot be formed without overflow.

        (`errno` 3)
            Failure to converge in ``lapackeig.zhseqr``.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``cubic_complex`` attempts to find the roots of the cubic equation

        uz^3+rz^2+sz+t = 0,

    where u, r, s and t are complex coefficients with u != 0.
    The roots are located by finding the eigenvalues of the associated
    3*3 (upper Hessenberg) companion matrix H given by

        H = [table omitted].

    The eigenvalues are obtained by a call to ``lapackeig.zhseqr``.
    Further details can be found in [Further Comments].

    To obtain the roots of a quadratic equation, ``quadratic_complex``
    can be used.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    _zero = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (3,),
    )
    _errest = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    _b_zeros.cubic_complex(
        u,
        r,
        s,
        t,
        _zero,
        _errest,
    )
    _zero_py = _zero
    _errest_py = _errest
    _return_data = CubicComplexReturnData(
        _zero_py,
        _errest_py,
    )
    return _return_data

QuarticComplexReturnData = _collections.namedtuple(
    'QuarticComplexReturnData',
    ['zero', 'errest'],
)

def quartic_complex(e, a, b, c, d):
    u"""
    All zeros of complex quartic equation.

    ``quartic_complex`` determines the roots of a quartic equation with
    complex coefficients.

    For full information please refer to the NAG Library document for
    c02an

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/c02/c02anf.html

    Parameters
    ----------
    e : complex
        e, the coefficient of z^4.

    a : complex
        a, the coefficient of z^3.

    b : complex
        b, the coefficient of z^2.

    c : complex
        c, the coefficient of z.

    d : complex
        d, the constant coefficient.

    Returns
    -------
    zero : complex, ndarray, shape (4)
        `zero`[i-1] contains the ith root.

    errest : float, ndarray, shape (4)
        `errest`[i-1] contains an approximate error estimate for the ith
        root.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `e` = (0.0,0.0).

            Constraint: `e` != (0.0,0.0)

        (`errno` 2)
            The companion matrix H cannot be formed without overflow.

        (`errno` 3)
            Failure to converge in ``lapackeig.zhseqr``.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``quartic_complex`` attempts to find the roots of the quartic
    equation

        ez^4+az^3+bz^2+cz+d = 0,

    where e, a, b, c and d are complex coefficients with e != 0.
    The roots are located by finding the eigenvalues of the associated
    4*4 (upper Hessenberg) companion matrix H given by

        H = [table omitted].

    The eigenvalues are obtained by a call to ``lapackeig.zhseqr``.
    Further details can be found in [Further Comments].

    To obtain the roots of a cubic equation, ``cubic_complex`` can be
    used.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    _zero = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (4,),
    )
    _errest = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    _b_zeros.quartic_complex(
        e,
        a,
        b,
        c,
        d,
        _zero,
        _errest,
    )
    _zero_py = _zero
    _errest_py = _errest
    _return_data = QuarticComplexReturnData(
        _zero_py,
        _errest_py,
    )
    return _return_data
