# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,misplaced-comparison-constant,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.1 `sparse` Chapter.

``sparse`` - Large Scale Linear Systems

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module provides functions for the solution of large sparse systems
of simultaneous linear equations.
These include **iterative** methods for real nonsymmetric and symmetric,
complex non-Hermitian and Hermitian linear systems and **direct**
methods for general real linear systems.
Further direct methods are currently available in submodule ``matop``
and submodule ``linsys``.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11intro.html
"""

# NAG Copyright 2017-2020.

import collections as _collections
import logging as _logging
import sys as _sys

from ..base import utils

def real_gen_basic_setup(method, precon, norm, weight, iterm, n, m, tol, maxitn, anorm, sigmax, monit):
    u"""
    Real sparse nonsymmetric linear systems, setup for
    ``real_gen_basic_solver``.

    ``real_gen_basic_setup`` is a setup function, the first in a suite
    of three functions for the iterative solution of a real general
    (nonsymmetric) system of simultaneous linear equations.
    ``real_gen_basic_setup`` must be called before
    ``real_gen_basic_solver``, the iterative solver.
    The third function in the suite, ``real_gen_basic_diag``, can be
    used to return additional information about the computation.

    These functions are suitable for the solution of large sparse
    general (nonsymmetric) systems of equations.

    For full information please refer to the NAG Library document for
    f11bd

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11bdf.html

    Parameters
    ----------
    method : str
        The iterative method to be used.

        `method` = 'RGMRES'

            Restarted generalized minimum residual method.

        `method` = 'CGS'

            Conjugate gradient squared method.

        `method` = 'BICGSTAB'

            Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

            Transpose-free quasi-minimal residual method.

    precon : str, length 1
        Determines whether preconditioning is used.

        `precon` = 'N'

            No preconditioning.

        `precon` = 'P'

            Preconditioning.

    norm : str, length 1
        Defines the matrix and vector norm to be used in the termination
        criteria.

        `norm` = '1'

            l_1 norm.

        `norm` = 'I'

            l_infinity norm.

        `norm` = '2'

            l_2 norm.

        Suggested value:

            if `iterm` = 1, `norm` = 'I'

            if `iterm` = 2, `norm` = '2'

    weight : str, length 1
        Specifies whether a vector w of user-supplied weights is to be
        used in the computation of the vector norms required in
        termination criterion (2) (`iterm` = 1): ||v||_p^(w) =
        ||v^(w)||_p, where v_i^(w) = w_i v_i, for i = 1,2,...,n. The
        suffix p = 1,2, infinity denotes the vector norm used, as
        specified by the argument `norm`. Note that weights cannot be
        used when `iterm` = 2, i.e., when criterion (3) is used.

        `weight` = 'W'

            User-supplied weights are to be used and must be supplied on
            initial entry to ``real_gen_basic_solver``.

        `weight` = 'N'

            All weights are implicitly set equal to one. Weights do not
            need to be supplied on initial entry to
            ``real_gen_basic_solver``.

        Suggested value: `weight` = 'N'.

    iterm : int
        Defines the termination criterion to be used.

        `iterm` = 1

            Use the termination criterion defined in (2).

        `iterm` = 2

            Use the termination criterion defined in (3).

        Suggested value: `iterm` = 1.

    n : int
        n, the order of the matrix A.

    m : int
        If `method` = 'RGMRES', `m` is the dimension m of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method.

        Otherwise, `m` is not referenced.

    tol : float
        The tolerance tau for the termination criterion. If `tol` <=
        0.0, tau = max(sqrt(epsilon), sqrt(n)*epsilon) is used, where
        epsilon is the machine precision. Otherwise tau = max(`tol`,
        10*epsilon, sqrt(n)*epsilon) is used.

    maxitn : int
        The maximum number of iterations.

    anorm : float
        If `anorm` > 0.0, the value of ||A||_p to be used in the
        termination criterion (2) (`iterm` = 1).

        If `anorm` <= 0.0, `iterm` = 1 and `norm` = '1' or 'I', then
        ||A||_1 = ||A||_infinity is estimated internally by
        ``real_gen_basic_solver``.

        If `iterm` = 2, `anorm` is not referenced.

    sigmax : float
        If `iterm` = 2, the largest singular value sigma_1 of the
        preconditioned iteration matrix; otherwise, `sigmax` is not
        referenced.

        If `sigmax` <= 0.0, `iterm` = 2 and `method` = 'RGMRES', the
        value of sigma_1 will be estimated internally.

    monit : int
        If `monit` > 0, the frequency at which a monitoring step is
        executed by ``real_gen_basic_solver``: if `method` = 'CGS' or
        'TFQMR', a monitoring step is executed every `monit` iterations;
        otherwise, a monitoring step is executed every `monit`
        super-iterations (groups of up to m or l iterations for RGMRES
        or BI-CGSTAB(l), respectively).

        There are some additional computational costs involved in
        monitoring the solution and residual vectors when the
        BI-CGSTAB(l) method is used with l > 1.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` -12)
            On entry, `monit` = *<value>* and `maxitn` = *<value>*.

            Constraint: `monit` <= `maxitn`.

        (`errno` -11)
            On entry, `iterm` = 2, `method` = *<value>* and `sigmax` =
            *<value>*.

            Constraint: if `iterm` = 2 and `method` = 'CGS' or
            'BICGSTAB', `sigmax` > 0.0.

        (`errno` -10)
            On entry, `iterm` = 1, `norm` = '2' and `anorm` = *<value>*.

            Constraint: if `iterm` = 1 and `norm` = '2', `anorm` > 0.0.

        (`errno` -9)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` > 0.

        (`errno` -8)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` -7)
            On entry, `m` = *<value>* and `method` = *<value>*.

            Constraint: if `method` = 'RGMRES' or 'BICGSTAB', `m` > 0.

        (`errno` -7)
            On entry, `m` = *<value>*, `n` = *<value>* and `method` =
            *<value>*.

            Constraint: if `method` = 'RGMRES', `m` <= min(`n`, 50). If
            `method` = 'BICGSTAB', `m` <= min(`n`, 10).

        (`errno` -6)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` -5)
            On entry, `iterm` = *<value>*.

            Constraint: `iterm` = 1 or 2.

        (`errno` -5)
            On entry, `iterm` = 2 and `method` = 'TFQMR'.

            Constraint: if `iterm` = 2, `method` != 'TFQMR'.

        (`errno` -5)
            On entry, `iterm` = 2 and `norm` = *<value>*.

            Constraint: if `iterm` = 2, `norm` = '2'.

        (`errno` -5)
            On entry, `iterm` = 2 and `weight` = *<value>*.

            Constraint: if `iterm` = 2, `weight` = 'N'.

        (`errno` -4)
            On entry, `weight` = *<value>*.

            Constraint: `weight` = 'N' or 'W'.

        (`errno` -3)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = '1', 'I' or '2'.

        (`errno` -2)
            On entry, `precon` = *<value>*.

            Constraint: `precon` = 'N' or 'P'.

        (`errno` -1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS', 'BICGSTAB' or
            'TFQMR'.

        (`errno` 1)
            ``real_gen_basic_setup`` has been called out of sequence:
            either ``real_gen_basic_setup`` has been called twice or
            ``real_gen_basic_solver`` has not terminated its current
            task.

    Notes
    -----
    The suite consisting of the functions ``real_gen_basic_setup``,
    ``real_gen_basic_solver`` and ``real_gen_basic_diag`` is designed to
    solve the general (nonsymmetric) system of simultaneous linear
    equations Ax = b of order n, where n is large and the coefficient
    matrix A is sparse.

    ``real_gen_basic_setup`` is a setup function which must be called
    before ``real_gen_basic_solver``, the iterative solver.
    The third function in the suite, ``real_gen_basic_diag``, can be
    used to return additional information about the computation.
    A choice of methods is available:

        restarted generalized minimum residual method (RGMRES);

        conjugate gradient squared method (CGS);

        bi-conjugate gradient stabilized (l) method (BI-CGSTAB(l));

        transpose-free quasi-minimal residual method (TFQMR).

    Restarted Generalized Minimum Residual Method (RGMRES)
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    The restarted generalized minimum residual method (RGMRES) (see Saad
    and Schultz (1986), Barrett `et al.` (1994) and Dias da Cunha and
    Hopkins (1994)) starts from the residual r_0 = b-Ax_0, where x_0 is
    an initial estimate for the solution (often x_0 = 0).
    An orthogonal basis for the Krylov subspace span{A^kr_0}, for k =
    0,1,...,, is generated explicitly: this is referred to as Arnoldi's
    method (see Arnoldi (1951)).
    The solution is then expanded onto the orthogonal basis so as to
    minimize the residual norm ||b-Ax||_2.
    The lack of symmetry of A implies that the orthogonal basis is
    generated by applying a 'long' recurrence relation, whose length
    increases linearly with the iteration count.
    For all but the most trivial problems, computational and storage
    costs can quickly become prohibitive as the iteration count
    increases. RGMRES limits these costs by employing a restart
    strategy: every m iterations at most, the Arnoldi process is
    restarted from r_l = b-Ax_l, where the subscript l denotes the last
    available iterate.
    Each group of m iterations is referred to as a 'super-iteration'.
    The value of m is chosen in advance and is fixed throughout the
    computation.
    Unfortunately, an optimum value of m cannot easily be predicted.

    Conjugate Gradient Squared Method (CGS)
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    The conjugate gradient squared method (CGS) (see Sonneveld (1989),
    Barrett `et al.` (1994) and Dias da Cunha and Hopkins (1994)) is a
    development of the bi-conjugate gradient method where the
    nonsymmetric Lanczos method is applied to reduce the coefficients
    matrix to real tridiagonal form: two bi-orthogonal sequences of
    vectors are generated starting from the residual r_0 = b-Ax_0, where
    x_0 is an initial estimate for the solution (often x_0 = 0) and from
    the `shadow residual` r^_0 corresponding to the arbitrary problem
    A^Tx^= b^, where b^ can be any vector, but in practice is chosen so
    that r_0 = r^_0.
    In the course of the iteration, the residual and shadow residual r_i
    = P_i(A)r_0 and r^_i = P_i(A^T)r^_0 are generated, where P_i is a
    polynomial of order i, and bi-orthogonality is exploited by
    computing the vector product rho_i = (r^_i,r_i) = (P_i(A^T)r^_0,
    P_i(A)r_0) = (r^_0,P_i^2(A)r_0).
    Applying the 'contraction' operator P_i(A) twice, the iteration
    coefficients can still be recovered without advancing the solution
    of the shadow problem, which is of no interest.
    The CGS method often provides fast convergence; however, there is no
    reason why the contraction operator should also reduce the once
    reduced vector P_i(A)r_0: this may well lead to a highly irregular
    convergence which may result in large cancellation errors.

    Bi-Conjugate Gradient Stabilized (l) Method (BI-CGSTAB(l))
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    The bi-conjugate gradient stabilized (l) method (BI-CGSTAB(l)) (see
    Van der Vorst (1989), Sleijpen and Fokkema (1993) and Dias da Cunha
    and Hopkins (1994)) is similar to the CGS method above.
    However, instead of generating the sequence {P_i^2(A)r_0}, it
    generates the sequence {Q_i(A)P_i(A)r_0}, where the Q_i(A) are
    polynomials chosen to minimize the residual `after` the application
    of the contraction operator P_i(A).
    Two main steps can be identified for each iteration: an OR
    (Orthogonal Residuals) step where a basis of order l is generated by
    a Bi-CG iteration and an MR (Minimum Residuals) step where the
    residual is minimized over the basis generated, by a method akin to
    GMRES.
    For l = 1, the method corresponds to the BI-CGSTAB method of Van der
    Vorst (1989).
    For l > 1, more information about complex eigenvalues of the
    iteration matrix can be taken into account, and this may lead to
    improved convergence and robustness.
    However, as l increases, numerical instabilities may arise.
    For this reason, a maximum value of l = 10 is imposed, but probably
    l = 4 is sufficient in most cases.

    Transpose-free Quasi-minimal Residual Method (TFQMR)
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    The transpose-free quasi-minimal residual method (TFQMR) (see Freund
    and Nachtigal (1991) and Freund (1993)) is conceptually derived from
    the CGS method.
    The residual is minimized over the space of the residual vectors
    generated by the CGS iterations under the simplifying assumption
    that residuals are almost orthogonal.
    In practice, this is not the case but theoretical analysis has
    proved the validity of the method.
    This has the effect of remedying the rather irregular convergence
    behaviour with wild oscillations in the residual norm that can
    degrade the numerical performance and robustness of the CGS method.
    In general, the TFQMR method can be expected to converge at least as
    fast as the CGS method, in terms of number of iterations, although
    each iteration involves a higher operation count.
    When the CGS method exhibits irregular convergence, the TFQMR method
    can produce much smoother, almost monotonic convergence curves.
    However, the close relationship between the CGS and TFQMR method
    implies that the `overall` speed of convergence is similar for both
    methods.
    In some cases, the TFQMR method may converge faster than the CGS
    method.

    General Considerations
    ~~~~~~~~~~~~~~~~~~~~~~
    For each method, a sequence of solution iterates {x_i} is generated
    such that, hopefully, the sequence of the residual norms {||r_i||}
    converges to a required tolerance.
    Note that, in general, convergence, when it occurs, is not
    monotonic.

    In the RGMRES and BI-CGSTAB(l) methods above, your program must
    provide the **maximum** number of basis vectors used, m or l,
    respectively; however, a **smaller** number of basis vectors may be
    generated and used when the stability of the solution process
    requires this (see [Further Comments]).

    Faster convergence can be achieved using a **preconditioner** (see
    Golub and Van Loan (1996) and Barrett `et al.` (1994)).
    A preconditioner maps the original system of equations onto a
    different system, say

        A_x_= b_,

    with, hopefully, better characteristics with respect to its speed of
    convergence: for example, the condition number of the coefficients
    matrix can be improved or eigenvalues in its spectrum can be made to
    coalesce.
    An orthogonal basis for the Krylov subspace span{A_^kr__0}, for k =
    0,1,...,, is generated and the solution proceeds as outlined above.
    The algorithms used are such that the solution and residual iterates
    of the original system are produced, not their preconditioned
    counterparts.
    Note that an unsuitable preconditioner or no preconditioning at all
    may result in a very slow rate, or lack, of convergence.
    However, preconditioning involves a trade-off between the reduction
    in the number of iterations required for convergence and the
    additional computational costs per iteration.
    Also, setting up a preconditioner may involve non-negligible
    overheads.

    A `left` preconditioner M^-1 can be used by the RGMRES, CGS and
    TFQMR methods, such that A_= M^-1A~I_n in (1), where I_n is the
    identity matrix of order n; a `right` preconditioner M^-1 can be
    used by the BI-CGSTAB(l) method, such that A_= AM^-1~I_n.
    These are formal definitions, used only in the design of the
    algorithms; in practice, only the means to compute the
    matrix--vector products v = Au and v = A^Tu (the latter only being
    required when an estimate of ||A||_1 or ||A||_infinity is computed
    internally), and to solve the preconditioning equations Mv = u are
    required, i.e., explicit information about M, or its inverse is not
    required at any stage.

    The first termination criterion

        ||r_k||_p <= tau(||b||_p+||A||_p*||x_k||_p)

    is available for all four methods.
    In (2), p = 1, infinity or 2 and tau denotes a user-specified
    tolerance subject to max(10, sqrt(n)), epsilon <= tau < 1, where
    epsilon is the machine precision.
    Facilities are provided for the estimation of the norm of the
    coefficients matrix ||A||_1 or ||A||_infinity, when this is not
    known in advance, by applying Higham's method (see Higham (1988)).
    Note that ||A||_2 cannot be estimated internally.
    This criterion uses an error bound derived from **backward** error
    analysis to ensure that the computed solution is the exact solution
    of a problem as close to the original as the termination tolerance
    requires.
    Termination criteria employing bounds derived from **forward** error
    analysis are not used because any such criteria would require
    information about the condition number kappa(A) which is not easily
    obtainable.

    The second termination criterion

        ||r__k||_2 <= tau(||r__0||_2+sigma_1(A_)*||Delta x__k||_2)

    is available for all methods except TFQMR.
    In (3), sigma_1(A_) = ||A_||_2 is the largest singular value of the
    (preconditioned) iteration matrix A_.
    This termination criterion monitors the progress of the solution of
    the preconditioned system of equations and is less expensive to
    apply than criterion (2) for the BI-CGSTAB(l) method with l > 1.
    Only the RGMRES method provides facilities to estimate sigma_1(A_)
    internally, when this is not supplied (see [Further Comments]).

    Termination criterion (2) is the recommended choice, despite its
    additional costs per iteration when using the BI-CGSTAB(l) method
    with l > 1.
    Also, if the norm of the initial estimate is much larger than the
    norm of the solution, that is, if ||x_0||≫||x||, a dramatic loss of
    significant digits could result in complete lack of convergence.
    The use of criterion (2) will enable the detection of such a
    situation, and the iteration will be restarted at a suitable point.
    No such restart facilities are provided for criterion (3).

    Optionally, a vector w of user-specified weights can be used in the
    computation of the vector norms in termination criterion (2), i.e.,
    ||v||_p^(w) = ||v^(w)||_p, where (v^(w))_i = w_i v_i, for i =
    1,2,...,n.
    Note that the use of weights increases the computational costs.

    The sequence of calls to the functions comprising the suite is
    enforced: first, the setup function ``real_gen_basic_setup`` must be
    called, followed by the solver ``real_gen_basic_solver``.
    ``real_gen_basic_diag`` can be called either when
    ``real_gen_basic_solver`` is carrying out a monitoring step or after
    ``real_gen_basic_solver`` has completed its tasks.
    Incorrect sequencing will raise an error condition.

    In general, it is not possible to recommend one method in preference
    to another. RGMRES is often used in the solution of systems arising
    from PDEs.
    On the other hand, it can easily stagnate when the size m of the
    orthogonal basis is too small, or the preconditioner is not good
    enough. CGS can be the fastest method, but the computed residuals
    can exhibit instability which may greatly affect the convergence and
    quality of the solution. BI-CGSTAB(l) seems robust and reliable, but
    it can be slower than the other methods: if a preconditioner is used
    and l > 1, BI-CGSTAB(l) computes the solution of the preconditioned
    system x__k = Mx_k: the preconditioning equations must be solved to
    obtain the required solution.
    The algorithm employed limits to 10% or less, when no intermediate
    monitoring is requested, the number of times the preconditioner has
    to be thus applied compared with the total number of applications of
    the preconditioner. TFQMR can be viewed as a more robust variant of
    the CGS method: it shares the CGS method speed but avoids the CGS
    fluctuations in the residual, which may give rise to instability.
    Also, when the termination criterion (2) is used, the CGS,
    BI-CGSTAB(l) and TFQMR methods will restart the iteration
    automatically when necessary in order to solve the given problem.

    References
    ----------
    Arnoldi, W, 1951, `The principle of minimized iterations in the
    solution of the matrix eigenvalue problem`, Quart. Appl. Math. (9),
    17--29

    Barrett, R, Berry, M, Chan, T F, Demmel, J, Donato, J, Dongarra, J,
    Eijkhout, V, Pozo, R, Romine, C and Van der Vorst, H, 1994,
    `Templates for the Solution of Linear Systems: Building Blocks for
    Iterative Methods`, SIAM, Philadelphia

    Dias da Cunha, R and Hopkins, T, 1994, `PIM 1.1 --- the parallel
    iterative method package for systems of linear equations user's
    guide --- Fortran 77 version`, Technical Report, Computing
    Laboratory, University of Kent at Canterbury, Kent, UK

    Freund, R W, 1993, `A transpose-free quasi-minimal residual
    algorithm for non-Hermitian linear systems`, SIAM J. Sci. Comput.
    (14), 470--482

    Freund, R W and Nachtigal, N, 1991, `QMR: a Quasi-Minimal Residual
    Method for Non-Hermitian Linear Systems`, Numer. Math. (60),
    315--339

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396

    Saad, Y and Schultz, M, 1986, `GMRES: a generalized minimal residual
    algorithm for solving nonsymmetric linear systems`, SIAM J. Sci.
    Statist. Comput. (7), 856--869

    Sleijpen, G L G and Fokkema, D R, 1993, `BiCGSTAB(l) for linear
    equations involving matrices with complex spectrum`, ETNA (1),
    11--32

    Sonneveld, P, 1989, `CGS, a fast Lanczos-type solver for
    nonsymmetric linear systems`, SIAM J. Sci. Statist. Comput. (10),
    36--52

    Van der Vorst, H, 1989, `Bi-CGSTAB, a fast and smoothly converging
    variant of Bi-CG for the solution of nonsymmetric linear systems`,
    SIAM J. Sci. Statist. Comput. (13), 631--644
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_basic_setup'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11bdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, iterm.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._BYREF_FLOAT64, # Input, anorm.
        utils._BYREF_FLOAT64, # Input, sigmax.
        utils._BYREF_INT, # Input, monit.
        utils._BYREF_INT, # Output, lwreq.
        utils._EngineFloat64ArrayType.c_type, # Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _weight = utils._EngineCharScalarType(
        weight, {'fun_name': fun_name, 'entity': 'weight'}, exp_chrlen=1,
    ).to_ctypes()
    _iterm = utils._EngineIntScalarType(
        iterm, {'fun_name': fun_name, 'entity': 'iterm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _anorm = utils._EngineFloat64ScalarType(
        anorm, {'fun_name': fun_name, 'entity': 'anorm'},
    ).to_ctypes()
    _sigmax = utils._EngineFloat64ScalarType(
        sigmax, {'fun_name': fun_name, 'entity': 'sigmax'},
    ).to_ctypes()
    _monit = utils._EngineIntScalarType(
        monit, {'fun_name': fun_name, 'entity': 'monit'},
    ).to_ctypes()
    _lwreq = utils._EngineIntScalarType.empty_ctypes()
    _lwork = utils._EngineIntScalarType.c_type(100)
    _lwork_py = utils._EngineIntScalarType.to_py(_lwork)
    _work = utils._EngineFloat64ArrayType.empty_ctypes(_lwork_py)
    _comm_py = {}
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _norm,
        _weight,
        _iterm,
        _n,
        _m,
        _tol,
        _maxitn,
        _anorm,
        _sigmax,
        _monit,
        _lwreq,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _norm._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    if _ierr.value == utils._IERR_SUCCESS:
        _lwreq_py = utils._EngineIntScalarType.to_py(_lwreq)
        _lwork = utils._EngineIntScalarType(
            _lwreq_py, {'fun_name': fun_name, 'entity': '_lwork'},
        ).to_ctypes()
        _lwork_py = utils._EngineIntScalarType.to_py(_lwork)
        _work = utils._EngineFloat64ArrayType.empty_ctypes(_lwork_py)
        _comm_py['work'] = _work
        engine_ctypes(
            _en_data,
            _method,
            _precon,
            _norm,
            _weight,
            _iterm,
            _n,
            _m,
            _tol,
            _maxitn,
            _anorm,
            _sigmax,
            _monit,
            _lwreq,
            _work,
            _lwork,
            _errbuf,
            _ierr,
            _method._length_,
            _precon._length_,
            _norm._length_,
            _weight._length_,
            _errbuf._length_,
        )
        utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _monit_py = utils._EngineIntScalarType.to_py(_monit)
        _maxitn_py = utils._EngineIntScalarType.to_py(_maxitn)
        _method_py = utils._EngineCharScalarType.to_py(_method)
        _sigmax_py = utils._EngineFloat64ScalarType.to_py(_sigmax)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _tol_py = utils._EngineFloat64ScalarType.to_py(_tol)
        _m_py = utils._EngineIntScalarType.to_py(_m)
        _n_py = utils._EngineIntScalarType.to_py(_n)
        _iterm_py = utils._EngineIntScalarType.to_py(_iterm)
        _norm_py = utils._EngineCharScalarType.to_py(_norm)
        _weight_py = utils._EngineCharScalarType.to_py(_weight)
        _precon_py = utils._EngineCharScalarType.to_py(_precon)
        if _ierr_py == -12 and _ifmt == -12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, monit = {:s} and maxitn = {:s}.'.format(
                        utils._STR_FMT_INT.format(_monit_py),
                        utils._STR_FMT_INT.format(_maxitn_py),
                    ),
                    'Constraint: monit <= maxitn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -11 and _ifmt == -11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2, method = \'{:s}\' and'.format(
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'sigmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_sigmax_py),
                    ),
                    'Constraint: if iterm = 2 and method = \'CGS\' or \'BICGSTAB\', sigmax > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -10 and _ifmt == -10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 1, norm = \'2\' and anorm = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_anorm_py),
                    ),
                    'Constraint: if iterm = 1 and norm = \'2\', anorm > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -9 and _ifmt == -9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT.format(_maxitn_py),
                    ),
                    'Constraint: maxitn > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -8 and _ifmt == -8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_tol_py),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -7 and _ifmt == -7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and method = \'{:s}\'.'.format(
                        utils._STR_FMT_INT.format(_m_py),
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'Constraint: if method = \'RGMRES\' or \'BICGSTAB\', m > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -7 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}, n = {:s} and'.format(
                        utils._STR_FMT_INT.format(_m_py),
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'Constraint: if method = \'RGMRES\', m <= min(n, 50).',
                    'If method = \'BICGSTAB\', m <= min(n, 10).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == -6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == -5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = {:s}.'.format(
                        utils._STR_FMT_INT.format(_iterm_py),
                    ),
                    'Constraint: iterm = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and method = \'TFQMR\'.',
                    'Constraint: if iterm = 2, method != \'TFQMR\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: if iterm = 2, norm = \'2\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_weight_py),
                    ),
                    'Constraint: if iterm = 2, weight = \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -4 and _ifmt == -4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_weight_py),
                    ),
                    'Constraint: weight = \'N\' or \'W\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -3 and _ifmt == -3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: norm = \'1\', \'I\' or \'2\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == -2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_precon_py),
                    ),
                    'Constraint: precon = \'N\' or \'P\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\', \'BICGSTAB\' or \'TFQMR\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'real_gen_basic_setup has been called out of sequence:',
                    'either real_gen_basic_setup has been called twice or',
                    'real_gen_basic_solver has not terminated its current task.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = _comm_py
    _err_class.manage()
    return return_data

def real_gen_basic_solver(irevcm, u, v, wgt, comm):
    u"""
    Real sparse nonsymmetric linear systems, preconditioned RGMRES, CGS,
    Bi-CGSTAB or TFQMR method.

    ``real_gen_basic_solver`` is an iterative solver for a real general
    (nonsymmetric) system of simultaneous linear equations;
    ``real_gen_basic_solver`` is the second in a suite of three
    functions, where the first function, ``real_gen_basic_setup``, must
    be called prior to ``real_gen_basic_solver`` to set up the suite,
    and the third function in the suite, ``real_gen_basic_diag``, can be
    used to return additional information about the computation.

    These functions are suitable for the solution of large sparse
    general (nonsymmetric) systems of equations.

    For full information please refer to the NAG Library document for
    f11be

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11bef.html

    Parameters
    ----------
    irevcm : int
        `On initial entry`: `irevcm` = 0, otherwise an error condition
        will be raised.

        `On intermediate entry`: must either be unchanged from its
        previous exit value, or can have one of the following values.

        `irevcm` = 5

            Tidy termination: the computation will terminate at the end
            of the current iteration. Further reverse communication
            exits may occur depending on when the termination request is
            issued. ``real_gen_basic_solver`` will then return with the
            termination code `irevcm` = 4. Note that before calling
            ``real_gen_basic_solver`` with `irevcm` = 5 the calling
            program must have performed the tasks required by the value
            of `irevcm` returned by the previous call to
            ``real_gen_basic_solver``, otherwise subsequently returned
            values may be invalid.

        `irevcm` = 6

            Immediate termination: ``real_gen_basic_solver`` will return
            immediately with termination code `irevcm` = 4 and with any
            useful information available. This includes the last iterate
            of the solution. The residual vector is generally not
            available.

            Immediate termination may be useful, for example, when
            errors are detected during matrix-vector multiplication or
            during the solution of the preconditioning equation.

        Changing `irevcm` to any other value between calls will result
        in an error.

    u : float, shape (n), modified in place
        `On initial entry`: an initial estimate, x_0, of the solution of
        the system of equations Ax = b.

        `On intermediate entry`: must remain unchanged.

        `On intermediate exit`: the returned value of `irevcm`
        determines the contents of `u` in the following way:

            if `irevcm` = -1, 1 or 2, `u` holds the vector u on which
            the operation specified by `irevcm` is to be carried out;

            if `irevcm` = 3, `u` holds the current iterate of the
            solution vector.

        `On final exit`: if `errno` = 3 or `errno` < 0, the array `u` is
        unchanged from the initial entry to ``real_gen_basic_solver``.

        If `errno` = 1, the array `u` is unchanged from the last entry
        to ``real_gen_basic_solver``.

        Otherwise, `u` holds the last available iterate of the solution
        of the system of equations, for all returned values of errno.

    v : float, shape (n), modified in place
        `On initial entry`: the right-hand side b of the system of
        equations Ax = b.

        `On intermediate entry`: the returned value of `irevcm`
        determines the contents of `v` in the following way:

            if `irevcm` = -1, 1 or 2, `v` must store the vector v, the
            result of the operation specified by the value of `irevcm`
            returned by the previous call to ``real_gen_basic_solver``;

            if `irevcm` = 3, `v` must remain unchanged.

        `On intermediate exit`: if `irevcm` = 3, `v` holds the current
        iterate of the residual vector. Note that this is an
        approximation to the true residual vector. Otherwise, it does
        not contain any useful information.

        `On final exit`: if `errno` = 3 or `errno` < 0, the array `v` is
        unchanged from the initial entry to ``real_gen_basic_solver``.

        If `errno` = 1, the array `v` is unchanged from the last entry
        to ``real_gen_basic_solver``.

        If the function exits successfully or `errno` = 2, the array `v`
        contains the true residual vector of the system of equations
        (see also [Raises/Warns]).

        Otherwise, `v` stores the last available iterate of the residual
        vector unless `errno` = 8 is returned on last entry, in which
        case `v` is set to 0.0.

    wgt : None or float, shape (n)
        The user-supplied weights, if these are to be used in the
        computation of the vector norms in the termination criterion
        (see [Notes for real_gen_basic_setup] and [Parameters for
        real_gen_basic_setup]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``real_gen_basic_setup``.

    Returns
    -------
    irevcm : int
        `On intermediate exit`: has the following meanings.

        `irevcm` = -1

            The calling program must compute the matrix-vector product v
            = A^Tu, where u and v are stored in `u` and `v`,
            respectively; RGMRES, CGS and BI-CGSTAB(l) methods return
            `irevcm` = -1 only if the matrix norm ||A||_1 or
            ||A||_infinity is estimated internally using Higham's
            method. This can only happen if iterm = 1 in
            ``real_gen_basic_setup``.

        `irevcm` = 1

            The calling program must compute the matrix-vector product v
            = Au, where u and v are stored in `u` and `v`, respectively.

        `irevcm` = 2

            The calling program must solve the preconditioning equation
            Mv = u, where u and v are stored in `u` and `v`,
            respectively.

        `irevcm` = 3

            Monitoring step: the solution and residual at the current
            iteration are returned in the arrays `u` and `v`,
            respectively. No action by the calling program is required.
            ``real_gen_basic_diag`` can be called at this step to return
            additional information.

        `On final exit`: `irevcm` = 4: ``real_gen_basic_solver`` has
        completed its tasks. The value of errno determines whether the
        iteration has been successfully completed, errors have been
        detected or the calling program has requested termination.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On intermediate re-entry, `irevcm` = *<value>*.

            Constraint: either `irevcm` must be unchanged from its
            previous exit value or `irevcm` = 5 or 6.

        (`errno` -1)
            On initial entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0.

        (`errno` 3)
            Either ``real_gen_basic_setup`` was not called before
            calling ``real_gen_basic_solver`` or it has returned an
            error.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            ``real_gen_basic_solver`` has already completed its tasks.
            You need to set a new problem.

        (`errno` 2)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 2)
            User-requested tidy termination. The required accuracy has
            not been achieved. However, a reasonable accuracy may have
            been achieved.

        (`errno` 4)
            User-requested tidy termination. The solution has not
            converged after *<value>* iterations.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            Algorithm breakdown at iteration no. *<value>*.

        (`errno` 8)
            User-requested immediate termination.

        (`errno` 10)
            The weights in array `wgt` are all zero.

    Notes
    -----
    ``real_gen_basic_solver`` solves the general (nonsymmetric) system
    of linear simultaneous equations Ax = b of order n, where n is large
    and the coefficient matrix A is sparse, using one of four available
    methods: RGMRES, the preconditioned restarted generalized minimum
    residual method (see Saad and Schultz (1986)); CGS, the
    preconditioned conjugate gradient squared method (see Sonneveld
    (1989)); BI-CGSTAB(l), the bi-conjugate gradient stabilized method
    of order l (see Van der Vorst (1989) and Sleijpen and Fokkema
    (1993)); or TFQMR, the transpose-free quasi-minimal residual method
    (see Freund and Nachtigal (1991) and Freund (1993)).

    For a general description of the methods employed you are referred
    to [Notes for real_gen_basic_setup].

    ``real_gen_basic_solver`` can solve the system after the first
    function in the suite, ``real_gen_basic_setup``, has been called to
    initialize the computation and specify the method of solution.
    The third function in the suite, ``real_gen_basic_diag``, can be
    used to return additional information generated by the computation,
    during monitoring steps and after ``real_gen_basic_solver`` has
    completed its tasks.

    ``real_gen_basic_solver`` uses **reverse communication**, i.e., it
    returns repeatedly to the calling program with the argument `irevcm`
    (see [Parameters]) set to specified values which require the calling
    program to carry out one of the following tasks:

    -   compute the matrix-vector product v = Au or v = A^Tu (the four
        methods require the matrix transpose-vector product only if
        ||A||_1 or ||A||_infinity is estimated internally by Higham's
        method (see Higham (1988)));

    -   solve the preconditioning equation Mv = u;

    -   notify the completion of the computation;

    -   allow the calling program to monitor the solution.

    Through the argument `irevcm` the calling program can cause
    immediate or tidy termination of the execution.
    On final exit, the last iterates of the solution and of the residual
    vectors of the original system of equations are returned.

    Reverse communication has the following advantages.

    (1) Maximum flexibility in the representation and storage of sparse
        matrices: all matrix operations are performed outside the solver
        function, thereby avoiding the need for a complicated interface
        with enough flexibility to cope with all types of storage
        schemes and sparsity patterns. This applies also to
        preconditioners.

    (#) Enhanced user interaction: you can closely monitor the progress
        of the solution and tidy or immediate termination can be
        requested. This is useful, for example, when alternative
        termination criteria are to be employed or in case of failure of
        the external functions used to perform matrix operations.

    References
    ----------
    Freund, R W, 1993, `A transpose-free quasi-minimal residual
    algorithm for non-Hermitian linear systems`, SIAM J. Sci. Comput.
    (14), 470--482

    Freund, R W and Nachtigal, N, 1991, `QMR: a Quasi-Minimal Residual
    Method for Non-Hermitian Linear Systems`, Numer. Math. (60),
    315--339

    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396

    Saad, Y and Schultz, M, 1986, `GMRES: a generalized minimal residual
    algorithm for solving nonsymmetric linear systems`, SIAM J. Sci.
    Statist. Comput. (7), 856--869

    Sleijpen, G L G and Fokkema, D R, 1993, `BiCGSTAB(l) for linear
    equations involving matrices with complex spectrum`, ETNA (1),
    11--32

    Sonneveld, P, 1989, `CGS, a fast Lanczos-type solver for
    nonsymmetric linear systems`, SIAM J. Sci. Statist. Comput. (10),
    36--52

    Van der Vorst, H, 1989, `Bi-CGSTAB, a fast and smoothly converging
    variant of Bi-CG for the solution of nonsymmetric linear systems`,
    SIAM J. Sci. Statist. Comput. (13), 631--644
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_basic_solver'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11beft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input/Output, irevcm.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, u[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, v[*].
        utils._EngineFloat64ArrayType.c_type, # Input, wgt[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _irevcm = utils._EngineIntScalarType(
        irevcm, {'fun_name': fun_name, 'entity': 'irevcm'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _vl_n_py = int(utils._EngineFloat64ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (8,), validate=False))
    _u_cls = utils._EngineFloat64ArrayType(
        u, {'fun_name': fun_name, 'entity': 'u'},
        exp_shape=(_vl_n_py,),
    )
    _u = _u_cls.to_ctypes()
    _v_cls = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_vl_n_py,),
    )
    _v = _v_cls.to_ctypes()
    if utils._EngineFloat64ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (6,), validate=False) > 2.0:
        if wgt is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'wgt'},
                'must not be None when weight = \'W\'.',
            )
    _wgt = utils._EngineFloat64ArrayType(
        wgt, {'fun_name': fun_name, 'entity': 'wgt'},
        exp_shape=(_vl_n_py,),
        check_contents=utils._EngineFloat64ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (6,), validate=False) > 2.0,
        allow_none=True,
    ).to_ctypes(can_reference_contents=utils._EngineFloat64ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (6,), validate=False) > 2.0)
    _lwork_py = utils._EngineFloat64ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _irevcm,
        _u,
        _v,
        _wgt,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _irevcm_py = utils._EngineIntScalarType.to_py(_irevcm)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _irevcm_py != 4 or _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On intermediate re-entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: either irevcm must be unchanged from its previous exit value or',
                    'irevcm = 5 or 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On initial entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irevcm = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'real_gen_basic_solver has already completed its tasks. You need to set a new problem.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested tidy termination.',
                    'The required accuracy has not been achieved.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Either real_gen_basic_setup was not called before calling real_gen_basic_solver',
                    'or it has returned an error.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested tidy termination.',
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99992:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithm breakdown at iteration no. {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99991:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested immediate termination.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99988:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The weights in array wgt are all zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _u_cls.refresh_py(_u)
        _v_cls.refresh_py(_v)
        return_data = _irevcm_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['irevcm'],
            )(
                _irevcm_py,
            )
    _err_class.manage()
    return return_data

def real_gen_basic_diag(comm):
    u"""
    Real sparse nonsymmetric linear systems, diagnostic for
    ``real_gen_basic_solver``.

    ``real_gen_basic_diag`` is the third in a suite of three functions
    for the iterative solution of a real general (nonsymmetric) system
    of simultaneous linear equations (see Golub and Van Loan (1996)).
    ``real_gen_basic_diag`` returns information about the computations
    during an iteration and/or after this has been completed.
    The first function of the suite, ``real_gen_basic_setup``, is a
    setup function; the second function, ``real_gen_basic_solver``, is
    the iterative solver itself.

    These three functions are suitable for the solution of large sparse
    general (nonsymmetric) systems of equations.

    For full information please refer to the NAG Library document for
    f11bf

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11bff.html

    Parameters
    ----------
    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``real_gen_basic_setup``.

    Returns
    -------
    itn : int
        The number of iterations carried out by
        ``real_gen_basic_solver``.

    stplhs : float
        The current value of the left-hand side of the termination
        criterion used by ``real_gen_basic_solver``.

    stprhs : float
        The current value of the right-hand side of the termination
        criterion used by ``real_gen_basic_solver``.

    anorm : float
        If iterm = 1 in the previous call to ``real_gen_basic_setup``,
        then `anorm` contains ||A||_p, where p = 1, 2 or infinity,
        either supplied or, in the case of 1 or infinity, estimated by
        ``real_gen_basic_solver``; otherwise `anorm` = 0.0.

    sigmax : float
        If iterm = 2 in the previous call to ``real_gen_basic_setup``,
        the current estimate of the largest singular value sigma_1(A_)
        of the preconditioned iteration matrix, either when it has been
        supplied to ``real_gen_basic_setup`` or it has been estimated by
        ``real_gen_basic_solver`` (see also [Notes for
        real_gen_basic_setup] and [Parameters for
        real_gen_basic_setup]); otherwise, `sigmax` = 0.0 is returned.

    Raises
    ------
    NagValueError
        (`errno` 1)
            ``real_gen_basic_diag`` has been called out of sequence.

    Notes
    -----
    ``real_gen_basic_diag`` returns information about the solution
    process.
    It can be called either during a monitoring step of
    ``real_gen_basic_solver`` or after ``real_gen_basic_solver`` has
    completed its tasks.
    Calling ``real_gen_basic_diag`` at any other time will result in an
    error condition being raised.

    For further information you should read the documentation for
    ``real_gen_basic_setup`` and ``real_gen_basic_solver``.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_basic_diag'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11bfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, itn.
        utils._BYREF_FLOAT64, # Output, stplhs.
        utils._BYREF_FLOAT64, # Output, stprhs.
        utils._BYREF_FLOAT64, # Output, anorm.
        utils._BYREF_FLOAT64, # Output, sigmax.
        utils._EngineFloat64ArrayType.c_type, # Input, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _stplhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _stprhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _anorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _sigmax = utils._EngineFloat64ScalarType.empty_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _lwork_py = utils._EngineFloat64ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _itn,
        _stplhs,
        _stprhs,
        _anorm,
        _sigmax,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'real_gen_basic_diag has been called out of sequence.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        _stplhs_py = utils._EngineFloat64ScalarType.to_py(_stplhs)
        _stprhs_py = utils._EngineFloat64ScalarType.to_py(_stprhs)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _sigmax_py = utils._EngineFloat64ScalarType.to_py(_sigmax)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['itn', 'stplhs', 'stprhs', 'anorm', 'sigmax'],
        )(
            _itn_py,
            _stplhs_py,
            _stprhs_py,
            _anorm_py,
            _sigmax_py,
        )
    _err_class.manage()
    return return_data

def complex_gen_basic_setup(method, precon, norm, weight, iterm, n, m, tol, maxitn, anorm, sigmax, monit):
    u"""
    Complex sparse non-Hermitian linear systems, setup for
    ``complex_gen_basic_solver``.

    ``complex_gen_basic_setup`` is a setup function, the first in a
    suite of three functions for the iterative solution of a complex
    general (non-Hermitian) system of simultaneous linear equations.
    ``complex_gen_basic_setup`` must be called before
    ``complex_gen_basic_solver``, the iterative solver.
    The third function in the suite, ``complex_gen_basic_diag``, can be
    used to return additional information about the computation.

    These three functions are suitable for the solution of large sparse
    general (non-Hermitian) systems of equations.

    For full information please refer to the NAG Library document for
    f11br

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11brf.html

    Parameters
    ----------
    method : str
        The iterative method to be used.

        `method` = 'RGMRES'

            Restarted generalized minimum residual method.

        `method` = 'CGS'

            Conjugate gradient squared method.

        `method` = 'BICGSTAB'

            Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

            Transpose-free quasi-minimal residual method.

    precon : str, length 1
        Determines whether preconditioning is used.

        `precon` = 'N'

            No preconditioning.

        `precon` = 'P'

            Preconditioning.

    norm : str, length 1
        Defines the matrix and vector norm to be used in the termination
        criteria.

        `norm` = '1'

            l_1 norm.

        `norm` = 'I'

            l_infinity norm.

        `norm` = '2'

            l_2 norm.

        Suggested value:

            if `iterm` = 1, `norm` = 'I'

            if `iterm` = 2, `norm` = '2'

    weight : str, length 1
        Specifies whether a vector w of user-supplied weights is to be
        used in the computation of the vector norms required in
        termination criterion (2) (`iterm` = 1): ||v||_p^(w) =
        ||v^(w)||_p, where v_i^(w) = w_i v_i, for i = 1,2,...,n. The
        suffix p = 1,2, infinity denotes the vector norm used, as
        specified by the argument `norm`. Note that weights cannot be
        used when `iterm` = 2, i.e., when criterion (3) is used.

        `weight` = 'W'

            User-supplied weights are to be used and must be supplied on
            initial entry to ``complex_gen_basic_solver``.

        `weight` = 'N'

            All weights are implicitly set equal to one. Weights do not
            need to be supplied on initial entry to
            ``complex_gen_basic_solver``.

        Suggested value: `weight` = 'N'.

    iterm : int
        Defines the termination criterion to be used.

        `iterm` = 1

            Use the termination criterion defined in (2).

        `iterm` = 2

            Use the termination criterion defined in (3).

        Suggested value: `iterm` = 1.

    n : int
        n, the order of the matrix A.

    m : int
        If `method` = 'RGMRES', `m` is the dimension m of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method.

        Otherwise, `m` is not referenced.

    tol : float
        The tolerance tau for the termination criterion. If `tol` <=
        0.0, tau = max(sqrt(epsilon), sqrt(n)*epsilon) is used, where
        epsilon is the machine precision. Otherwise tau = max(`tol`,
        10*epsilon, sqrt(n)*epsilon) is used.

    maxitn : int
        The maximum number of iterations.

    anorm : float
        If `anorm` > 0.0, the value of ||A||_p to be used in the
        termination criterion (2) (`iterm` = 1).

        If `anorm` <= 0.0, `iterm` = 1 and `norm` = '1' or 'I', then
        ||A||_1 = ||A||_infinity is estimated internally by
        ``complex_gen_basic_solver``.

        If `iterm` = 2, `anorm` is not referenced.

    sigmax : float
        If `iterm` = 2, the largest singular value sigma_1 of the
        preconditioned iteration matrix; otherwise, `sigmax` is not
        referenced.

        If `sigmax` <= 0.0, `iterm` = 2 and `method` = 'RGMRES', the
        value of sigma_1 will be estimated internally.

    monit : int
        If `monit` > 0, the frequency at which a monitoring step is
        executed by ``complex_gen_basic_solver``: if `method` = 'CGS' or
        'TFQMR', a monitoring step is executed every `monit` iterations;
        otherwise, a monitoring step is executed every `monit`
        super-iterations (groups of up to m or l iterations for RGMRES
        or BI-CGSTAB(l), respectively).

        There are some additional computational costs involved in
        monitoring the solution and residual vectors when the
        BI-CGSTAB(l) method is used with l > 1.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` -12)
            On entry, `monit` = *<value>* and `maxitn` = *<value>*.

            Constraint: `monit` <= `maxitn`.

        (`errno` -11)
            On entry, `iterm` = 2, `method` = *<value>* and `sigmax` =
            *<value>*.

            Constraint: if `iterm` = 2 and `method` = 'CGS' or
            'BICGSTAB', `sigmax` > 0.0.

        (`errno` -10)
            On entry, `iterm` = 1, `norm` = '2' and `anorm` = *<value>*.

            Constraint: if `iterm` = 1 and `norm` = '2', `anorm` > 0.0.

        (`errno` -9)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` > 0.

        (`errno` -8)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` -7)
            On entry, `m` = *<value>*, `n` = *<value>* and `method` =
            *<value>*.

            Constraint: if `method` = 'RGMRES', `m` <= min(`n`, 50). If
            `method` = 'BICGSTAB', `m` <= min(`n`, 10).

        (`errno` -7)
            On entry, `m` = *<value>* and `method` = *<value>*.

            Constraint: if `method` = 'RGMRES' or 'BICGSTAB', `m` > 0.

        (`errno` -6)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` -5)
            On entry, `iterm` = 2 and `method` = 'TFQMR'.

            Constraint: if `iterm` = 2, `method` != 'TFQMR'.

        (`errno` -5)
            On entry, `iterm` = 2 and `norm` = *<value>*.

            Constraint: if `iterm` = 2, `norm` = '2'.

        (`errno` -5)
            On entry, `iterm` = 2 and `weight` = *<value>*.

            Constraint: if `iterm` = 2, `weight` = 'N'.

        (`errno` -5)
            On entry, `iterm` = *<value>*.

            Constraint: `iterm` = 1 or 2.

        (`errno` -4)
            On entry, `weight` = *<value>*.

            Constraint: `weight` = 'N' or 'W'.

        (`errno` -3)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = '1', 'I' or '2'.

        (`errno` -2)
            On entry, `precon` = *<value>*.

            Constraint: `precon` = 'N' or 'P'.

        (`errno` -1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS', 'BICGSTAB' or
            'TFQMR'.

        (`errno` 1)
            ``complex_gen_basic_setup`` has been called out of sequence:
            either ``complex_gen_basic_setup`` has been called twice or
            ``complex_gen_basic_solver`` has not terminated its current
            task.

    Notes
    -----
    The suite consisting of the functions ``complex_gen_basic_setup``,
    ``complex_gen_basic_solver`` and ``complex_gen_basic_diag`` is
    designed to solve the general (non-Hermitian) system of simultaneous
    linear equations Ax = b of order n, where n is large and the
    coefficient matrix A is sparse.

    ``complex_gen_basic_setup`` is a setup function which must be called
    before ``complex_gen_basic_solver``, the iterative solver.
    The third function in the suite, ``complex_gen_basic_diag``, can be
    used to return additional information about the computation.
    A choice of methods is available:

        restarted generalized minimum residual method (RGMRES);

        conjugate gradient squared method (CGS);

        bi-conjugate gradient stabilized (l) method (BI-CGSTAB(l));

        transpose-free quasi-minimal residual method (TFQMR).

    Restarted Generalized Minimum Residual Method (RGMRES)
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    The restarted generalized minimum residual method (RGMRES) (see Saad
    and Schultz (1986), Barrett `et al.` (1994) and Dias da Cunha and
    Hopkins (1994)) starts from the residual r_0 = b-Ax_0, where x_0 is
    an initial estimate for the solution (often x_0 = 0).
    An orthogonal basis for the Krylov subspace span{A^kr_0}, for k =
    0,1,...,, is generated explicitly: this is referred to as Arnoldi's
    method (see Arnoldi (1951)).
    The solution is then expanded onto the orthogonal basis so as to
    minimize the residual norm ||b-Ax||_2.
    The lack of symmetry of A implies that the orthogonal basis is
    generated by applying a 'long' recurrence relation, whose length
    increases linearly with the iteration count.
    For all but the most trivial problems, computational and storage
    costs can quickly become prohibitive as the iteration count
    increases. RGMRES limits these costs by employing a restart
    strategy: every m iterations at most, the Arnoldi process is
    restarted from r_l = b-Ax_l, where the subscript l denotes the last
    available iterate.
    Each group of m iterations is referred to as a 'super-iteration'.
    The value of m is chosen in advance and is fixed throughout the
    computation.
    Unfortunately, an optimum value of m cannot easily be predicted.

    Conjugate Gradient Squared Method (CGS)
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    The conjugate gradient squared method (CGS) (see Sonneveld (1989),
    Barrett `et al.` (1994) and Dias da Cunha and Hopkins (1994)) is a
    development of the bi-conjugate gradient method where the
    nonsymmetric Lanczos method is applied to reduce the coefficients
    matrix to tridiagonal form: two bi-orthogonal sequences of vectors
    are generated starting from the residual r_0 = b-Ax_0, where x_0 is
    an initial estimate for the solution (often x_0 = 0) and from the
    `shadow residual` r^_0 corresponding to the arbitrary problem A^Hx^=
    b^, where b^ can be any vector, but in practice is chosen so that
    r_0 = r^_0.
    In the course of the iteration, the residual and shadow residual r_i
    = P_i(A)r_0 and r^_i = P_i(A^H)r^_0 are generated, where P_i is a
    polynomial of order i, and bi-orthogonality is exploited by
    computing the vector product rho_i = (r^_i,r_i) = (P_i(A^H)r^_0,
    P_i(A)r_0) = (r^_0,P_i^2(A)r_0).
    Applying the 'contraction' operator P_i(A) twice, the iteration
    coefficients can still be recovered without advancing the solution
    of the shadow problem, which is of no interest.
    The CGS method often provides fast convergence; however, there is no
    reason why the contraction operator should also reduce the once
    reduced vector P_i(A)r_0: this may well lead to a highly irregular
    convergence which may result in large cancellation errors.

    Bi-Conjugate Gradient Stabilized (l) Method (BI-CGSTAB(l))
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    The bi-conjugate gradient stabilized (l) method (BI-CGSTAB(l)) (see
    Van der Vorst (1989), Sleijpen and Fokkema (1993) and Dias da Cunha
    and Hopkins (1994)) is similar to the CGS method above.
    However, instead of generating the sequence {P_i^2(A)r_0}, it
    generates the sequence {Q_i(A)P_i(A)r_0}, where the Q_i(A) are
    polynomials chosen to minimize the residual `after` the application
    of the contraction operator P_i(A).
    Two main steps can be identified for each iteration: an OR
    (Orthogonal Residuals) step where a basis of order l is generated by
    a Bi-CG iteration and an MR (Minimum Residuals) step where the
    residual is minimized over the basis generated, by a method akin to
    GMRES.
    For l = 1, the method corresponds to the BI-CGSTAB method of Van der
    Vorst (1989).
    For l > 1, more information about complex eigenvalues of the
    iteration matrix can be taken into account, and this may lead to
    improved convergence and robustness.
    However, as l increases, numerical instabilities may arise.
    For this reason, a maximum value of l = 10 is imposed, but probably
    l = 4 is sufficient in most cases.

    Transpose-free Quasi-minimal Residual Method (TFQMR)
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    The transpose-free quasi-minimal residual method (TFQMR) (see Freund
    and Nachtigal (1991) and Freund (1993)) is conceptually derived from
    the CGS method.
    The residual is minimized over the space of the residual vectors
    generated by the CGS iterations under the simplifying assumption
    that residuals are almost orthogonal.
    In practice, this is not the case but theoretical analysis has
    proved the validity of the method.
    This has the effect of remedying the rather irregular convergence
    behaviour with wild oscillations in the residual norm that can
    degrade the numerical performance and robustness of the CGS method.
    In general, the TFQMR method can be expected to converge at least as
    fast as the CGS method, in terms of number of iterations, although
    each iteration involves a higher operation count.
    When the CGS method exhibits irregular convergence, the TFQMR method
    can produce much smoother, almost monotonic convergence curves.
    However, the close relationship between the CGS and TFQMR method
    implies that the `overall` speed of convergence is similar for both
    methods.
    In some cases, the TFQMR method may converge faster than the CGS
    method.

    General Considerations
    ~~~~~~~~~~~~~~~~~~~~~~
    For each method, a sequence of solution iterates {x_i} is generated
    such that, hopefully, the sequence of the residual norms {||r_i||}
    converges to a required tolerance.
    Note that, in general, convergence, when it occurs, is not
    monotonic.

    In the RGMRES and BI-CGSTAB(l) methods above, your program must
    provide the **maximum** number of basis vectors used, m or l,
    respectively; however, a **smaller** number of basis vectors may be
    generated and used when the stability of the solution process
    requires this (see [Further Comments]).

    Faster convergence can be achieved using a **preconditioner** (see
    Golub and Van Loan (1996) and Barrett `et al.` (1994)).
    A preconditioner maps the original system of equations onto a
    different system, say

        A_x_= b_,

    with, hopefully, better characteristics with respect to its speed of
    convergence: for example, the condition number of the coefficients
    matrix can be improved or eigenvalues in its spectrum can be made to
    coalesce.
    An orthogonal basis for the Krylov subspace span{A_^kr__0}, for k =
    0,1,...,, is generated and the solution proceeds as outlined above.
    The algorithms used are such that the solution and residual iterates
    of the original system are produced, not their preconditioned
    counterparts.
    Note that an unsuitable preconditioner or no preconditioning at all
    may result in a very slow rate, or lack, of convergence.
    However, preconditioning involves a trade-off between the reduction
    in the number of iterations required for convergence and the
    additional computational costs per iteration.
    Also, setting up a preconditioner may involve non-negligible
    overheads.

    A `left` preconditioner M^-1 can be used by the RGMRES, CGS and
    TFQMR methods, such that A_= M^-1A~I_n in (1), where I_n is the
    identity matrix of order n; a `right` preconditioner M^-1 can be
    used by the BI-CGSTAB(l) method, such that A_= AM^-1~I_n.
    These are formal definitions, used only in the design of the
    algorithms; in practice, only the means to compute the
    matrix--vector products v = Au and v = A^Hu (the latter only being
    required when an estimate of ||A||_1 or ||A||_infinity is computed
    internally), and to solve the preconditioning equations Mv = u are
    required, i.e., explicit information about M, or its inverse is not
    required at any stage.

    The first termination criterion

        ||r_k||_p <= tau(||b||_p+||A||_p*||x_k||_p)

    is available for all four methods.
    In (2), p = 1, infinity or 2 and tau denotes a user-specified
    tolerance subject to max(10, sqrt(n)), epsilon <= tau < 1, where
    epsilon is the machine precision.
    Facilities are provided for the estimation of the norm of the
    coefficients matrix ||A||_1 or ||A||_infinity, when this is not
    known in advance, by applying Higham's method (see Higham (1988)).
    Note that ||A||_2 cannot be estimated internally.
    This criterion uses an error bound derived from **backward** error
    analysis to ensure that the computed solution is the exact solution
    of a problem as close to the original as the termination tolerance
    requires.
    Termination criteria employing bounds derived from **forward** error
    analysis are not used because any such criteria would require
    information about the condition number kappa(A) which is not easily
    obtainable.

    The second termination criterion

        ||r__k||_2 <= tau(||r__0||_2+sigma_1(A_)*||Delta x__k||_2)

    is available for all methods except TFQMR.
    In (3), sigma_1(A_) = ||A_||_2 is the largest singular value of the
    (preconditioned) iteration matrix A_.
    This termination criterion monitors the progress of the solution of
    the preconditioned system of equations and is less expensive to
    apply than criterion (2) for the BI-CGSTAB(l) method with l > 1.
    Only the RGMRES method provides facilities to estimate sigma_1(A_)
    internally, when this is not supplied (see [Further Comments]).

    Termination criterion (2) is the recommended choice, despite its
    additional costs per iteration when using the BI-CGSTAB(l) method
    with l > 1.
    Also, if the norm of the initial estimate is much larger than the
    norm of the solution, that is, if ||x_0||≫||x||, a dramatic loss of
    significant digits could result in complete lack of convergence.
    The use of criterion (2) will enable the detection of such a
    situation, and the iteration will be restarted at a suitable point.
    No such restart facilities are provided for criterion (3).

    Optionally, a vector w of user-specified weights can be used in the
    computation of the vector norms in termination criterion (2), i.e.,
    ||v||_p^(w) = ||v^(w)||_p, where (v^(w))_i = w_i v_i, for i =
    1,2,...,n.
    Note that the use of weights increases the computational costs.

    The sequence of calls to the functions comprising the suite is
    enforced: first, the setup function ``complex_gen_basic_setup`` must
    be called, followed by the solver ``complex_gen_basic_solver``.
    ``complex_gen_basic_diag`` can be called either when
    ``complex_gen_basic_solver`` is carrying out a monitoring step or
    after ``complex_gen_basic_solver`` has completed its tasks.
    Incorrect sequencing will raise an error condition.

    In general, it is not possible to recommend one method in preference
    to another. RGMRES is often used in the solution of systems arising
    from PDEs.
    On the other hand, it can easily stagnate when the size m of the
    orthogonal basis is too small, or the preconditioner is not good
    enough. CGS can be the fastest method, but the computed residuals
    can exhibit instability which may greatly affect the convergence and
    quality of the solution. BI-CGSTAB(l) seems robust and reliable, but
    it can be slower than the other methods: if a preconditioner is used
    and l > 1, BI-CGSTAB(l) computes the solution of the preconditioned
    system x__k = Mx_k: the preconditioning equations must be solved to
    obtain the required solution.
    The algorithm employed limits to 10% or less, when no intermediate
    monitoring is requested, the number of times the preconditioner has
    to be thus applied compared with the total number of applications of
    the preconditioner. TFQMR can be viewed as a more robust variant of
    the CGS method: it shares the CGS method speed but avoids the CGS
    fluctuations in the residual, which may give rise to instability.
    Also, when the termination criterion (2) is used, the CGS,
    BI-CGSTAB(l) and TFQMR methods will restart the iteration
    automatically when necessary in order to solve the given problem.

    References
    ----------
    Arnoldi, W, 1951, `The principle of minimized iterations in the
    solution of the matrix eigenvalue problem`, Quart. Appl. Math. (9),
    17--29

    Barrett, R, Berry, M, Chan, T F, Demmel, J, Donato, J, Dongarra, J,
    Eijkhout, V, Pozo, R, Romine, C and Van der Vorst, H, 1994,
    `Templates for the Solution of Linear Systems: Building Blocks for
    Iterative Methods`, SIAM, Philadelphia

    Dias da Cunha, R and Hopkins, T, 1994, `PIM 1.1 --- the parallel
    iterative method package for systems of linear equations user's
    guide --- Fortran 77 version`, Technical Report, Computing
    Laboratory, University of Kent at Canterbury, Kent, UK

    Freund, R W, 1993, `A transpose-free quasi-minimal residual
    algorithm for non-Hermitian linear systems`, SIAM J. Sci. Comput.
    (14), 470--482

    Freund, R W and Nachtigal, N, 1991, `QMR: a Quasi-Minimal Residual
    Method for Non-Hermitian Linear Systems`, Numer. Math. (60),
    315--339

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396

    Saad, Y and Schultz, M, 1986, `GMRES: a generalized minimal residual
    algorithm for solving nonsymmetric linear systems`, SIAM J. Sci.
    Statist. Comput. (7), 856--869

    Sleijpen, G L G and Fokkema, D R, 1993, `BiCGSTAB(l) for linear
    equations involving matrices with complex spectrum`, ETNA (1),
    11--32

    Sonneveld, P, 1989, `CGS, a fast Lanczos-type solver for
    nonsymmetric linear systems`, SIAM J. Sci. Statist. Comput. (10),
    36--52

    Van der Vorst, H, 1989, `Bi-CGSTAB, a fast and smoothly converging
    variant of Bi-CG for the solution of nonsymmetric linear systems`,
    SIAM J. Sci. Statist. Comput. (13), 631--644
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_basic_setup'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11brft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, iterm.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._BYREF_FLOAT64, # Input, anorm.
        utils._BYREF_FLOAT64, # Input, sigmax.
        utils._BYREF_INT, # Input, monit.
        utils._BYREF_INT, # Output, lwreq.
        utils._EngineComplex128ArrayType.c_type, # Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _weight = utils._EngineCharScalarType(
        weight, {'fun_name': fun_name, 'entity': 'weight'}, exp_chrlen=1,
    ).to_ctypes()
    _iterm = utils._EngineIntScalarType(
        iterm, {'fun_name': fun_name, 'entity': 'iterm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _anorm = utils._EngineFloat64ScalarType(
        anorm, {'fun_name': fun_name, 'entity': 'anorm'},
    ).to_ctypes()
    _sigmax = utils._EngineFloat64ScalarType(
        sigmax, {'fun_name': fun_name, 'entity': 'sigmax'},
    ).to_ctypes()
    _monit = utils._EngineIntScalarType(
        monit, {'fun_name': fun_name, 'entity': 'monit'},
    ).to_ctypes()
    _lwreq = utils._EngineIntScalarType.empty_ctypes()
    _lwork = utils._EngineIntScalarType.c_type(120)
    _lwork_py = utils._EngineIntScalarType.to_py(_lwork)
    _work = utils._EngineComplex128ArrayType.empty_ctypes(_lwork_py)
    _comm_py = {}
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _norm,
        _weight,
        _iterm,
        _n,
        _m,
        _tol,
        _maxitn,
        _anorm,
        _sigmax,
        _monit,
        _lwreq,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _norm._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    if _ierr.value == utils._IERR_SUCCESS:
        _lwreq_py = utils._EngineIntScalarType.to_py(_lwreq)
        _lwork = utils._EngineIntScalarType(
            _lwreq_py, {'fun_name': fun_name, 'entity': '_lwork'},
        ).to_ctypes()
        _lwork_py = utils._EngineIntScalarType.to_py(_lwork)
        _work = utils._EngineComplex128ArrayType.empty_ctypes(_lwork_py)
        _comm_py['work'] = _work
        engine_ctypes(
            _en_data,
            _method,
            _precon,
            _norm,
            _weight,
            _iterm,
            _n,
            _m,
            _tol,
            _maxitn,
            _anorm,
            _sigmax,
            _monit,
            _lwreq,
            _work,
            _lwork,
            _errbuf,
            _ierr,
            _method._length_,
            _precon._length_,
            _norm._length_,
            _weight._length_,
            _errbuf._length_,
        )
        utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _monit_py = utils._EngineIntScalarType.to_py(_monit)
        _maxitn_py = utils._EngineIntScalarType.to_py(_maxitn)
        _method_py = utils._EngineCharScalarType.to_py(_method)
        _sigmax_py = utils._EngineFloat64ScalarType.to_py(_sigmax)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _tol_py = utils._EngineFloat64ScalarType.to_py(_tol)
        _m_py = utils._EngineIntScalarType.to_py(_m)
        _n_py = utils._EngineIntScalarType.to_py(_n)
        _norm_py = utils._EngineCharScalarType.to_py(_norm)
        _weight_py = utils._EngineCharScalarType.to_py(_weight)
        _iterm_py = utils._EngineIntScalarType.to_py(_iterm)
        _precon_py = utils._EngineCharScalarType.to_py(_precon)
        if _ierr_py == -12 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, monit = {:s} and maxitn = {:s}.'.format(
                        utils._STR_FMT_INT.format(_monit_py),
                        utils._STR_FMT_INT.format(_maxitn_py),
                    ),
                    'Constraint: monit <= maxitn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -11 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2, method = \'{:s}\' and'.format(
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'sigmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_sigmax_py),
                    ),
                    'Constraint: if iterm = 2 and method = \'CGS\' or \'BICGSTAB\', sigmax > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -10 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 1, norm = \'2\' and anorm = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_anorm_py),
                    ),
                    'Constraint: if iterm = 1 and norm = \'2\', anorm > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -9 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT.format(_maxitn_py),
                    ),
                    'Constraint: maxitn > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -8 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_tol_py),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -7 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}, n = {:s} and'.format(
                        utils._STR_FMT_INT.format(_m_py),
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'Constraint: if method = \'RGMRES\', m <= min(n, 50).',
                    'If method = \'BICGSTAB\', m <= min(n, 10).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -7 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and method = \'{:s}\'.'.format(
                        utils._STR_FMT_INT.format(_m_py),
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'Constraint: if method = \'RGMRES\' or \'BICGSTAB\', m > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and method = \'TFQMR\'.',
                    'Constraint: if iterm = 2, method != \'TFQMR\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 99919:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: if iterm = 2, norm = \'2\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 99929:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_weight_py),
                    ),
                    'Constraint: if iterm = 2, weight = \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = {:s}.'.format(
                        utils._STR_FMT_INT.format(_iterm_py),
                    ),
                    'Constraint: iterm = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -4 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_weight_py),
                    ),
                    'Constraint: weight = \'N\' or \'W\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -3 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: norm = \'1\', \'I\' or \'2\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_precon_py),
                    ),
                    'Constraint: precon = \'N\' or \'P\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\', \'BICGSTAB\' or \'TFQMR\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'complex_gen_basic_setup has been called out of sequence:',
                    'either complex_gen_basic_setup has been called twice or',
                    'complex_gen_basic_solver has not terminated its current task.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = _comm_py
    _err_class.manage()
    return return_data

def complex_gen_basic_solver(irevcm, u, v, wgt, comm):
    u"""
    Complex sparse non-Hermitian linear systems, preconditioned RGMRES,
    CGS, Bi-CGSTAB or TFQMR method.

    ``complex_gen_basic_solver`` is an iterative solver for a complex
    general (non-Hermitian) system of simultaneous linear equations;
    ``complex_gen_basic_solver`` is the second in a suite of three
    functions, where the first function, ``complex_gen_basic_setup``,
    must be called prior to ``complex_gen_basic_solver`` to set up the
    suite, and the third function in the suite,
    ``complex_gen_basic_diag``, can be used to return additional
    information about the computation.

    These three functions are suitable for the solution of large sparse
    general (non-Hermitian) systems of equations.

    For full information please refer to the NAG Library document for
    f11bs

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11bsf.html

    Parameters
    ----------
    irevcm : int
        `On initial entry`: `irevcm` = 0, otherwise an error condition
        will be raised.

        `On intermediate entry`: must either be unchanged from its
        previous exit value, or can have one of the following values.

        `irevcm` = 5

            Tidy termination: the computation will terminate at the end
            of the current iteration. Further reverse communication
            exits may occur depending on when the termination request is
            issued. ``complex_gen_basic_solver`` will then return with
            the termination code `irevcm` = 4. Note that before calling
            ``complex_gen_basic_solver`` with `irevcm` = 5 the calling
            program must have performed the tasks required by the value
            of `irevcm` returned by the previous call to
            ``complex_gen_basic_solver``, otherwise subsequently
            returned values may be invalid.

        `irevcm` = 6

            Immediate termination: ``complex_gen_basic_solver`` will
            return immediately with termination code `irevcm` = 4 and
            with any useful information available. This includes the
            last iterate of the solution.

            Immediate termination may be useful, for example, when
            errors are detected during matrix-vector multiplication or
            during the solution of the preconditioning equation.

        Changing `irevcm` to any other value between calls will result
        in an error.

    u : complex, shape (n), modified in place
        `On initial entry`: an initial estimate, x_0, of the solution of
        the system of equations Ax = b.

        `On intermediate entry`: must remain unchanged.

        `On intermediate exit`: the returned value of `irevcm`
        determines the contents of `u` as follows.

        If `irevcm` = -1, 1 or 2, `u` holds the vector u on which the
        operation specified by `irevcm` is to be carried out.

        If `irevcm` = 3, `u` holds the current iterate of the solution
        vector.

        `On final exit`: if `errno` = 3 or `errno` < 0, the array `u` is
        unchanged from the initial entry to
        ``complex_gen_basic_solver``.

        If `errno` = 1, the array `u` is unchanged from the last entry
        to ``complex_gen_basic_solver``.

        Otherwise, `u` holds the last available iterate of the solution
        of the system of equations, for all returned values of errno.

    v : complex, shape (n), modified in place
        `On initial entry`: the right-hand side b of the system of
        equations Ax = b.

        `On intermediate entry`: the returned value of `irevcm`
        determines the contents of `v` as follows.

        If `irevcm` = -1, 1 or 2, `v` must store the vector v, the
        result of the operation specified by the value of `irevcm`
        returned by the previous call to ``complex_gen_basic_solver``.

        If `irevcm` = 3, `v` must remain unchanged.

        `On intermediate exit`: if `irevcm` = 3, `v` holds the current
        iterate of the residual vector. Note that this is an
        approximation to the true residual vector. Otherwise, it does
        not contain any useful information.

        `On final exit`: if `errno` = 3 or `errno` < 0, the array `v` is
        unchanged from the initial entry to
        ``complex_gen_basic_solver``.

        If `errno` = 1, the array `v` is unchanged from the last entry
        to ``complex_gen_basic_solver``.

        If the function exits successfully or `errno` = 2, the array `v`
        contains the true residual vector of the system of equations
        (see also [Raises/Warns]).

        Otherwise, `v` stores the last available iterate of the residual
        vector unless `errno` = 8 is returned on last entry, in which
        case `v` is set to 0.0.

    wgt : None or float, shape (n)
        The user-supplied weights, if these are to be used in the
        computation of the vector norms in the termination criterion
        (see [Notes for complex_gen_basic_setup] and [Parameters for
        complex_gen_basic_setup]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``complex_gen_basic_setup``.

    Returns
    -------
    irevcm : int
        `On intermediate exit`: has the following meanings.

        `irevcm` = -1

            The calling program must compute the matrix-vector product v
            = A^Hu, where u and v are stored in `u` and `v`,
            respectively; RGMRES, CGS and BI-CGSTAB(l) methods return
            `irevcm` = -1 only if the matrix norm ||A||_1 or
            ||A||_infinity is estimated internally using Higham's
            method. This can only happen if iterm = 1 in
            ``complex_gen_basic_setup``.

        `irevcm` = 1

            The calling program must compute the matrix-vector product v
            = Au, where u and v are stored in `u` and `v`, respectively.

        `irevcm` = 2

            The calling program must solve the preconditioning equation
            Mv = u, where u and v are stored in `u` and `v`,
            respectively.

        `irevcm` = 3

            Monitoring step: the solution and residual at the current
            iteration are returned in the arrays `u` and `v`,
            respectively. No action by the calling program is required.
            ``complex_gen_basic_diag`` can be called at this step to
            return additional information.

        `On final exit`: `irevcm` = 4: ``complex_gen_basic_solver`` has
        completed its tasks. The value of errno determines whether the
        iteration has been successfully completed, errors have been
        detected or the calling program has requested termination.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On intermediate re-entry, `irevcm` = *<value>*.

            Constraint: either `irevcm` must be unchanged from its
            previous exit value or `irevcm` = 5 or 6.

        (`errno` -1)
            On initial entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0.

        (`errno` 3)
            Either ``complex_gen_basic_setup`` was not called before
            calling ``complex_gen_basic_solver`` or it has returned an
            error.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            ``complex_gen_basic_solver`` has already completed its
            tasks. You need to set a new problem.

        (`errno` 2)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 2)
            User-requested termination: the required accuracy could not
            be obtained. However, a reasonable accuracy may have been
            achieved.

        (`errno` 4)
            User-requested tidy termination. The solution has not
            converged after *<value>* iterations.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            Algorithm breakdown at iteration no. *<value>*.

        (`errno` 8)
            User-requested immediate termination.

        (`errno` 10)
            The weights in array `wgt` are all zero.

    Notes
    -----
    ``complex_gen_basic_solver`` solves the general (non-Hermitian)
    system of linear simultaneous equations Ax = b of order n, where n
    is large and the coefficient matrix A is sparse, using one of four
    available methods: RGMRES, the preconditioned restarted generalized
    minimum residual method (see Saad and Schultz (1986)); CGS, the
    preconditioned conjugate gradient squared method (see Sonneveld
    (1989)); BI-CGSTAB(l), the bi-conjugate gradient stabilized method
    of order l (see Van der Vorst (1989) and Sleijpen and Fokkema
    (1993)); or TFQMR, the transpose-free quasi-minimal residual method
    (see Freund and Nachtigal (1991) and Freund (1993)).

    For a general description of the methods employed you are referred
    to [Notes for complex_gen_basic_setup].

    ``complex_gen_basic_solver`` can solve the system after the first
    function in the suite, ``complex_gen_basic_setup``, has been called
    to initialize the computation and specify the method of solution.
    The third function in the suite, ``complex_gen_basic_diag``, can be
    used to return additional information generated by the computation
    during monitoring steps and after ``complex_gen_basic_solver`` has
    completed its tasks.

    ``complex_gen_basic_solver`` uses **reverse communication**, i.e.,
    it returns repeatedly to the calling program with the argument
    `irevcm` (see [Parameters]) set to specified values which require
    the calling program to carry out one of the following tasks:

        compute the matrix-vector product v = Au or v = A^Hu (the four
        methods require the matrix transpose-vector product only if
        ||A||_1 or ||A||_infinity is estimated internally by Higham's
        method (see Higham (1988)));

        solve the preconditioning equation Mv = u;

        notify the completion of the computation;

        allow the calling program to monitor the solution.

    Through the argument `irevcm` the calling program can cause
    immediate or tidy termination of the execution.
    On final exit, the last iterates of the solution and of the residual
    vectors of the original system of equations are returned.

    Reverse communication has the following advantages.

    (1) Maximum flexibility in the representation and storage of sparse
        matrices: all matrix operations are performed outside the solver
        function, thereby avoiding the need for a complicated interface
        with enough flexibility to cope with all types of storage
        schemes and sparsity patterns. This applies also to
        preconditioners.

    (#) Enhanced user interaction: you can closely monitor the progress
        of the solution and tidy or immediate termination can be
        requested. This is useful, for example, when alternative
        termination criteria are to be employed or in case of failure of
        the external functions used to perform matrix operations.

    References
    ----------
    Freund, R W, 1993, `A transpose-free quasi-minimal residual
    algorithm for non-Hermitian linear systems`, SIAM J. Sci. Comput.
    (14), 470--482

    Freund, R W and Nachtigal, N, 1991, `QMR: a Quasi-Minimal Residual
    Method for Non-Hermitian Linear Systems`, Numer. Math. (60),
    315--339

    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396

    Saad, Y and Schultz, M, 1986, `GMRES: a generalized minimal residual
    algorithm for solving nonsymmetric linear systems`, SIAM J. Sci.
    Statist. Comput. (7), 856--869

    Sleijpen, G L G and Fokkema, D R, 1993, `BiCGSTAB(l) for linear
    equations involving matrices with complex spectrum`, ETNA (1),
    11--32

    Sonneveld, P, 1989, `CGS, a fast Lanczos-type solver for
    nonsymmetric linear systems`, SIAM J. Sci. Statist. Comput. (10),
    36--52

    Van der Vorst, H, 1989, `Bi-CGSTAB, a fast and smoothly converging
    variant of Bi-CG for the solution of nonsymmetric linear systems`,
    SIAM J. Sci. Statist. Comput. (13), 631--644
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_basic_solver'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11bsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input/Output, irevcm.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, u[*].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, v[*].
        utils._EngineFloat64ArrayType.c_type, # Input, wgt[*].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _irevcm = utils._EngineIntScalarType(
        irevcm, {'fun_name': fun_name, 'entity': 'irevcm'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _vl_n_py = int(utils._EngineComplex128ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (9,), validate=False).real)
    _u_cls = utils._EngineComplex128ArrayType(
        u, {'fun_name': fun_name, 'entity': 'u'},
        exp_shape=(_vl_n_py,),
    )
    _u = _u_cls.to_ctypes()
    _v_cls = utils._EngineComplex128ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_vl_n_py,),
    )
    _v = _v_cls.to_ctypes()
    if utils._EngineComplex128ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (7,), validate=False).real > 2.0:
        if wgt is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'wgt'},
                'must not be None when weight = \'W\'.',
            )
    _wgt = utils._EngineFloat64ArrayType(
        wgt, {'fun_name': fun_name, 'entity': 'wgt'},
        exp_shape=(_vl_n_py,),
        check_contents=utils._EngineComplex128ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (7,), validate=False).real > 2.0,
        allow_none=True,
    ).to_ctypes(can_reference_contents=utils._EngineComplex128ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (7,), validate=False).real > 2.0)
    _lwork_py = utils._EngineComplex128ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _irevcm,
        _u,
        _v,
        _wgt,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _irevcm_py = utils._EngineIntScalarType.to_py(_irevcm)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _irevcm_py != 4 or _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On intermediate re-entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: either irevcm must be unchanged from its previous exit value or',
                    'irevcm = 5 or 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On initial entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irevcm = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'complex_gen_basic_solver has already completed its tasks. You need to set a new problem.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested termination: the required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Either complex_gen_basic_setup was not called before calling complex_gen_basic_solver',
                    'or it has returned an error.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested tidy termination.',
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99992:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithm breakdown at iteration no. {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99991:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested immediate termination.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99988:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The weights in array wgt are all zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _u_cls.refresh_py(_u)
        _v_cls.refresh_py(_v)
        return_data = _irevcm_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['irevcm'],
            )(
                _irevcm_py,
            )
    _err_class.manage()
    return return_data

def complex_gen_basic_diag(comm):
    u"""
    Complex sparse non-Hermitian linear systems, diagnostic for
    ``complex_gen_basic_solver``.

    ``complex_gen_basic_diag`` is the third in a suite of three
    functions for the iterative solution of a complex general
    (non-Hermitian) system of simultaneous linear equations (see Golub
    and Van Loan (1996)). ``complex_gen_basic_diag`` returns information
    about the computations during an iteration and/or after this has
    been completed.
    The first function of the suite, ``complex_gen_basic_setup``, is a
    setup function; the second function, ``complex_gen_basic_solver``,
    is the iterative solver itself.

    These three functions are suitable for the solution of large sparse
    general (non-Hermitian) systems of equations.

    For full information please refer to the NAG Library document for
    f11bt

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11btf.html

    Parameters
    ----------
    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``complex_gen_basic_setup``.

    Returns
    -------
    itn : int
        The number of iterations carried out by
        ``complex_gen_basic_solver``.

    stplhs : float
        The current value of the left-hand side of the termination
        criterion used by ``complex_gen_basic_solver``.

    stprhs : float
        The current value of the right-hand side of the termination
        criterion used by ``complex_gen_basic_solver``.

    anorm : float
        If iterm = 1 in the previous call to
        ``complex_gen_basic_setup``, then `anorm` contains ||A||_p,
        where p = 1, 2 or infinity, either supplied or, in the case of 1
        or infinity, estimated by ``complex_gen_basic_solver``;
        otherwise `anorm` = 0.0.

    sigmax : float
        If iterm = 2 in the previous call to
        ``complex_gen_basic_setup``, the current estimate of the largest
        singular value sigma_1(A_) of the preconditioned iteration
        matrix when it is used by the termination criterion in
        ``complex_gen_basic_solver``, either when it has been supplied
        to ``complex_gen_basic_setup`` or it has been estimated by
        ``complex_gen_basic_solver`` (see also [Notes for
        complex_gen_basic_setup] and [Parameters for
        complex_gen_basic_setup]); otherwise, `sigmax` = 0.0 is
        returned.

    Raises
    ------
    NagValueError
        (`errno` 1)
            ``complex_gen_basic_diag`` has been called out of sequence.

    Notes
    -----
    ``complex_gen_basic_diag`` returns information about the solution
    process.
    It can be called either during a monitoring step of
    ``complex_gen_basic_solver`` or after ``complex_gen_basic_solver``
    has completed its tasks.
    Calling ``complex_gen_basic_diag`` at any other time will result in
    an error condition being raised.

    For further information you should read the documentation for
    ``complex_gen_basic_setup`` and ``complex_gen_basic_solver``.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_basic_diag'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11btft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, itn.
        utils._BYREF_FLOAT64, # Output, stplhs.
        utils._BYREF_FLOAT64, # Output, stprhs.
        utils._BYREF_FLOAT64, # Output, anorm.
        utils._BYREF_FLOAT64, # Output, sigmax.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _stplhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _stprhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _anorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _sigmax = utils._EngineFloat64ScalarType.empty_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _lwork_py = utils._EngineComplex128ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _itn,
        _stplhs,
        _stprhs,
        _anorm,
        _sigmax,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'complex_gen_basic_diag has been called out of sequence.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        _stplhs_py = utils._EngineFloat64ScalarType.to_py(_stplhs)
        _stprhs_py = utils._EngineFloat64ScalarType.to_py(_stprhs)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _sigmax_py = utils._EngineFloat64ScalarType.to_py(_sigmax)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['itn', 'stplhs', 'stprhs', 'anorm', 'sigmax'],
        )(
            _itn_py,
            _stplhs_py,
            _stprhs_py,
            _anorm_py,
            _sigmax_py,
        )
    _err_class.manage()
    return return_data

def real_gen_precon_ilu(n, nnz, a, la, irow, icol, lfill, dtol, pstrat, milu, ipivp, ipivq, istr, idiag):
    u"""
    Real sparse nonsymmetric linear systems, incomplete LU
    factorization.

    ``real_gen_precon_ilu`` computes an incomplete LU factorization of a
    real sparse nonsymmetric matrix, represented in coordinate storage
    format.
    This factorization may be used as a preconditioner in combination
    with ``real_gen_solve_ilu`` or ``real_gen_basic_solver``.

    For full information please refer to the NAG Library document for
    f11da

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11daf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : float, shape (`la`), modified in place
        `On entry`: the nonzero elements in the matrix A, ordered by
        increasing row index, and by increasing column index within each
        row. Multiple entries for the same row and column indices are
        not permitted. The function ``real_gen_sort`` may be used to
        order the elements in this way.

        `On exit`: the first `nnz` entries of `a` contain the nonzero
        elements of A and the next `nnzc` entries contain the elements
        of the matrix C. Matrix elements are ordered by increasing row
        index, and by increasing column index within each row.

    la : int
        These arrays must be of sufficient size to store both A (`nnz`
        elements) and C (`nnzc` elements).

    irow : int, shape (`la`), modified in place
        `On entry`: the row indices of the nonzero elements supplied in
        `a`.

        `On exit`: the row indices of the nonzero elements returned in
        `a`.

    icol : int, shape (`la`), modified in place
        `On entry`: the column indices of the nonzero elements supplied
        in `a`.

        `On exit`: the column indices of the nonzero elements returned
        in `a`.

    lfill : int
        If `lfill` >= 0 its value is the maximum level of fill allowed
        in the decomposition (see [Further Comments]). A negative value
        of `lfill` indicates that `dtol` will be used to control the
        fill instead.

    dtol : float
        If `lfill` < 0, `dtol` is used as a drop tolerance to control
        the fill-in (see [Further Comments]); otherwise `dtol` is not
        referenced.

    pstrat : str, length 1
        Specifies the pivoting strategy to be adopted.

        `pstrat` = 'N'

            No pivoting is carried out.

        `pstrat` = 'U'

            Pivoting is carried out according to the user-defined input
            values of `ipivp` and `ipivq`.

        `pstrat` = 'P'

            Partial pivoting by columns for stability is carried out.

        `pstrat` = 'C'

            Complete pivoting by rows for sparsity, and by columns for
            stability, is carried out.

        Suggested value: `pstrat` = 'C'.

    milu : str, length 1
        Indicates whether or not the factorization should be modified to
        preserve row-sums (see [Further Comments]).

        `milu` = 'M'

            The factorization is modified.

        `milu` = 'N'

            The factorization is not modified.

    ipivp : int, shape (`n`), modified in place
        `On entry`: if `pstrat` = 'U', `ipivp`[k-1] and `ipivq`[k-1]
        must specify the row and column indices of the element used as a
        pivot at elimination stage k. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        `On exit`: the pivot indices. If `ipivp`[k-1] = i and
        `ipivq`[k-1] = j then the element in row i and column j was used
        as the pivot at elimination stage k.

    ipivq : int, shape (`n`), modified in place
        `On entry`: if `pstrat` = 'U', `ipivp`[k-1] and `ipivq`[k-1]
        must specify the row and column indices of the element used as a
        pivot at elimination stage k. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        `On exit`: the pivot indices. If `ipivp`[k-1] = i and
        `ipivq`[k-1] = j then the element in row i and column j was used
        as the pivot at elimination stage k.

    istr : int, shape (`n`+1), modified in place
        `On exit`: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix C. `istr`[`n`]-1 is the address of the last nonzero
        element in C plus one.

    idiag : int, shape (`n`), modified in place
        `On exit`: `idiag`[i-1], for i = 1,2,...,`n`, holds the index of
        arrays `a`, `irow` and `icol` which holds the diagonal element
        in row i of the matrix C.

    Returns
    -------
    nnzc : int
        The number of nonzero elements in the matrix C.

    npivm : int
        If `npivm` > 0 it gives the number of pivots which were modified
        during the factorization to ensure that M exists.

        If `npivm` = -1 no pivot modifications were required, but a
        local restart occurred (see [Algorithmic Details]).

        The quality of the preconditioner will generally depend on the
        returned value of `npivm`.

        If `npivm` is large the preconditioner may not be satisfactory.

        In this case it may be advantageous to call
        ``real_gen_precon_ilu`` again with an increased value of
        `lfill`, a reduced value of `dtol`, or set `pstrat` = 'C'.

        See also [Further Comments].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `milu` = *<value>*.

            Constraint: `milu` = 'M' or 'N'.

        (`errno` 1)
            On entry, `pstrat` = *<value>*.

            Constraint: `pstrat` = 'N', 'U', 'P' or 'C'.

        (`errno` 1)
            On entry, `dtol` = *<value>*.

            Constraint: `dtol` >= 0.0.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 2)
            On entry, the location (`irow`[i-1],`icol`[i-1]) is a
            duplicate: i = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, a user-supplied value of `ipivq` is repeated.

        (`errno` 3)
            On entry, a user-supplied value of `ipivq` lies outside the
            range [1,`n`].

        (`errno` 3)
            On entry, a user-supplied value of `ipivp` is repeated.

        (`errno` 3)
            On entry, a user-supplied value of `ipivp` lies outside the
            range [1,`n`].

        (`errno` 4)
            The number of nonzero entries in the decomposition is too
            large. The decomposition has been terminated before
            completion. Either increase `la`, or reduce the fill by
            reducing `lfill`, or increasing `dtol`.

        (`errno` 5)
            A serious error has occurred in an internal call to an
            auxiliary function. Check all function calls and array
            sizes. Seek expert help.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``real_gen_precon_ilu`` computes an incomplete LU factorization (see
    Meijerink and Van der Vorst (1977) and Meijerink and Van der Vorst
    (1981)) of a real sparse nonsymmetric n*n matrix A.
    The factorization is intended primarily for use as a preconditioner
    with one of the iterative solvers ``real_gen_solve_ilu`` or
    ``real_gen_basic_solver``.

    The decomposition is written in the form

        A = M+R

    where

        M = PLDUQ

    and L is lower triangular with unit diagonal elements, D is
    diagonal, U is upper triangular with unit diagonals, P and Q are
    permutation matrices, and R is a remainder matrix.

    The amount of fill-in occurring in the factorization can vary from
    zero to complete fill, and can be controlled by specifying either
    the maximum level of fill `lfill`, or the drop tolerance `dtol`.

    The argument `pstrat` defines the pivoting strategy to be used.
    The options currently available are no pivoting, user-defined
    pivoting, partial pivoting by columns for stability, and complete
    pivoting by rows for sparsity and by columns for stability.
    The factorization may optionally be modified to preserve the
    row-sums of the original matrix.

    The sparse matrix A is represented in coordinate storage (CS) format
    (see [the F11 Introduction]).
    The array `a` stores all the nonzero elements of the matrix A, while
    arrays `irow` and `icol` store the corresponding row and column
    indices respectively.
    Multiple nonzero elements may not be specified for the same row and
    column index.

    The preconditioning matrix M is returned in terms of the CS
    representation of the matrix

        C = L+D^-1+U-2I.

    Further algorithmic details are given in [Algorithmic Details].

    References
    ----------
    Meijerink, J and Van der Vorst, H, 1977, `An iterative solution
    method for linear systems of which the coefficient matrix is a
    symmetric M-matrix`, Math. Comput. (31), 148--162

    Meijerink, J and Van der Vorst, H, 1981, `Guidelines for the usage
    of incomplete decompositions in solving sets of linear equations as
    they occur in practical problems`, J. Comput. Phys. (44), 134--155

    Salvini, S A and Shaw, G J, 1996, `An evaluation of new NAG Library
    solvers for large sparse unsymmetric linear systems`, NAG Technical
    Report TR2/96
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_precon_ilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11daft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input/Output, irow[la].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[la].
        utils._BYREF_INT, # Input, lfill.
        utils._BYREF_FLOAT64, # Input, dtol.
        utils._BYREF_STRING, # Input, pstrat(len=1).
        utils._BYREF_STRING, # Input, milu(len=1).
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[n].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[n].
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._EngineIntArrayType.c_type, # Output, idiag[n].
        utils._BYREF_INT, # Output, nnzc.
        utils._BYREF_INT, # Output, npivm.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._BYREF_INT, # Input, liwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pstrat.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of milu.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _lfill = utils._EngineIntScalarType(
        lfill, {'fun_name': fun_name, 'entity': 'lfill'},
    ).to_ctypes()
    _dtol = utils._EngineFloat64ScalarType(
        dtol, {'fun_name': fun_name, 'entity': 'dtol'},
    ).to_ctypes()
    _pstrat = utils._EngineCharScalarType(
        pstrat, {'fun_name': fun_name, 'entity': 'pstrat'}, exp_chrlen=1,
    ).to_ctypes()
    _milu = utils._EngineCharScalarType(
        milu, {'fun_name': fun_name, 'entity': 'milu'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _pstrat_py = utils._EngineCharScalarType.to_py(_pstrat)
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_n_py,),
        check_contents=_pstrat_py in ('U', 'u'),
    )
    _ipivp = _ipivp_cls.to_ctypes(can_reference_contents=_pstrat_py in ('U', 'u'))
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_n_py,),
        check_contents=_pstrat_py in ('U', 'u'),
    )
    _ipivq = _ipivq_cls.to_ctypes(can_reference_contents=_pstrat_py in ('U', 'u'))
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _idiag_cls = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _idiag = _idiag_cls.to_ctypes(can_reference_contents=False)
    _nnzc = utils._EngineIntScalarType.empty_ctypes()
    _npivm = utils._EngineIntScalarType.empty_ctypes()
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _liwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _lfill,
        _dtol,
        _pstrat,
        _milu,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _nnzc,
        _npivm,
        _iwork,
        _liwork,
        _errbuf,
        _ierr,
        _pstrat._length_,
        _milu._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, milu = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: milu = \'M\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pstrat = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: pstrat = \'N\', \'U\', \'P\' or \'C\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dtol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dtol >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[i-1],icol[i-1]) is a duplicate:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivq is repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivq lies outside the range [1,n].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivp is repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivp lies outside the range [1,n].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of nonzero entries in the decomposition is too large.',
                    'The decomposition has been terminated before completion.',
                    'Either increase la, or reduce the fill by reducing lfill,',
                    'or increasing dtol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call to an auxiliary function.',
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _istr_cls.refresh_py(_istr)
        _idiag_cls.refresh_py(_idiag)
        _nnzc_py = utils._EngineIntScalarType.to_py(_nnzc)
        _npivm_py = utils._EngineIntScalarType.to_py(_npivm)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nnzc', 'npivm'],
        )(
            _nnzc_py,
            _npivm_py,
        )
    _err_class.manage()
    return return_data

def real_gen_precon_ilu_solve(trans, n, a, la, irow, icol, ipivp, ipivq, istr, idiag, check, y, x):
    u"""
    Solution of linear system involving incomplete LU preconditioning
    matrix generated by ``real_gen_precon_ilu``.

    ``real_gen_precon_ilu_solve`` solves a system of linear equations
    involving the incomplete LU preconditioning matrix generated by
    ``real_gen_precon_ilu``.

    For full information please refer to the NAG Library document for
    f11db

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11dbf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether or not the matrix M is transposed.

        `trans` = 'N'

            Mx = y is solved.

        `trans` = 'T'

            M^Tx = y is solved.

    n : int
        n, the order of the matrix M. This **must** be the same value as
        was supplied in the preceding call to ``real_gen_precon_ilu``.

    a : float, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``real_gen_precon_ilu``.

    la : int
        This **must** be the same value returned by the preceding call
        to ``real_gen_precon_ilu``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

    ipivp : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

    ipivq : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

    idiag : int, shape (`n`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

    check : str, length 1
        Specifies whether or not the CS representation of the matrix M
        should be checked.

        `check` = 'C'

            Checks are carried on the values of `n`, `irow`, `icol`,
            `ipivp`, `ipivq`, `istr` and `idiag`.

        `check` = 'N'

            None of these checks are carried out.

        See also [Further Comments].

    y : float, shape (`n`)
        The right-hand side vector y.

    x : float, shape (`n`), modified in place
        `On exit`: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `ipivq`[i-1] is a repeated value: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `ipivq`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `ipivq`[i-1] >= 1 and `ipivq`[i-1] <= `n`.

        (`errno` 3)
            On entry, `ipivp`[i-1] is a repeated value: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `ipivp`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `ipivp`[i-1] >= 1 and `ipivp`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr`[i-1] is inconsistent with `irow`: i =
            *<value>*.

        (`errno` 3)
            On entry, `idiag`[i-1] appears to be incorrect: i =
            *<value>*.

        (`errno` 3)
            On entry, the location (`irow`[i-1],`icol`[i-1]) is a
            duplicate: i = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `icol`[i-1] = *<value>*, and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr` appears to be invalid.

    Notes
    -----
    ``real_gen_precon_ilu_solve`` solves a system of linear equations

        Mx = y, or M^Tx = y,

    according to the value of the argument `trans`, where the matrix M =
    PLDUQ, corresponds to an incomplete LU decomposition of a sparse
    matrix stored in coordinate storage (CS) format (see [the F11
    Introduction]), as generated by ``real_gen_precon_ilu``.

    In the above decomposition L is a lower triangular sparse matrix
    with unit diagonal elements, D is a diagonal matrix, U is an upper
    triangular sparse matrix with unit diagonal elements and, P and Q
    are permutation matrices. L, D and U are supplied to
    ``real_gen_precon_ilu_solve`` through the matrix

        C = L+D^-1+U-2I

    which is an `n` by `n` sparse matrix, stored in CS format, as
    returned by ``real_gen_precon_ilu``.
    The permutation matrices P and Q are returned from
    ``real_gen_precon_ilu`` via the arrays `ipivp` and `ipivq`.

    It is envisaged that a common use of ``real_gen_precon_ilu_solve``
    will be to carry out the preconditioning step required in the
    application of ``real_gen_basic_solver`` to sparse linear systems.
    ``real_gen_precon_ilu_solve`` is used for this purpose by the Black
    Box function ``real_gen_solve_ilu``.

    ``real_gen_precon_ilu_solve`` may also be used in combination with
    ``real_gen_precon_ilu`` to solve a sparse system of linear equations
    directly (see [Further Comments]).
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_precon_ilu_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[n].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._EngineIntArrayType.c_type, # Input, idiag[n].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_n_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_n_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _idiag = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _a,
        _la,
        _irow,
        _icol,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _check,
        _y,
        _x,
        _errbuf,
        _ierr,
        _trans._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ipivq[i-1] is a repeated value:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ipivq[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ipivq[i-1] >= 1 and ipivq[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ipivp[i-1] is a repeated value:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ipivp[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ipivp[i-1] >= 1 and ipivp[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr[i-1] is inconsistent with irow:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, idiag[i-1] appears to be incorrect:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[i-1],icol[i-1]) is a duplicate:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr appears to be invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def real_gen_solve_ilu(method, n, nnz, a, la, irow, icol, ipivp, ipivq, istr, idiag, b, m, tol, maxitn, x):
    u"""
    Solution of real sparse nonsymmetric linear system, RGMRES, CGS,
    Bi-CGSTAB or TFQMR method, preconditioner computed by
    ``real_gen_precon_ilu``.

    ``real_gen_solve_ilu`` solves a real sparse nonsymmetric system of
    linear equations, represented in coordinate storage format, using a
    restarted generalized minimal residual (RGMRES), conjugate gradient
    squared (CGS), stabilized bi-conjugate gradient (BI-CGSTAB), or
    transpose-free quasi-minimal residual (TFQMR) method, with
    incomplete LU preconditioning.

    For full information please refer to the NAG Library document for
    f11dc

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11dcf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'RGMRES'

            Restarted generalized minimum residual method.

        `method` = 'CGS'

            Conjugate gradient squared method.

        `method` = 'BICGSTAB'

            Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

            Transpose-free quasi-minimal residual method.

    n : int
        n, the order of the matrix A. This **must** be the same value as
        was supplied in the preceding call to ``real_gen_precon_ilu``.

    nnz : int
        The number of nonzero elements in the matrix A. This **must** be
        the same value as was supplied in the preceding call to
        ``real_gen_precon_ilu``.

    a : float, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``real_gen_precon_ilu``.

    la : int
        This **must** be the same value as was supplied in the preceding
        call to ``real_gen_precon_ilu``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    ipivp : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    ipivq : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    idiag : int, shape (`n`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    b : float, shape (`n`)
        The right-hand side vector b.

    m : int
        If `method` = 'RGMRES', `m` is the dimension of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method; otherwise, `m` is not referenced.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : float, shape (`n`), modified in place
        `On entry`: an initial approximation to the solution vector x.

        `On exit`: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` >= 1 and `m` <= min(`n`, *<value>*).

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS' or 'BICGSTAB'.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[i-1],`icol`[i-1]) is a
            duplicate: i = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            The CS representation of the preconditioner is invalid.

        (`errno` 6)
            Algorithmic breakdown. A solution is returned, although it
            is possible that it is completely inaccurate.

        (`errno` 7)
            A serious error, code *<value>*, has occurred in an internal
            call. Check all function calls and array sizes. Seek expert
            help.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``real_gen_solve_ilu`` solves a real sparse nonsymmetric linear
    system of equations:

        Ax = b,

    using a preconditioned RGMRES (see Saad and Schultz (1986)), CGS
    (see Sonneveld (1989)), BI-CGSTAB(l) (see Van der Vorst (1989) and
    Sleijpen and Fokkema (1993)), or TFQMR (see Freund and Nachtigal
    (1991) and Freund (1993)) method.

    ``real_gen_solve_ilu`` uses the incomplete LU factorization
    determined by ``real_gen_precon_ilu`` as the preconditioning matrix.
    A call to ``real_gen_solve_ilu`` must always be preceded by a call
    to ``real_gen_precon_ilu``.
    Alternative preconditioners for the same storage scheme are
    available by calling ``real_gen_solve_jacssor``.

    The matrix A, and the preconditioning matrix M, are represented in
    coordinate storage (CS) format (see [the F11 Introduction]) in the
    arrays `a`, `irow` and `icol`, as returned from
    ``real_gen_precon_ilu``.
    The array `a` holds the nonzero entries in these matrices, while
    `irow` and `icol` hold the corresponding row and column indices.

    ``real_gen_solve_ilu`` is a Black Box function which calls
    ``real_gen_basic_setup``, ``real_gen_basic_solver`` and
    ``real_gen_basic_diag``.
    If you wish to use an alternative storage scheme, preconditioner, or
    termination criterion, or require additional diagnostic information,
    you should call these underlying functions directly.

    References
    ----------
    Freund, R W, 1993, `A transpose-free quasi-minimal residual
    algorithm for non-Hermitian linear systems`, SIAM J. Sci. Comput.
    (14), 470--482

    Freund, R W and Nachtigal, N, 1991, `QMR: a Quasi-Minimal Residual
    Method for Non-Hermitian Linear Systems`, Numer. Math. (60),
    315--339

    Saad, Y and Schultz, M, 1986, `GMRES: a generalized minimal residual
    algorithm for solving nonsymmetric linear systems`, SIAM J. Sci.
    Statist. Comput. (7), 856--869

    Salvini, S A and Shaw, G J, 1996, `An evaluation of new NAG Library
    solvers for large sparse unsymmetric linear systems`, NAG Technical
    Report TR2/96

    Sleijpen, G L G and Fokkema, D R, 1993, `BiCGSTAB(l) for linear
    equations involving matrices with complex spectrum`, ETNA (1),
    11--32

    Sonneveld, P, 1989, `CGS, a fast Lanczos-type solver for
    nonsymmetric linear systems`, SIAM J. Sci. Statist. Comput. (10),
    36--52

    Van der Vorst, H, 1989, `Bi-CGSTAB, a fast and smoothly converging
    variant of Bi-CG for the solution of nonsymmetric linear systems`,
    SIAM J. Sci. Statist. Comput. (13), 631--644
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_solve_ilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[n].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._EngineIntArrayType.c_type, # Input, idiag[n].
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_n_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_n_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _idiag = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _b,
        _m,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: m >= 1 and m <= min(n, {:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\' or \'BICGSTAB\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[i-1],icol[i-1]) is a duplicate:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The CS representation of the preconditioner is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99985:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99984:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithmic breakdown. A solution is returned, although',
                    'it is possible that it is completely inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error, code {:s}, has occurred in an internal'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'call. Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error, code {:s}, has occurred in an internal'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'call. Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_gen_precon_ssor_solve(trans, n, nnz, a, irow, icol, rdiag, omega, check, y, x):
    u"""
    Solution of linear system involving preconditioning matrix generated
    by applying SSOR to real sparse nonsymmetric matrix.

    ``real_gen_precon_ssor_solve`` solves a system of linear equations
    involving the preconditioning matrix corresponding to SSOR applied
    to a real sparse nonsymmetric matrix, represented in coordinate
    storage format.

    For full information please refer to the NAG Library document for
    f11dd

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11ddf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether or not the matrix M is transposed.

        `trans` = 'N'

            Mx = y is solved.

        `trans` = 'T'

            M^Tx = y is solved.

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : float, shape (`nnz`)
        The nonzero elements in the matrix A, ordered by increasing row
        index, and by increasing column index within each row. Multiple
        entries for the same row and column indices are not permitted.
        The function ``real_gen_sort`` may be used to order the elements
        in this way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    rdiag : float, shape (`n`)
        The elements of the diagonal matrix D^-1, where D is the
        diagonal part of A.

    omega : float
        The relaxation parameter omega.

    check : str, length 1
        Specifies whether or not the CS representation of the matrix M
        should be checked.

        `check` = 'C'

            Checks are carried on the values of `n`, `nnz`, `irow`,
            `icol` and `omega`.

        `check` = 'N'

            None of these checks are carried out.

        See also [Further Comments].

    y : float, shape (`n`)
        The right-hand side vector y.

    x : float, shape (`n`), modified in place
        `On exit`: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 2)
            On entry, `omega` = *<value>*.

            Constraint: 0.0 < `omega` < 2.0.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `nnz` <= `n`^2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: 1 <= `nnz` <= `n`^2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 4)
            The matrix A has no diagonal entry in row *<value>*.

    Notes
    -----
    ``real_gen_precon_ssor_solve`` solves a system of linear equations

        Mx = y, or M^Tx = y,

    according to the value of the argument `trans`, where the matrix

        M = 1/omega(2-omega)(D+omega L)D^-1(D+omega U)

    corresponds to symmetric successive-over-relaxation (SSOR) (see
    Young (1971)) applied to a linear system Ax = b, where A is a real
    sparse nonsymmetric matrix stored in coordinate storage (CS) format
    (see [the F11 Introduction]).

    In the definition of M given above D is the diagonal part of A, L is
    the strictly lower triangular part of A, U is the strictly upper
    triangular part of A, and omega is a user-defined relaxation
    parameter.

    It is envisaged that a common use of ``real_gen_precon_ssor_solve``
    will be to carry out the preconditioning step required in the
    application of ``real_gen_basic_solver`` to sparse linear systems.
    ``real_gen_precon_ssor_solve`` is also used for this purpose by the
    Black Box function ``real_gen_solve_jacssor``.

    References
    ----------
    Young, D, 1971, `Iterative Solution of Large Linear Systems`,
    Academic Press, New York
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_precon_ssor_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11ddft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._EngineFloat64ArrayType.c_type, # Input, rdiag[n].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _rdiag = utils._EngineFloat64ArrayType(
        rdiag, {'fun_name': fun_name, 'entity': 'rdiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _rdiag,
        _omega,
        _check,
        _y,
        _x,
        _iwork,
        _errbuf,
        _ierr,
        _trans._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < omega < 2.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 1 <= nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def real_gen_solve_jacssor(method, precon, n, nnz, a, irow, icol, omega, b, m, tol, maxitn, x):
    u"""
    Solution of real sparse nonsymmetric linear system, RGMRES, CGS,
    Bi-CGSTAB, or TFQMR method, Jacobi or SSOR preconditioner (Black
    Box).

    ``real_gen_solve_jacssor`` solves a real sparse nonsymmetric system
    of linear equations, represented in coordinate storage format, using
    a restarted generalized minimal residual (RGMRES), conjugate
    gradient squared (CGS), stabilized bi-conjugate gradient
    (BI-CGSTAB), or transpose-free quasi-minimal residual (TFQMR)
    method, without preconditioning, with Jacobi, or with SSOR
    preconditioning.

    For full information please refer to the NAG Library document for
    f11de

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11def.html

    Parameters
    ----------
    method : str
        The iterative method to be used.

        `method` = 'RGMRES'

            Restarted generalized minimum residual method.

        `method` = 'CGS'

            Conjugate gradient squared method.

        `method` = 'BICGSTAB'

            Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

            Transpose-free quasi-minimal residual method.

    precon : str, length 1
        Specifies the type of preconditioning to be used.

        `precon` = 'N'

            No preconditioning.

        `precon` = 'J'

            Jacobi.

        `precon` = 'S'

            Symmetric successive-over-relaxation.

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : float, shape (`nnz`)
        The nonzero elements of the matrix A, ordered by increasing row
        index, and by increasing column index within each row. Multiple
        entries for the same row and column indices are not permitted.
        The function ``real_gen_sort`` may be used to order the elements
        in this way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in `a`.

    omega : float
        If `precon` = 'S', `omega` is the relaxation parameter omega to
        be used in the SSOR method. Otherwise `omega` need not be
        initialized and is not referenced.

    b : float, shape (`n`)
        The right-hand side vector b.

    m : int
        If `method` = 'RGMRES', `m` is the dimension of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method.

        Otherwise, `m` is not referenced.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : float, shape (`n`), modified in place
        `On entry`: an initial approximation to the solution vector x.

        `On exit`: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` >= 1 and `m` <= MIN(`n`,*<value>*).

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS' or 'BICGSTAB'.

        (`errno` 1)
            On entry, `omega` = *<value>*.

            Constraint: `omega` > 0.0 and `omega` < 2.0.

        (`errno` 1)
            On entry, `precon` != 'N', 'J' or 'S': `precon` = *<value>*.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            The matrix A has a zero diagonal entry in row *<value>*.

        (`errno` 3)
            The matrix A has no diagonal entry in row *<value>*.

        (`errno` 6)
            Algorithmic breakdown. A solution is returned, although it
            is possible that it is completely inaccurate.

        (`errno` 7)
            A serious error has occurred in an internal call: IREVCM =
            *<value>*. Check all function calls and array sizes. Seek
            expert help.

        (`errno` 7)
            A serious error has occurred in an internal call: errno =
            *<value>*. Check all function calls and array sizes. Seek
            expert help.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``real_gen_solve_jacssor`` solves a real sparse nonsymmetric system
    of linear equations

        Ax = b,

    using an RGMRES (see Saad and Schultz (1986)), CGS (see Sonneveld
    (1989)), BI-CGSTAB(l) (see Van der Vorst (1989) and Sleijpen and
    Fokkema (1993)), or TFQMR (see Freund and Nachtigal (1991) and
    Freund (1993)) method.

    The function allows the following choices for the preconditioner:

        no preconditioning;

        Jacobi preconditioning (see Young (1971));

        symmetric successive-over-relaxation (SSOR) preconditioning (see
        Young (1971)).

    For incomplete LU (ILU) preconditioning see ``real_gen_solve_ilu``.

    The matrix A is represented in coordinate storage (CS) format (see
    [the F11 Introduction]) in the arrays `a`, `irow` and `icol`.
    The array `a` holds the nonzero entries in the matrix, while `irow`
    and `icol` hold the corresponding row and column indices.

    ``real_gen_solve_jacssor`` is a Black Box function which calls
    ``real_gen_basic_setup``, ``real_gen_basic_solver`` and
    ``real_gen_basic_diag``.
    If you wish to use an alternative storage scheme, preconditioner, or
    termination criterion, or require additional diagnostic information,
    you should call these underlying functions directly.

    References
    ----------
    Freund, R W, 1993, `A transpose-free quasi-minimal residual
    algorithm for non-Hermitian linear systems`, SIAM J. Sci. Comput.
    (14), 470--482

    Freund, R W and Nachtigal, N, 1991, `QMR: a Quasi-Minimal Residual
    Method for Non-Hermitian Linear Systems`, Numer. Math. (60),
    315--339

    Saad, Y and Schultz, M, 1986, `GMRES: a generalized minimal residual
    algorithm for solving nonsymmetric linear systems`, SIAM J. Sci.
    Statist. Comput. (7), 856--869

    Sleijpen, G L G and Fokkema, D R, 1993, `BiCGSTAB(l) for linear
    equations involving matrices with complex spectrum`, ETNA (1),
    11--32

    Sonneveld, P, 1989, `CGS, a fast Lanczos-type solver for
    nonsymmetric linear systems`, SIAM J. Sci. Statist. Comput. (10),
    36--52

    Van der Vorst, H, 1989, `Bi-CGSTAB, a fast and smoothly converging
    variant of Bi-CG for the solution of nonsymmetric linear systems`,
    SIAM J. Sci. Statist. Comput. (13), 631--644

    Young, D, 1971, `Iterative Solution of Large Linear Systems`,
    Academic Press, New York
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_solve_jacssor'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11deft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _omega,
        _b,
        _m,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _iwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: m >= 1 and m <= MIN(n,{:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\' or \'BICGSTAB\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: omega > 0.0 and omega < 2.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon != \'N\', \'J\' or \'S\': precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has a zero diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99983:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99982:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99979:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithmic breakdown. A solution is returned, although',
                    'it is possible that it is completely inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99980:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call:',
                    'IREVCM = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call:',
                    'errno = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_gen_precon_bdilu(n, nnz, a, la, irow, icol, nb, istb, indb, lindb, lfill, dtol, pstrat, milu, ipivp, ipivq, istr, idiag, npivm):
    u"""
    Real sparse nonsymmetric linear system, incomplete LU factorization
    of local or overlapping diagonal blocks.

    ``real_gen_precon_bdilu`` computes a block diagonal incomplete LU
    factorization of a real sparse nonsymmetric matrix, represented in
    coordinate storage format.
    The diagonal blocks may be composed of arbitrary rows and the
    corresponding columns, and may overlap.
    This factorization can be used to provide a block Jacobi or additive
    Schwarz preconditioner, for use in combination with
    ``real_gen_solve_bdilu`` or ``real_gen_basic_solver``.

    For full information please refer to the NAG Library document for
    f11df

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11dff.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : float, shape (`la`), modified in place
        `On entry`: the nonzero elements in the matrix A, ordered by
        increasing row index, and by increasing column index within each
        row. Multiple entries for the same row and column indices are
        not permitted. The function ``real_gen_sort`` may be used to
        order the elements in this way.

        `On exit`: the first `nnz` entries of `a` contain the nonzero
        elements of A and the next `nnzc` entries contain the elements
        of the matrices C_b, for b = 1,2,...,`nb` stored consecutively.
        Within each block the matrix elements are ordered by increasing
        row index, and by increasing column index within each row.

    la : int
        These arrays must be of sufficient size to store both A (`nnz`
        elements) and C (`nnzc` elements).

        Note: the minimum value for `la` is only appropriate if `lfill`
        and `dtol` are set such that minimal fill-in occurs. If this is
        not the case then we recommend that `la` is set much larger than
        the minimum value indicated in the constraint.

    irow : int, shape (`la`), modified in place
        `On entry`: the row indices of the nonzero elements supplied in
        `a`.

        `On exit`: the row indices of the nonzero elements returned in
        `a`.

    icol : int, shape (`la`), modified in place
        `On entry`: the column indices of the nonzero elements supplied
        in `a`.

        `On exit`: the column indices of the nonzero elements returned
        in `a`.

    nb : int
        The number of diagonal blocks to factorize.

    istb : int, shape (`nb`+1)
        `istb`[b-1], for b = 1,2,...,`nb`, holds the indices in arrays
        `indb`, `ipivp`, `ipivq` and `idiag` that, on successful exit
        from this function, define block b. `istb`[`nb`] holds the sum
        of the number of rows in all blocks plus `istb`[0].

    indb : int, shape (`lindb`)
        `indb` must hold the row indices appearing in each diagonal
        block, stored consecutively. Thus the elements
        `indb`[`istb`[b-1]-1] to `indb`[`istb`[b]-2] are the row indices
        in the bth block, for b = 1,2,...,`nb`.

    lindb : int
        The dimension of the arrays `indb`, `ipivp`, `ipivq` and
        `idiag`.

    lfill : int, shape (`nb`)
        If `lfill`[b-1] >= 0 its value is the maximum level of fill
        allowed in the decomposition of the block (see [Further
        Comments]). A negative value of `lfill`[b-1] indicates that
        `dtol`[b-1] will be used to control the fill in the block
        instead.

    dtol : float, shape (`nb`)
        If `lfill`[b-1] < 0 then `dtol`[b-1] is used as a drop tolerance
        in the block to control the fill-in (see [Further Comments]);
        otherwise `dtol`[b-1] is not referenced.

    pstrat : str, length 1, shape (`nb`)
        `pstrat`[b-1], for b = 1,2,...,`nb`, specifies the pivoting
        strategy to be adopted in the block as follows:

        `pstrat`[b-1] = 'N'

            No pivoting is carried out.

        `pstrat`[b-1] = 'U'

            Pivoting is carried out according to the user-defined input
            values of `ipivp` and `ipivq`.

        `pstrat`[b-1] = 'P'

            Partial pivoting by columns for stability is carried out.

        `pstrat`[b-1] = 'C'

            Complete pivoting by rows for sparsity, and by columns for
            stability, is carried out.

        Suggested value: `pstrat`[b-1] = 'C', for b = 1,2,...,`nb`.

    milu : str, length 1, shape (`nb`)
        `milu`[b-1], for b = 1,2,...,`nb`, indicates whether or not the
        factorization in the block should be modified to preserve
        row-sums (see [Further Comments]).

        `milu`[b-1] = 'M'

            The factorization is modified.

        `milu`[b-1] = 'N'

            The factorization is not modified.

    ipivp : int, shape (`lindb`), modified in place
        `On entry`: if `pstrat`[b-1] = 'U', `ipivp`[`istb`[b-1]+k-2] and
        `ipivq`[`istb`[b-1]+k-2] must specify the row and column indices
        of the element used as a pivot at elimination stage k of the
        factorization of the block. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        `On exit`: the row and column indices of the pivot elements,
        arranged consecutively for each block, as for `indb`. If
        `ipivp`[`istb`[b-1]+k-2] = i and `ipivq`[`istb`[b-1]+k-2] = j,
        the element in row i and column j of A_b was used as the pivot
        at elimination stage k.

    ipivq : int, shape (`lindb`), modified in place
        `On entry`: if `pstrat`[b-1] = 'U', `ipivp`[`istb`[b-1]+k-2] and
        `ipivq`[`istb`[b-1]+k-2] must specify the row and column indices
        of the element used as a pivot at elimination stage k of the
        factorization of the block. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        `On exit`: the row and column indices of the pivot elements,
        arranged consecutively for each block, as for `indb`. If
        `ipivp`[`istb`[b-1]+k-2] = i and `ipivq`[`istb`[b-1]+k-2] = j,
        the element in row i and column j of A_b was used as the pivot
        at elimination stage k.

    istr : int, shape (`lindb`+1), modified in place
        `On exit`: `istr`[`istb`[b-1]+k-2], gives the index in the
        arrays `a`, `irow` and `icol` of row k of the matrix C_b, for k
        = 1,2,...,`istb`[b]-`istb`[b-1], for b = 1,2,...,`nb`.

        `istr`[`istb`[`nb`]-1] contains `nnz`+`nnzc`+1.

    idiag : int, shape (`lindb`), modified in place
        `On exit`: `idiag`[`istb`[b-1]+k-2], gives the index in the
        arrays `a`, `irow` and `icol` of the diagonal element in row k
        of the matrix C_b, for k = 1,2,...,`istb`[b]-`istb`[b-1], for b
        = 1,2,...,`nb`.

    npivm : int, shape (`nb`), modified in place
        `On exit`: if `npivm`[b-1] > 0 it gives the number of pivots
        which were modified during the factorization to ensure that M_b
        exists.

        If `npivm`[b-1] = -1 no pivot modifications were required, but a
        local restart occurred (see [Further Comments for
        real_gen_precon_ilu]). The quality of the preconditioner will
        generally depend on the returned values of `npivm`[b-1], for b =
        1,2,...,`nb`.

        If `npivm`[b-1] is large, for some block, the preconditioner may
        not be satisfactory.

        In this case it may be advantageous to call
        ``real_gen_precon_bdilu`` again with an increased value of
        `lfill`[b-1], a reduced value of `dtol`[b-1], or `pstrat`[b-1] =
        'C'.

    Returns
    -------
    nnzc : int
        The sum total number of nonzero elements in the matrices C_b,
        for b = 1,2,...,`nb`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nb` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `nb` <= `n`.

        (`errno` 1)
            On entry, `istb`[0] = *<value>*.

            Constraint: `istb`[0] >= 1.

        (`errno` 1)
            On entry, for b = *<value>*, `istb`[b] = *<value>* and
            `istb`[b-1] = *<value>*.

            Constraint: `istb`[b] > `istb`[b-1], for b = 1,2,...,`nb`.

        (`errno` 1)
            On entry, `indb`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `indb`[m-1] <= `n`, for m =
            1,2,...,`istb`[`nb`]-1

        (`errno` 1)
            On entry, `lindb` = *<value>*, `istb`[`nb`]-1 = *<value>*
            and `nb` = *<value>*.

            Constraint: `lindb` >= `istb`[`nb`]-1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `dtol`[*<value>*] = *<value>*.

            Constraint: `dtol`[b-1] >= 0.0, for b = 1,2,...,`nb`.

        (`errno` 1)
            On entry, `pstrat`[*<value>*] = *<value>*.

            Constraint: `pstrat`[b-1] = 'N', 'U', 'P' or 'C' for all b.

        (`errno` 1)
            On entry, `milu`[*<value>*] = *<value>*.

            Constraint: `milu`[b-1] = 'M' or 'N' for all b.

        (`errno` 2)
            On entry, `irow`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `irow`[i-1] <= `n`, for i = 1,2,...,`nnz`.

        (`errno` 2)
            On entry, `icol`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `icol`[j-1] <= `n`, for j = 1,2,...,`nnz`.

        (`errno` 2)
            On entry, element *<value>* of `a` was out of order.

        (`errno` 2)
            On entry, location *<value>* of (`irow`,`icol`) was a
            duplicate.

        (`errno` 3)
            On entry, the user-supplied value of `ipivp` for block
            *<value>* lies outside its range.

        (`errno` 3)
            On entry, the user-supplied value of `ipivp` for block
            *<value>* was repeated.

        (`errno` 3)
            On entry, the user-supplied value of `ipivq` for block
            *<value>* lies outside its range.

        (`errno` 3)
            On entry, the user-supplied value of `ipivq` for block
            *<value>* was repeated.

        (`errno` 4)
            The number of nonzero entries in the decomposition is too
            large.

            The decomposition has been terminated before completion.

            Either increase `la`, or reduce the fill by reducing
            `lfill`, or increasing `dtol`.

    Notes
    -----
    ``real_gen_precon_bdilu`` computes an incomplete LU factorization
    (see Meijerink and Van der Vorst (1977) and Meijerink and Van der
    Vorst (1981)) of the (possibly overlapping) diagonal blocks A_b, for
    b = 1,2,...,`nb`, of a real sparse nonsymmetric n*n matrix A.
    The factorization is intended primarily for use as a block Jacobi or
    additive Schwarz preconditioner (see Saad (1996)), with one of the
    iterative solvers ``real_gen_solve_bdilu`` and
    ``real_gen_basic_solver``.

    The `nb` diagonal blocks need not consist of consecutive rows and
    columns of A, but may be composed of arbitrarily indexed rows, and
    the corresponding columns, as defined in the arguments `indb` and
    `istb`.
    Any given row or column index may appear in more than one diagonal
    block, resulting in overlap.
    Each diagonal block A_b, for b = 1,2,...,`nb`, is factorized as:

        A_b = M_b+R_b

    where

        M_b = P_b L_b D_b U_b Q_b

    and L_b is lower triangular with unit diagonal elements, D_b is
    diagonal, U_b is upper triangular with unit diagonals, P_b and Q_b
    are permutation matrices, and R_b is a remainder matrix.

    The amount of fill-in occurring in the factorization of block b can
    vary from zero to complete fill, and can be controlled by specifying
    either the maximum level of fill `lfill`[b-1], or the drop tolerance
    `dtol`[b-1].

    The parameter `pstrat`[b-1] defines the pivoting strategy to be used
    in block b.
    The options currently available are no pivoting, user-defined
    pivoting, partial pivoting by columns for stability, and complete
    pivoting by rows for sparsity and by columns for stability.
    The factorization may optionally be modified to preserve the
    row-sums of the original block matrix.

    The sparse matrix A is represented in coordinate storage (CS) format
    (see [the F11 Introduction]).
    The array `a` stores all the nonzero elements of the matrix A, while
    arrays `irow` and `icol` store the corresponding row and column
    indices respectively.
    Multiple nonzero elements may not be specified for the same row and
    column index.

    The preconditioning matrices M_b, for b = 1,2,...,`nb`, are returned
    in terms of the CS representations of the matrices

        C_b = L_b+D^-1_b+U_b-2I.

    References
    ----------
    Meijerink, J and Van der Vorst, H, 1977, `An iterative solution
    method for linear systems of which the coefficient matrix is a
    symmetric M-matrix`, Math. Comput. (31), 148--162

    Meijerink, J and Van der Vorst, H, 1981, `Guidelines for the usage
    of incomplete decompositions in solving sets of linear equations as
    they occur in practical problems`, J. Comput. Phys. (44), 134--155

    Saad, Y, 1996, `Iterative Methods for Sparse Linear Systems`, PWS
    Publishing Company, Boston, MA
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_precon_bdilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input/Output, irow[la].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[la].
        utils._BYREF_INT, # Input, nb.
        utils._EngineIntArrayType.c_type, # Input, istb[nb+1].
        utils._EngineIntArrayType.c_type, # Input, indb[lindb].
        utils._BYREF_INT, # Input, lindb.
        utils._EngineIntArrayType.c_type, # Input, lfill[nb].
        utils._EngineFloat64ArrayType.c_type, # Input, dtol[nb].
        utils._EngineCharArrayType.c_type, # Input, pstrat(len=1)[nb].
        utils._EngineCharArrayType.c_type, # Input, milu(len=1)[nb].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[lindb].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[lindb].
        utils._EngineIntArrayType.c_type, # Output, istr[lindb+1].
        utils._EngineIntArrayType.c_type, # Output, idiag[lindb].
        utils._BYREF_INT, # Output, nnzc.
        utils._EngineIntArrayType.c_type, # Output, npivm[nb].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._BYREF_INT, # Input, liwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pstrat.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of milu.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _nb = utils._EngineIntScalarType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
    ).to_ctypes()
    _nb_py = utils._EngineIntScalarType.to_py(_nb)
    _istb = utils._EngineIntArrayType(
        istb, {'fun_name': fun_name, 'entity': 'istb'},
        exp_shape=(_nb_py+1,),
    ).to_ctypes()
    _lindb = utils._EngineIntScalarType(
        lindb, {'fun_name': fun_name, 'entity': 'lindb'},
    ).to_ctypes()
    _lindb_py = utils._EngineIntScalarType.to_py(_lindb)
    _indb = utils._EngineIntArrayType(
        indb, {'fun_name': fun_name, 'entity': 'indb'},
        exp_shape=(_lindb_py,),
    ).to_ctypes()
    _lfill = utils._EngineIntArrayType(
        lfill, {'fun_name': fun_name, 'entity': 'lfill'},
        exp_shape=(_nb_py,),
    ).to_ctypes()
    _dtol = utils._EngineFloat64ArrayType(
        dtol, {'fun_name': fun_name, 'entity': 'dtol'},
        exp_shape=(_nb_py,),
    ).to_ctypes()
    _pstrat_cls = utils._EngineCharArrayType(
        pstrat, {'fun_name': fun_name, 'entity': 'pstrat'}, exp_chrlen=1,
        exp_shape=(_nb_py,),
    )
    _pstrat = _pstrat_cls.to_ctypes()
    _pstrat_chrlen = _pstrat_cls.chrlen
    _milu_cls = utils._EngineCharArrayType(
        milu, {'fun_name': fun_name, 'entity': 'milu'}, exp_chrlen=1,
        exp_shape=(_nb_py,),
    )
    _milu = _milu_cls.to_ctypes()
    _milu_chrlen = _milu_cls.chrlen
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_lindb_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_lindb_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_lindb_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _idiag_cls = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_lindb_py,),
        check_contents=False,
    )
    _idiag = _idiag_cls.to_ctypes(can_reference_contents=False)
    _nnzc = utils._EngineIntScalarType.empty_ctypes()
    _npivm_cls = utils._EngineIntArrayType(
        npivm, {'fun_name': fun_name, 'entity': 'npivm'},
        exp_shape=(_nb_py,),
        check_contents=False,
    )
    _npivm = _npivm_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _liwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _nb,
        _istb,
        _indb,
        _lindb,
        _lfill,
        _dtol,
        _pstrat,
        _milu,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _nnzc,
        _npivm,
        _iwork,
        _liwork,
        _errbuf,
        _ierr,
        _pstrat_chrlen,
        _milu_chrlen,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nb = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= nb <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1012:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istb[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: istb[0] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1013:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, for b = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[b] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'istb[b-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: istb[b] > istb[b-1], for b = 1,2,...,nb.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1014:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, indb[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= indb[m-1] <= n, for m = 1,2,...,istb[nb]-1',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1015:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lindb = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[nb]-1 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lindb >= istb[nb]-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1016:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1017:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1018:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1020:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dtol[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: dtol[b-1] >= 0.0, for b = 1,2,...,nb.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1021:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pstrat[{:s}] = \'{:s}\'.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'Constraint: pstrat[b-1] = \'N\', \'U\', \'P\' or \'C\' for all b.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1022:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, milu[{:s}] = \'{:s}\'.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'Constraint: milu[b-1] = \'M\' or \'N\' for all b.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irow[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irow[i-1] <= n, for i = 1,2,...,nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icol[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= icol[j-1] <= n, for j = 1,2,...,nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1112:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, element {:s} of a was out of order.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1113:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, location {:s} of (irow,icol) was a duplicate.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1210:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivp for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'lies outside its range.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1211:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivp for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'was repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1212:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivq for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'lies outside its range.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1213:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivq for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'was repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1310:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of nonzero entries in the decomposition is too large.',
                    'The decomposition has been terminated before completion.',
                    'Either increase la, or reduce the fill by reducing lfill,',
                    'or increasing dtol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _istr_cls.refresh_py(_istr)
        _idiag_cls.refresh_py(_idiag)
        _npivm_cls.refresh_py(_npivm)
        _nnzc_py = utils._EngineIntScalarType.to_py(_nnzc)
        return_data = _nnzc_py
    _err_class.manage()
    return return_data

def real_gen_solve_bdilu(method, n, nnz, a, la, irow, icol, nb, istb, indb, lindb, ipivp, ipivq, istr, idiag, b, m, tol, maxitn, x):
    u"""
    Solution of real sparse nonsymmetric linear system, RGMRES, CGS,
    Bi-CGSTAB or TFQMR method, incomplete LU block diagonal
    preconditioner computed by ``real_gen_precon_bdilu``.

    ``real_gen_solve_bdilu`` solves a real sparse nonsymmetric system of
    linear equations, represented in coordinate storage format, using a
    restarted generalized minimal residual (RGMRES), conjugate gradient
    squared (CGS), stabilized bi-conjugate gradient (BI-CGSTAB), or
    transpose-free quasi-minimal residual (TFQMR) method, with block
    Jacobi or additive Schwarz preconditioning.

    For full information please refer to the NAG Library document for
    f11dg

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11dgf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'RGMRES'

            Restarted generalized minimum residual method.

        `method` = 'CGS'

            Conjugate gradient squared method.

        `method` = 'BICGSTAB'

            Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

            Transpose-free quasi-minimal residual method.

    n : int
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    nnz : int
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    a : float, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    la : int
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    nb : int
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    istb : int, shape (`nb`+1)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    indb : int, shape (`lindb`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    lindb : int
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    ipivp : int, shape (`lindb`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    ipivq : int, shape (`lindb`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    istr : int, shape (`lindb`+1)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    idiag : int, shape (`lindb`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    b : float, shape (`n`)
        The right-hand side vector b.

    m : int
        If `method` = 'RGMRES', `m` is the dimension of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method.

        Otherwise, `m` is not referenced.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),sqrt(n)epsilon) is
        used, where epsilon is the machine precision. Otherwise tau =
        max(`tol`,10 epsilon, sqrt(n)epsilon) is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : float, shape (`n`), modified in place
        `On entry`: an initial approximation to the solution vector x.

        `On exit`: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nb` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `nb` <= `n`.

        (`errno` 1)
            On entry, `istb`[0] = *<value>*.

            Constraint: `istb`[0] >= 1.

        (`errno` 1)
            On entry, for b = *<value>*, `istb`[b] = *<value>* and
            `istb`[b-1] = *<value>*.

            Constraint: `istb`[b] > `istb`[b-1], for b = 1,2,...,`nb`.

        (`errno` 1)
            On entry, `indb`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `indb`[m-1] <= `n`, for m =
            1,2,...,`istb`[`nb`]-1

        (`errno` 1)
            On entry, `lindb` = *<value>*, `istb`[`nb`]-1 = *<value>*
            and `nb` = *<value>*.

            Constraint: `lindb` >= `istb`[`nb`]-1.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS' or 'BICGSTAB'.

        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: if `method` = 'RGMRES', 1 <= `m` <= min(`n`,
            *<value>*).

            If `method` = 'BICGSTAB', 1 <= `m` <= min(`n`, *<value>*).

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 2)
            On entry, `irow`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `irow`[i-1] <= `n`, for i = 1,2,...,`nnz`.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``real_gen_precon_bdilu`` and ``real_gen_solve_bdilu``.

        (`errno` 2)
            On entry, `icol`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `icol`[i-1] <= `n`, for i = 1,2,...,`nnz`.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``real_gen_precon_bdilu`` and ``real_gen_solve_bdilu``.

        (`errno` 2)
            On entry, element *<value>* of `a` was out of order.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``real_gen_precon_bdilu`` and ``real_gen_solve_bdilu``.

        (`errno` 2)
            On entry, location *<value>* of (`irow`,`icol`) was a
            duplicate.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``real_gen_precon_bdilu`` and ``real_gen_solve_bdilu``.

        (`errno` 3)
            The CS representation of the preconditioner is invalid.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``real_gen_precon_bdilu`` and ``real_gen_solve_bdilu``.

        (`errno` 4)
            The required accuracy could not be obtained. However a
            reasonable accuracy may have been achieved. You should check
            the output value of `rnorm` for acceptability. This error
            code usually implies that your problem has been fully and
            satisfactorily solved to within or close to the accuracy
            available on your system. Further iterations are unlikely to
            improve on this situation.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            Algorithmic breakdown. A solution is returned, although it
            is possible that it is completely inaccurate.

    Notes
    -----
    ``real_gen_solve_bdilu`` solves a real sparse nonsymmetric linear
    system of equations:

        Ax = b,

    using a preconditioned RGMRES (see Saad and Schultz (1986)), CGS
    (see Sonneveld (1989)), BI-CGSTAB(l) (see Van der Vorst (1989) and
    Sleijpen and Fokkema (1993)), or TFQMR (see Freund and Nachtigal
    (1991) and Freund (1993)) method.

    ``real_gen_solve_bdilu`` uses the incomplete (possibly overlapping)
    block LU factorization determined by ``real_gen_precon_bdilu`` as
    the preconditioning matrix.
    A call to ``real_gen_solve_bdilu`` must always be preceded by a call
    to ``real_gen_precon_bdilu``.
    Alternative preconditioners for the same storage scheme are
    available by calling ``real_gen_solve_ilu`` or
    ``real_gen_solve_jacssor``.

    The matrix A, and the preconditioning matrix M, are represented in
    coordinate storage (CS) format (see [the F11 Introduction]) in the
    arrays `a`, `irow` and `icol`, as returned from
    ``real_gen_precon_bdilu``.
    The array `a` holds the nonzero entries in these matrices, while
    `irow` and `icol` hold the corresponding row and column indices.

    ``real_gen_solve_bdilu`` is a Black Box function which calls
    ``real_gen_basic_setup``, ``real_gen_basic_solver`` and
    ``real_gen_basic_diag``.
    If you wish to use an alternative storage scheme, preconditioner, or
    termination criterion, or require additional diagnostic information,
    you should call these underlying functions directly.

    References
    ----------
    Freund, R W, 1993, `A transpose-free quasi-minimal residual
    algorithm for non-Hermitian linear systems`, SIAM J. Sci. Comput.
    (14), 470--482

    Freund, R W and Nachtigal, N, 1991, `QMR: a Quasi-Minimal Residual
    Method for Non-Hermitian Linear Systems`, Numer. Math. (60),
    315--339

    Saad, Y and Schultz, M, 1986, `GMRES: a generalized minimal residual
    algorithm for solving nonsymmetric linear systems`, SIAM J. Sci.
    Statist. Comput. (7), 856--869

    Salvini, S A and Shaw, G J, 1996, `An evaluation of new NAG Library
    solvers for large sparse unsymmetric linear systems`, NAG Technical
    Report TR2/96

    Sleijpen, G L G and Fokkema, D R, 1993, `BiCGSTAB(l) for linear
    equations involving matrices with complex spectrum`, ETNA (1),
    11--32

    Sonneveld, P, 1989, `CGS, a fast Lanczos-type solver for
    nonsymmetric linear systems`, SIAM J. Sci. Statist. Comput. (10),
    36--52

    Van der Vorst, H, 1989, `Bi-CGSTAB, a fast and smoothly converging
    variant of Bi-CG for the solution of nonsymmetric linear systems`,
    SIAM J. Sci. Statist. Comput. (13), 631--644
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_solve_bdilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._BYREF_INT, # Input, nb.
        utils._EngineIntArrayType.c_type, # Input, istb[nb+1].
        utils._EngineIntArrayType.c_type, # Input, indb[lindb].
        utils._BYREF_INT, # Input, lindb.
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[lindb].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[lindb].
        utils._EngineIntArrayType.c_type, # Input, istr[lindb+1].
        utils._EngineIntArrayType.c_type, # Input, idiag[lindb].
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _nb = utils._EngineIntScalarType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
    ).to_ctypes()
    _nb_py = utils._EngineIntScalarType.to_py(_nb)
    _istb = utils._EngineIntArrayType(
        istb, {'fun_name': fun_name, 'entity': 'istb'},
        exp_shape=(_nb_py+1,),
    ).to_ctypes()
    _lindb = utils._EngineIntScalarType(
        lindb, {'fun_name': fun_name, 'entity': 'lindb'},
    ).to_ctypes()
    _lindb_py = utils._EngineIntScalarType.to_py(_lindb)
    _indb = utils._EngineIntArrayType(
        indb, {'fun_name': fun_name, 'entity': 'indb'},
        exp_shape=(_lindb_py,),
    ).to_ctypes()
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_lindb_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_lindb_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_lindb_py+1,),
    ).to_ctypes()
    _idiag = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_lindb_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _nb,
        _istb,
        _indb,
        _lindb,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _b,
        _m,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1012:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1013:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1015:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nb = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= nb <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1016:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istb[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: istb[0] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1017:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, for b = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[b] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'istb[b-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: istb[b] > istb[b-1], for b = 1,2,...,nb.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1018:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, indb[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= indb[m-1] <= n, for m = 1,2,...,istb[nb]-1',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1019:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lindb = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[nb]-1 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lindb >= istb[nb]-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1020:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\' or \'BICGSTAB\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1021:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if method = \'RGMRES\', 1 <= m <= min(n, {:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'If method = \'BICGSTAB\', 1 <= m <= min(n, {:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1022:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1023:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irow[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irow[i-1] <= n, for i = 1,2,...,nnz.',
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to real_gen_precon_bdilu and real_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icol[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= icol[i-1] <= n, for i = 1,2,...,nnz.',
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to real_gen_precon_bdilu and real_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1112:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, element {:s} of a was out of order.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to real_gen_precon_bdilu and real_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1113:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, location {:s} of (irow,icol) was a duplicate.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to real_gen_precon_bdilu and real_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1210:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The CS representation of the preconditioner is invalid.',
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to real_gen_precon_bdilu and real_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1310:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained. However a reasonable accuracy',
                    'may have been achieved. You should check the output value of rnorm for',
                    'acceptability. This error code usually implies that your problem has been',
                    'fully and satisfactorily solved to within or close to the accuracy available',
                    'on your system. Further iterations are unlikely to improve on this',
                    'situation.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1410:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 1510:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithmic breakdown. A solution is returned, although',
                    'it is possible that it is completely inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
    _err_class.manage()
    return return_data

def real_gen_precon_jacobi(store, trans, init, niter, n, nnz, a, irow, icol, check, b, x, diag):
    u"""
    Real, sparse, symmetric or nonsymmetric, linear systems, line Jacobi
    preconditioner.

    ``real_gen_precon_jacobi`` computes the **approximate** solution of
    a real, symmetric or nonsymmetric, sparse system of linear equations
    applying a number of Jacobi iterations.
    It is expected that ``real_gen_precon_jacobi`` will be used as a
    preconditioner for the iterative solution of real sparse systems of
    equations.

    For full information please refer to the NAG Library document for
    f11dk

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11dkf.html

    Parameters
    ----------
    store : str, length 1
        Specifies whether the matrix A is stored using symmetric
        coordinate storage (SCS) (applicable only to a symmetric matrix
        A) or coordinate storage (CS) (applicable to both symmetric and
        non-symmetric matrices).

        `store` = 'N'

            The complete matrix A is stored in CS format.

        `store` = 'S'

            The lower triangle of the symmetric matrix A is stored in
            SCS format.

    trans : str, length 1
        If `store` = 'N', specifies whether the approximate solution of
        Ax = b or of A^Tx = b is required.

        `trans` = 'N'

            The approximate solution of Ax = b is calculated.

        `trans` = 'T'

            The approximate solution of A^Tx = b is calculated.

        Suggested value: if the matrix A is symmetric and stored in CS
        format, it is recommended that `trans` = 'N' for reasons of
        efficiency.

    init : str, length 1
        On first entry, `init` should be set to 'I', unless the diagonal
        elements of A are already stored in the array `diag`. If `diag`
        already contains the diagonal of A, it must be set to 'N'.

        `init` = 'N'

            `diag` must contain the diagonal of A.

        `init` = 'I'

            `diag` will store the diagonal of A on exit.

        Suggested value: `init` = 'I' on first entry; `init` = 'N',
        subsequently, unless `diag` has been overwritten.

    niter : int
        The number of Jacobi iterations requested.

    n : int
        n, the order of the matrix A.

    nnz : int
        If `store` = 'N', the number of nonzero elements in the matrix
        A.

        If `store` = 'S', the number of nonzero elements in the lower
        triangle of the matrix A.

    a : float, shape (`nnz`)
        If `store` = 'N', the nonzero elements in the matrix A (CS
        format).

        If `store` = 'S', the nonzero elements in the lower triangle of
        the matrix A (SCS format).

        In both cases, the elements of either A or of its lower triangle
        must be ordered by increasing row index and by increasing column
        index within each row.

        Multiple entries for the same row and columns indices are not
        permitted.

        The function ``real_gen_sort`` or ``real_symm_sort`` may be used
        to reorder the elements in this way for CS and SCS storage,
        respectively.

    irow : int, shape (`nnz`)
        If `store` = 'N', the row indices of the nonzero elements
        supplied in `a`.

        If `store` = 'S', the row and column indices of the nonzero
        elements of the lower triangle of the matrix A supplied in `a`.

    icol : int, shape (`nnz`)
        If `store` = 'N', the column indices of the nonzero elements
        supplied in `a`.

        If `store` = 'S', the row and column indices of the nonzero
        elements of the lower triangle of the matrix A supplied in `a`.

    check : str, length 1
        Specifies whether or not the CS or SCS representation of the
        matrix A should be checked.

        `check` = 'C'

            Checks are carried out on the values of `n`, `nnz`, `irow`,
            `icol`; if `init` = 'N', `diag` is also checked.

        `check` = 'N'

            None of these checks are carried out.

        See also [Further Comments].

    b : float, shape (`n`)
        The right-hand side vector b.

    x : float, shape (`n`), modified in place
        `On exit`: the approximate solution vector x_`niter`.

    diag : float, shape (`n`), modified in place
        `On entry`: if `init` = 'N', the diagonal elements of A.

        `On exit`: if `init` = 'N', unchanged on exit.

        If `init` = 'I', the diagonal elements of A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `niter` = *<value>*.

            Constraint: `niter` >= 1.

        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'N' or 'C'.

        (`errno` 1)
            On entry, `init` = *<value>*.

            Constraint: `init` = 'N' or 'I'.

        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 1)
            On entry, `store` = *<value>*.

            Constraint: `store` = 'N' or 'S'.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `n`.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 3)
            On entry, I = *<value>*, `irow`[I-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[I-1] >= 1 and `irow`[I-1] <= `n`.

        (`errno` 4)
            On entry, the element `diag`[I-1] is zero: I = *<value>*.

        (`errno` 5)
            On entry, the diagonal element of the Ith row is zero or
            missing: I = *<value>*.

    Notes
    -----
    ``real_gen_precon_jacobi`` computes the **approximate** solution of
    the real sparse system of linear equations Ax = b using `niter`
    iterations of the Jacobi algorithm (see also Golub and Van Loan
    (1996) and Young (1971)):

        x_{k+1} = x_k+D^-1(b-Ax_k)

    where k = 1,2,...,`niter` and x_0 = 0.

    ``real_gen_precon_jacobi`` can be used both for nonsymmetric and
    symmetric systems of equations.
    For symmetric matrices, either all nonzero elements of the matrix A
    can be supplied using coordinate storage (CS), or only the nonzero
    elements of the lower triangle of A, using symmetric coordinate
    storage (SCS) (see [the F11 Introduction]).

    It is expected that ``real_gen_precon_jacobi`` will be used as a
    preconditioner for the iterative solution of real sparse systems of
    equations, using either the suite comprising the functions
    ``real_symm_basic_setup``, ``real_symm_basic_solver`` and
    ``real_symm_basic_diag``, for symmetric systems, or the suite
    comprising the functions ``real_gen_basic_setup``,
    ``real_gen_basic_solver`` and ``real_gen_basic_diag``, for
    nonsymmetric systems of equations.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Young, D, 1971, `Iterative Solution of Large Linear Systems`,
    Academic Press, New York
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_precon_jacobi'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, store(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, init(len=1).
        utils._BYREF_INT, # Input, niter.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, diag[n].
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of store.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of init.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _store = utils._EngineCharScalarType(
        store, {'fun_name': fun_name, 'entity': 'store'}, exp_chrlen=1,
    ).to_ctypes()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _init = utils._EngineCharScalarType(
        init, {'fun_name': fun_name, 'entity': 'init'}, exp_chrlen=1,
    ).to_ctypes()
    _niter = utils._EngineIntScalarType(
        niter, {'fun_name': fun_name, 'entity': 'niter'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _diag_cls = utils._EngineFloat64ArrayType(
        diag, {'fun_name': fun_name, 'entity': 'diag'},
        exp_shape=(_n_py,),
    )
    _diag = _diag_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _store,
        _trans,
        _init,
        _niter,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _check,
        _b,
        _x,
        _diag,
        _work,
        _errbuf,
        _ierr,
        _store._length_,
        _trans._length_,
        _init._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, niter = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: niter >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'N\' or \'C\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, init = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: init = \'N\' or \'I\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, store = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: store = \'N\' or \'S\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'irow[I-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[I-1] >= 1 and irow[I-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the element diag[I-1] is zero:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the diagonal element of the Ith row is zero or missing:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _diag_cls.refresh_py(_diag)
        return_data = None
    _err_class.manage()
    return return_data

def complex_gen_precon_ilu(n, nnz, a, la, irow, icol, lfill, dtol, pstrat, milu, ipivp, ipivq, istr, idiag):
    u"""
    Complex sparse non-Hermitian linear systems, incomplete LU
    factorization.

    ``complex_gen_precon_ilu`` computes an incomplete LU factorization
    of a complex sparse non-Hermitian matrix, represented in coordinate
    storage format.
    This factorization may be used as a preconditioner in combination
    with ``complex_gen_solve_ilu`` or ``complex_gen_basic_solver``.

    For full information please refer to the NAG Library document for
    f11dn

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11dnf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : complex, shape (`la`), modified in place
        `On entry`: the nonzero elements in the matrix A, ordered by
        increasing row index, and by increasing column index within each
        row. Multiple entries for the same row and column indices are
        not permitted. The function ``complex_gen_sort`` may be used to
        order the elements in this way.

        `On exit`: the first `nnz` entries of `a` contain the nonzero
        elements of A and the next `nnzc` entries contain the elements
        of the matrix C. Matrix elements are ordered by increasing row
        index, and by increasing column index within each row.

    la : int
        These arrays must be of sufficient size to store both A (`nnz`
        elements) and C (`nnzc` elements).

    irow : int, shape (`la`), modified in place
        `On entry`: the row indices of the nonzero elements supplied in
        `a`.

        `On exit`: the row indices of the nonzero elements returned in
        `a`.

    icol : int, shape (`la`), modified in place
        `On entry`: the column indices of the nonzero elements supplied
        in `a`.

        `On exit`: the column indices of the nonzero elements returned
        in `a`.

    lfill : int
        If `lfill` >= 0 its value is the maximum level of fill allowed
        in the decomposition (see [Further Comments]). A negative value
        of `lfill` indicates that `dtol` will be used to control the
        fill instead.

    dtol : float
        If `lfill` < 0, `dtol` is used as a drop tolerance to control
        the fill-in (see [Further Comments]); otherwise `dtol` is not
        referenced.

    pstrat : str, length 1
        Specifies the pivoting strategy to be adopted.

        `pstrat` = 'N'

            No pivoting is carried out.

        `pstrat` = 'U'

            Pivoting is carried out according to the user-defined input
            values of `ipivp` and `ipivq`.

        `pstrat` = 'P'

            Partial pivoting by columns for stability is carried out.

        `pstrat` = 'C'

            Complete pivoting by rows for sparsity, and by columns for
            stability, is carried out.

        Suggested value: `pstrat` = 'C'.

    milu : str, length 1
        Indicates whether or not the factorization should be modified to
        preserve row-sums (see [Further Comments]).

        `milu` = 'M'

            The factorization is modified.

        `milu` = 'N'

            The factorization is not modified.

    ipivp : int, shape (`n`), modified in place
        `On entry`: if `pstrat` = 'U', `ipivp`[k-1] and `ipivq`[k-1]
        must specify the row and column indices of the element used as a
        pivot at elimination stage k. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        `On exit`: the pivot indices. If `ipivp`[k-1] = i and
        `ipivq`[k-1] = j then the element in row i and column j was used
        as the pivot at elimination stage k.

    ipivq : int, shape (`n`), modified in place
        `On entry`: if `pstrat` = 'U', `ipivp`[k-1] and `ipivq`[k-1]
        must specify the row and column indices of the element used as a
        pivot at elimination stage k. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        `On exit`: the pivot indices. If `ipivp`[k-1] = i and
        `ipivq`[k-1] = j then the element in row i and column j was used
        as the pivot at elimination stage k.

    istr : int, shape (`n`+1), modified in place
        `On exit`: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix C. `istr`[`n`]-1 is the address of the last nonzero
        element in C plus one.

    idiag : int, shape (`n`), modified in place
        `On exit`: `idiag`[i-1], for i = 1,2,...,`n`, holds the index of
        arrays `a`, `irow` and `icol` which holds the diagonal element
        in row i of the matrix C.

    Returns
    -------
    nnzc : int
        The number of nonzero elements in the matrix C.

    npivm : int
        If `npivm` > 0 it gives the number of pivots which were modified
        during the factorization to ensure that M exists.

        If `npivm` = -1 no pivot modifications were required, but a
        local restart occurred (see [Algorithmic Details]).

        The quality of the preconditioner will generally depend on the
        returned value of `npivm`.

        If `npivm` is large the preconditioner may not be satisfactory.

        In this case it may be advantageous to call
        ``complex_gen_precon_ilu`` again with an increased value of
        `lfill`, a reduced value of `dtol`, or set `pstrat` = 'C'.

        See also [Further Comments].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `milu` != 'M' or 'N': `milu` = *<value>*.

        (`errno` 1)
            On entry, `pstrat` != 'N', 'U', 'P' or 'C': `pstrat` =
            *<value>*.

        (`errno` 1)
            On entry, `dtol` = *<value>*.

            Constraint: `dtol` >= 0.0

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`

        (`errno` 2)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and `n` =
            *<value>*.

            Constraint: 1 <= `icol`[i-1] <= `n`, for i = 1,2,...,`nnz`

        (`errno` 2)
            On entry, I = *<value>*, `irow`[I-1] = *<value>* and `n` =
            *<value>*.

            Constraint: 1 <= `irow`[i-1] <= `n`, for i = 1,2,...,`nnz`

        (`errno` 3)
            On entry, a user-supplied value of `ipivq` is repeated.

        (`errno` 3)
            On entry, a user-supplied value of `ipivq` lies outside the
            range [1,`n`].

        (`errno` 3)
            On entry, a user-supplied value of `ipivp` is repeated.

        (`errno` 3)
            On entry, a user-supplied value of `ipivp` lies outside the
            range [1,`n`].

        (`errno` 4)
            The number of nonzero entries in the decomposition is too
            large. The decomposition has been terminated before
            completion. Either increase `la`, or reduce the fill by
            reducing `lfill`, or increasing `dtol`.

        (`errno` 5)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    Notes
    -----
    ``complex_gen_precon_ilu`` computes an incomplete LU factorization
    (see Meijerink and Van der Vorst (1977) and Meijerink and Van der
    Vorst (1981)) of a complex sparse non-Hermitian n*n matrix A.
    The factorization is intended primarily for use as a preconditioner
    with one of the iterative solvers ``complex_gen_solve_ilu`` or
    ``complex_gen_basic_solver``.

    The decomposition is written in the form

        A = M+R,

    where

        M = PLDUQ

    and L is lower triangular with unit diagonal elements, D is
    diagonal, U is upper triangular with unit diagonals, P and Q are
    permutation matrices, and R is a remainder matrix.

    The amount of fill-in occurring in the factorization can vary from
    zero to complete fill, and can be controlled by specifying either
    the maximum level of fill `lfill`, or the drop tolerance `dtol`.

    The argument `pstrat` defines the pivoting strategy to be used.
    The options currently available are no pivoting, user-defined
    pivoting, partial pivoting by columns for stability, and complete
    pivoting by rows for sparsity and by columns for stability.
    The factorization may optionally be modified to preserve the
    row-sums of the original matrix.

    The sparse matrix A is represented in coordinate storage (CS) format
    (see [the F11 Introduction]).
    The array `a` stores all the nonzero elements of the matrix A, while
    arrays `irow` and `icol` store the corresponding row and column
    indices respectively.
    Multiple nonzero elements may not be specified for the same row and
    column index.

    The preconditioning matrix M is returned in terms of the CS
    representation of the matrix

        C = L+D^-1+U-2I.

    Further algorithmic details are given in [Algorithmic Details].

    References
    ----------
    Meijerink, J and Van der Vorst, H, 1977, `An iterative solution
    method for linear systems of which the coefficient matrix is a
    symmetric M-matrix`, Math. Comput. (31), 148--162

    Meijerink, J and Van der Vorst, H, 1981, `Guidelines for the usage
    of incomplete decompositions in solving sets of linear equations as
    they occur in practical problems`, J. Comput. Phys. (44), 134--155
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_precon_ilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input/Output, irow[la].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[la].
        utils._BYREF_INT, # Input, lfill.
        utils._BYREF_FLOAT64, # Input, dtol.
        utils._BYREF_STRING, # Input, pstrat(len=1).
        utils._BYREF_STRING, # Input, milu(len=1).
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[n].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[n].
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._EngineIntArrayType.c_type, # Output, idiag[n].
        utils._BYREF_INT, # Output, nnzc.
        utils._BYREF_INT, # Output, npivm.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._BYREF_INT, # Input, liwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pstrat.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of milu.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _lfill = utils._EngineIntScalarType(
        lfill, {'fun_name': fun_name, 'entity': 'lfill'},
    ).to_ctypes()
    _dtol = utils._EngineFloat64ScalarType(
        dtol, {'fun_name': fun_name, 'entity': 'dtol'},
    ).to_ctypes()
    _pstrat = utils._EngineCharScalarType(
        pstrat, {'fun_name': fun_name, 'entity': 'pstrat'}, exp_chrlen=1,
    ).to_ctypes()
    _milu = utils._EngineCharScalarType(
        milu, {'fun_name': fun_name, 'entity': 'milu'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _pstrat_py = utils._EngineCharScalarType.to_py(_pstrat)
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_n_py,),
        check_contents=_pstrat_py in ('U', 'u'),
    )
    _ipivp = _ipivp_cls.to_ctypes(can_reference_contents=_pstrat_py in ('U', 'u'))
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_n_py,),
        check_contents=_pstrat_py in ('U', 'u'),
    )
    _ipivq = _ipivq_cls.to_ctypes(can_reference_contents=_pstrat_py in ('U', 'u'))
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _idiag_cls = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _idiag = _idiag_cls.to_ctypes(can_reference_contents=False)
    _nnzc = utils._EngineIntScalarType.empty_ctypes()
    _npivm = utils._EngineIntScalarType.empty_ctypes()
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _liwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _lfill,
        _dtol,
        _pstrat,
        _milu,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _nnzc,
        _npivm,
        _iwork,
        _liwork,
        _errbuf,
        _ierr,
        _pstrat._length_,
        _milu._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, milu != \'M\' or \'N\': milu = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pstrat != \'N\', \'U\', \'P\' or \'C\': pstrat = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dtol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dtol >= 0.0',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, icol[I-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= icol[i-1] <= n, for i = 1,2,...,nnz',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, irow[I-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irow[i-1] <= n, for i = 1,2,...,nnz',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivq is repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivq lies outside the range [1,n].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivp is repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivp lies outside the range [1,n].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of nonzero entries in the decomposition is too large.',
                    'The decomposition has been terminated before completion.',
                    'Either increase la, or reduce the fill by reducing lfill,',
                    'or increasing dtol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _istr_cls.refresh_py(_istr)
        _idiag_cls.refresh_py(_idiag)
        _nnzc_py = utils._EngineIntScalarType.to_py(_nnzc)
        _npivm_py = utils._EngineIntScalarType.to_py(_npivm)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nnzc', 'npivm'],
        )(
            _nnzc_py,
            _npivm_py,
        )
    _err_class.manage()
    return return_data

def complex_gen_precon_ilu_solve(trans, n, a, la, irow, icol, ipivp, ipivq, istr, idiag, check, y, x):
    u"""
    Solution of complex linear system involving incomplete LU
    preconditioning matrix generated by ``complex_gen_precon_ilu``.

    ``complex_gen_precon_ilu_solve`` solves a system of complex linear
    equations involving the incomplete LU preconditioning matrix
    generated by ``complex_gen_precon_ilu``.

    For full information please refer to the NAG Library document for
    f11dp

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11dpf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether or not the matrix M is transposed.

        `trans` = 'N'

            Mx = y is solved.

        `trans` = 'T'

            M^Tx = y is solved.

    n : int
        n, the order of the matrix M. This **must** be the same value as
        was supplied in the preceding call to
        ``complex_gen_precon_ilu``.

    a : complex, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``complex_gen_precon_ilu``.

    la : int
        This **must** be the same value supplied in the preceding call
        to ``complex_gen_precon_ilu``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

    ipivp : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

    ipivq : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

    idiag : int, shape (`n`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

    check : str, length 1
        Specifies whether or not the CS representation of the matrix M
        should be checked.

        `check` = 'C'

            Checks are carried on the values of `n`, `irow`, `icol`,
            `ipivp`, `ipivq`, `istr` and `idiag`.

        `check` = 'N'

            None of these checks are carried out.

        See also [Further Comments].

    y : complex, shape (`n`)
        The right-hand side vector y.

    x : complex, shape (`n`), modified in place
        `On exit`: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `ipivq`[i-1] is a repeated value: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `ipivq`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `ipivq`[i-1] >= 1 and `ipivq`[i-1] <= `n`.

        (`errno` 3)
            On entry, `ipivp`[i-1] is a repeated value: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `ipivp`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `ipivp`[i-1] >= 1 and `ipivp`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr`[i-1] is inconsistent with `irow`: i =
            *<value>*.

        (`errno` 3)
            On entry, `idiag`[i-1] appears to be incorrect: i =
            *<value>*.

        (`errno` 3)
            On entry, the location (`irow`[i-1],`icol`[i-1]) is a
            duplicate: i = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `icol`[i-1] = *<value>*, and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr` appears to be invalid.

    Notes
    -----
    ``complex_gen_precon_ilu_solve`` solves a system of complex linear
    equations

        Mx = y, or M^Tx = y,

    according to the value of the argument `trans`, where the matrix M =
    PLDUQ corresponds to an incomplete LU decomposition of a complex
    sparse matrix stored in coordinate storage (CS) format (see [the F11
    Introduction]), as generated by ``complex_gen_precon_ilu``.

    In the above decomposition L is a lower triangular sparse matrix
    with unit diagonal elements, D is a diagonal matrix, U is an upper
    triangular sparse matrix with unit diagonal elements and, P and Q
    are permutation matrices. L, D and U are supplied to
    ``complex_gen_precon_ilu_solve`` through the matrix

        C = L+D^-1+U-2I

    which is an `n` by `n` sparse matrix, stored in CS format, as
    returned by ``complex_gen_precon_ilu``.
    The permutation matrices P and Q are returned from
    ``complex_gen_precon_ilu`` via the arrays `ipivp` and `ipivq`.

    It is envisaged that a common use of
    ``complex_gen_precon_ilu_solve`` will be to carry out the
    preconditioning step required in the application of
    ``complex_gen_basic_solver`` to sparse complex linear systems.
    ``complex_gen_precon_ilu_solve`` is used for this purpose by the
    Black Box function ``complex_gen_solve_ilu``.

    ``complex_gen_precon_ilu_solve`` may also be used in combination
    with ``complex_gen_precon_ilu`` to solve a sparse system of complex
    linear equations directly (see [Further Comments]).
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_precon_ilu_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[n].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._EngineIntArrayType.c_type, # Input, idiag[n].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineComplex128ArrayType.c_type, # Input, y[n].
        utils._EngineComplex128ArrayType.c_type, # Output, x[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_n_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_n_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _idiag = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _a,
        _la,
        _irow,
        _icol,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _check,
        _y,
        _x,
        _errbuf,
        _ierr,
        _trans._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ipivq[i-1] is a repeated value:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ipivq[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ipivq[i-1] >= 1 and ipivq[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ipivp[i-1] is a repeated value:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ipivp[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ipivp[i-1] >= 1 and ipivp[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr[i-1] is inconsistent with irow:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, idiag[i-1] appears to be incorrect:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[i-1],icol[i-1]) is a duplicate:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr appears to be invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def complex_gen_solve_ilu(method, n, nnz, a, la, irow, icol, ipivp, ipivq, istr, idiag, b, m, tol, maxitn, x):
    u"""
    Solution of complex sparse non-Hermitian linear system, RGMRES, CGS,
    Bi-CGSTAB or TFQMR method, preconditioner computed by
    ``complex_gen_precon_ilu`` (Black Box).

    ``complex_gen_solve_ilu`` solves a complex sparse non-Hermitian
    system of linear equations, represented in coordinate storage
    format, using a restarted generalized minimal residual (RGMRES),
    conjugate gradient squared (CGS), stabilized bi-conjugate gradient
    (BI-CGSTAB), or transpose-free quasi-minimal residual (TFQMR)
    method, with incomplete LU preconditioning.

    For full information please refer to the NAG Library document for
    f11dq

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11dqf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'RGMRES'

            Restarted generalized minimum residual method.

        `method` = 'CGS'

            Conjugate gradient squared method.

        `method` = 'BICGSTAB'

            Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

            Transpose-free quasi-minimal residual method.

    n : int
        n, the order of the matrix A. This **must** be the same value as
        was supplied in the preceding call to
        ``complex_gen_precon_ilu``.

    nnz : int
        The number of nonzero elements in the matrix A. This **must** be
        the same value as was supplied in the preceding call to
        ``complex_gen_precon_ilu``.

    a : complex, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``complex_gen_precon_ilu``.

    la : int
        This **must** be the same value as was supplied in the preceding
        call to ``complex_gen_precon_ilu``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    ipivp : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    ipivq : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    idiag : int, shape (`n`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    b : complex, shape (`n`)
        The right-hand side vector b.

    m : int
        If `method` = 'RGMRES', `m` is the dimension of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method.

        Otherwise, `m` is not referenced.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : complex, shape (`n`), modified in place
        `On entry`: an initial approximation to the solution vector x.

        `On exit`: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` >= 1 and `m` <= min(`n`, *<value>*).

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS' or 'BICGSTAB'.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[i-1],`icol`[i-1]) is a
            duplicate: i = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>*, and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            The CS representation of the preconditioner is invalid.

        (`errno` 6)
            Algorithmic breakdown. A solution is returned, although it
            is possible that it is completely inaccurate.

        (`errno` 7)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

    Notes
    -----
    ``complex_gen_solve_ilu`` solves a complex sparse non-Hermitian
    linear system of equations

        Ax = b,

    using a preconditioned RGMRES (see Saad and Schultz (1986)), CGS
    (see Sonneveld (1989)), BI-CGSTAB(l) (see Van der Vorst (1989) and
    Sleijpen and Fokkema (1993)), or TFQMR (see Freund and Nachtigal
    (1991) and Freund (1993)) method.

    ``complex_gen_solve_ilu`` uses the incomplete LU factorization
    determined by ``complex_gen_precon_ilu`` as the preconditioning
    matrix.
    A call to ``complex_gen_solve_ilu`` must always be preceded by a
    call to ``complex_gen_precon_ilu``.
    Alternative preconditioners for the same storage scheme are
    available by calling ``complex_gen_solve_jacssor``.

    The matrix A, and the preconditioning matrix M, are represented in
    coordinate storage (CS) format (see [the F11 Introduction]) in the
    arrays `a`, `irow` and `icol`, as returned from
    ``complex_gen_precon_ilu``.
    The array `a` holds the nonzero entries in these matrices, while
    `irow` and `icol` hold the corresponding row and column indices.

    ``complex_gen_solve_ilu`` is a Black Box function which calls
    ``complex_gen_basic_setup``, ``complex_gen_basic_solver`` and
    ``complex_gen_basic_diag``.
    If you wish to use an alternative storage scheme, preconditioner, or
    termination criterion, or require additional diagnostic information,
    you should call these underlying functions directly.

    References
    ----------
    Freund, R W, 1993, `A transpose-free quasi-minimal residual
    algorithm for non-Hermitian linear systems`, SIAM J. Sci. Comput.
    (14), 470--482

    Freund, R W and Nachtigal, N, 1991, `QMR: a Quasi-Minimal Residual
    Method for Non-Hermitian Linear Systems`, Numer. Math. (60),
    315--339

    Saad, Y and Schultz, M, 1986, `GMRES: a generalized minimal residual
    algorithm for solving nonsymmetric linear systems`, SIAM J. Sci.
    Statist. Comput. (7), 856--869

    Sleijpen, G L G and Fokkema, D R, 1993, `BiCGSTAB(l) for linear
    equations involving matrices with complex spectrum`, ETNA (1),
    11--32

    Sonneveld, P, 1989, `CGS, a fast Lanczos-type solver for
    nonsymmetric linear systems`, SIAM J. Sci. Statist. Comput. (10),
    36--52

    Van der Vorst, H, 1989, `Bi-CGSTAB, a fast and smoothly converging
    variant of Bi-CG for the solution of nonsymmetric linear systems`,
    SIAM J. Sci. Statist. Comput. (13), 631--644
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_solve_ilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[n].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._EngineIntArrayType.c_type, # Input, idiag[n].
        utils._EngineComplex128ArrayType.c_type, # Input, b[n].
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineComplex128ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_n_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_n_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _idiag = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _b,
        _m,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: m >= 1 and m <= min(n, {:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\' or \'BICGSTAB\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[i-1],icol[i-1]) is a duplicate:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The CS representation of the preconditioner is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99985:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99984:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithmic breakdown. A solution is returned, although',
                    'it is possible that it is completely inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_gen_precon_ssor_solve(trans, n, nnz, a, irow, icol, rdiag, omega, check, y, x):
    u"""
    Solution of linear system involving preconditioning matrix generated
    by applying SSOR to complex sparse non-Hermitian matrix.

    ``complex_gen_precon_ssor_solve`` solves a system of linear
    equations involving the preconditioning matrix corresponding to SSOR
    applied to a complex sparse non-Hermitian matrix, represented in
    coordinate storage format.

    For full information please refer to the NAG Library document for
    f11dr

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11drf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether or not the matrix M is transposed.

        `trans` = 'N'

            Mx = y is solved.

        `trans` = 'T'

            M^Hx = y is solved.

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : complex, shape (`nnz`)
        The nonzero elements in the matrix A, ordered by increasing row
        index, and by increasing column index within each row. Multiple
        entries for the same row and column indices are not permitted.
        The function ``complex_gen_sort`` may be used to order the
        elements in this way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in `a`.

    rdiag : complex, shape (`n`)
        The elements of the diagonal matrix D^-1, where D is the
        diagonal part of A.

    omega : float
        The relaxation parameter omega.

    check : str, length 1
        Specifies whether or not the CS representation of the matrix M
        should be checked.

        `check` = 'C'

            Checks are carried on the values of `n`, `nnz`, `irow`,
            `icol` and `omega`.

        `check` = 'N'

            None of these checks are carried out.

        See also [Further Comments].

    y : complex, shape (`n`)
        The right-hand side vector y.

    x : complex, shape (`n`), modified in place
        `On exit`: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` != 'C' or 'N': `check` = *<value>*.

        (`errno` 1)
            On entry, `trans` != 'N' or 'T': `trans` = *<value>*.

        (`errno` 2)
            On entry, `omega` = *<value>*.

            Constraint: 0.0 < `omega` < 2.0

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 4)
            The matrix A has no diagonal entry in row *<value>*.

    Notes
    -----
    ``complex_gen_precon_ssor_solve`` solves a system of linear
    equations

        Mx = y, or M^Hx = y,

    according to the value of the argument `trans`, where the matrix

        M = 1/omega(2-omega)(D+omega L)D^-1(D+omega U)

    corresponds to symmetric successive-over-relaxation (SSOR) Young
    (1971) applied to a linear system Ax = b, where A is a complex
    sparse non-Hermitian matrix stored in coordinate storage (CS) format
    (see [the F11 Introduction]).

    In the definition of M given above D is the diagonal part of A, L is
    the strictly lower triangular part of A, U is the strictly upper
    triangular part of A, and omega is a user-defined relaxation
    parameter.

    It is envisaged that a common use of
    ``complex_gen_precon_ssor_solve`` will be to carry out the
    preconditioning step required in the application of
    ``complex_gen_basic_solver`` to sparse linear systems.
    ``complex_gen_precon_ssor_solve`` is also used for this purpose by
    the Black Box function ``complex_gen_solve_jacssor``.

    References
    ----------
    Young, D, 1971, `Iterative Solution of Large Linear Systems`,
    Academic Press, New York
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_precon_ssor_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11drft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._EngineComplex128ArrayType.c_type, # Input, rdiag[n].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineComplex128ArrayType.c_type, # Input, y[n].
        utils._EngineComplex128ArrayType.c_type, # Output, x[n].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _rdiag = utils._EngineComplex128ArrayType(
        rdiag, {'fun_name': fun_name, 'entity': 'rdiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _rdiag,
        _omega,
        _check,
        _y,
        _x,
        _iwork,
        _errbuf,
        _ierr,
        _trans._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check != \'C\' or \'N\': check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans != \'N\' or \'T\': trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < omega < 2.0',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def complex_gen_solve_jacssor(method, precon, n, nnz, a, irow, icol, omega, b, m, tol, maxitn, x):
    u"""
    Solution of complex sparse non-Hermitian linear system, RGMRES, CGS,
    Bi-CGSTAB or TFQMR method, Jacobi or SSOR preconditioner Black Box.

    ``complex_gen_solve_jacssor`` solves a complex sparse non-Hermitian
    system of linear equations, represented in coordinate storage
    format, using a restarted generalized minimal residual (RGMRES),
    conjugate gradient squared (CGS), stabilized bi-conjugate gradient
    (BI-CGSTAB), or transpose-free quasi-minimal residual (TFQMR)
    method, without preconditioning, with Jacobi, or with SSOR
    preconditioning.

    For full information please refer to the NAG Library document for
    f11ds

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11dsf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'RGMRES'

            Restarted generalized minimum residual method.

        `method` = 'CGS'

            Conjugate gradient squared method.

        `method` = 'BICGSTAB'

            Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

            Transpose-free quasi-minimal residual method.

    precon : str, length 1
        Specifies the type of preconditioning to be used.

        `precon` = 'N'

            No preconditioning.

        `precon` = 'J'

            Jacobi.

        `precon` = 'S'

            Symmetric successive-over-relaxation (SSOR).

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : complex, shape (`nnz`)
        The nonzero elements of the matrix A, ordered by increasing row
        index, and by increasing column index within each row. Multiple
        entries for the same row and column indices are not permitted.
        The function ``complex_gen_sort`` may be used to order the
        elements in this way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in `a`.

    omega : float
        If `precon` = 'S', `omega` is the relaxation parameter omega to
        be used in the SSOR method. Otherwise `omega` need not be
        initialized and is not referenced.

    b : complex, shape (`n`)
        The right-hand side vector b.

    m : int
        If `method` = 'RGMRES', `m` is the dimension of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method.

        Otherwise, `m` is not referenced.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : complex, shape (`n`), modified in place
        `On entry`: an initial approximation to the solution vector x.

        `On exit`: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: 0 < `m` <= min(`n`, *<value>*).

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS' or 'BICGSTAB'.

        (`errno` 1)
            On entry, `omega` = *<value>*.

            Constraint: 0.0 < `omega` < 2.0

        (`errno` 1)
            On entry, `precon` = *<value>*.

            Constraint: `precon` = 'N', 'J' or 'S'.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            The matrix A has a zero diagonal entry in row *<value>*.

        (`errno` 3)
            The matrix A has no diagonal entry in row *<value>*.

        (`errno` 6)
            Algorithmic breakdown. A solution is returned, although it
            is possible that it is completely inaccurate.

        (`errno` 7)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

    Notes
    -----
    ``complex_gen_solve_jacssor`` solves a complex sparse non-Hermitian
    system of linear equations:

        Ax = b,

    using an RGMRES (see Saad and Schultz (1986)), CGS (see Sonneveld
    (1989)), BI-CGSTAB(l) (see Van der Vorst (1989) and Sleijpen and
    Fokkema (1993)), or TFQMR (see Freund and Nachtigal (1991) and
    Freund (1993)) method.

    ``complex_gen_solve_jacssor`` allows the following choices for the
    preconditioner:

    -   no preconditioning;

    -   Jacobi preconditioning (see Young (1971));

    -   symmetric successive-over-relaxation (SSOR) preconditioning (see
        Young (1971)).

    For incomplete LU (ILU) preconditioning see
    ``complex_gen_solve_ilu``.

    The matrix A is represented in coordinate storage (CS) format (see
    [the F11 Introduction]) in the arrays `a`, `irow` and `icol`.
    The array `a` holds the nonzero entries in the matrix, while `irow`
    and `icol` hold the corresponding row and column indices.

    ``complex_gen_solve_jacssor`` is a Black Box function which calls
    ``complex_gen_basic_setup``, ``complex_gen_basic_solver`` and
    ``complex_gen_basic_diag``.
    If you wish to use an alternative storage scheme, preconditioner, or
    termination criterion, or require additional diagnostic information,
    you should call these underlying functions directly.

    References
    ----------
    Freund, R W, 1993, `A transpose-free quasi-minimal residual
    algorithm for non-Hermitian linear systems`, SIAM J. Sci. Comput.
    (14), 470--482

    Freund, R W and Nachtigal, N, 1991, `QMR: a Quasi-Minimal Residual
    Method for Non-Hermitian Linear Systems`, Numer. Math. (60),
    315--339

    Saad, Y and Schultz, M, 1986, `GMRES: a generalized minimal residual
    algorithm for solving nonsymmetric linear systems`, SIAM J. Sci.
    Statist. Comput. (7), 856--869

    Sleijpen, G L G and Fokkema, D R, 1993, `BiCGSTAB(l) for linear
    equations involving matrices with complex spectrum`, ETNA (1),
    11--32

    Sonneveld, P, 1989, `CGS, a fast Lanczos-type solver for
    nonsymmetric linear systems`, SIAM J. Sci. Statist. Comput. (10),
    36--52

    Van der Vorst, H, 1989, `Bi-CGSTAB, a fast and smoothly converging
    variant of Bi-CG for the solution of nonsymmetric linear systems`,
    SIAM J. Sci. Statist. Comput. (13), 631--644

    Young, D, 1971, `Iterative Solution of Large Linear Systems`,
    Academic Press, New York
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_solve_jacssor'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._EngineComplex128ArrayType.c_type, # Input, b[n].
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineComplex128ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _omega,
        _b,
        _m,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _iwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 0 < m <= min(n, {:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\' or \'BICGSTAB\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < omega < 2.0',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: precon = \'N\', \'J\' or \'S\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has a zero diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99983:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99982:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99979:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithmic breakdown. A solution is returned, although',
                    'it is possible that it is completely inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99980:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_gen_precon_bdilu(n, nnz, a, la, irow, icol, nb, istb, indb, lindb, lfill, dtol, pstrat, milu, ipivp, ipivq, istr, idiag, npivm):
    u"""
    Complex, sparse, non-Hermitian linear system, incomplete LU
    factorization of local or overlapping diagonal blocks.

    ``complex_gen_precon_bdilu`` computes a block diagonal incomplete LU
    factorization of a complex sparse non-Hermitian matrix, represented
    in coordinate storage format.
    The diagonal blocks may be composed of arbitrary rows and the
    corresponding columns, and may overlap.
    This factorization can be used to provide a block Jacobi or additive
    Schwarz preconditioner, for use in combination with
    ``complex_gen_solve_bdilu`` or ``complex_gen_basic_solver``.

    For full information please refer to the NAG Library document for
    f11dt

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11dtf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : complex, shape (`la`), modified in place
        `On entry`: the nonzero elements in the matrix A, ordered by
        increasing row index, and by increasing column index within each
        row. Multiple entries for the same row and column indices are
        not permitted. The function ``complex_gen_sort`` may be used to
        order the elements in this way.

        `On exit`: the first `nnz` entries of `a` contain the nonzero
        elements of A and the next `nnzc` entries contain the elements
        of the matrices C_b, for b = 1,2,...,`nb` stored consecutively.
        Within each block the matrix elements are ordered by increasing
        row index, and by increasing column index within each row.

    la : int
        These arrays must be of sufficient size to store both A (`nnz`
        elements) and C (`nnzc` elements).

        Note: the minimum value for `la` is only appropriate if `lfill`
        and `dtol` are set such that minimal fill-in occurs. If this is
        not the case then we recommend that `la` is set much larger than
        the minimum value indicated in the constraint.

    irow : int, shape (`la`), modified in place
        `On entry`: the row indices of the nonzero elements supplied in
        `a`.

        `On exit`: the row indices of the nonzero elements returned in
        `a`.

    icol : int, shape (`la`), modified in place
        `On entry`: the column indices of the nonzero elements supplied
        in `a`.

        `On exit`: the column indices of the nonzero elements returned
        in `a`.

    nb : int
        The number of diagonal blocks to factorize.

    istb : int, shape (`nb`+1)
        `istb`[b-1], for b = 1,2,...,`nb`, holds the indices in arrays
        `indb`, `ipivp`, `ipivq` and `idiag` that, on successful exit
        from this function, define block b. `istb`[`nb`] holds the sum
        of the number of rows in all blocks plus `istb`[0].

    indb : int, shape (`lindb`)
        `indb` must hold the row indices appearing in each diagonal
        block, stored consecutively. Thus the elements
        `indb`[`istb`[b-1]-1] to `indb`[`istb`[b]-2] are the row indices
        in the bth block, for b = 1,2,...,`nb`.

    lindb : int
        The dimension of the arrays `indb`, `ipivp`, `ipivq` and
        `idiag`.

    lfill : int, shape (`nb`)
        If `lfill`[b-1] >= 0 its value is the maximum level of fill
        allowed in the decomposition of the block b (see [Further
        Comments]). A negative value of `lfill`[b-1] indicates that
        `dtol`[b-1] will be used to control the fill in block b instead.

    dtol : float, shape (`nb`)
        If `lfill`[b-1] < 0 then `dtol`[b-1] is used as a drop tolerance
        in block b to control the fill-in (see [Further Comments]);
        otherwise `dtol`[b-1] is not referenced.

    pstrat : str, length 1, shape (`nb`)
        `pstrat`[b-1], for b = 1,2,...,`nb`, specifies the pivoting
        strategy to be adopted in the block as follows:

        `pstrat`[b-1] = 'N'

            No pivoting is carried out.

        `pstrat`[b-1] = 'U'

            Pivoting is carried out according to the user-defined input
            values of `ipivp` and `ipivq`.

        `pstrat`[b-1] = 'P'

            Partial pivoting by columns for stability is carried out.

        `pstrat`[b-1] = 'C'

            Complete pivoting by rows for sparsity, and by columns for
            stability, is carried out.

        Suggested value: `pstrat`[b-1] = 'C', for b = 1,2,...,`nb`.

    milu : str, length 1, shape (`nb`)
        `milu`[b-1], for b = 1,2,...,`nb`, indicates whether or not the
        factorization in block b should be modified to preserve row-sums
        (see [Further Comments]).

        `milu`[b-1] = 'M'

            The factorization is modified.

        `milu`[b-1] = 'N'

            The factorization is not modified.

    ipivp : int, shape (`lindb`), modified in place
        `On entry`: if `pstrat`[b-1] = 'U', `ipivp`[`istb`[b-1]+k-2] and
        `ipivq`[`istb`[b-1]+k-2] must specify the row and column indices
        of the element used as a pivot at elimination stage k of the
        factorization of the block. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        `On exit`: the row and column indices of the pivot elements,
        arranged consecutively for each block, as for `indb`. If
        `ipivp`[`istb`[b-1]+k-2] = i and `ipivq`[`istb`[b-1]+k-2] = j,
        the element in row i and column j of A_b was used as the pivot
        at elimination stage k.

    ipivq : int, shape (`lindb`), modified in place
        `On entry`: if `pstrat`[b-1] = 'U', `ipivp`[`istb`[b-1]+k-2] and
        `ipivq`[`istb`[b-1]+k-2] must specify the row and column indices
        of the element used as a pivot at elimination stage k of the
        factorization of the block. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        `On exit`: the row and column indices of the pivot elements,
        arranged consecutively for each block, as for `indb`. If
        `ipivp`[`istb`[b-1]+k-2] = i and `ipivq`[`istb`[b-1]+k-2] = j,
        the element in row i and column j of A_b was used as the pivot
        at elimination stage k.

    istr : int, shape (`lindb`+1), modified in place
        `On exit`: `istr`[`istb`[b-1]+k-2], gives the index in the
        arrays `a`, `irow` and `icol` of row k of the matrix C_b, for k
        = 1,2,...,`istb`[b]-`istb`[b-1], for b = 1,2,...,`nb`.

        `istr`[`istb`[`nb`]-1] contains `nnz`+`nnzc`+1.

    idiag : int, shape (`lindb`), modified in place
        `On exit`: `idiag`[`istb`[b-1]+k-2], gives the index in the
        arrays `a`, `irow` and `icol` of the diagonal element in row k
        of the matrix C_b, for k = 1,2,...,`istb`[b]-`istb`[b-1], for b
        = 1,2,...,`nb`.

    npivm : int, shape (`nb`), modified in place
        `On exit`: if `npivm`[b-1] > 0 it gives the number of pivots
        which were modified during the factorization to ensure that M_b
        exists.

        If `npivm`[b-1] = -1 no pivot modifications were required, but a
        local restart occurred (see [Further Comments for
        complex_gen_precon_ilu]). The quality of the preconditioner will
        generally depend on the returned values of `npivm`[b-1], for b =
        1,2,...,`nb`.

        If `npivm`[b-1] is large, for some b, the preconditioner may not
        be satisfactory.

        In this case it may be advantageous to call
        ``complex_gen_precon_bdilu`` again with an increased value of
        `lfill`[b-1], a reduced value of `dtol`[b-1], or `pstrat`[b-1] =
        'C'.

    Returns
    -------
    nnzc : int
        The sum total number of nonzero elements in the matrices C_b,
        for b = 1,2,...,`nb`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nb` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `nb` <= `n`.

        (`errno` 1)
            On entry, `istb`[0] = *<value>*.

            Constraint: `istb`[0] >= 1.

        (`errno` 1)
            On entry, for b = *<value>*, `istb`[b] = *<value>* and
            `istb`[b-1] = *<value>*.

            Constraint: `istb`[b] > `istb`[b-1], for b = 1,2,...,`nb`.

        (`errno` 1)
            On entry, `indb`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `indb`[m-1] <= `n`, for m =
            1,2,...,`istb`[`nb`]-1

        (`errno` 1)
            On entry, `lindb` = *<value>*, `istb`[`nb`]-1 = *<value>*
            and `nb` = *<value>*.

            Constraint: `lindb` >= `istb`[`nb`]-1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `dtol`[*<value>*] = *<value>*.

            Constraint: `dtol`[b-1] >= 0.0, for b = 1,2,...,`nb`.

        (`errno` 1)
            On entry, `pstrat`[*<value>*] = *<value>*.

            Constraint: `pstrat`[b-1] = 'N', 'U', 'P' or 'C' for all b.

        (`errno` 1)
            On entry, `milu`[*<value>*] = *<value>*.

            Constraint: `milu`[b-1] = 'M' or 'N' for all b.

        (`errno` 2)
            On entry, `irow`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `irow`[i-1] <= `n`, for i = 1,2,...,`nnz`.

        (`errno` 2)
            On entry, `icol`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `icol`[j-1] <= `n`, for j = 1,2,...,`nnz`.

        (`errno` 2)
            On entry, element *<value>* of `a` was out of order.

        (`errno` 2)
            On entry, location *<value>* of (`irow`,`icol`) was a
            duplicate.

        (`errno` 3)
            On entry, the user-supplied value of `ipivp` for block
            *<value>* lies outside its range.

        (`errno` 3)
            On entry, the user-supplied value of `ipivp` for block
            *<value>* was repeated.

        (`errno` 3)
            On entry, the user-supplied value of `ipivq` for block
            *<value>* lies outside its range.

        (`errno` 3)
            On entry, the user-supplied value of `ipivq` for block
            *<value>* was repeated.

        (`errno` 4)
            The number of nonzero entries in the decomposition is too
            large.

            The decomposition has been terminated before completion.

            Either increase `la`, or reduce the fill by reducing
            `lfill`, or increasing `dtol`.

    Notes
    -----
    ``complex_gen_precon_bdilu`` computes an incomplete LU factorization
    (see Meijerink and Van der Vorst (1977) and Meijerink and Van der
    Vorst (1981)) of the (possibly overlapping) diagonal blocks A_b, b =
    1,2,...,`nb`, of a complex sparse non-Hermitian n*n matrix A.
    The factorization is intended primarily for use as a block Jacobi or
    additive Schwarz preconditioner (see Saad (1996)), with one of the
    iterative solvers ``complex_gen_solve_bdilu`` and
    ``complex_gen_basic_solver``.

    The `nb` diagonal blocks need not consist of consecutive rows and
    columns of A, but may be composed of arbitrarily indexed rows, and
    the corresponding columns, as defined in the arguments `indb` and
    `istb`.
    Any given row or column index may appear in more than one diagonal
    block, resulting in overlap.
    Each diagonal block A_b, b = 1,2,...,`nb`, is factorized as:

        A_b = M_b+R_b

    where

        M_b = P_b L_b D_b U_b Q_b

    and L_b is lower triangular with unit diagonal elements, D_b is
    diagonal, U_b is upper triangular with unit diagonals, P_b and Q_b
    are permutation matrices, and R_b is a remainder matrix.

    The amount of fill-in occurring in the factorization of block b can
    vary from zero to complete fill, and can be controlled by specifying
    either the maximum level of fill `lfill`[b-1], or the drop tolerance
    `dtol`[b-1].

    The parameter `pstrat`[b-1] defines the pivoting strategy to be used
    in block b.
    The options currently available are no pivoting, user-defined
    pivoting, partial pivoting by columns for stability, and complete
    pivoting by rows for sparsity and by columns for stability.
    The factorization may optionally be modified to preserve the
    row-sums of the original block matrix.

    The sparse matrix A is represented in coordinate storage (CS) format
    (see [the F11 Introduction]).
    The array `a` stores all the nonzero elements of the matrix A, while
    arrays `irow` and `icol` store the corresponding row and column
    indices respectively.
    Multiple nonzero elements may not be specified for the same row and
    column index.

    The preconditioning matrices M_b, b = 1,2,...,`nb`, are returned in
    terms of the CS representations of the matrices

        C_b = L_b+D^-1_b+U_b-2I.

    References
    ----------
    Meijerink, J and Van der Vorst, H, 1977, `An iterative solution
    method for linear systems of which the coefficient matrix is a
    symmetric M-matrix`, Math. Comput. (31), 148--162

    Meijerink, J and Van der Vorst, H, 1981, `Guidelines for the usage
    of incomplete decompositions in solving sets of linear equations as
    they occur in practical problems`, J. Comput. Phys. (44), 134--155

    Saad, Y, 1996, `Iterative Methods for Sparse Linear Systems`, PWS
    Publishing Company, Boston, MA
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_precon_bdilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dtft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input/Output, irow[la].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[la].
        utils._BYREF_INT, # Input, nb.
        utils._EngineIntArrayType.c_type, # Input, istb[nb+1].
        utils._EngineIntArrayType.c_type, # Input, indb[lindb].
        utils._BYREF_INT, # Input, lindb.
        utils._EngineIntArrayType.c_type, # Input, lfill[nb].
        utils._EngineFloat64ArrayType.c_type, # Input, dtol[nb].
        utils._EngineCharArrayType.c_type, # Input, pstrat(len=1)[nb].
        utils._EngineCharArrayType.c_type, # Input, milu(len=1)[nb].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[lindb].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[lindb].
        utils._EngineIntArrayType.c_type, # Output, istr[lindb+1].
        utils._EngineIntArrayType.c_type, # Output, idiag[lindb].
        utils._BYREF_INT, # Output, nnzc.
        utils._EngineIntArrayType.c_type, # Output, npivm[nb].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._BYREF_INT, # Input, liwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pstrat.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of milu.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _nb = utils._EngineIntScalarType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
    ).to_ctypes()
    _nb_py = utils._EngineIntScalarType.to_py(_nb)
    _istb = utils._EngineIntArrayType(
        istb, {'fun_name': fun_name, 'entity': 'istb'},
        exp_shape=(_nb_py+1,),
    ).to_ctypes()
    _lindb = utils._EngineIntScalarType(
        lindb, {'fun_name': fun_name, 'entity': 'lindb'},
    ).to_ctypes()
    _lindb_py = utils._EngineIntScalarType.to_py(_lindb)
    _indb = utils._EngineIntArrayType(
        indb, {'fun_name': fun_name, 'entity': 'indb'},
        exp_shape=(_lindb_py,),
    ).to_ctypes()
    _lfill = utils._EngineIntArrayType(
        lfill, {'fun_name': fun_name, 'entity': 'lfill'},
        exp_shape=(_nb_py,),
    ).to_ctypes()
    _dtol = utils._EngineFloat64ArrayType(
        dtol, {'fun_name': fun_name, 'entity': 'dtol'},
        exp_shape=(_nb_py,),
    ).to_ctypes()
    _pstrat_cls = utils._EngineCharArrayType(
        pstrat, {'fun_name': fun_name, 'entity': 'pstrat'}, exp_chrlen=1,
        exp_shape=(_nb_py,),
    )
    _pstrat = _pstrat_cls.to_ctypes()
    _pstrat_chrlen = _pstrat_cls.chrlen
    _milu_cls = utils._EngineCharArrayType(
        milu, {'fun_name': fun_name, 'entity': 'milu'}, exp_chrlen=1,
        exp_shape=(_nb_py,),
    )
    _milu = _milu_cls.to_ctypes()
    _milu_chrlen = _milu_cls.chrlen
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_lindb_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_lindb_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_lindb_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _idiag_cls = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_lindb_py,),
        check_contents=False,
    )
    _idiag = _idiag_cls.to_ctypes(can_reference_contents=False)
    _nnzc = utils._EngineIntScalarType.empty_ctypes()
    _npivm_cls = utils._EngineIntArrayType(
        npivm, {'fun_name': fun_name, 'entity': 'npivm'},
        exp_shape=(_nb_py,),
        check_contents=False,
    )
    _npivm = _npivm_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _liwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _nb,
        _istb,
        _indb,
        _lindb,
        _lfill,
        _dtol,
        _pstrat,
        _milu,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _nnzc,
        _npivm,
        _iwork,
        _liwork,
        _errbuf,
        _ierr,
        _pstrat_chrlen,
        _milu_chrlen,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nb = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= nb <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1012:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istb[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: istb[0] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1013:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, for b = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[b] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'istb[b-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: istb[b] > istb[b-1], for b = 1,2,...,nb.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1014:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, indb[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= indb[m-1] <= n, for m = 1,2,...,istb[nb]-1',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1015:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lindb = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[nb]-1 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lindb >= istb[nb]-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1016:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1017:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1018:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1020:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dtol[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: dtol[b-1] >= 0.0, for b = 1,2,...,nb.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1021:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pstrat[{:s}] = \'{:s}\'.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'Constraint: pstrat[b-1] = \'N\', \'U\', \'P\' or \'C\' for all b.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1022:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, milu[{:s}] = \'{:s}\'.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'Constraint: milu[b-1] = \'M\' or \'N\' for all b.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irow[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irow[i-1] <= n, for i = 1,2,...,nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icol[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= icol[j-1] <= n, for j = 1,2,...,nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1112:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, element {:s} of a was out of order.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1113:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, location {:s} of (irow,icol) was a duplicate.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1210:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivp for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'lies outside its range.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1211:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivp for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'was repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1212:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivq for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'lies outside its range.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1213:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivq for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'was repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1310:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of nonzero entries in the decomposition is too large.',
                    'The decomposition has been terminated before completion.',
                    'Either increase la, or reduce the fill by reducing lfill,',
                    'or increasing dtol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _istr_cls.refresh_py(_istr)
        _idiag_cls.refresh_py(_idiag)
        _npivm_cls.refresh_py(_npivm)
        _nnzc_py = utils._EngineIntScalarType.to_py(_nnzc)
        return_data = _nnzc_py
    _err_class.manage()
    return return_data

def complex_gen_solve_bdilu(method, n, nnz, a, la, irow, icol, nb, istb, indb, lindb, ipivp, ipivq, istr, idiag, b, m, tol, maxitn, x):
    u"""
    Solution of complex, sparse, non-Hermitian linear system, RGMRES,
    CGS, Bi-CGSTAB or TFQMR method, incomplete LU block diagonal
    preconditioner computed by ``complex_gen_precon_bdilu``.

    ``complex_gen_solve_bdilu`` solves a complex sparse non-Hermitian
    system of linear equations, represented in coordinate storage
    format, using a restarted generalized minimal residual (RGMRES),
    conjugate gradient squared (CGS), stabilized bi-conjugate gradient
    (BI-CGSTAB), or transpose-free quasi-minimal residual (TFQMR)
    method, with block Jacobi or additive Schwarz preconditioning.

    For full information please refer to the NAG Library document for
    f11du

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11duf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'RGMRES'

            Restarted generalized minimum residual method.

        `method` = 'CGS'

            Conjugate gradient squared method.

        `method` = 'BICGSTAB'

            Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

            Transpose-free quasi-minimal residual method.

    n : int
        The value of `n` as supplied to a previous call to
        ``complex_gen_precon_bdilu``

    nnz : int
        The value of `nnz` as supplied to a previous call to
        ``complex_gen_precon_bdilu``

    a : complex, shape (`la`)
        The value of `a` returned by a previous call to
        ``complex_gen_precon_bdilu``

    la : int
        The value of `la` as supplied to a previous call to
        ``complex_gen_precon_bdilu``

    irow : int, shape (`la`)
        The value of `irow` returned by a previous call to
        ``complex_gen_precon_bdilu``

    icol : int, shape (`la`)
        The value of `icol` returned by a previous call to
        ``complex_gen_precon_bdilu``

    nb : int
        The value of `nb` as supplied to a previous call to
        ``complex_gen_precon_bdilu``

    istb : int, shape (`nb`+1)
        The value of `istb` as supplied to a previous call to
        ``complex_gen_precon_bdilu``

    indb : int, shape (`lindb`)
        The value of `indb` as supplied to a previous call to
        ``complex_gen_precon_bdilu``

    lindb : int
        The value of `lindb` as supplied to a previous call to
        ``complex_gen_precon_bdilu``

    ipivp : int, shape (`lindb`), modified in place
        The value of `ipivp` returned by a previous call to
        ``complex_gen_precon_bdilu``

    ipivq : int, shape (`lindb`), modified in place
        The value of `ipivq` returned by a previous call to
        ``complex_gen_precon_bdilu``

    istr : int, shape (`lindb`+1)
        The value of `istr` returned by a previous call to
        ``complex_gen_precon_bdilu``

    idiag : int, shape (`lindb`)
        The value of `idiag` returned by a previous call to
        ``complex_gen_precon_bdilu``

    b : complex, shape (`n`)
        The right-hand side vector b.

    m : int
        If `method` = 'RGMRES', `m` is the dimension of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method.

        Otherwise, `m` is not referenced.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),sqrt(n)epsilon) is
        used, where epsilon is the machine precision. Otherwise tau =
        max(`tol`,10 epsilon, sqrt(n)epsilon) is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : complex, shape (`n`), modified in place
        `On entry`: an initial approximation to the solution vector x.

        `On exit`: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nb` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `nb` <= `n`.

        (`errno` 1)
            On entry, `istb`[0] = *<value>*.

            Constraint: `istb`[0] >= 1.

        (`errno` 1)
            On entry, for b = *<value>*, `istb`[b] = *<value>* and
            `istb`[b-1] = *<value>*.

            Constraint: `istb`[b] > `istb`[b-1], for b = 1,2,...,`nb`.

        (`errno` 1)
            On entry, `indb`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `indb`[m-1] <= `n`, for m =
            1,2,...,`istb`[`nb`]-1

        (`errno` 1)
            On entry, `lindb` = *<value>*, `istb`[`nb`]-1 = *<value>*
            and `nb` = *<value>*.

            Constraint: `lindb` >= `istb`[`nb`]-1.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS' or 'BICGSTAB'.

        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: if `method` = 'RGMRES', 1 <= `m` <= min(`n`,
            *<value>*).

            If `method` = 'BICGSTAB', 1 <= `m` <= min(`n`, *<value>*).

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 2)
            On entry, `irow`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `irow`[i-1] <= `n`, for i = 1,2,...,`nnz`.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``complex_gen_precon_bdilu`` and
            ``complex_gen_solve_bdilu``.

        (`errno` 2)
            On entry, `icol`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `icol`[i-1] <= `n`, for i = 1,2,...,`nnz`.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``complex_gen_precon_bdilu`` and
            ``complex_gen_solve_bdilu``.

        (`errno` 2)
            On entry, element *<value>* of `a` was out of order.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``complex_gen_precon_bdilu`` and
            ``complex_gen_solve_bdilu``.

        (`errno` 2)
            On entry, location *<value>* of (`irow`,`icol`) was a
            duplicate.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``complex_gen_precon_bdilu`` and
            ``complex_gen_solve_bdilu``.

        (`errno` 3)
            The CS representation of the preconditioner is invalid.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``complex_gen_precon_bdilu`` and
            ``complex_gen_solve_bdilu``.

        (`errno` 4)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            Algorithmic breakdown. A solution is returned, although it
            is possible that it is completely inaccurate.

    Notes
    -----
    ``complex_gen_solve_bdilu`` solves a complex sparse non-Hermitian
    linear system of equations

        Ax = b,

    using a preconditioned RGMRES (see Saad and Schultz (1986)), CGS
    (see Sonneveld (1989)), BI-CGSTAB(l) (see Van der Vorst (1989) and
    Sleijpen and Fokkema (1993)), or TFQMR (see Freund and Nachtigal
    (1991) and Freund (1993)) method.

    ``complex_gen_solve_bdilu`` uses the incomplete (possibly
    overlapping) block LU factorization determined by
    ``complex_gen_precon_bdilu`` as the preconditioning matrix.
    A call to ``complex_gen_solve_bdilu`` must always be preceded by a
    call to ``complex_gen_precon_bdilu``.
    Alternative preconditioners for the same storage scheme are
    available by calling ``complex_gen_solve_ilu`` or
    ``complex_gen_solve_jacssor``.

    The matrix A, and the preconditioning matrix M, are represented in
    coordinate storage (CS) format (see [the F11 Introduction]) in the
    arrays `a`, `irow` and `icol`, as returned from
    ``complex_gen_precon_bdilu``.
    The array `a` holds the nonzero entries in these matrices, while
    `irow` and `icol` hold the corresponding row and column indices.

    ``complex_gen_solve_bdilu`` is a Black Box function which calls
    ``complex_gen_basic_setup``, ``complex_gen_basic_solver`` and
    ``complex_gen_basic_diag``.
    If you wish to use an alternative storage scheme, preconditioner, or
    termination criterion, or require additional diagnostic information,
    you should call these underlying functions directly.

    References
    ----------
    Freund, R W, 1993, `A transpose-free quasi-minimal residual
    algorithm for non-Hermitian linear systems`, SIAM J. Sci. Comput.
    (14), 470--482

    Freund, R W and Nachtigal, N, 1991, `QMR: a Quasi-Minimal Residual
    Method for Non-Hermitian Linear Systems`, Numer. Math. (60),
    315--339

    Saad, Y and Schultz, M, 1986, `GMRES: a generalized minimal residual
    algorithm for solving nonsymmetric linear systems`, SIAM J. Sci.
    Statist. Comput. (7), 856--869

    Sleijpen, G L G and Fokkema, D R, 1993, `BiCGSTAB(l) for linear
    equations involving matrices with complex spectrum`, ETNA (1),
    11--32

    Sonneveld, P, 1989, `CGS, a fast Lanczos-type solver for
    nonsymmetric linear systems`, SIAM J. Sci. Statist. Comput. (10),
    36--52

    Van der Vorst, H, 1989, `Bi-CGSTAB, a fast and smoothly converging
    variant of Bi-CG for the solution of nonsymmetric linear systems`,
    SIAM J. Sci. Statist. Comput. (13), 631--644
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_solve_bdilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11duft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._BYREF_INT, # Input, nb.
        utils._EngineIntArrayType.c_type, # Input, istb[nb+1].
        utils._EngineIntArrayType.c_type, # Input, indb[lindb].
        utils._BYREF_INT, # Input, lindb.
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[lindb].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[lindb].
        utils._EngineIntArrayType.c_type, # Input, istr[lindb+1].
        utils._EngineIntArrayType.c_type, # Input, idiag[lindb].
        utils._EngineComplex128ArrayType.c_type, # Input, b[n].
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineComplex128ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _nb = utils._EngineIntScalarType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
    ).to_ctypes()
    _nb_py = utils._EngineIntScalarType.to_py(_nb)
    _istb = utils._EngineIntArrayType(
        istb, {'fun_name': fun_name, 'entity': 'istb'},
        exp_shape=(_nb_py+1,),
    ).to_ctypes()
    _lindb = utils._EngineIntScalarType(
        lindb, {'fun_name': fun_name, 'entity': 'lindb'},
    ).to_ctypes()
    _lindb_py = utils._EngineIntScalarType.to_py(_lindb)
    _indb = utils._EngineIntArrayType(
        indb, {'fun_name': fun_name, 'entity': 'indb'},
        exp_shape=(_lindb_py,),
    ).to_ctypes()
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_lindb_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_lindb_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_lindb_py+1,),
    ).to_ctypes()
    _idiag = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_lindb_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _nb,
        _istb,
        _indb,
        _lindb,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _b,
        _m,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1012:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1013:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1015:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nb = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= nb <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1016:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istb[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: istb[0] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1017:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, for b = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[b] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'istb[b-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: istb[b] > istb[b-1], for b = 1,2,...,nb.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1018:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, indb[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= indb[m-1] <= n, for m = 1,2,...,istb[nb]-1',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1019:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lindb = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[nb]-1 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lindb >= istb[nb]-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1020:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\' or \'BICGSTAB\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1021:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if method = \'RGMRES\', 1 <= m <= min(n, {:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'If method = \'BICGSTAB\', 1 <= m <= min(n, {:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1022:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1023:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irow[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irow[i-1] <= n, for i = 1,2,...,nnz.',
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to complex_gen_precon_bdilu and complex_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icol[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= icol[i-1] <= n, for i = 1,2,...,nnz.',
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to complex_gen_precon_bdilu and complex_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1112:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, element {:s} of a was out of order.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to complex_gen_precon_bdilu and complex_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1113:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, location {:s} of (irow,icol) was a duplicate.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to complex_gen_precon_bdilu and complex_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1210:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The CS representation of the preconditioner is invalid.',
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to complex_gen_precon_bdilu and complex_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1310:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1410:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 1510:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithmic breakdown. A solution is returned, although',
                    'it is possible that it is completely inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
    _err_class.manage()
    return return_data

def complex_gen_precon_jacobi(store, trans, init, niter, n, nnz, a, irow, icol, check, b, x, diag):
    u"""
    Complex, sparse, Hermitian or non-Hermitian, linear systems, line
    Jacobi preconditioner.

    ``complex_gen_precon_jacobi`` computes the **approximate** solution
    of a complex, Hermitian or non-Hermitian, sparse system of linear
    equations applying a number of Jacobi iterations.
    It is expected that ``complex_gen_precon_jacobi`` will be used as a
    preconditioner for the iterative solution of complex sparse systems
    of equations.

    For full information please refer to the NAG Library document for
    f11dx

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11dxf.html

    Parameters
    ----------
    store : str, length 1
        Specifies whether the matrix A is stored using symmetric
        coordinate storage (SCS) (applicable only to a Hermitian matrix
        A) or coordinate storage (CS) (applicable to both Hermitian and
        non-Hermitian matrices).

        `store` = 'N'

            The complete matrix A is stored in CS format.

        `store` = 'S'

            The lower triangle of the Hermitian matrix A is stored in
            SCS format.

    trans : str, length 1
        If `store` = 'N', specifies whether the approximate solution of
        Ax = b or of A^Hx = b is required.

        `trans` = 'N'

            The approximate solution of Ax = b is calculated.

        `trans` = 'T'

            The approximate solution of A^Hx = b is calculated.

        Suggested value: if the matrix A is Hermitian and stored in CS
        format, it is recommended that `trans` = 'N' for reasons of
        efficiency.

    init : str, length 1
        On first entry, `init` should be set to 'I', unless the diagonal
        elements of A are already stored in the array `diag`. If `diag`
        already contains the diagonal of A, it must be set to 'N'.

        `init` = 'N'

            `diag` must contain the diagonal of A.

        `init` = 'I'

            `diag` will store the diagonal of A on exit.

        Suggested value: `init` = 'I' on first entry; `init` = 'N',
        subsequently, unless `diag` has been overwritten.

    niter : int
        The number of Jacobi iterations requested.

    n : int
        n, the order of the matrix A.

    nnz : int
        If `store` = 'N', the number of nonzero elements in the matrix
        A.

        If `store` = 'S', the number of nonzero elements in the lower
        triangle of the matrix A.

    a : complex, shape (`nnz`)
        If `store` = 'N', the nonzero elements in the matrix A (CS
        format).

        If `store` = 'S', the nonzero elements in the lower triangle of
        the matrix A (SCS format).

        In both cases, the elements of either A or of its lower triangle
        must be ordered by increasing row index and by increasing column
        index within each row.

        Multiple entries for the same row and columns indices are not
        permitted.

        The function ``complex_gen_sort`` or ``complex_herm_sort`` may
        be used to reorder the elements in this way for CS and SCS
        storage, respectively.

    irow : int, shape (`nnz`)
        If `store` = 'N', the row indices of the nonzero elements
        supplied in `a`.

        If `store` = 'S', the row and column indices of the nonzero
        elements of the lower triangle of the matrix A supplied in `a`.

    icol : int, shape (`nnz`)
        If `store` = 'N', the column indices of the nonzero elements
        supplied in `a`.

        If `store` = 'S', the row and column indices of the nonzero
        elements of the lower triangle of the matrix A supplied in `a`.

    check : str, length 1
        Specifies whether or not the CS or SCS representation of the
        matrix A should be checked.

        `check` = 'C'

            Checks are carried out on the values of `n`, `nnz`, `irow`,
            `icol`; if `init` = 'N', `diag` is also checked.

        `check` = 'N'

            None of these checks are carried out.

        See also [Further Comments].

    b : complex, shape (`n`)
        The right-hand side vector b.

    x : complex, shape (`n`), modified in place
        `On exit`: the approximate solution vector x_`niter`.

    diag : complex, shape (`n`), modified in place
        `On entry`: if `init` = 'N', the diagonal elements of A.

        `On exit`: if `init` = 'N', unchanged on exit.

        If `init` = 'I', the diagonal elements of A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `niter` = *<value>*.

            Constraint: `niter` >= 1.

        (`errno` 1)
            On entry, `check` != 'N' or 'C': `check` = *<value>*.

        (`errno` 1)
            On entry, `init` != 'N' or 'I': `init` = *<value>*.

        (`errno` 1)
            On entry, `trans` != 'N' or 'T': `trans` = *<value>*.

        (`errno` 1)
            On entry, `store` != 'N' or 'S': `store` = *<value>*.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 3)
            On entry, I = *<value>*, `irow`[I-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[I-1] >= 1 and `irow`[I-1] <= `n`.

        (`errno` 4)
            On entry, the element `diag`[I-1] is zero: I = *<value>*.

        (`errno` 5)
            On entry, the diagonal element of the Ith row is zero or
            missing: I = *<value>*.

    Notes
    -----
    ``complex_gen_precon_jacobi`` computes the **approximate** solution
    of the complex sparse system of linear equations Ax = b using
    `niter` iterations of the Jacobi algorithm (see also Golub and Van
    Loan (1996) and Young (1971)):

        x_{k+1} = x_k+D^-1(b-Ax_k)

    where k = 1,2,...,`niter` and x_0 = 0.

    ``complex_gen_precon_jacobi`` can be used both for non-Hermitian and
    Hermitian systems of equations.
    For Hermitian matrices, either all nonzero elements of the matrix A
    can be supplied using coordinate storage (CS), or only the nonzero
    elements of the lower triangle of A, using symmetric coordinate
    storage (SCS) (see [the F11 Introduction]).

    It is expected that ``complex_gen_precon_jacobi`` will be used as a
    preconditioner for the iterative solution of complex sparse systems
    of equations, using either the suite comprising the functions
    ``complex_herm_basic_setup``, ``complex_herm_basic_solver`` and
    ``complex_herm_basic_diag``, for Hermitian systems, or the suite
    comprising the functions ``complex_gen_basic_setup``,
    ``complex_gen_basic_solver`` and ``complex_gen_basic_diag``, for
    non-Hermitian systems of equations.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Young, D, 1971, `Iterative Solution of Large Linear Systems`,
    Academic Press, New York
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_precon_jacobi'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dxft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, store(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, init(len=1).
        utils._BYREF_INT, # Input, niter.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineComplex128ArrayType.c_type, # Input, b[n].
        utils._EngineComplex128ArrayType.c_type, # Output, x[n].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, diag[n].
        utils._EngineComplex128ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of store.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of init.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _store = utils._EngineCharScalarType(
        store, {'fun_name': fun_name, 'entity': 'store'}, exp_chrlen=1,
    ).to_ctypes()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _init = utils._EngineCharScalarType(
        init, {'fun_name': fun_name, 'entity': 'init'}, exp_chrlen=1,
    ).to_ctypes()
    _niter = utils._EngineIntScalarType(
        niter, {'fun_name': fun_name, 'entity': 'niter'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _diag_cls = utils._EngineComplex128ArrayType(
        diag, {'fun_name': fun_name, 'entity': 'diag'},
        exp_shape=(_n_py,),
    )
    _diag = _diag_cls.to_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _store,
        _trans,
        _init,
        _niter,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _check,
        _b,
        _x,
        _diag,
        _work,
        _errbuf,
        _ierr,
        _store._length_,
        _trans._length_,
        _init._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, niter = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: niter >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check != \'N\' or \'C\': check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, init != \'N\' or \'I\': init = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans != \'N\' or \'T\': trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, store != \'N\' or \'S\': store = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, icol[I-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'irow[I-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[I-1] >= 1 and irow[I-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the element diag[I-1] is zero:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the diagonal element of the Ith row is zero or missing:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _diag_cls.refresh_py(_diag)
        return_data = None
    _err_class.manage()
    return return_data

def real_symm_basic_setup(method, precon, sigcmp, norm, weight, iterm, n, tol, maxitn, anorm, sigmax, sigtol, maxits, monit):
    u"""
    Real sparse symmetric linear systems, setup for
    ``real_symm_basic_solver``.

    ``real_symm_basic_setup`` is a setup function, the first in a suite
    of three functions for the iterative solution of a symmetric system
    of simultaneous linear equations. ``real_symm_basic_setup`` must be
    called before the iterative solver, ``real_symm_basic_solver``.
    The third function in the suite, ``real_symm_basic_diag``, can be
    used to return additional information about the computation.

    These three functions are suitable for the solution of large sparse
    symmetric systems of equations.

    For full information please refer to the NAG Library document for
    f11gd

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11gdf.html

    Parameters
    ----------
    method : str
        The iterative method to be used.

        `method` = 'CG'

            Conjugate gradient method (CG).

        `method` = 'SYMMLQ'

            Lanczos method (SYMMLQ).

        `method` = 'MINRES'

            Minimum residual method (MINRES).

    precon : str, length 1
        Determines whether preconditioning is used.

        `precon` = 'N'

            No preconditioning.

        `precon` = 'P'

            Preconditioning.

    sigcmp : str, length 1
        Determines whether an estimate of sigma_1(A_) = ||E^-1AE^-T||_2,
        the largest singular value of the preconditioned matrix of the
        coefficients, is to be computed using the bisection method on
        the sequence of tridiagonal matrices {T_k} generated during the
        iteration. Note that A_= A when a preconditioner is not used.

        If `sigmax` > 0.0 (see `sigmax`), i.e., when sigma_1(A_) is
        supplied, the value of `sigcmp` is ignored.

        `sigcmp` = 'S'

            sigma_1(A_) is to be computed using the bisection method.

        `sigcmp` = 'N'

            The bisection method is not used.

            If the termination criterion (3) is used, requiring
            sigma_1(A_), an inexpensive estimate is computed and used
            (see [Notes]).

        It is not used if `method` = 'MINRES'.

        Suggested value: `sigcmp` = 'N'.

    norm : str, length 1
        If `method` = 'CG' or 'SYMMLQ', `norm` defines the matrix and
        vector norm to be used in the termination criteria.

        `norm` = '1'

            Use the l_1 norm.

        `norm` = 'I'

            Use the l_infinity norm.

        `norm` = '2'

            Use the l_2 norm.

        It has no effect if `method` = 'MINRES'.

        Suggested value:

            if `iterm` = 1, `norm` = 'I'

            if `iterm` = 2, `norm` = '2'

    weight : str, length 1
        Specifies whether a vector w of user-supplied weights is to be
        used in the vector norms used in the computation of termination
        criterion (2) (`iterm` = 1): ||v||_p^(w) = ||v^(w)||_p, where
        v_i^(w) = w_i v_i, for i = 1,2,...,n. The suffix p = 1,2,
        infinity denotes the vector norm used, as specified by the
        argument `norm`. Note that weights cannot be used when `iterm` =
        2, i.e., when criterion (3) is used.

        `weight` = 'W'

            User-supplied weights are to be used and must be supplied on
            initial entry to ``real_symm_basic_solver``.

        `weight` = 'N'

            All weights are implicitly set equal to one. Weights do not
            need to be supplied on initial entry to
            ``real_symm_basic_solver``.

        It has no effect if `method` = 'MINRES'.

        Suggested value: `weight` = 'N'.

    iterm : int
        Defines the termination criterion to be used.

        `iterm` = 1

            Use the termination criterion defined in (2) (both conjugate
            gradient and Lanczos (SYMMLQ) methods).

        `iterm` = 2

            Use the termination criterion defined in (3) (Lanczos method
            (SYMMLQ) only).

        It has no effect if `method` = 'MINRES'.

        Suggested value: `iterm` = 1.

    n : int
        n, the order of the matrix A.

    tol : float
        The tolerance tau for the termination criterion.

        If `tol` <= 0.0, tau = max(sqrt(epsilon), sqrt(n)epsilon) is
        used, where epsilon is the machine precision.

        Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon) is used.

    maxitn : int
        The maximum number of iterations.

    anorm : float
        If `anorm` > 0.0, the value of ||A||_p to be used in the
        termination criterion (2) (`iterm` = 1).

        If `anorm` <= 0.0, `iterm` = 1 and `norm` = '1' or 'I', ||A||_1
        = ||A||_infinity is estimated internally by
        ``real_symm_basic_solver``.

        If `iterm` = 2, `anorm` is not referenced.

        It has no effect if `method` = 'MINRES'.

    sigmax : float
        If `sigmax` > 0.0, the value of sigma_1(A_) = ||E^-1AE^-T||_2.

        If `sigmax` <= 0.0, sigma_1(A_) is estimated by
        ``real_symm_basic_solver`` when either `sigcmp` = 'S' or
        termination criterion (3) (`iterm` = 2) is employed, though it
        will be used only in the latter case.

        Otherwise, or if `method` = 'MINRES', `sigmax` is not
        referenced.

    sigtol : float
        The tolerance used in assessing the convergence of the estimate
        of sigma_1(A_) = ||A_||_2 when the bisection method is used.

        If `sigtol` <= 0.0, the default value `sigtol` = 0.01 is used.

        The actual value used is max(`sigtol`, epsilon).

        If `sigcmp` = 'N' or `sigmax` > 0.0, `sigtol` is not referenced.

        It has no effect if `method` = 'MINRES'.

        Suggested value: `sigtol` = 0.01 should be sufficient in most
        cases.

    maxits : int
        The maximum iteration number k = `maxits` for which sigma_1(T_k)
        is computed by bisection (see also [Notes]). If `sigcmp` = 'N'
        or `sigmax` > 0.0, or if `method` = 'MINRES', `maxits` is not
        referenced.

        Suggested value: `maxits` = min(10, n) when `sigtol` is of the
        order of its default value (0.01).

    monit : int
        If `monit` > 0, the frequency at which a monitoring step is
        executed by ``real_symm_basic_solver``: the current solution and
        residual iterates will be returned by ``real_symm_basic_solver``
        and a call to ``real_symm_basic_diag`` made possible every
        `monit` iterations, starting from the (`monit`)th. Otherwise, no
        monitoring takes place.

        There are some additional computational costs involved in
        monitoring the solution and residual vectors when the Lanczos
        method (SYMMLQ) is used.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` -14)
            On entry, `monit` = *<value>* and `maxitn` = *<value>*.

            Constraint: `monit` <= `maxitn`.

        (`errno` -13)
            On entry, `sigcmp` = 'S', `sigmax` <= 0.0, `maxits` =
            *<value>* and `maxitn` = *<value>*.

            Constraint: if `sigcmp` = 'S' and `sigmax` <= 0.0, `maxits`
            <= `maxitn`.

        (`errno` -13)
            On entry, `sigcmp` = 'S', `sigmax` <= 0.0 and `maxits` =
            *<value>*.

            Constraint: if `sigcmp` = 'S' and `sigmax` <= 0.0, `maxits`
            >= 1.

        (`errno` -12)
            On entry, `sigcmp` = 'S', `sigmax` <= 0.0 and `sigtol` =
            *<value>*.

            Constraint: if `sigcmp` = 'S' and `sigmax` <= 0.0, `sigtol`
            < 1.0.

        (`errno` -10)
            On entry, `iterm` = 1, `norm` = '2' and `anorm` = *<value>*.

            Constraint: if `iterm` = 1 and `norm` = '2', `anorm` > 0.0.

        (`errno` -9)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` > 0.

        (`errno` -8)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` -7)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` -6)
            On entry, `iterm` = 2 and `method` = 'CG'.

            Constraint: if `iterm` = 2, `method` != 'CG'.

        (`errno` -6)
            On entry, `iterm` = 2 and `norm` = *<value>*.

            Constraint: if `iterm` = 2, `norm` = '2'.

        (`errno` -6)
            On entry, `iterm` = 2 and `weight` = *<value>*.

            Constraint: if `iterm` = 2, `weight` = 'N'.

        (`errno` -6)
            On entry, `iterm` = *<value>*.

            Constraint: `iterm` = 1 or 2.

        (`errno` -5)
            On entry, `weight` = *<value>*.

            Constraint: `weight` = 'N' or 'W'.

        (`errno` -4)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = '1', 'I' or '2'.

        (`errno` -3)
            On entry, `sigcmp` = *<value>*.

            Constraint: `sigcmp` = 'S' or 'N'.

        (`errno` -2)
            On entry, `precon` = *<value>*.

            Constraint: `precon` = 'N' or 'P'.

        (`errno` -1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'CG', 'SYMMLQ' or 'MINRES'.

        (`errno` 1)
            ``real_symm_basic_setup`` has been called out of sequence:
            either ``real_symm_basic_setup`` has been called twice or
            ``real_symm_basic_solver`` has not terminated its current
            task.

    Notes
    -----
    The suite consisting of the functions ``real_symm_basic_setup``,
    ``real_symm_basic_solver`` and ``real_symm_basic_diag`` is designed
    to solve the symmetric system of simultaneous linear equations Ax =
    b of order n, where n is large and the matrix of the coefficients A
    is sparse.

    ``real_symm_basic_setup`` is a setup function which must be called
    before ``real_symm_basic_solver``, the iterative solver.
    The third function in the suite, ``real_symm_basic_diag`` can be
    used to return additional information about the computation.
    One of the following methods can be used:

    (1) **Conjugate Gradient Method (CG)**

        For this method (see Hestenes and Stiefel (1952), Golub and Van
        Loan (1996), Barrett `et al.` (1994) and Dias da Cunha and
        Hopkins (1994)), the matrix A should ideally be positive
        definite.
        The application of the Conjugate Gradient method to indefinite
        matrices may lead to failure or to lack of convergence.

    (#) **Lanczos Method (SYMMLQ)**

        This method, based upon the algorithm SYMMLQ (see Paige and
        Saunders (1975) and Barrett `et al.` (1994)), is suitable for
        both positive definite and indefinite matrices.
        It is more robust than the Conjugate Gradient method but less
        efficient when A is positive definite.

    (#) **Minimum Residual Method (MINRES)**

        This method may be used when the matrix is indefinite.
        It seeks to reduce the norm of the residual at each iteration
        and often takes fewer iterations than the other methods.
        It does however require slightly more memory.

    The CG and SYMMLQ methods start from the residual r_0 = b-Ax_0,
    where x_0 is an initial estimate for the solution (often x_0 = 0),
    and generate an orthogonal basis for the Krylov subspace
    span{A^kr_0}, for k = 0,1,...,, by means of three-term recurrence
    relations (see Golub and Van Loan (1996)).
    A sequence of symmetric tridiagonal matrices {T_k} is also
    generated.
    Here and in the following, the index k denotes the iteration count.
    The resulting symmetric tridiagonal systems of equations are usually
    more easily solved than the original problem.
    A sequence of solution iterates {x_k} is thus generated such that
    the sequence of the norms of the residuals {||r_k||} converges to a
    required tolerance.
    Note that, in general, the convergence is not monotonic.

    In exact arithmetic, after n iterations, this process is equivalent
    to an orthogonal reduction of A to symmetric tridiagonal form, T_n =
    Q^TAQ; the solution x_n would thus achieve exact convergence.
    In finite-precision arithmetic, cancellation and round-off errors
    accumulate causing loss of orthogonality.
    These methods must, therefore, be viewed as genuinely iterative
    methods, able to converge to a solution **within a prescribed
    tolerance**.

    The orthogonal basis is not formed explicitly in either method.
    The basic difference between the Conjugate Gradient and Lanczos
    methods lies in the method of solution of the resulting symmetric
    tridiagonal systems of equations: the conjugate gradient method is
    equivalent to carrying out an LDL^T (Cholesky) factorization whereas
    the Lanczos method (SYMMLQ) uses an LQ factorization.

    Faster convergence for all the methods can be achieved using a
    **preconditioner** (see Golub and Van Loan (1996) and Barrett `et
    al.` (1994)).
    A preconditioner maps the original system of equations onto a
    different system, say

        A_x_= b_,

    with, hopefully, better characteristics with respect to its speed of
    convergence: for example, the condition number of the matrix of the
    coefficients can be improved or eigenvalues in its spectrum can be
    made to coalesce.
    An orthogonal basis for the Krylov subspace span{A_^kr__0}, for k =
    0,1,...,, is generated and the solution proceeds as outlined above.
    The algorithms used are such that the solution and residual iterates
    of the original system are produced, not their preconditioned
    counterparts.
    Note that an unsuitable preconditioner or no preconditioning at all
    may result in a very slow rate, or lack, of convergence.
    However, preconditioning involves a trade-off between the reduction
    in the number of iterations required for convergence and the
    additional computational costs per iteration.
    Also, setting up a preconditioner may involve non-negligible
    overheads.

    A preconditioner must be **symmetric and positive definite**, i.e.,
    representable by M = EE^T, where M is nonsingular, and such that A_=
    E^-1AE^-T~I_n in (1), where I_n is the identity matrix of order n.
    Also, we can define r_= E^-1r and x_= E^Tx.
    These are formal definitions, used only in the design of the
    algorithms; in practice, only the means to compute the matrix-vector
    products v = Au and to solve the preconditioning equations Mv = u
    are required, that is, explicit information about M, E or their
    inverses is not required at any stage.

    The first termination criterion

        ||r_k||_p <= tau(||b||_p+||A||_p*||x_k||_p)

    is available for both conjugate gradient and Lanczos (SYMMLQ)
    methods.
    In (2), p = 1, infinity or 2 and tau denotes a user-specified
    tolerance subject to max(10, sqrt(n))epsilon <= tau < 1, where
    epsilon is the machine precision.
    Facilities are provided for the estimation of the norm of the matrix
    of the coefficients ||A||_1 = ||A||_infinity, when this is not known
    in advance, used in (2), by applying Higham's method (see Higham
    (1988)).
    Note that ||A||_2 cannot be estimated internally.
    This criterion uses an error bound derived from **backward** error
    analysis to ensure that the computed solution is the exact solution
    of a problem as close to the original as the termination tolerance
    requires.
    Termination criteria employing bounds derived from **forward** error
    analysis could be used, but any such criteria would require
    information about the condition number kappa(A) which is not easily
    obtainable.

    The second termination criterion

        ||r__k||_2 <= tau max(1.0,
        ||b||_2/||r_0||_2)(||r__0||_2+sigma_1(A_)*||Delta x__k||_2)

    is available only for the Lanczos method (SYMMLQ).
    In (3), sigma_1(A_) = ||A_||_2 is the largest singular value of the
    (preconditioned) iteration matrix A_.
    This termination criterion monitors the progress of the solution of
    the preconditioned system of equations and is less expensive to
    apply than criterion (2).
    When sigma_1(A_) is not supplied, facilities are provided for its
    estimation by sigma_1(A_)~max_ksigma_1(T_k).
    The interlacing property sigma_1(T_{k-1}) <= sigma_1(T_k) and
    Gerschgorin's theorem provide lower and upper bounds from which
    sigma_1(T_k) can be easily computed by bisection.
    Alternatively, the less expensive estimate
    sigma_1(A_)~max_k||T_k||_1 can be used, where sigma_1(A_) <=
    ||T_k||_1 by Gerschgorin's theorem.
    Note that only order of magnitude estimates are required by the
    termination criterion.

    Termination criterion (2) is the recommended choice, despite its
    (small) additional costs per iteration when using the Lanczos method
    (SYMMLQ).
    Also, if the norm of the initial estimate is much larger than the
    norm of the solution, that is, if ||x_0||≫||x||, a dramatic loss of
    significant digits could result in complete lack of convergence.
    The use of criterion (2) will enable the detection of such a
    situation, and the iteration will be restarted at a suitable point.
    No such restart facilities are provided for criterion (3).

    Optionally, a vector w of user-specified weights can be used in the
    computation of the vector norms in termination criterion (2), i.e.,
    ||v||_p^(w) = ||v^(w)||_p, where (v^(w))_i = w_i v_i, for i =
    1,2,...,n.
    Note that the use of weights increases the computational costs.

    The MINRES algorithm terminates when the norm of the residual of the
    preconditioned system F, ||F||_2 <= tau*||A_||_2*||x_k||_2, where A_
    is the preconditioned matrix.

    The termination criteria discussed are not robust in the presence of
    a non-trivial nullspace of A, i.e., when A is singular.
    It is then possible for ||x_k||_p to grow without limit, spuriously
    satisfying the termination criterion.
    If singularity is suspected, more robust functions can be found in
    submodule ``opt``.

    The sequence of calls to the functions comprising the suite is
    enforced: first, the setup function ``real_symm_basic_setup`` must
    be called, followed by the solver ``real_symm_basic_solver``.
    The diagnostic function ``real_symm_basic_diag`` can be called
    either when ``real_symm_basic_solver`` is carrying out a monitoring
    step or after ``real_symm_basic_solver`` has completed its tasks.
    Incorrect sequencing will raise an error condition.

    References
    ----------
    Barrett, R, Berry, M, Chan, T F, Demmel, J, Donato, J, Dongarra, J,
    Eijkhout, V, Pozo, R, Romine, C and Van der Vorst, H, 1994,
    `Templates for the Solution of Linear Systems: Building Blocks for
    Iterative Methods`, SIAM, Philadelphia

    Dias da Cunha, R and Hopkins, T, 1994, `PIM 1.1 --- the parallel
    iterative method package for systems of linear equations user's
    guide --- Fortran 77 version`, Technical Report, Computing
    Laboratory, University of Kent at Canterbury, Kent, UK

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Hestenes, M and Stiefel, E, 1952, `Methods of conjugate gradients
    for solving linear systems`, J. Res. Nat. Bur. Stand. (49), 409--436

    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396

    Paige, C C and Saunders, M A, 1975, `Solution of sparse indefinite
    systems of linear equations`, SIAM J. Numer. Anal. (12), 617--629
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_basic_setup'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11gdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_STRING, # Input, sigcmp(len=1).
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, iterm.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._BYREF_FLOAT64, # Input, anorm.
        utils._BYREF_FLOAT64, # Input, sigmax.
        utils._BYREF_FLOAT64, # Input, sigtol.
        utils._BYREF_INT, # Input, maxits.
        utils._BYREF_INT, # Input, monit.
        utils._BYREF_INT, # Output, lwreq.
        utils._EngineFloat64ArrayType.c_type, # Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of sigcmp.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _sigcmp = utils._EngineCharScalarType(
        sigcmp, {'fun_name': fun_name, 'entity': 'sigcmp'}, exp_chrlen=1,
    ).to_ctypes()
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _weight = utils._EngineCharScalarType(
        weight, {'fun_name': fun_name, 'entity': 'weight'}, exp_chrlen=1,
    ).to_ctypes()
    _iterm = utils._EngineIntScalarType(
        iterm, {'fun_name': fun_name, 'entity': 'iterm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _anorm = utils._EngineFloat64ScalarType(
        anorm, {'fun_name': fun_name, 'entity': 'anorm'},
    ).to_ctypes()
    _sigmax = utils._EngineFloat64ScalarType(
        sigmax, {'fun_name': fun_name, 'entity': 'sigmax'},
    ).to_ctypes()
    _sigtol = utils._EngineFloat64ScalarType(
        sigtol, {'fun_name': fun_name, 'entity': 'sigtol'},
    ).to_ctypes()
    _maxits = utils._EngineIntScalarType(
        maxits, {'fun_name': fun_name, 'entity': 'maxits'},
    ).to_ctypes()
    _monit = utils._EngineIntScalarType(
        monit, {'fun_name': fun_name, 'entity': 'monit'},
    ).to_ctypes()
    _lwreq = utils._EngineIntScalarType.empty_ctypes()
    _lwork = utils._EngineIntScalarType.c_type(120)
    _lwork_py = utils._EngineIntScalarType.to_py(_lwork)
    _work = utils._EngineFloat64ArrayType.empty_ctypes(_lwork_py)
    _comm_py = {}
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _sigcmp,
        _norm,
        _weight,
        _iterm,
        _n,
        _tol,
        _maxitn,
        _anorm,
        _sigmax,
        _sigtol,
        _maxits,
        _monit,
        _lwreq,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _sigcmp._length_,
        _norm._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    if _ierr.value == utils._IERR_SUCCESS:
        _lwreq_py = utils._EngineIntScalarType.to_py(_lwreq)
        _lwork = utils._EngineIntScalarType(
            _lwreq_py, {'fun_name': fun_name, 'entity': '_lwork'},
        ).to_ctypes()
        _lwork_py = utils._EngineIntScalarType.to_py(_lwork)
        _work = utils._EngineFloat64ArrayType.empty_ctypes(_lwork_py)
        _comm_py['work'] = _work
        engine_ctypes(
            _en_data,
            _method,
            _precon,
            _sigcmp,
            _norm,
            _weight,
            _iterm,
            _n,
            _tol,
            _maxitn,
            _anorm,
            _sigmax,
            _sigtol,
            _maxits,
            _monit,
            _lwreq,
            _work,
            _lwork,
            _errbuf,
            _ierr,
            _method._length_,
            _precon._length_,
            _sigcmp._length_,
            _norm._length_,
            _weight._length_,
            _errbuf._length_,
        )
        utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _monit_py = utils._EngineIntScalarType.to_py(_monit)
        _maxitn_py = utils._EngineIntScalarType.to_py(_maxitn)
        _maxits_py = utils._EngineIntScalarType.to_py(_maxits)
        _sigtol_py = utils._EngineFloat64ScalarType.to_py(_sigtol)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _tol_py = utils._EngineFloat64ScalarType.to_py(_tol)
        _n_py = utils._EngineIntScalarType.to_py(_n)
        _norm_py = utils._EngineCharScalarType.to_py(_norm)
        _weight_py = utils._EngineCharScalarType.to_py(_weight)
        _iterm_py = utils._EngineIntScalarType.to_py(_iterm)
        _sigcmp_py = utils._EngineCharScalarType.to_py(_sigcmp)
        _precon_py = utils._EngineCharScalarType.to_py(_precon)
        _method_py = utils._EngineCharScalarType.to_py(_method)
        if _ierr_py == -14 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, monit = {:s} and maxitn = {:s}.'.format(
                        utils._STR_FMT_INT.format(_monit_py),
                        utils._STR_FMT_INT.format(_maxitn_py),
                    ),
                    'Constraint: monit <= maxitn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -13 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigcmp = \'S\', sigmax <= 0.0,',
                    'maxits = {:s} and maxitn = {:s}.'.format(
                        utils._STR_FMT_INT.format(_maxits_py),
                        utils._STR_FMT_INT.format(_maxitn_py),
                    ),
                    'Constraint: if sigcmp = \'S\' and sigmax <= 0.0,',
                    'maxits <= maxitn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -13 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigcmp = \'S\', sigmax <= 0.0 and',
                    'maxits = {:s}.'.format(
                        utils._STR_FMT_INT.format(_maxits_py),
                    ),
                    'Constraint: if sigcmp = \'S\' and sigmax <= 0.0,',
                    'maxits >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -12 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigcmp = \'S\', sigmax <= 0.0 and',
                    'sigtol = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_sigtol_py),
                    ),
                    'Constraint: if sigcmp = \'S\' and sigmax <= 0.0,',
                    'sigtol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -10 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 1, norm = \'2\' and anorm = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_anorm_py),
                    ),
                    'Constraint: if iterm = 1 and norm = \'2\', anorm > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -9 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT.format(_maxitn_py),
                    ),
                    'Constraint: maxitn > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -8 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_tol_py),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -7 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and method = \'CG\'.',
                    'Constraint: if iterm = 2, method != \'CG\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == 99919:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: if iterm = 2, norm = \'2\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == 99929:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_weight_py),
                    ),
                    'Constraint: if iterm = 2, weight = \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = {:s}.'.format(
                        utils._STR_FMT_INT.format(_iterm_py),
                    ),
                    'Constraint: iterm = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_weight_py),
                    ),
                    'Constraint: weight = \'N\' or \'W\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -4 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: norm = \'1\', \'I\' or \'2\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -3 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigcmp = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_sigcmp_py),
                    ),
                    'Constraint: sigcmp = \'S\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_precon_py),
                    ),
                    'Constraint: precon = \'N\' or \'P\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'Constraint: method = \'CG\', \'SYMMLQ\' or \'MINRES\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'real_symm_basic_setup has been called out of sequence:',
                    'either real_symm_basic_setup has been called twice or',
                    'real_symm_basic_solver has not terminated its current task.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = _comm_py
    _err_class.manage()
    return return_data

def real_symm_basic_solver(irevcm, u, v, wgt, comm):
    u"""
    Real sparse symmetric linear systems, preconditioned conjugate
    gradient or Lanczos method or the MINRES algorithm.

    ``real_symm_basic_solver`` is an iterative solver for a symmetric
    system of simultaneous linear equations; ``real_symm_basic_solver``
    is the second in a suite of three functions, where the first
    function, ``real_symm_basic_setup``, must be called prior to
    ``real_symm_basic_solver`` to set up the suite, and the third
    function in the suite, ``real_symm_basic_diag``, can be used to
    return additional information about the computation.

    These three functions are suitable for the solution of large sparse
    symmetric systems of equations.

    For full information please refer to the NAG Library document for
    f11ge

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11gef.html

    Parameters
    ----------
    irevcm : int
        `On initial entry`: `irevcm` = 0, otherwise an error condition
        will be raised.

        `On intermediate entry`: must either be unchanged from its
        previous exit value, or can have one of the following values.

        `irevcm` = 5

            Tidy termination: the computation will terminate at the end
            of the current iteration. Further reverse communication
            exits may occur depending on when the termination request is
            issued. ``real_symm_basic_solver`` will then return with the
            termination code `irevcm` = 4. Note that before calling
            ``real_symm_basic_solver`` with `irevcm` = 5 the calling
            program must have performed the tasks required by the value
            of `irevcm` returned by the previous call to
            ``real_symm_basic_solver``, otherwise subsequently returned
            values may be invalid.

        `irevcm` = 6

            Immediate termination: ``real_symm_basic_solver`` will
            return immediately with termination code `irevcm` = 4 and
            with any useful information available. This includes the
            last iterate of the solution and, for conjugate gradient
            only, the last iterate of the residual vector. The residual
            vector is generally not available when the Lanczos method
            (SYMMLQ) is used. ``real_symm_basic_solver`` will then
            return with the termination code `irevcm` = 4.

            Immediate termination may be useful, for example, when
            errors are detected during matrix-vector multiplication or
            during the solution of the preconditioning equation.

        Changing `irevcm` to any other value between calls will result
        in an error.

    u : float, shape (n), modified in place
        `On initial entry`: an initial estimate, x_0, of the solution of
        the system of equations Ax = b.

        `On intermediate entry`: must remain unchanged.

        `On intermediate exit`: the returned value of `irevcm`
        determines the contents of `u` in the following way.

        If `irevcm` = 1 or 2, `u` holds the vector u on which the
        operation specified by `irevcm` is to be carried out.

        If `irevcm` = 3, `u` holds the current iterate of the solution
        vector.

        `On final exit`: if `errno` = 3 or -i, the array `u` is
        unchanged from the initial entry to ``real_symm_basic_solver``.
        If `errno` = 1, the array `u` is unchanged from the last entry
        to ``real_symm_basic_solver``. Otherwise, `u` holds the last
        iterate of the solution of the system of equations, for all
        returned values of errno.

    v : float, shape (n), modified in place
        `On initial entry`: the right-hand side b of the system of
        equations Ax = b.

        `On intermediate entry`: the returned value of `irevcm`
        determines the contents of `v` in the following way.

        If `irevcm` = 1 or 2, `v` must store the vector v, the result of
        the operation specified by the value of `irevcm` returned by the
        previous call to ``real_symm_basic_solver``.

        If `irevcm` = 3, `v` must remain unchanged.

        `On intermediate exit`: if `irevcm` = 3, `v` holds the current
        iterate of the residual vector. Note that this is an
        approximation to the true residual vector. Otherwise, it does
        not contain any useful information.

        `On final exit`: if `errno` = 3 or -i, the array `v` is
        unchanged from the last entry to ``real_symm_basic_solver``. If
        `errno` = 1, the array `v` is unchanged from the initial entry
        to ``real_symm_basic_solver``. If the function exits
        successfully or `errno` = 2, the array `v` contains the true
        residual vector of the system of equations (see also
        [Raises/Warns]). Otherwise, `v` stores the last iterate of the
        residual vector unless the Lanczos method (SYMMLQ) was used and
        `errno` >= 5, in which case `v` is set to 0.0.

    wgt : float, shape (n)
        The user-supplied weights, if these are to be used in the
        computation of the vector norms in the termination criterion
        (see [Notes for real_symm_basic_setup] and [Parameters for
        real_symm_basic_setup]).

        Weights are NOT used in the MINRES algorithm.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``real_symm_basic_setup``.

    Returns
    -------
    irevcm : int
        `On intermediate exit`: has the following meanings.

        `irevcm` = 1

            The calling program must compute the matrix-vector product v
            = Au, where u and v are stored in `u` and `v`, respectively.

        `irevcm` = 2

            The calling program must solve the preconditioning equation
            Mv = u, where u and v are stored in `u` and `v`,
            respectively.

        `irevcm` = 3

            Monitoring step: the solution and residual at the current
            iteration are returned in the arrays `u` and `v`,
            respectively. No action by the calling program is required.
            To return additional information ``real_symm_basic_diag``
            can be called at this step.

        `On final exit`: if `irevcm` = 4, ``real_symm_basic_solver`` has
        completed its tasks. The value of errno determines whether the
        iteration has been successfully completed, errors have been
        detected or the calling program has requested termination.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On intermediate re-entry, `irevcm` = *<value>*.

            Constraint: either `irevcm` must be unchanged from its
            previous exit value or `irevcm` = 5 or 6.

        (`errno` -1)
            On initial entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0.

        (`errno` 3)
            Either ``real_symm_basic_setup`` was not called before
            calling this function or it has returned an error.

        (`errno` 6)
            The preconditioner appears not to be positive definite. The
            computation cannot continue.

        (`errno` 9)
            The matrix of the coefficients A appears to be singular. The
            computation cannot continue.

        (`errno` 10)
            The weights in array `wgt` are all zero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            ``real_symm_basic_solver`` has already completed its tasks.
            You need to set a new problem.

        (`errno` 2)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 2)
            User-requested termination: the required accuracy could not
            be obtained. However, a reasonable accuracy may have been
            achieved.

        (`errno` 4)
            User-requested tidy termination. The solution has not
            converged after *<value>* iterations.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 7)
            The matrix of the coefficients A appears not to be positive
            definite. The computation cannot continue.

        (`errno` 8)
            User-requested immediate termination.

    Notes
    -----
    ``real_symm_basic_solver`` solves the symmetric system of linear
    simultaneous equations Ax = b using the preconditioned conjugate
    gradient method (see Hestenes and Stiefel (1952), Golub and Van Loan
    (1996), Barrett `et al.` (1994) and Dias da Cunha and Hopkins
    (1994)), a preconditioned Lanczos method based upon the algorithm
    SYMMLQ (see Paige and Saunders (1975) and Barrett `et al.` (1994)),
    or the MINRES algorithm (see Paige and Saunders (1975)).

    For a general description of the methods employed you are referred
    to [Notes for real_symm_basic_setup].

    ``real_symm_basic_solver`` can solve the system after the first
    function in the suite, ``real_symm_basic_setup``, has been called to
    initialize the computation and specify the method of solution.
    The third function in the suite, ``real_symm_basic_diag``, can be
    used to return additional information generated by the computation
    during monitoring steps and after ``real_symm_basic_solver`` has
    completed its tasks.

    ``real_symm_basic_solver`` uses **reverse communication**, i.e.,
    ``real_symm_basic_solver`` returns repeatedly to the calling program
    with the argument `irevcm` (see [Parameters]) set to specified
    values which require the calling program to carry out a specific
    task: either to compute the matrix-vector product v = Au; to solve
    the preconditioning equation Mv = u; to notify the completion of the
    computation; or, to allow the calling program to monitor the
    solution.
    Through the argument `irevcm` the calling program can cause
    immediate or tidy termination of the execution.
    On final exit, the last iterates of the solution and of the residual
    vectors of the original system of equations are returned.

    Reverse communication has the following advantages.

    (1) Maximum flexibility in the representation and storage of sparse
        matrices: all matrix operations are performed outside the solver
        function, thereby avoiding the need for a complicated interface
        with enough flexibility to cope with all types of storage
        schemes and sparsity patterns. This applies also to
        preconditioners.

    (#) Enhanced user interaction: you can closely monitor the solution
        and tidy or immediate termination can be requested. This is
        useful, for example, when alternative termination criteria are
        to be employed or in case of failure of the external functions
        used to perform matrix operations.

    References
    ----------
    Barrett, R, Berry, M, Chan, T F, Demmel, J, Donato, J, Dongarra, J,
    Eijkhout, V, Pozo, R, Romine, C and Van der Vorst, H, 1994,
    `Templates for the Solution of Linear Systems: Building Blocks for
    Iterative Methods`, SIAM, Philadelphia

    Dias da Cunha, R and Hopkins, T, 1994, `PIM 1.1 --- the parallel
    iterative method package for systems of linear equations user's
    guide --- Fortran 77 version`, Technical Report, Computing
    Laboratory, University of Kent at Canterbury, Kent, UK

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Hestenes, M and Stiefel, E, 1952, `Methods of conjugate gradients
    for solving linear systems`, J. Res. Nat. Bur. Stand. (49), 409--436

    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396

    Paige, C C and Saunders, M A, 1975, `Solution of sparse indefinite
    systems of linear equations`, SIAM J. Numer. Anal. (12), 617--629
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_basic_solver'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11geft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input/Output, irevcm.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, u[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, v[*].
        utils._EngineFloat64ArrayType.c_type, # Input, wgt[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _irevcm = utils._EngineIntScalarType(
        irevcm, {'fun_name': fun_name, 'entity': 'irevcm'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _lwork_py = utils._EngineFloat64ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    if _lwork_py >= 20:
        if int(utils._EngineFloat64ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (3,), validate=False)) == 1739512 and int(utils._EngineFloat64ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (19,), validate=False)) == 840653:
            _vl_n_py = int(utils._EngineFloat64ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (9,), validate=False))
        else:
            _vl_n_py = 0
    else:
        _vl_n_py = 0
    _u_cls = utils._EngineFloat64ArrayType(
        u, {'fun_name': fun_name, 'entity': 'u'},
        exp_shape=(_vl_n_py,),
    )
    _u = _u_cls.to_ctypes()
    _v_cls = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_vl_n_py,),
    )
    _v = _v_cls.to_ctypes()
    _wgt = utils._EngineFloat64ArrayType(
        wgt, {'fun_name': fun_name, 'entity': 'wgt'},
        exp_shape=(_vl_n_py,),
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _irevcm,
        _u,
        _v,
        _wgt,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _irevcm_py = utils._EngineIntScalarType.to_py(_irevcm)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _irevcm_py != 4 or _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On intermediate re-entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: either irevcm must be unchanged from its previous exit value or',
                    'irevcm = 5 or 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On initial entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irevcm = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'real_symm_basic_solver has already completed its tasks. You need to set a new problem.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested termination: the required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Either real_symm_basic_setup was not called before calling this function',
                    'or it has returned an error.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested tidy termination.',
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The preconditioner appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99991:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients A appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99990:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested immediate termination.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients A appears to be singular.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The weights in array wgt are all zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _u_cls.refresh_py(_u)
        _v_cls.refresh_py(_v)
        return_data = _irevcm_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['irevcm'],
            )(
                _irevcm_py,
            )
    _err_class.manage()
    return return_data

def real_symm_basic_diag(comm):
    u"""
    Real sparse symmetric linear systems, diagnostic for
    ``real_symm_basic_solver``.

    ``real_symm_basic_diag`` is the third in a suite of three functions
    for the iterative solution of a symmetric system of simultaneous
    linear equations (see Golub and Van Loan (1996)).
    ``real_symm_basic_diag`` returns information about the computations
    during an iteration and/or after this has been completed.
    The first function of the suite, ``real_symm_basic_setup``, is a
    setup function, the second function, ``real_symm_basic_solver`` is
    the proper iterative solver.

    These three functions are suitable for the solution of large sparse
    symmetric systems of equations.

    For full information please refer to the NAG Library document for
    f11gf

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11gff.html

    Parameters
    ----------
    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``real_symm_basic_setup``.

    Returns
    -------
    itn : int
        The number of iterations carried out by
        ``real_symm_basic_solver``.

    stplhs : float
        The current value of the left-hand side of the termination
        criterion used by ``real_symm_basic_solver``.

    stprhs : float
        The current value of the right-hand side of the termination
        criterion used by ``real_symm_basic_solver``.

    anorm : float
        For CG and SYMMLQ methods, the norm ||A||_1 = ||A||_infinity
        when either it has been supplied to ``real_symm_basic_setup`` or
        it has been estimated by ``real_symm_basic_solver`` (see also
        [Notes for real_symm_basic_setup] and [Parameters for
        real_symm_basic_setup]). Otherwise, `anorm` = 0.0 is returned.

        For MINRES method, an estimate of the infinity norm of the
        preconditioned matrix operator.

    sigmax : float
        For CG and SYMMLQ methods, the current estimate of the largest
        singular value sigma_1(A_) of the preconditioned iteration
        matrix A_= E^-1AE^-T, when either it has been supplied to
        ``real_symm_basic_setup`` or it has been estimated by
        ``real_symm_basic_solver`` (see also [Notes for
        real_symm_basic_setup] and [Parameters for
        real_symm_basic_setup]). Note that if `its` < `itn` then
        `sigmax` contains the final estimate. If, on final exit from
        ``real_symm_basic_solver``, `its` = `itn`, the estimation of
        sigma_1(A_) may have not converged; in this case you should look
        at the value returned in `sigerr`. Otherwise, `sigmax` = 0.0 is
        returned.

        For MINRES method, an estimate of the final transformed
        residual.

    its : int
        For CG and SYMMLQ methods, the number of iterations employed so
        far in the computation of the estimate of sigma_1(A_), the
        largest singular value of the preconditioned matrix A_=
        E^-1AE^-T, when sigma_1(A_) has been estimated by
        ``real_symm_basic_solver`` using the bisection method (see also
        [Notes for real_symm_basic_setup], [Parameters for
        real_symm_basic_setup] and [Further Comments]). Otherwise, `its`
        = 0 is returned.

    sigerr : float
        For CG and SYMMLQ methods, if sigma_1(A_) has been estimated by
        ``real_symm_basic_solver`` using bisection,

            `sigerr` = max(|sigma_1^(k)-sigma_1^(k-1)|/sigma_1^(k),
            |sigma_1^(k)-sigma_1^(k-2)|/sigma_1^(k)),

        where k = `its` denotes the iteration number. The estimation has
        converged if `sigerr` <= sigtol where sigtol is an input
        argument to ``real_symm_basic_setup``. Otherwise, `sigerr` = 0.0
        is returned.

        For MINRES method, an estimate of the condition number of the
        preconditioned matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            ``real_symm_basic_diag`` has been called out of sequence.

    Notes
    -----
    ``real_symm_basic_diag`` returns information about the solution
    process.
    It can be called both during a monitoring step of the solver
    ``real_symm_basic_solver``, or after this solver has completed its
    tasks.
    Calling ``real_symm_basic_diag`` at any other time will result in an
    error condition being raised.

    For further information you should read the documentation for
    ``real_symm_basic_setup`` and ``real_symm_basic_solver``.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_basic_diag'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11gfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, itn.
        utils._BYREF_FLOAT64, # Output, stplhs.
        utils._BYREF_FLOAT64, # Output, stprhs.
        utils._BYREF_FLOAT64, # Output, anorm.
        utils._BYREF_FLOAT64, # Output, sigmax.
        utils._BYREF_INT, # Output, its.
        utils._BYREF_FLOAT64, # Output, sigerr.
        utils._EngineFloat64ArrayType.c_type, # Input, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _stplhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _stprhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _anorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _sigmax = utils._EngineFloat64ScalarType.empty_ctypes()
    _its = utils._EngineIntScalarType.empty_ctypes()
    _sigerr = utils._EngineFloat64ScalarType.empty_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _lwork_py = utils._EngineFloat64ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _itn,
        _stplhs,
        _stprhs,
        _anorm,
        _sigmax,
        _its,
        _sigerr,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'real_symm_basic_diag has been called out of sequence.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        _stplhs_py = utils._EngineFloat64ScalarType.to_py(_stplhs)
        _stprhs_py = utils._EngineFloat64ScalarType.to_py(_stprhs)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _sigmax_py = utils._EngineFloat64ScalarType.to_py(_sigmax)
        _its_py = utils._EngineIntScalarType.to_py(_its)
        _sigerr_py = utils._EngineFloat64ScalarType.to_py(_sigerr)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['itn', 'stplhs', 'stprhs', 'anorm', 'sigmax', 'its', 'sigerr'],
        )(
            _itn_py,
            _stplhs_py,
            _stprhs_py,
            _anorm_py,
            _sigmax_py,
            _its_py,
            _sigerr_py,
        )
    _err_class.manage()
    return return_data

def complex_herm_basic_setup(method, precon, sigcmp, norm, weight, iterm, n, tol, maxitn, anorm, sigmax, sigtol, maxits, monit):
    u"""
    Complex sparse Hermitian linear systems, setup for
    ``complex_herm_basic_solver``.

    ``complex_herm_basic_setup`` is a setup function, the first in a
    suite of three functions for the iterative solution of a complex
    Hermitian system of simultaneous linear equations.
    ``complex_herm_basic_setup`` must be called before
    ``complex_herm_basic_solver``, the iterative solver.
    The third function in the suite, ``complex_herm_basic_diag``, can be
    used to return additional information about the computation.

    These three functions are suitable for the solution of large sparse
    complex Hermitian systems of equations.

    For full information please refer to the NAG Library document for
    f11gr

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11grf.html

    Parameters
    ----------
    method : str
        The iterative method to be used.

        `method` = 'CG'

            Conjugate gradient method.

        `method` = 'SYMMLQ'

            Lanczos method (SYMMLQ).

    precon : str, length 1
        Determines whether preconditioning is used.

        `precon` = 'N'

            No preconditioning.

        `precon` = 'P'

            Preconditioning.

    sigcmp : str, length 1
        Determines whether an estimate of sigma_1(A_) = ||E^-1AE^-H||_2,
        the largest singular value of the preconditioned matrix of the
        coefficients, is to be computed using the bisection method on
        the sequence of tridiagonal matrices {T_k} generated during the
        iteration. Note that A_= A when a preconditioner is not used.

        If `sigmax` > 0.0 (see `sigmax`), i.e., when sigma_1(A_) is
        supplied, the value of `sigcmp` is ignored.

        `sigcmp` = 'S'

            sigma_1(A_) is to be computed using the bisection method.

        `sigcmp` = 'N'

            The bisection method is not used.

            If the termination criterion (3) is used, requiring
            sigma_1(A_), an inexpensive estimate is computed and used
            (see [Notes]).

        Suggested value: `sigcmp` = 'N'.

    norm : str, length 1
        Defines the matrix and vector norm to be used in the termination
        criteria.

        `norm` = '1'

            Use the l_1 norm.

        `norm` = 'I'

            Use the l_infinity norm.

        `norm` = '2'

            Use the l_2 norm.

        Suggested value:

            if `iterm` = 1, `norm` = 'I'

            if `iterm` = 2, `norm` = '2'

    weight : str, length 1
        Specifies whether a vector w of user-supplied weights is to be
        used in the vector norms used in the computation of termination
        criterion (2) (`iterm` = 1): ||v||_p^(w) = ||v^(w)||_p, where
        v_i^(w) = w_i v_i, for i = 1,2,...,n. The suffix p = 1,2,
        infinity denotes the vector norm used, as specified by the
        argument `norm`. Note that weights cannot be used when `iterm` =
        2, i.e., when criterion (3) is used.

        `weight` = 'W'

            User-supplied weights are to be used and must be supplied on
            initial entry to ``complex_herm_basic_solver``.

        `weight` = 'N'

            All weights are implicitly set equal to one. Weights do not
            need to be supplied on initial entry to
            ``complex_herm_basic_solver``.

        Suggested value: `weight` = 'N'.

    iterm : int
        Defines the termination criterion to be used.

        `iterm` = 1

            Use the termination criterion defined in (2) (both conjugate
            gradient and Lanczos (SYMMLQ) methods).

        `iterm` = 2

            Use the termination criterion defined in (3) (Lanczos method
            (SYMMLQ) only).

        Suggested value: `iterm` = 1.

    n : int
        n, the order of the matrix A.

    tol : float
        The tolerance tau for the termination criterion.

        If `tol` <= 0.0, tau = max(sqrt(epsilon), sqrt(n)epsilon) is
        used, where epsilon is the machine precision.

        Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon) is used.

    maxitn : int
        The maximum number of iterations.

    anorm : float
        If `anorm` > 0.0, the value of ||A||_p to be used in the
        termination criterion (2) (`iterm` = 1).

        If `anorm` <= 0.0, `iterm` = 1 and `norm` = '1' or 'I', ||A||_1
        = ||A||_infinity is estimated internally by
        ``complex_herm_basic_solver``.

        If `iterm` = 2, `anorm` is not referenced.

    sigmax : float
        If `sigmax` > 0.0, the value of sigma_1(A_) = ||E^-1AE^-H||_2.

        If `sigmax` <= 0.0, sigma_1(A_) is estimated by
        ``complex_herm_basic_solver`` when either `sigcmp` = 'S' or
        termination criterion (3) (`iterm` = 2) is employed, though it
        will be used only in the latter case.

        Otherwise, `sigmax` is not referenced.

    sigtol : float
        The tolerance used in assessing the convergence of the estimate
        of sigma_1(A_) = ||A_||_2 when the bisection method is used.

        If `sigtol` <= 0.0, the default value `sigtol` = 0.01 is used.

        The actual value used is max(`sigtol`, epsilon).

        If `sigcmp` = 'N' or `sigmax` > 0.0, `sigtol` is not referenced.

        Suggested value: `sigtol` = 0.01 should be sufficient in most
        cases.

    maxits : int
        The maximum iteration number k = `maxits` for which sigma_1(T_k)
        is computed by bisection (see also [Notes]). If `sigcmp` = 'N'
        or `sigmax` > 0.0, `maxits` is not referenced.

        Suggested value: `maxits` = min(10, n) when `sigtol` is of the
        order of its default value (0.01).

    monit : int
        If `monit` > 0, the frequency at which a monitoring step is
        executed by ``complex_herm_basic_solver``: the current solution
        and residual iterates will be returned by
        ``complex_herm_basic_solver`` and a call to
        ``complex_herm_basic_diag`` made possible every `monit`
        iterations, starting from iteration number `monit`. Otherwise,
        no monitoring takes place. There are some additional
        computational costs involved in monitoring the solution and
        residual vectors when the Lanczos method (SYMMLQ) is used.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` -14)
            On entry, `monit` = *<value>* and `maxitn` = *<value>*.

            Constraint: `monit` <= `maxitn`.

        (`errno` -13)
            On entry, `maxits` = *<value>*, `maxitn` = *<value>*,
            `sigcmp` = *<value>*, and `sigmax` = *<value>*.

            Constraint: if `sigcmp` = 'S' and `sigmax` <= 0.0, 1 <=
            `maxits` <= `maxitn`.

        (`errno` -12)
            On entry, `sigcmp` = *<value>*, `sigtol` = *<value>*, and
            `sigmax` = *<value>*.

            Constraint: if `sigcmp` = 'S' and `sigmax` <= 0.0, `sigtol`
            < 1.0.

        (`errno` -10)
            On entry, `norm` = *<value>*, `iterm` = *<value>*, and
            `anorm` = *<value>*.

            Constraint: if `iterm` = 1 and `norm` = '2' or 'I', `anorm`
            > 0.0.

        (`errno` -9)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` > 0.

        (`errno` -8)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` -7)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` -6)
            On entry, `iterm` = *<value>*.

            Constraint: `iterm` = 1 or 2.

        (`errno` -6)
            On entry, `iterm` = *<value>* and `method` = *<value>*.

            Constraint: if `iterm` = 2, `method` = 'SYMMLQ'.

        (`errno` -6)
            On entry, `iterm` = *<value>* and `norm` = *<value>*.

            Constraint: if `iterm` = 2, `norm` = '2' or 'I'.

        (`errno` -5)
            On entry, `weight` = *<value>*.

            Constraint: `weight` = 'N' or 'W'.

        (`errno` -5)
            On entry, `iterm` = *<value>* and `weight` = *<value>*.

            Constraint: if `iterm` = 2, `weight` = 'N'.

        (`errno` -4)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = '1', 'I' or '2'.

        (`errno` -3)
            On entry, `sigcmp` = *<value>*.

            Constraint: `sigcmp` = 'N' or 'S'.

        (`errno` -2)
            On entry, `precon` = *<value>*.

            Constraint: `precon` = 'N' or 'P'.

        (`errno` -1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'CG' or 'SYMMLQ'.

        (`errno` 1)
            ``complex_herm_basic_setup`` has been called out of
            sequence: either ``complex_herm_basic_setup`` has been
            called twice or ``complex_herm_basic_solver`` has not
            terminated its current task.

    Notes
    -----
    The suite consisting of the functions ``complex_herm_basic_setup``,
    ``complex_herm_basic_solver`` and ``complex_herm_basic_diag`` is
    designed to solve the complex Hermitian system of simultaneous
    linear equations Ax = b of order n, where n is large and the matrix
    of the coefficients A is sparse.

    ``complex_herm_basic_setup`` is a setup function which must be
    called before the iterative solver ``complex_herm_basic_solver``.
    ``complex_herm_basic_diag``, the third function in the suite, can be
    used to return additional information about the computation.
    Either of two methods can be used:

    (1) **Conjugate Gradient Method (CG)**

        For this method (see Hestenes and Stiefel (1952), Golub and Van
        Loan (1996), Barrett `et al.` (1994) and Dias da Cunha and
        Hopkins (1994)), the matrix A should ideally be positive
        definite.
        The application of the Conjugate Gradient method to indefinite
        matrices may lead to failure or to lack of convergence.

    (#) **Lanczos Method (SYMMLQ)**

        This method, based upon the algorithm SYMMLQ (see Paige and
        Saunders (1975) and Barrett `et al.` (1994)), is suitable for
        both positive definite and indefinite matrices.
        It is more robust than the Conjugate Gradient method but less
        efficient when A is positive definite.

    Both CG and SYMMLQ methods start from the residual r_0 = b-Ax_0,
    where x_0 is an initial estimate for the solution (often x_0 = 0),
    and generate an orthogonal basis for the Krylov subspace
    span{A^kr_0}, for k = 0,1,...,, by means of three-term recurrence
    relations (see Golub and Van Loan (1996)).
    A sequence of real symmetric tridiagonal matrices {T_k} is also
    generated.
    Here and in the following, the index k denotes the iteration count.
    The resulting real symmetric tridiagonal systems of equations are
    usually more easily solved than the original problem.
    A sequence of solution iterates {x_k} is thus generated such that
    the sequence of the norms of the residuals {||r_k||} converges to a
    required tolerance.
    Note that, in general, the convergence is not monotonic.

    In exact arithmetic, after n iterations, this process is equivalent
    to an orthogonal reduction of A to real symmetric tridiagonal form,
    T_n = Q^HAQ; the solution x_n would thus achieve exact convergence.
    In finite-precision arithmetic, cancellation and round-off errors
    accumulate causing loss of orthogonality.
    These methods must, therefore, be viewed as genuinely iterative
    methods, able to converge to a solution **within a prescribed
    tolerance**.

    The orthogonal basis is not formed explicitly in either method.
    The basic difference between the two methods lies in the method of
    solution of the resulting real symmetric tridiagonal systems of
    equations: the conjugate gradient method is equivalent to carrying
    out an LDL^H (Cholesky) factorization whereas the Lanczos method
    (SYMMLQ) uses an LQ factorization.

    Faster convergence can be achieved using a **preconditioner** (see
    Golub and Van Loan (1996) and Barrett `et al.` (1994)).
    A preconditioner maps the original system of equations onto a
    different system, say

        A_x_= b_,

    with, hopefully, better characteristics with respect to its speed of
    convergence: for example, the condition number of the matrix of the
    coefficients can be improved or eigenvalues in its spectrum can be
    made to coalesce.
    An orthogonal basis for the Krylov subspace span{A_^kr__0}, for k =
    0,1,...,, is generated and the solution proceeds as outlined above.
    The algorithms used are such that the solution and residual iterates
    of the original system are produced, not their preconditioned
    counterparts.
    Note that an unsuitable preconditioner or no preconditioning at all
    may result in a very slow rate, or lack, of convergence.
    However, preconditioning involves a trade-off between the reduction
    in the number of iterations required for convergence and the
    additional computational costs per iteration.
    Also, setting up a preconditioner may involve non-negligible
    overheads.

    A preconditioner must be **Hermitian and positive definite**, i.e.,
    representable by M = EE^H, where M is nonsingular, and such that A_=
    E^-1AE^-H~I_n in (1), where I_n is the identity matrix of order n.
    Also, we can define r_= E^-1r and x_= E^Hx.
    These are formal definitions, used only in the design of the
    algorithms; in practice, only the means to compute the matrix-vector
    products v = Au and to solve the preconditioning equations Mv = u
    are required, that is, explicit information about M, E or their
    inverses is not required at any stage.

    The first termination criterion

        ||r_k||_p <= tau(||b||_p+||A||_p*||x_k||_p)

    is available for both conjugate gradient and Lanczos (SYMMLQ)
    methods.
    In (2), p = 1, infinity or 2 and tau denotes a user-specified
    tolerance subject to max(10, sqrt(n))epsilon <= tau < 1, where
    epsilon is the machine precision.
    Facilities are provided for the estimation of the norm of the matrix
    of the coefficients ||A||_1 = ||A||_infinity, when this is not known
    in advance, used in (2), by applying Higham's method (see Higham
    (1988)).
    Note that ||A||_2 cannot be estimated internally.
    This criterion uses an error bound derived from **backward** error
    analysis to ensure that the computed solution is the exact solution
    of a problem as close to the original as the termination tolerance
    requires.
    Termination criteria employing bounds derived from **forward** error
    analysis could be used, but any such criteria would require
    information about the condition number kappa(A) which is not easily
    obtainable.

    The second termination criterion

        ||r__k||_2 <= tau max(1.0,
        ||b||_2/||r_0||_2)(||r__0||_2+sigma_1(A_)*||Delta x__k||_2)

    is available only for the Lanczos method (SYMMLQ).
    In (3), sigma_1(A_) = ||A_||_2 is the largest singular value of the
    (preconditioned) iteration matrix A_.
    This termination criterion monitors the progress of the solution of
    the preconditioned system of equations and is less expensive to
    apply than criterion (2).
    When sigma_1(A_) is not supplied, facilities are provided for its
    estimation by sigma_1(A_)~max_ksigma_1(T_k).
    The interlacing property sigma_1(T_{k-1}) <= sigma_1(T_k) and
    Gerschgorin's theorem provide lower and upper bounds from which
    sigma_1(T_k) can be easily computed by bisection.
    Alternatively, the less expensive estimate
    sigma_1(A_)~max_k||T_k||_1 can be used, where sigma_1(A_) <=
    ||T_k||_1 by Gerschgorin's theorem.
    Note that only order of magnitude estimates are required by the
    termination criterion.

    Termination criterion (2) is the recommended choice, despite its
    (small) additional costs per iteration when using the Lanczos method
    (SYMMLQ).
    Also, if the norm of the initial estimate is much larger than the
    norm of the solution, that is, if ||x_0||≫||x||, a dramatic loss of
    significant digits could result in complete lack of convergence.
    The use of criterion (2) will enable the detection of such a
    situation, and the iteration will be restarted at a suitable point.
    No such restart facilities are provided for criterion (3).

    Optionally, a vector w of user-specified weights can be used in the
    computation of the vector norms in termination criterion (2), i.e.,
    ||v||_p^(w) = ||v^(w)||_p, where (v^(w))_i = w_i v_i, for i =
    1,2,...,n.
    Note that the use of weights increases the computational costs.

    The sequence of calls to the functions comprising the suite is
    enforced: first, the setup function ``complex_herm_basic_setup``
    must be called, followed by the solver
    ``complex_herm_basic_solver``. ``complex_herm_basic_diag`` can be
    called either when ``complex_herm_basic_solver`` is carrying out a
    monitoring step or after ``complex_herm_basic_solver`` has completed
    its tasks.
    Incorrect sequencing will raise an error condition.

    References
    ----------
    Barrett, R, Berry, M, Chan, T F, Demmel, J, Donato, J, Dongarra, J,
    Eijkhout, V, Pozo, R, Romine, C and Van der Vorst, H, 1994,
    `Templates for the Solution of Linear Systems: Building Blocks for
    Iterative Methods`, SIAM, Philadelphia

    Dias da Cunha, R and Hopkins, T, 1994, `PIM 1.1 --- the parallel
    iterative method package for systems of linear equations user's
    guide --- Fortran 77 version`, Technical Report, Computing
    Laboratory, University of Kent at Canterbury, Kent, UK

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Hestenes, M and Stiefel, E, 1952, `Methods of conjugate gradients
    for solving linear systems`, J. Res. Nat. Bur. Stand. (49), 409--436

    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396

    Paige, C C and Saunders, M A, 1975, `Solution of sparse indefinite
    systems of linear equations`, SIAM J. Numer. Anal. (12), 617--629
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_basic_setup'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11grft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_STRING, # Input, sigcmp(len=1).
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, iterm.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._BYREF_FLOAT64, # Input, anorm.
        utils._BYREF_FLOAT64, # Input, sigmax.
        utils._BYREF_FLOAT64, # Input, sigtol.
        utils._BYREF_INT, # Input, maxits.
        utils._BYREF_INT, # Input, monit.
        utils._BYREF_INT, # Output, lwreq.
        utils._EngineComplex128ArrayType.c_type, # Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of sigcmp.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _sigcmp = utils._EngineCharScalarType(
        sigcmp, {'fun_name': fun_name, 'entity': 'sigcmp'}, exp_chrlen=1,
    ).to_ctypes()
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _weight = utils._EngineCharScalarType(
        weight, {'fun_name': fun_name, 'entity': 'weight'}, exp_chrlen=1,
    ).to_ctypes()
    _iterm = utils._EngineIntScalarType(
        iterm, {'fun_name': fun_name, 'entity': 'iterm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _anorm = utils._EngineFloat64ScalarType(
        anorm, {'fun_name': fun_name, 'entity': 'anorm'},
    ).to_ctypes()
    _sigmax = utils._EngineFloat64ScalarType(
        sigmax, {'fun_name': fun_name, 'entity': 'sigmax'},
    ).to_ctypes()
    _sigtol = utils._EngineFloat64ScalarType(
        sigtol, {'fun_name': fun_name, 'entity': 'sigtol'},
    ).to_ctypes()
    _maxits = utils._EngineIntScalarType(
        maxits, {'fun_name': fun_name, 'entity': 'maxits'},
    ).to_ctypes()
    _monit = utils._EngineIntScalarType(
        monit, {'fun_name': fun_name, 'entity': 'monit'},
    ).to_ctypes()
    _lwreq = utils._EngineIntScalarType.empty_ctypes()
    _lwork = utils._EngineIntScalarType.c_type(120)
    _lwork_py = utils._EngineIntScalarType.to_py(_lwork)
    _work = utils._EngineComplex128ArrayType.empty_ctypes(_lwork_py)
    _comm_py = {}
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _sigcmp,
        _norm,
        _weight,
        _iterm,
        _n,
        _tol,
        _maxitn,
        _anorm,
        _sigmax,
        _sigtol,
        _maxits,
        _monit,
        _lwreq,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _sigcmp._length_,
        _norm._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    if _ierr.value == utils._IERR_SUCCESS:
        _lwreq_py = utils._EngineIntScalarType.to_py(_lwreq)
        _lwork = utils._EngineIntScalarType(
            _lwreq_py, {'fun_name': fun_name, 'entity': '_lwork'},
        ).to_ctypes()
        _lwork_py = utils._EngineIntScalarType.to_py(_lwork)
        _work = utils._EngineComplex128ArrayType.empty_ctypes(_lwork_py)
        _comm_py['work'] = _work
        engine_ctypes(
            _en_data,
            _method,
            _precon,
            _sigcmp,
            _norm,
            _weight,
            _iterm,
            _n,
            _tol,
            _maxitn,
            _anorm,
            _sigmax,
            _sigtol,
            _maxits,
            _monit,
            _lwreq,
            _work,
            _lwork,
            _errbuf,
            _ierr,
            _method._length_,
            _precon._length_,
            _sigcmp._length_,
            _norm._length_,
            _weight._length_,
            _errbuf._length_,
        )
        utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -14 and _ifmt == -14:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, monit = {:s} and maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: monit <= maxitn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -13 and _ifmt == -13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxits = {:s}, maxitn = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'sigcmp = \'{:s}\', and sigmax = {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: if sigcmp = \'S\' and sigmax <= 0.0, 1 <= maxits <= maxitn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -12 and _ifmt == -12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigcmp = \'{:s}\', sigtol = {:s},'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and sigmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if sigcmp = \'S\' and sigmax <= 0.0, sigtol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -10 and _ifmt == -10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, norm = \'{:s}\', iterm = {:s},'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and anorm = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if iterm = 1 and norm = \'2\' or \'I\', anorm > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -9 and _ifmt == -9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -8 and _ifmt == -8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -7 and _ifmt == -7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == -6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iterm = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = {:s} and method = \'{:s}\'.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if iterm = 2, method = \'SYMMLQ\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = {:s} and norm = \'{:s}\'.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if iterm = 2, norm = \'2\' or \'I\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == -5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'N\' or \'W\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = {:s} and weight = \'{:s}\'.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if iterm = 2, weight = \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -4 and _ifmt == -4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: norm = \'1\', \'I\' or \'2\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -3 and _ifmt == -3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigcmp = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigcmp = \'N\' or \'S\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == -2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: precon = \'N\' or \'P\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'CG\' or \'SYMMLQ\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'complex_herm_basic_setup has been called out of sequence:',
                    'either complex_herm_basic_setup has been called twice or',
                    'complex_herm_basic_solver has not terminated its current task.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = _comm_py
    _err_class.manage()
    return return_data

def complex_herm_basic_solver(irevcm, u, v, wgt, comm):
    u"""
    Complex sparse Hermitian linear systems, preconditioned conjugate
    gradient or Lanczos.

    ``complex_herm_basic_solver`` is an iterative solver for a complex
    Hermitian system of simultaneous linear equations;
    ``complex_herm_basic_solver`` is the second in a suite of three
    functions, where the first function, ``complex_herm_basic_setup``,
    must be called prior to ``complex_herm_basic_solver`` to set up the
    suite, and the third function in the suite,
    ``complex_herm_basic_diag``, can be used to return additional
    information about the computation.

    These three functions are suitable for the solution of large sparse
    complex Hermitian systems of equations.

    For full information please refer to the NAG Library document for
    f11gs

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11gsf.html

    Parameters
    ----------
    irevcm : int
        `On initial entry`: `irevcm` = 0, otherwise an error condition
        will be raised.

        `On intermediate entry`: `irevcm` must either be unchanged from
        its previous exit value, or can have one of the following
        values.

        `irevcm` = 5

            Tidy termination: the computation will terminate at the end
            of the current iteration. Further reverse communication
            exits may occur depending on when the termination request is
            issued. ``complex_herm_basic_solver`` will then return with
            the termination code `irevcm` = 4. Note that before calling
            ``complex_herm_basic_solver`` with `irevcm` = 5 the calling
            program must have performed the tasks required by the value
            of `irevcm` returned by the previous call to
            ``complex_herm_basic_solver``, otherwise subsequently
            returned values may be invalid.

        `irevcm` = 6

            Immediate termination: ``complex_herm_basic_solver`` will
            return immediately with termination code `irevcm` = 4 and
            with any useful information available. This includes the
            last iterate of the solution and, for conjugate gradient
            only, the last iterate of the residual vector. The residual
            vector is generally not available when the Lanczos method
            (SYMMLQ) is used. ``complex_herm_basic_solver`` will then
            return with the termination code `irevcm` = 4.

            Immediate termination may be useful, for example, when
            errors are detected during matrix-vector multiplication or
            during the solution of the preconditioning equation.

        Changing `irevcm` to any other value between calls will result
        in an error.

    u : complex, shape (n), modified in place
        `On initial entry`: an initial estimate, x_0, of the solution of
        the system of equations Ax = b.

        `On intermediate entry`: must remain unchanged.

        `On intermediate exit`: the returned value of `irevcm`
        determines the contents of `u` in the following way.

        If `irevcm` = 1 or 2, `u` holds the vector u on which the
        operation specified by `irevcm` is to be carried out.

        If `irevcm` = 3, `u` holds the current iterate of the solution
        vector.

        `On final exit`: if `errno` = 3 or -i, the array `u` is
        unchanged from the initial entry to
        ``complex_herm_basic_solver``. If `errno` = 1, the array `u` is
        unchanged from the last entry to ``complex_herm_basic_solver``.
        Otherwise, `u` holds the last iterate of the solution of the
        system of equations, for all returned values of errno.

    v : complex, shape (n), modified in place
        `On initial entry`: the right-hand side b of the system of
        equations Ax = b.

        `On intermediate entry`: the returned value of `irevcm`
        determines the contents of `v` in the following way.

        If `irevcm` = 1 or 2, `v` must store the vector v, the result of
        the operation specified by the value of `irevcm` returned by the
        previous call to ``complex_herm_basic_solver``

        If `irevcm` = 3, `v` must remain unchanged.

        `On intermediate exit`: if `irevcm` = 3, `v` holds the current
        iterate of the residual vector. Note that this is an
        approximation to the true residual vector. Otherwise, it does
        not contain any useful information.

        `On final exit`: if `errno` = 3 or 0, the array `v` is unchanged
        from the last entry to ``complex_herm_basic_solver``. If `errno`
        = 1, the array `v` is unchanged from the initial entry to
        ``complex_herm_basic_solver``. If the function exits
        successfully or `errno` = 2, the array `v` contains the true
        residual vector of the system of equations (see also
        [Raises/Warns]). Otherwise, `v` stores the last iterate of the
        residual vector unless the Lanczos method (SYMMLQ) was used and
        `errno` >= 5, in which case `v` is set to 0.0.

    wgt : float, shape (n)
        The user-supplied weights, if these are to be used in the
        computation of the vector norms in the termination criterion
        (see [Notes for complex_herm_basic_setup] and [Parameters for
        complex_herm_basic_setup]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``complex_herm_basic_setup``.

    Returns
    -------
    irevcm : int
        `On intermediate exit`: has the following meanings.

        `irevcm` = 1

            The calling program must compute the matrix-vector product v
            = Au, where u and v are stored in `u` and `v`, respectively.

        `irevcm` = 2

            The calling program must solve the preconditioning equation
            Mv = u, where u and v are stored in `u` and `v`,
            respectively.

        `irevcm` = 3

            Monitoring step: the solution and residual at the current
            iteration are returned in the arrays `u` and `v`,
            respectively. No action by the calling program is required.
            To return additional information ``complex_herm_basic_diag``
            can be called at this step.

        `On final exit`: if `irevcm` = 4, ``complex_herm_basic_solver``
        has completed its tasks. The value of errno determines whether
        the iteration has been successfully completed, errors have been
        detected or the calling program has requested termination.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On intermediate re-entry, `irevcm` = *<value>*.

            Constraint: either `irevcm` must be unchanged from its
            previous exit value or `irevcm` = 5 or 6.

        (`errno` -1)
            On initial entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0.

        (`errno` 3)
            Either ``complex_herm_basic_setup`` was not called before
            calling ``complex_herm_basic_solver`` or it has returned an
            error.

        (`errno` 9)
            The matrix of the coefficients A appears to be singular. The
            computation cannot continue.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            ``complex_herm_basic_solver`` has already completed its
            tasks. You need to set a new problem.

        (`errno` 2)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 2)
            User-requested termination: the required accuracy could not
            be obtained. However, a reasonable accuracy may have been
            achieved.

        (`errno` 4)
            User-requested tidy termination. The solution has not
            converged after *<value>* iterations.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            The preconditioner appears not to be positive definite. The
            computation cannot continue.

        (`errno` 7)
            The matrix of the coefficients A appears not to be positive
            definite. The computation cannot continue.

        (`errno` 8)
            User-requested immediate termination.

        (`errno` 10)
            The weights in array `wgt` are all zero.

    Notes
    -----
    ``complex_herm_basic_solver`` solves the complex Hermitian system of
    linear simultaneous equations Ax = b using either the preconditioned
    conjugate gradient method (see Hestenes and Stiefel (1952), Golub
    and Van Loan (1996), Barrett `et al.` (1994) and Dias da Cunha and
    Hopkins (1994)) or a preconditioned Lanczos method based upon the
    algorithm SYMMLQ (see Paige and Saunders (1975) and Barrett `et al.`
    (1994)).

    For a general description of the methods employed you are referred
    to [Notes for complex_herm_basic_setup].

    ``complex_herm_basic_solver`` can solve the system after the first
    function in the suite, ``complex_herm_basic_setup``, has been called
    to initialize the computation and specify the method of solution.
    The third function in the suite, ``complex_herm_basic_diag``, can be
    used to return additional information generated by the computation
    during monitoring steps and after ``complex_herm_basic_solver`` has
    completed its tasks.

    ``complex_herm_basic_solver`` uses **reverse communication**, i.e.,
    ``complex_herm_basic_solver`` returns repeatedly to the calling
    program with the argument `irevcm` (see [Parameters]) set to
    specified values which require the calling program to carry out a
    specific task: either to compute the matrix-vector product v = Au;
    to solve the preconditioning equation Mv = u; to notify the
    completion of the computation; or, to allow the calling program to
    monitor the solution.
    Through the argument `irevcm` the calling program can cause
    immediate or tidy termination of the execution.
    On final exit, the last iterates of the solution and of the residual
    vectors of the original system of equations are returned.

    Reverse communication has the following advantages.

    (1) Maximum flexibility in the representation and storage of sparse
        matrices. All matrix operations are performed outside the solver
        function, thereby avoiding the need for a complicated interface
        with enough flexibility to cope with all types of storage
        schemes and sparsity patterns. This applies also to
        preconditioners.

    (#) Enhanced user interaction: you can closely monitor the progress
        of the solution and tidy or immediate termination can be
        requested. This is useful, for example, when alternative
        termination criteria are to be employed or in case of failure of
        the external functions used to perform matrix operations.

    References
    ----------
    Barrett, R, Berry, M, Chan, T F, Demmel, J, Donato, J, Dongarra, J,
    Eijkhout, V, Pozo, R, Romine, C and Van der Vorst, H, 1994,
    `Templates for the Solution of Linear Systems: Building Blocks for
    Iterative Methods`, SIAM, Philadelphia

    Dias da Cunha, R and Hopkins, T, 1994, `PIM 1.1 --- the parallel
    iterative method package for systems of linear equations user's
    guide --- Fortran 77 version`, Technical Report, Computing
    Laboratory, University of Kent at Canterbury, Kent, UK

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Hestenes, M and Stiefel, E, 1952, `Methods of conjugate gradients
    for solving linear systems`, J. Res. Nat. Bur. Stand. (49), 409--436

    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396

    Paige, C C and Saunders, M A, 1975, `Solution of sparse indefinite
    systems of linear equations`, SIAM J. Numer. Anal. (12), 617--629
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_basic_solver'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11gsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input/Output, irevcm.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, u[*].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, v[*].
        utils._EngineFloat64ArrayType.c_type, # Input, wgt[*].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _irevcm = utils._EngineIntScalarType(
        irevcm, {'fun_name': fun_name, 'entity': 'irevcm'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _lwork_py = utils._EngineComplex128ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    if _lwork_py >= 20:
        if int(utils._EngineComplex128ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (4,), validate=False).real) == 1739512 and int(utils._EngineComplex128ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (19,), validate=False).real) == 840653:
            _vl_n_py = int(utils._EngineComplex128ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (10,), validate=False).real)
        else:
            _vl_n_py = 0
    else:
        _vl_n_py = 0
    _u_cls = utils._EngineComplex128ArrayType(
        u, {'fun_name': fun_name, 'entity': 'u'},
        exp_shape=(_vl_n_py,),
    )
    _u = _u_cls.to_ctypes()
    _v_cls = utils._EngineComplex128ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_vl_n_py,),
    )
    _v = _v_cls.to_ctypes()
    _wgt = utils._EngineFloat64ArrayType(
        wgt, {'fun_name': fun_name, 'entity': 'wgt'},
        exp_shape=(_vl_n_py,),
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _irevcm,
        _u,
        _v,
        _wgt,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _irevcm_py = utils._EngineIntScalarType.to_py(_irevcm)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _irevcm_py != 4 or _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On intermediate re-entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: either irevcm must be unchanged from its previous exit value or',
                    'irevcm = 5 or 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On initial entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irevcm = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'complex_herm_basic_solver has already completed its tasks. You need to set a new problem.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested termination: the required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Either complex_herm_basic_setup was not called before calling complex_herm_basic_solver',
                    'or it has returned an error.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested tidy termination.',
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99992:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The preconditioner appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99991:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients A appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99990:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested immediate termination.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients A appears to be singular.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99988:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The weights in array wgt are all zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _u_cls.refresh_py(_u)
        _v_cls.refresh_py(_v)
        return_data = _irevcm_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['irevcm'],
            )(
                _irevcm_py,
            )
    _err_class.manage()
    return return_data

def complex_herm_basic_diag(comm):
    u"""
    Complex sparse Hermitian linear systems, diagnostic for
    ``complex_herm_basic_solver``.

    ``complex_herm_basic_diag`` is the third in a suite of three
    functions for the iterative solution of a complex Hermitian system
    of simultaneous linear equations (see Golub and Van Loan (1996)).
    ``complex_herm_basic_diag`` returns information about the
    computations during an iteration and/or after this has been
    completed.
    The first function of the suite, ``complex_herm_basic_setup``, is a
    setup function, the second function, ``complex_herm_basic_solver``
    is the proper iterative solver.

    These three functions are suitable for the solution of large sparse
    complex Hermitian systems of equations.

    For full information please refer to the NAG Library document for
    f11gt

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11gtf.html

    Parameters
    ----------
    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``complex_herm_basic_solver``.

    Returns
    -------
    itn : int
        The number of iterations carried out by
        ``complex_herm_basic_solver``.

    stplhs : float
        The current value of the left-hand side of the termination
        criterion used by ``complex_herm_basic_solver``.

    stprhs : float
        The current value of the right-hand side of the termination
        criterion used by ``complex_herm_basic_solver``.

    anorm : float
        The norm ||A||_1 = ||A||_infinity when either it has been
        supplied to ``complex_herm_basic_setup`` or it has been
        estimated by ``complex_herm_basic_solver`` (see also [Notes for
        complex_herm_basic_setup] and [Parameters for
        complex_herm_basic_setup]).

        Otherwise, `anorm` = 0.0 is returned.

    sigmax : float
        The current estimate of the largest singular value sigma_1(A_)
        of the preconditioned iteration matrix A_= E^-1AE^(-H), when
        either it has been supplied to ``complex_herm_basic_setup`` or
        it has been estimated by ``complex_herm_basic_solver`` (see also
        [Notes for complex_herm_basic_setup] and [Parameters for
        complex_herm_basic_setup]). Note that if `its` < `itn` then
        `sigmax` contains the final estimate. If, on final exit from
        ``complex_herm_basic_solver``, `its` = `itn`, the estimation of
        sigma_1(A_) may have not converged: in this case you should look
        at the value returned in `sigerr`. Otherwise, `sigmax` = 0.0 is
        returned.

    its : int
        The number of iterations employed so far in the computation of
        the estimate of sigma_1(A_), the largest singular value of the
        preconditioned matrix A_= E^-1AE^(-H), when sigma_1(A_) has been
        estimated by ``complex_herm_basic_solver`` using the bisection
        method (see also [Notes for complex_herm_basic_setup],
        [Parameters for complex_herm_basic_setup] and [Further
        Comments]). Otherwise, `its` = 0 is returned.

    sigerr : float
        If sigma_1(A_) has been estimated by
        ``complex_herm_basic_solver`` using bisection,

            `sigerr` = max(|sigma_1^(k)-sigma_1^(k-1)|/sigma_1^(k),
            |sigma_1^(k)-sigma_1^(k-2)|/sigma_1^(k)),

        where k = `its` denotes the iteration number. The estimation has
        converged if `sigerr` <= sigtol where sigtol is an input
        argument to ``complex_herm_basic_setup``.

        Otherwise, `sigerr` = 0.0 is returned.

    Raises
    ------
    NagValueError
        (`errno` 1)
            ``complex_herm_basic_diag`` has been called out of sequence.

    Notes
    -----
    ``complex_herm_basic_diag`` returns information about the solution
    process.
    It can be called both during a monitoring step of the solver
    ``complex_herm_basic_solver`` or after this solver has completed its
    tasks.
    Calling ``complex_herm_basic_diag`` at any other time will result in
    an error condition being raised.

    For further information you should read the documentation for
    ``complex_herm_basic_setup`` and ``complex_herm_basic_solver``.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_basic_diag'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11gtft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, itn.
        utils._BYREF_FLOAT64, # Output, stplhs.
        utils._BYREF_FLOAT64, # Output, stprhs.
        utils._BYREF_FLOAT64, # Output, anorm.
        utils._BYREF_FLOAT64, # Output, sigmax.
        utils._BYREF_INT, # Output, its.
        utils._BYREF_FLOAT64, # Output, sigerr.
        utils._EngineComplex128ArrayType.c_type, # Input, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _stplhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _stprhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _anorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _sigmax = utils._EngineFloat64ScalarType.empty_ctypes()
    _its = utils._EngineIntScalarType.empty_ctypes()
    _sigerr = utils._EngineFloat64ScalarType.empty_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _lwork_py = utils._EngineComplex128ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _itn,
        _stplhs,
        _stprhs,
        _anorm,
        _sigmax,
        _its,
        _sigerr,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'complex_herm_basic_diag has been called out of sequence.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        _stplhs_py = utils._EngineFloat64ScalarType.to_py(_stplhs)
        _stprhs_py = utils._EngineFloat64ScalarType.to_py(_stprhs)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _sigmax_py = utils._EngineFloat64ScalarType.to_py(_sigmax)
        _its_py = utils._EngineIntScalarType.to_py(_its)
        _sigerr_py = utils._EngineFloat64ScalarType.to_py(_sigerr)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['itn', 'stplhs', 'stprhs', 'anorm', 'sigmax', 'its', 'sigerr'],
        )(
            _itn_py,
            _stplhs_py,
            _stprhs_py,
            _anorm_py,
            _sigmax_py,
            _its_py,
            _sigerr_py,
        )
    _err_class.manage()
    return return_data

def real_symm_precon_ichol(n, nnz, a, la, irow, icol, lfill, dtol, mic, dscale, pstrat, ipiv, istr):
    u"""
    Real sparse symmetric matrix, incomplete Cholesky factorization.

    ``real_symm_precon_ichol`` computes an incomplete Cholesky
    factorization of a real sparse symmetric matrix, represented in
    symmetric coordinate storage format.
    This factorization may be used as a preconditioner in combination
    with ``real_symm_solve_ichol`` or ``real_symm_basic_solver``.

    For full information please refer to the NAG Library document for
    f11ja

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11jaf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A.

    a : float, shape (`la`), modified in place
        `On entry`: the nonzero elements in the lower triangular part of
        the matrix A, ordered by increasing row index, and by increasing
        column index within each row. Multiple entries for the same row
        and column indices are not permitted. The function
        ``real_symm_sort`` may be used to order the elements in this
        way.

        `On exit`: the first `nnz` elements of `a` contain the nonzero
        elements of A and the next `nnzc` elements contain the elements
        of the lower triangular matrix C. Matrix elements are ordered by
        increasing row index, and by increasing column index within each
        row.

    la : int
        These arrays must be of sufficient size to store both A (`nnz`
        elements) and C (`nnzc` elements).

    irow : int, shape (`la`), modified in place
        `On entry`: the row indices of the nonzero elements supplied in
        `a`.

        `On exit`: the row indices of the nonzero elements returned in
        `a`.

    icol : int, shape (`la`), modified in place
        `On entry`: the column indices of the nonzero elements supplied
        in `a`.

        `On exit`: the column indices of the nonzero elements returned
        in `a`.

    lfill : int
        If `lfill` >= 0 its value is the maximum level of fill allowed
        in the decomposition (see [Further Comments]). A negative value
        of `lfill` indicates that `dtol` will be used to control the
        fill instead.

    dtol : float
        If `lfill` < 0, `dtol` is used as a drop tolerance to control
        the fill-in (see [Further Comments]); otherwise `dtol` is not
        referenced.

    mic : str, length 1
        Indicates whether or not the factorization should be modified to
        preserve row sums (see [Further Comments]).

        `mic` = 'M'

            The factorization is modified.

        `mic` = 'N'

            The factorization is not modified.

    dscale : float
        The diagonal scaling parameter. All diagonal elements are
        multiplied by the factor (1+`dscale`) at the start of the
        factorization. This can be used to ensure that the
        preconditioner is positive definite. See [Further Comments].

    pstrat : str, length 1
        Specifies the pivoting strategy to be adopted.

        `pstrat` = 'N'

            No pivoting is carried out.

        `pstrat` = 'M'

            Diagonal pivoting aimed at minimizing fill-in is carried
            out, using the Markowitz strategy.

        `pstrat` = 'U'

            Diagonal pivoting is carried out according to the
            user-defined input value of `ipiv`.

        Suggested value: `pstrat` = 'M'.

    ipiv : int, shape (`n`), modified in place
        `On entry`: if `pstrat` = 'U', `ipiv`[i-1] must specify the row
        index of the diagonal element used as a pivot at elimination
        stage i. Otherwise `ipiv` need not be initialized.

        `On exit`: the pivot indices. If `ipiv`[i-1] = j then the
        diagonal element in row j was used as the pivot at elimination
        stage i.

    istr : int, shape (`n`+1), modified in place
        `On exit`: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix C. `istr`[`n`]-1 is the address of the last nonzero
        element in C plus one.

    Returns
    -------
    nnzc : int
        The number of nonzero elements in the lower triangular matrix C.

    npivm : int
        The number of pivots which were modified during the
        factorization to ensure that M was positive definite. The
        quality of the preconditioner will generally depend on the
        returned value of `npivm`. If `npivm` is large the
        preconditioner may not be satisfactory. In this case it may be
        advantageous to call ``real_symm_precon_ichol`` again with an
        increased value of either `lfill` or `dscale`. See also [Further
        Comments].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `pstrat` != 'N', 'U' or 'M': `pstrat` = *<value>*.

        (`errno` 1)
            On entry, `mic` != 'M' or 'N': `mic` = *<value>*.

        (`errno` 1)
            On entry, `dtol` = *<value>*.

            Constraint: `dtol` >= 0.0.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 2)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, a user-supplied value of `ipiv` is repeated.

        (`errno` 3)
            On entry, a user-supplied value of `ipiv` lies outside the
            range [1,`n`].

        (`errno` 4)
            The number of nonzero entries in the decomposition is too
            large. The decomposition has been terminated before
            completion. Either increase `la`, or reduce the fill by
            setting `pstrat` = 'M', reducing `lfill`, or increasing
            `dtol`.

        (`errno` 5)
            A serious error has occurred in an internal call. Check all
            function calls and array sizes. Seek expert help.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``real_symm_precon_ichol`` computes an incomplete Cholesky
    factorization (see Meijerink and Van der Vorst (1977)) of a real
    sparse symmetric n*n matrix A.
    It is designed specifically for positive definite matrices, but may
    also work for some mildly indefinite cases.
    The factorization is intended primarily for use as a preconditioner
    with one of the symmetric iterative solvers
    ``real_symm_solve_ichol`` or ``real_symm_basic_solver``.

    The decomposition is written in the form

        A = M+R

    where

        M = PLDL^TP^T

    and P is a permutation matrix, L is lower triangular with unit
    diagonal elements, D is diagonal and R is a remainder matrix.

    The amount of fill-in occurring in the factorization can vary from
    zero to complete fill, and can be controlled by specifying either
    the maximum level of fill `lfill`, or the drop tolerance `dtol`.
    The factorization may be modified in order to preserve row sums, and
    the diagonal elements may be perturbed to ensure that the
    preconditioner is positive definite.
    Diagonal pivoting may optionally be employed, either with a
    user-defined ordering, or using the Markowitz strategy (see
    Markowitz (1957)), which aims to minimize fill-in.
    For further details see [Further Comments].

    The sparse matrix A is represented in symmetric coordinate storage
    (SCS) format (see [the F11 Introduction]).
    The array `a` stores all the nonzero elements of the lower
    triangular part of A, while arrays `irow` and `icol` store the
    corresponding row and column indices respectively.
    Multiple nonzero elements may not be specified for the same row and
    column index.

    The preconditioning matrix M is returned in terms of the SCS
    representation of the lower triangular matrix

        C = L+D^-1-I.

    References
    ----------
    Chan, T F, 1991, `Fourier analysis of relaxed incomplete
    factorization preconditioners`, SIAM J. Sci. Statist. Comput.
    (12(2)), 668--680

    Markowitz, H M, 1957, `The elimination form of the inverse and its
    application to linear programming`, Management Sci. (3), 255--269

    Meijerink, J and Van der Vorst, H, 1977, `An iterative solution
    method for linear systems of which the coefficient matrix is a
    symmetric M-matrix`, Math. Comput. (31), 148--162

    Salvini, S A and Shaw, G J, 1995, `An evaluation of new NAG Library
    solvers for large sparse symmetric linear systems`, NAG Technical
    Report TR1/95

    Van der Vorst, H A, 1990, `The convergence behaviour of
    preconditioned CG and CG-S in the presence of rounding errors`,
    Lecture Notes in Mathematics, (eds O Axelsson and L Y Kolotilina)
    (1457), Springer--Verlag
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_precon_ichol'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input/Output, irow[la].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[la].
        utils._BYREF_INT, # Input, lfill.
        utils._BYREF_FLOAT64, # Input, dtol.
        utils._BYREF_STRING, # Input, mic(len=1).
        utils._BYREF_FLOAT64, # Input, dscale.
        utils._BYREF_STRING, # Input, pstrat(len=1).
        utils._EngineIntArrayType.c_type, # Input/Output, ipiv[n].
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._BYREF_INT, # Output, nnzc.
        utils._BYREF_INT, # Output, npivm.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._BYREF_INT, # Input, liwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mic.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pstrat.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _lfill = utils._EngineIntScalarType(
        lfill, {'fun_name': fun_name, 'entity': 'lfill'},
    ).to_ctypes()
    _dtol = utils._EngineFloat64ScalarType(
        dtol, {'fun_name': fun_name, 'entity': 'dtol'},
    ).to_ctypes()
    _mic = utils._EngineCharScalarType(
        mic, {'fun_name': fun_name, 'entity': 'mic'}, exp_chrlen=1,
    ).to_ctypes()
    _dscale = utils._EngineFloat64ScalarType(
        dscale, {'fun_name': fun_name, 'entity': 'dscale'},
    ).to_ctypes()
    _pstrat = utils._EngineCharScalarType(
        pstrat, {'fun_name': fun_name, 'entity': 'pstrat'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
    )
    _ipiv = _ipiv_cls.to_ctypes()
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _nnzc = utils._EngineIntScalarType.empty_ctypes()
    _npivm = utils._EngineIntScalarType.empty_ctypes()
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _liwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _lfill,
        _dtol,
        _mic,
        _dscale,
        _pstrat,
        _ipiv,
        _istr,
        _nnzc,
        _npivm,
        _iwork,
        _liwork,
        _errbuf,
        _ierr,
        _mic._length_,
        _pstrat._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pstrat != \'N\', \'U\' or \'M\': pstrat = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mic != \'M\' or \'N\': mic = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dtol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dtol >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipiv is repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipiv lies outside the range [1,n].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of nonzero entries in the decomposition is too large.',
                    'The decomposition has been terminated before completion.',
                    'Either increase la, or reduce the fill by setting pstrat = \'M\',',
                    'reducing lfill, or increasing dtol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call.',
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _ipiv_cls.refresh_py(_ipiv)
        _istr_cls.refresh_py(_istr)
        _nnzc_py = utils._EngineIntScalarType.to_py(_nnzc)
        _npivm_py = utils._EngineIntScalarType.to_py(_npivm)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nnzc', 'npivm'],
        )(
            _nnzc_py,
            _npivm_py,
        )
    _err_class.manage()
    return return_data

def real_symm_precon_ichol_solve(n, a, la, irow, icol, ipiv, istr, check, y, x):
    u"""
    Solution of linear system involving incomplete Cholesky
    preconditioning matrix generated by ``real_symm_precon_ichol``.

    ``real_symm_precon_ichol_solve`` solves a system of linear equations
    involving the incomplete Cholesky preconditioning matrix generated
    by ``real_symm_precon_ichol``.

    For full information please refer to the NAG Library document for
    f11jb

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11jbf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix M. This **must** be the same value as
        was supplied in the preceding call to
        ``real_symm_precon_ichol``.

    a : float, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``real_symm_precon_ichol``.

    la : int
        This **must** be the same value returned by the preceding call
        to ``real_symm_precon_ichol``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    ipiv : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    check : str, length 1
        Specifies whether or not the input data should be checked.

        `check` = 'C'

            Checks are carried out on the values of `n`, `irow`, `icol`,
            `ipiv` and `istr`.

        `check` = 'N'

            No checks are carried out.

        See also [Further Comments].

    y : float, shape (`n`)
        The right-hand side vector y.

    x : float, shape (`n`), modified in place
        `On exit`: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `ipiv`[i-1] is a repeated value: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `ipiv`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `ipiv`[i-1] >= 1 and `ipiv`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr`[i-1] is inconsistent with `irow`: i =
            *<value>*.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr` appears to be invalid.

    Notes
    -----
    ``real_symm_precon_ichol_solve`` solves a system of linear equations

        Mx = y

    involving the preconditioning matrix M = PLDL^TP^T, corresponding to
    an incomplete Cholesky decomposition of a sparse symmetric matrix
    stored in symmetric coordinate storage (SCS) format (see [the F11
    Introduction]), as generated by ``real_symm_precon_ichol``.

    In the above decomposition L is a lower triangular sparse matrix
    with unit diagonal, D is a diagonal matrix and P is a permutation
    matrix. L and D are supplied to ``real_symm_precon_ichol_solve``
    through the matrix

        C = L+D^-1-I

    which is a lower triangular `n` by `n` sparse matrix, stored in SCS
    format, as returned by ``real_symm_precon_ichol``.
    The permutation matrix P is returned from ``real_symm_precon_ichol``
    via the array `ipiv`.

    It is envisaged that a common use of
    ``real_symm_precon_ichol_solve`` will be to carry out the
    preconditioning step required in the application of
    ``real_symm_basic_solver`` to sparse symmetric linear systems.
    ``real_symm_precon_ichol_solve`` is used for this purpose by the
    Black Box function ``real_symm_solve_ichol``.

    ``real_symm_precon_ichol_solve`` may also be used in combination
    with ``real_symm_precon_ichol`` to solve a sparse symmetric positive
    definite system of linear equations directly (see [Further
    Comments]).
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_precon_ichol_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipiv[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
    )
    _ipiv = _ipiv_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _la,
        _irow,
        _icol,
        _ipiv,
        _istr,
        _check,
        _y,
        _x,
        _errbuf,
        _ierr,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ipiv[i-1] is a repeated value:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ipiv[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ipiv[i-1] >= 1 and ipiv[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr[i-1] is inconsistent with irow:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'irow[i-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr appears to be invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ipiv_cls.refresh_py(_ipiv)
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def real_symm_solve_ichol(method, n, nnz, a, la, irow, icol, ipiv, istr, b, tol, maxitn, x):
    u"""
    Solution of real sparse symmetric linear system, conjugate
    gradient/Lanczos method, preconditioner computed by
    ``real_symm_precon_ichol`` (Black Box).

    ``real_symm_solve_ichol`` solves a real sparse symmetric system of
    linear equations, represented in symmetric coordinate storage
    format, using a conjugate gradient or Lanczos method, with
    incomplete Cholesky preconditioning.

    For full information please refer to the NAG Library document for
    f11jc

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11jcf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'CG'

            Conjugate gradient method.

        `method` = 'SYMMLQ'

            Lanczos method (SYMMLQ).

    n : int
        n, the order of the matrix A. This **must** be the same value as
        was supplied in the preceding call to
        ``real_symm_precon_ichol``.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A. This **must** be the same value as was supplied in
        the preceding call to ``real_symm_precon_ichol``.

    a : float, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``real_symm_precon_ichol``.

    la : int
        This **must** be the same value as was supplied in the preceding
        call to ``real_symm_precon_ichol``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    ipiv : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    b : float, shape (`n`)
        The right-hand side vector b.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : float, shape (`n`), modified in place
        `On entry`: an initial approximation to the solution vector x.

        `On exit`: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'CG' or 'SYMMLQ'.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[i-1],`icol`[i-1]) is a
            duplicate: i = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>*,
            `irow`[i-1] = *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `irow`[i-1].

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            The SCS representation of the preconditioner is invalid.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            The preconditioner appears not to be positive definite. The
            computation cannot continue.

        (`errno` 7)
            The matrix of the coefficients `a` appears not to be
            positive definite. The computation cannot continue.

        (`errno` 8)
            A serious error has occurred in an internal call: IREVCM =
            *<value>*. Check all function calls and array sizes. Seek
            expert help.

        (`errno` 8)
            A serious error has occurred in an internal call: errno =
            *<value>*. Check all function calls and array sizes. Seek
            expert help.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The required accuracy could not be obtained. However a
            reasonable accuracy has been achieved.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``real_symm_solve_ichol`` solves a real sparse symmetric linear
    system of equations

        Ax = b,

    using a preconditioned conjugate gradient method (see Meijerink and
    Van der Vorst (1977)), or a preconditioned Lanczos method based on
    the algorithm SYMMLQ (see Paige and Saunders (1975)).
    The conjugate gradient method is more efficient if A is positive
    definite, but may fail to converge for indefinite matrices.
    In this case the Lanczos method should be used instead.
    For further details see Barrett `et al.` (1994).

    ``real_symm_solve_ichol`` uses the incomplete Cholesky factorization
    determined by ``real_symm_precon_ichol`` as the preconditioning
    matrix.
    A call to ``real_symm_solve_ichol`` must always be preceded by a
    call to ``real_symm_precon_ichol``.
    Alternative preconditioners for the same storage scheme are
    available by calling ``real_symm_solve_jacssor``.

    The matrix A, and the preconditioning matrix M, are represented in
    symmetric coordinate storage (SCS) format (see [the F11
    Introduction]) in the arrays `a`, `irow` and `icol`, as returned
    from ``real_symm_precon_ichol``.
    The array `a` holds the nonzero entries in the lower triangular
    parts of these matrices, while `irow` and `icol` hold the
    corresponding row and column indices.

    References
    ----------
    Barrett, R, Berry, M, Chan, T F, Demmel, J, Donato, J, Dongarra, J,
    Eijkhout, V, Pozo, R, Romine, C and Van der Vorst, H, 1994,
    `Templates for the Solution of Linear Systems: Building Blocks for
    Iterative Methods`, SIAM, Philadelphia

    Meijerink, J and Van der Vorst, H, 1977, `An iterative solution
    method for linear systems of which the coefficient matrix is a
    symmetric M-matrix`, Math. Comput. (31), 148--162

    Paige, C C and Saunders, M A, 1975, `Solution of sparse indefinite
    systems of linear equations`, SIAM J. Numer. Anal. (12), 617--629

    Salvini, S A and Shaw, G J, 1995, `An evaluation of new NAG Library
    solvers for large sparse symmetric linear systems`, NAG Technical
    Report TR1/95
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_solve_ichol'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipiv[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
    )
    _ipiv = _ipiv_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _ipiv,
        _istr,
        _b,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'CG\' or \'SYMMLQ\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[i-1],icol[i-1]) is a duplicate:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[i-1] = {:s}, irow[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= irow[i-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The SCS representation of the preconditioner is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99986:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However a reasonable accuracy has been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The preconditioner appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients a appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call:',
                    'IREVCM = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call:',
                    'errno = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ipiv_cls.refresh_py(_ipiv)
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_symm_precon_ssor_solve(n, nnz, a, irow, icol, rdiag, omega, check, y, x):
    u"""
    Solution of linear system involving preconditioning matrix generated
    by applying SSOR to real sparse symmetric matrix.

    ``real_symm_precon_ssor_solve`` solves a system of linear equations
    involving the preconditioning matrix corresponding to SSOR applied
    to a real sparse symmetric matrix, represented in symmetric
    coordinate storage format.

    For full information please refer to the NAG Library document for
    f11jd

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11jdf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        A.

    a : float, shape (`nnz`)
        The nonzero elements in the lower triangular part of the matrix
        A, ordered by increasing row index, and by increasing column
        index within each row. Multiple entries for the same row and
        column indices are not permitted. The function
        ``real_symm_sort`` may be used to order the elements in this
        way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    rdiag : float, shape (`n`)
        The elements of the diagonal matrix D^-1, where D is the
        diagonal part of A.

    omega : float
        The relaxation parameter omega.

    check : str, length 1
        Specifies whether or not the input data should be checked.

        `check` = 'C'

            Checks are carried out on the values of `n`, `nnz`, `irow`,
            `icol` and `omega`.

        `check` = 'N'

            None of these checks are carried out.

        See also [Further Comments].

    y : float, shape (`n`)
        The right-hand side vector y.

    x : float, shape (`n`), modified in place
        `On exit`: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` != 'C' or 'N': `check` = *<value>*.

        (`errno` 2)
            On entry, `omega` = *<value>*.

            Constraint: 0.0 < `omega` < 2.0

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 4)
            The matrix A has no diagonal entry in row *<value>*.

    Notes
    -----
    ``real_symm_precon_ssor_solve`` solves a system of equations

        Mx = y

    involving the preconditioning matrix

        M = 1/omega(2-omega)(D+omega L)D^-1(D+omega L)^T

    corresponding to symmetric successive-over-relaxation (SSOR) (see
    Young (1971)) on a linear system Ax = b, where A is a sparse
    symmetric matrix stored in symmetric coordinate storage (SCS) format
    (see [the F11 Introduction]).

    In the definition of M given above D is the diagonal part of A, L is
    the strictly lower triangular part of A, and omega is a user-defined
    relaxation parameter.

    It is envisaged that a common use of ``real_symm_precon_ssor_solve``
    will be to carry out the preconditioning step required in the
    application of ``real_symm_basic_solver`` to sparse linear systems.
    ``real_symm_precon_ssor_solve`` is also used for this purpose by the
    Black Box function ``real_symm_solve_jacssor``.

    References
    ----------
    Young, D, 1971, `Iterative Solution of Large Linear Systems`,
    Academic Press, New York
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_precon_ssor_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._EngineFloat64ArrayType.c_type, # Input, rdiag[n].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _rdiag = utils._EngineFloat64ArrayType(
        rdiag, {'fun_name': fun_name, 'entity': 'rdiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _rdiag,
        _omega,
        _check,
        _y,
        _x,
        _iwork,
        _errbuf,
        _ierr,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check != \'C\' or \'N\': check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < omega < 2.0',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def real_symm_solve_jacssor(method, precon, n, nnz, a, irow, icol, omega, b, tol, maxitn, x):
    u"""
    Solution of real sparse symmetric linear system, conjugate
    gradient/Lanczos method, Jacobi or SSOR preconditioner (Black Box).

    ``real_symm_solve_jacssor`` solves a real sparse symmetric system of
    linear equations, represented in symmetric coordinate storage
    format, using a conjugate gradient or Lanczos method, without
    preconditioning, with Jacobi or with SSOR preconditioning.

    For full information please refer to the NAG Library document for
    f11je

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11jef.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'CG'

            Conjugate gradient method.

        `method` = 'SYMMLQ'

            Lanczos method (SYMMLQ).

    precon : str, length 1
        Specifies the type of preconditioning to be used.

        `precon` = 'N'

            No preconditioning.

        `precon` = 'J'

            Jacobi.

        `precon` = 'S'

            Symmetric successive-over-relaxation (SSOR).

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A.

    a : float, shape (`nnz`)
        The nonzero elements of the lower triangular part of the matrix
        A, ordered by increasing row index, and by increasing column
        index within each row. Multiple entries for the same row and
        column indices are not permitted. The function
        ``real_symm_sort`` may be used to order the elements in this
        way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    omega : float
        If `precon` = 'S', `omega` is the relaxation parameter omega to
        be used in the SSOR method. Otherwise `omega` need not be
        initialized.

    b : float, shape (`n`)
        The right-hand side vector b.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : float, shape (`n`), modified in place
        `On entry`: an initial approximation to the solution vector x.

        `On exit`: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'CG' or 'SYMMLQ'.

        (`errno` 1)
            On entry, `omega` = *<value>*.

            Constraint: `omega` > 0.0 and `omega` < 2.0.

        (`errno` 1)
            On entry, `precon` != 'N', 'J' or 'S': `precon` = *<value>*.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            The matrix A has a zero diagonal entry in row *<value>*.

        (`errno` 3)
            The matrix A has no diagonal entry in row *<value>*.

        (`errno` 4)
            The required accuracy could not be obtained. However a
            reasonable accuracy has been achieved.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            The preconditioner appears not to be positive definite. The
            computation cannot continue.

        (`errno` 7)
            The matrix of the coefficients `a` appears not to be
            positive definite. The computation cannot continue.

        (`errno` 8)
            A serious error has occurred in an internal call: IREVCM =
            *<value>*. Check all function calls and array sizes. Seek
            expert help.

        (`errno` 8)
            A serious error has occurred in an internal call: errno =
            *<value>*. Check all function calls and array sizes. Seek
            expert help.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``real_symm_solve_jacssor`` solves a real sparse symmetric linear
    system of equations

        Ax = b,

    using a preconditioned conjugate gradient method (see Barrett `et
    al.` (1994)), or a preconditioned Lanczos method based on the
    algorithm SYMMLQ (see Paige and Saunders (1975)).
    The conjugate gradient method is more efficient if A is positive
    definite, but may fail to converge for indefinite matrices.
    In this case the Lanczos method should be used instead.
    For further details see Barrett `et al.` (1994).

    The function allows the following choices for the preconditioner:

        no preconditioning;

        Jacobi preconditioning (see Young (1971));

        symmetric successive-over-relaxation (SSOR) preconditioning (see
        Young (1971)).

    For incomplete Cholesky (IC) preconditioning see
    ``real_symm_solve_ichol``.

    The matrix A is represented in symmetric coordinate storage (SCS)
    format (see [the F11 Introduction]) in the arrays `a`, `irow` and
    `icol`.
    The array `a` holds the nonzero entries in the lower triangular part
    of the matrix, while `irow` and `icol` hold the corresponding row
    and column indices.

    References
    ----------
    Barrett, R, Berry, M, Chan, T F, Demmel, J, Donato, J, Dongarra, J,
    Eijkhout, V, Pozo, R, Romine, C and Van der Vorst, H, 1994,
    `Templates for the Solution of Linear Systems: Building Blocks for
    Iterative Methods`, SIAM, Philadelphia

    Paige, C C and Saunders, M A, 1975, `Solution of sparse indefinite
    systems of linear equations`, SIAM J. Numer. Anal. (12), 617--629

    Young, D, 1971, `Iterative Solution of Large Linear Systems`,
    Academic Press, New York
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_solve_jacssor'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _omega,
        _b,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _iwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'CG\' or \'SYMMLQ\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: omega > 0.0 and omega < 2.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon != \'N\', \'J\' or \'S\': precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has a zero diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However a reasonable accuracy has been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The preconditioner appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients a appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99979:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call:',
                    'IREVCM = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99980:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call:',
                    'errno = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
    _err_class.manage()
    return return_data

def complex_herm_precon_ichol(n, nnz, a, la, irow, icol, lfill, dtol, mic, dscale, pstrat, ipiv, istr):
    u"""
    Complex sparse Hermitian matrix, incomplete Cholesky factorization.

    ``complex_herm_precon_ichol`` computes an incomplete Cholesky
    factorization of a complex sparse Hermitian matrix, represented in
    symmetric coordinate storage format.
    This factorization may be used as a preconditioner in combination
    with ``complex_herm_solve_ilu``.

    For full information please refer to the NAG Library document for
    f11jn

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11jnf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A.

    a : complex, shape (`la`), modified in place
        `On entry`: the nonzero elements in the lower triangular part of
        the matrix A, ordered by increasing row index, and by increasing
        column index within each row. Multiple entries for the same row
        and column indices are not permitted. The function
        ``complex_herm_sort`` may be used to order the elements in this
        way.

        `On exit`: the first `nnz` elements of `a` contain the nonzero
        elements of A and the next `nnzc` elements contain the elements
        of the lower triangular matrix C. Matrix elements are ordered by
        increasing row index, and by increasing column index within each
        row.

    la : int
        These arrays must be of sufficient size to store both A (`nnz`
        elements) and C (`nnzc` elements).

    irow : int, shape (`la`), modified in place
        `On entry`: the row indices of the nonzero elements supplied in
        `a`.

        `On exit`: the row indices of the nonzero elements returned in
        `a`.

    icol : int, shape (`la`), modified in place
        `On entry`: the column indices of the nonzero elements supplied
        in `a`.

        `On exit`: the column indices of the nonzero elements returned
        in `a`.

    lfill : int
        If `lfill` >= 0 its value is the maximum level of fill allowed
        in the decomposition (see [Further Comments]). A negative value
        of `lfill` indicates that `dtol` will be used to control the
        fill instead.

    dtol : float
        If `lfill` < 0, `dtol` is used as a drop tolerance to control
        the fill-in (see [Further Comments]); otherwise `dtol` is not
        referenced.

    mic : str, length 1
        Indicates whether or not the factorization should be modified to
        preserve row sums (see [Further Comments]).

        `mic` = 'M'

            The factorization is modified.

        `mic` = 'N'

            The factorization is not modified.

    dscale : float
        The diagonal scaling parameter. All diagonal elements are
        multiplied by the factor (1.0+`dscale`) at the start of the
        factorization. This can be used to ensure that the
        preconditioner is positive definite. See also [Further
        Comments].

    pstrat : str, length 1
        Specifies the pivoting strategy to be adopted.

        `pstrat` = 'N'

            No pivoting is carried out.

        `pstrat` = 'M'

            Diagonal pivoting aimed at minimizing fill-in is carried
            out, using the Markowitz strategy (see Markowitz (1957)).

        `pstrat` = 'U'

            Diagonal pivoting is carried out according to the
            user-defined input array `ipiv`.

        Suggested value: `pstrat` = 'M'.

    ipiv : int, shape (`n`), modified in place
        `On entry`: if `pstrat` = 'U', `ipiv`[i-1] must specify the row
        index of the diagonal element to be used as a pivot at
        elimination stage i. Otherwise `ipiv` need not be initialized.

        `On exit`: the pivot indices. If `ipiv`[i-1] = j, the diagonal
        element in row j was used as the pivot at elimination stage i.

    istr : int, shape (`n`+1), modified in place
        `On exit`: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix C. `istr`[`n`]-1 is the address of the last nonzero
        element in C plus one.

    Returns
    -------
    nnzc : int
        The number of nonzero elements in the lower triangular matrix C.

    npivm : int
        The number of pivots which were modified during the
        factorization to ensure that M was positive definite. The
        quality of the preconditioner will generally depend on the
        returned value of `npivm`. If `npivm` is large the
        preconditioner may not be satisfactory. In this case it may be
        advantageous to call ``complex_herm_precon_ichol`` again with an
        increased value of either `lfill` or `dscale`. See also [Further
        Comments].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `pstrat` != 'N', 'U' or 'M': `pstrat` = *<value>*.

        (`errno` 1)
            On entry, `mic` != 'M' or 'N': `mic` = *<value>*.

        (`errno` 1)
            On entry, `dtol` = *<value>*.

            Constraint: `dtol` >= 0.0

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 2)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 2)
            On entry, I = *<value>*, `irow`[I-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[I-1] >= 1 and `irow`[I-1] <= `n`.

        (`errno` 3)
            On entry, a user-supplied value of `ipiv` is repeated.

        (`errno` 3)
            On entry, a user-supplied value of `ipiv` lies outside the
            range [1,`n`].

        (`errno` 4)
            The number of nonzero entries in the decomposition is too
            large. The decomposition has been terminated before
            completion. Either increase `la`, or reduce the fill by
            setting `pstrat` = 'M', reducing `lfill`, or increasing
            `dtol`.

        (`errno` 5)
            A serious error has occurred in an internal call. Check all
            function calls and array sizes. Seek expert help.

    Notes
    -----
    ``complex_herm_precon_ichol`` computes an incomplete Cholesky
    factorization (see Meijerink and Van der Vorst (1977)) of a complex
    sparse Hermitian n*n matrix A.
    It is designed specifically for positive definite matrices, but may
    also work for some mildly indefinite cases.
    The factorization is intended primarily for use as a preconditioner
    with the complex Hermitian iterative solver
    ``complex_herm_solve_ilu``.

    The decomposition is written in the form

        A = M+R

    where

        M = PLDL^HP^T

    and P is a permutation matrix, L is lower triangular complex with
    unit diagonal elements, D is real diagonal and R is a remainder
    matrix.

    The amount of fill-in occurring in the factorization can vary from
    zero to complete fill, and can be controlled by specifying either
    the maximum level of fill `lfill`, or the drop tolerance `dtol`.
    The factorization may be modified in order to preserve row sums, and
    the diagonal elements may be perturbed to ensure that the
    preconditioner is positive definite.
    Diagonal pivoting may optionally be employed, either with a
    user-defined ordering, or using the Markowitz strategy (see
    Markowitz (1957)), which aims to minimize fill-in.
    For further details see [Further Comments].

    The sparse matrix A is represented in symmetric coordinate storage
    (SCS) format (see [the F11 Introduction]).
    The array `a` stores all the nonzero elements of the lower
    triangular part of A, while arrays `irow` and `icol` store the
    corresponding row and column indices respectively.
    Multiple nonzero elements may not be specified for the same row and
    column index.

    The preconditioning matrix M is returned in terms of the SCS
    representation of the lower triangular matrix

        C = L+D^-1-I.

    References
    ----------
    Chan, T F, 1991, `Fourier analysis of relaxed incomplete
    factorization preconditioners`, SIAM J. Sci. Statist. Comput.
    (12(2)), 668--680

    Markowitz, H M, 1957, `The elimination form of the inverse and its
    application to linear programming`, Management Sci. (3), 255--269

    Meijerink, J and Van der Vorst, H, 1977, `An iterative solution
    method for linear systems of which the coefficient matrix is a
    symmetric M-matrix`, Math. Comput. (31), 148--162

    Salvini, S A and Shaw, G J, 1995, `An evaluation of new NAG Library
    solvers for large sparse symmetric linear systems`, NAG Technical
    Report TR1/95

    Van der Vorst, H A, 1990, `The convergence behaviour of
    preconditioned CG and CG-S in the presence of rounding errors`,
    Lecture Notes in Mathematics, (eds O Axelsson and L Y Kolotilina)
    (1457), Springer--Verlag
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_precon_ichol'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input/Output, irow[la].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[la].
        utils._BYREF_INT, # Input, lfill.
        utils._BYREF_FLOAT64, # Input, dtol.
        utils._BYREF_STRING, # Input, mic(len=1).
        utils._BYREF_FLOAT64, # Input, dscale.
        utils._BYREF_STRING, # Input, pstrat(len=1).
        utils._EngineIntArrayType.c_type, # Input/Output, ipiv[n].
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._BYREF_INT, # Output, nnzc.
        utils._BYREF_INT, # Output, npivm.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._BYREF_INT, # Input, liwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mic.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pstrat.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _lfill = utils._EngineIntScalarType(
        lfill, {'fun_name': fun_name, 'entity': 'lfill'},
    ).to_ctypes()
    _dtol = utils._EngineFloat64ScalarType(
        dtol, {'fun_name': fun_name, 'entity': 'dtol'},
    ).to_ctypes()
    _mic = utils._EngineCharScalarType(
        mic, {'fun_name': fun_name, 'entity': 'mic'}, exp_chrlen=1,
    ).to_ctypes()
    _dscale = utils._EngineFloat64ScalarType(
        dscale, {'fun_name': fun_name, 'entity': 'dscale'},
    ).to_ctypes()
    _pstrat = utils._EngineCharScalarType(
        pstrat, {'fun_name': fun_name, 'entity': 'pstrat'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
    )
    _ipiv = _ipiv_cls.to_ctypes()
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _nnzc = utils._EngineIntScalarType.empty_ctypes()
    _npivm = utils._EngineIntScalarType.empty_ctypes()
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _liwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _lfill,
        _dtol,
        _mic,
        _dscale,
        _pstrat,
        _ipiv,
        _istr,
        _nnzc,
        _npivm,
        _iwork,
        _liwork,
        _errbuf,
        _ierr,
        _mic._length_,
        _pstrat._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pstrat != \'N\', \'U\' or \'M\': pstrat = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mic != \'M\' or \'N\': mic = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dtol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dtol >= 0.0',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, icol[I-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'irow[I-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[I-1] >= 1 and irow[I-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipiv is repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipiv lies outside the range [1,n].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of nonzero entries in the decomposition is too large.',
                    'The decomposition has been terminated before completion.',
                    'Either increase la, or reduce the fill by setting pstrat = \'M\',',
                    'reducing lfill, or increasing dtol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call.',
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _ipiv_cls.refresh_py(_ipiv)
        _istr_cls.refresh_py(_istr)
        _nnzc_py = utils._EngineIntScalarType.to_py(_nnzc)
        _npivm_py = utils._EngineIntScalarType.to_py(_npivm)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nnzc', 'npivm'],
        )(
            _nnzc_py,
            _npivm_py,
        )
    _err_class.manage()
    return return_data

def complex_herm_precon_ilu_solve(n, a, la, irow, icol, ipiv, istr, check, y, x):
    u"""
    Solution of complex linear system involving incomplete Cholesky
    preconditioning matrix generated by ``complex_herm_precon_ichol``.

    ``complex_herm_precon_ilu_solve`` solves a system of complex linear
    equations involving the incomplete Cholesky preconditioning matrix
    generated by ``complex_herm_precon_ichol``.

    For full information please refer to the NAG Library document for
    f11jp

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11jpf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix M. This **must** be the same value as
        was supplied in the preceding call to
        ``complex_herm_precon_ichol``.

    a : complex, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``complex_herm_precon_ichol``.

    la : int
        This **must** be the same value supplied in the preceding call
        to ``complex_herm_precon_ichol``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    ipiv : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    check : str, length 1
        Specifies whether or not the input data should be checked.

        `check` = 'C'

            Checks are carried out on the values of `n`, `irow`, `icol`,
            `ipiv` and `istr`.

        `check` = 'N'

            None of these checks are carried out.

    y : complex, shape (`n`)
        The right-hand side vector y.

    x : complex, shape (`n`), modified in place
        `On exit`: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `ipiv`[i-1] is a repeated value: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `ipiv`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `ipiv`[i-1] >= 1 and `ipiv`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr`[i-1] is inconsistent with `irow`: i =
            *<value>*.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr` appears to be invalid.

    Notes
    -----
    ``complex_herm_precon_ilu_solve`` solves a system of linear
    equations

        Mx = y

    involving the preconditioning matrix M = PLDL^HP^T, corresponding to
    an incomplete Cholesky decomposition of a complex sparse Hermitian
    matrix stored in symmetric coordinate storage (SCS) format (see [the
    F11 Introduction]), as generated by ``complex_herm_precon_ichol``.

    In the above decomposition L is a complex lower triangular sparse
    matrix with unit diagonal, D is a real diagonal matrix and P is a
    permutation matrix. L and D are supplied to
    ``complex_herm_precon_ilu_solve`` through the matrix

        C = L+D^-1-I

    which is a lower triangular n*n complex sparse matrix, stored in SCS
    format, as returned by ``complex_herm_precon_ichol``.
    The permutation matrix P is returned from
    ``complex_herm_precon_ichol`` via the array `ipiv`.

    ``complex_herm_precon_ilu_solve`` may also be used in combination
    with ``complex_herm_precon_ichol`` to solve a sparse complex
    Hermitian positive definite system of linear equations directly (see
    ``complex_herm_precon_ichol``).
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_precon_ilu_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipiv[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineComplex128ArrayType.c_type, # Input, y[n].
        utils._EngineComplex128ArrayType.c_type, # Output, x[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
    )
    _ipiv = _ipiv_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _la,
        _irow,
        _icol,
        _ipiv,
        _istr,
        _check,
        _y,
        _x,
        _errbuf,
        _ierr,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ipiv[i-1] is a repeated value:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ipiv[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ipiv[i-1] >= 1 and ipiv[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr[i-1] is inconsistent with irow:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr appears to be invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ipiv_cls.refresh_py(_ipiv)
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def complex_herm_solve_ilu(method, n, nnz, a, la, irow, icol, ipiv, istr, b, tol, maxitn, x):
    u"""
    Solution of complex sparse Hermitian linear system, conjugate
    gradient/Lanczos method, preconditioner computed by
    ``complex_herm_precon_ichol`` (Black Box).

    ``complex_herm_solve_ilu`` solves a complex sparse Hermitian system
    of linear equations, represented in symmetric coordinate storage
    format, using a conjugate gradient or Lanczos method, with
    incomplete Cholesky preconditioning.

    For full information please refer to the NAG Library document for
    f11jq

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11jqf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'CG'

            Conjugate gradient method.

        `method` = 'SYMMLQ'

            Lanczos method (SYMMLQ).

    n : int
        n, the order of the matrix A. This **must** be the same value as
        was supplied in the preceding call to
        ``complex_herm_precon_ichol``.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A. This **must** be the same value as was supplied in
        the preceding call to ``complex_herm_precon_ichol``.

    a : complex, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``complex_herm_precon_ichol``.

    la : int
        This **must** be the same value as was supplied in the preceding
        call to ``complex_herm_precon_ichol``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    ipiv : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    b : complex, shape (`n`)
        The right-hand side vector b.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : complex, shape (`n`), modified in place
        `On entry`: an initial approximation to the solution vector x.

        `On exit`: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'CG' or 'SYMMLQ'.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[i-1],`icol`[i-1]) is a
            duplicate: i = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>*,
            `irow`[i-1] = *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `irow`[i-1].

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr` appears to be invalid.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            The preconditioner appears not to be positive definite. The
            computation cannot continue.

        (`errno` 7)
            The matrix of the coefficients `a` appears not to be
            positive definite. The computation cannot continue.

        (`errno` 8)
            A serious error, code *<value>*, has occurred in an internal
            call. Check all function calls and array sizes. Seek expert
            help.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The required accuracy could not be obtained. However a
            reasonable accuracy has been achieved.

    Notes
    -----
    ``complex_herm_solve_ilu`` solves a complex sparse Hermitian linear
    system of equations

        Ax = b,

    using a preconditioned conjugate gradient method (see Meijerink and
    Van der Vorst (1977)), or a preconditioned Lanczos method based on
    the algorithm SYMMLQ (see Paige and Saunders (1975)).
    The conjugate gradient method is more efficient if A is positive
    definite, but may fail to converge for indefinite matrices.
    In this case the Lanczos method should be used instead.
    For further details see Barrett `et al.` (1994).

    ``complex_herm_solve_ilu`` uses the incomplete Cholesky
    factorization determined by ``complex_herm_precon_ichol`` as the
    preconditioning matrix.
    A call to ``complex_herm_solve_ilu`` must always be preceded by a
    call to ``complex_herm_precon_ichol``.
    Alternative preconditioners for the same storage scheme are
    available by calling ``complex_herm_solve_jacssor``.

    The matrix A and the preconditioning matrix M are represented in
    symmetric coordinate storage (SCS) format (see [the F11
    Introduction]) in the arrays `a`, `irow` and `icol`, as returned
    from ``complex_herm_precon_ichol``.
    The array `a` holds the nonzero entries in the lower triangular
    parts of these matrices, while `irow` and `icol` hold the
    corresponding row and column indices.

    References
    ----------
    Barrett, R, Berry, M, Chan, T F, Demmel, J, Donato, J, Dongarra, J,
    Eijkhout, V, Pozo, R, Romine, C and Van der Vorst, H, 1994,
    `Templates for the Solution of Linear Systems: Building Blocks for
    Iterative Methods`, SIAM, Philadelphia

    Meijerink, J and Van der Vorst, H, 1977, `An iterative solution
    method for linear systems of which the coefficient matrix is a
    symmetric M-matrix`, Math. Comput. (31), 148--162

    Paige, C C and Saunders, M A, 1975, `Solution of sparse indefinite
    systems of linear equations`, SIAM J. Numer. Anal. (12), 617--629
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_solve_ilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipiv[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._EngineComplex128ArrayType.c_type, # Input, b[n].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineComplex128ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
    )
    _ipiv = _ipiv_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _ipiv,
        _istr,
        _b,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'CG\' or \'SYMMLQ\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[i-1],icol[i-1]) is a duplicate:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[i-1] = {:s}, irow[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= irow[i-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'irow[i-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr appears to be invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99986:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However a reasonable accuracy has been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The preconditioner appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients a appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error, code {:s}, has occurred in an internal'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'call. Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error, code {:s}, has occurred in an internal'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'call. Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ipiv_cls.refresh_py(_ipiv)
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_herm_precon_ssor_solve(n, nnz, a, irow, icol, rdiag, omega, check, y, x):
    u"""
    Solution of linear system involving preconditioning matrix generated
    by applying SSOR to complex sparse Hermitian matrix.

    ``complex_herm_precon_ssor_solve`` solves a system of linear
    equations involving the preconditioning matrix corresponding to SSOR
    applied to a complex sparse Hermitian matrix, represented in
    symmetric coordinate storage format.

    For full information please refer to the NAG Library document for
    f11jr

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11jrf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A.

    a : complex, shape (`nnz`)
        The nonzero elements in the lower triangular part of the matrix
        A, ordered by increasing row index, and by increasing column
        index within each row. Multiple entries for the same row and
        column indices are not permitted. The function
        ``complex_herm_sort`` may be used to order the elements in this
        way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    rdiag : float, shape (`n`)
        The elements of the diagonal matrix D^-1, where D is the
        diagonal part of A. Note that since A is Hermitian the elements
        of D^-1 are necessarily real.

    omega : float
        The relaxation parameter omega.

    check : str, length 1
        Specifies whether or not the input data should be checked.

        `check` = 'C'

            Checks are carried out on the values of `n`, `nnz`, `irow`,
            `icol` and `omega`.

        `check` = 'N'

            None of these checks are carried out.

    y : complex, shape (`n`)
        The right-hand side vector y.

    x : complex, shape (`n`), modified in place
        `On exit`: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `omega` = *<value>*.

            Constraint: 0.0 < `omega` < 2.0.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>*,
            `irow`[I-1] = *<value>*.

            Constraint: 1 <= `icol`[i-1] <= `irow`[i-1].

        (`errno` 3)
            On entry, I = *<value>*, `irow`[I-1] = *<value>* and `n` =
            *<value>*.

            Constraint: 1 <= `irow`[i-1] <= `n`.

        (`errno` 4)
            The matrix A has no diagonal entry in row *<value>*.

    Notes
    -----
    ``complex_herm_precon_ssor_solve`` solves a system of equations

        Mx = y

    involving the preconditioning matrix

        M = 1/omega(2-omega)(D+omega L)D^-1(D+omega L)^H

    corresponding to symmetric successive-over-relaxation (SSOR) (see
    Young (1971)) on a linear system Ax = b, where A is a sparse complex
    Hermitian matrix stored in symmetric coordinate storage (SCS) format
    (see [the F11 Introduction]).

    In the definition of M given above D is the diagonal part of A, L is
    the strictly lower triangular part of A and omega is a user-defined
    relaxation parameter.
    Note that since A is Hermitian the matrix D is necessarily real.

    References
    ----------
    Young, D, 1971, `Iterative Solution of Large Linear Systems`,
    Academic Press, New York
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_precon_ssor_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jrft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._EngineFloat64ArrayType.c_type, # Input, rdiag[n].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineComplex128ArrayType.c_type, # Input, y[n].
        utils._EngineComplex128ArrayType.c_type, # Output, x[n].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _rdiag = utils._EngineFloat64ArrayType(
        rdiag, {'fun_name': fun_name, 'entity': 'rdiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _rdiag,
        _omega,
        _check,
        _y,
        _x,
        _iwork,
        _errbuf,
        _ierr,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < omega < 2.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s}, irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= icol[i-1] <= irow[i-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'irow[I-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def complex_herm_solve_jacssor(method, precon, n, nnz, a, irow, icol, omega, b, tol, maxitn, x, rdiag):
    u"""
    Solution of complex sparse Hermitian linear system, conjugate
    gradient/Lanczos method, Jacobi or SSOR preconditioner (Black Box).

    ``complex_herm_solve_jacssor`` solves a complex sparse Hermitian
    system of linear equations, represented in symmetric coordinate
    storage format, using a conjugate gradient or Lanczos method,
    without preconditioning, with Jacobi or with SSOR preconditioning.

    For full information please refer to the NAG Library document for
    f11js

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11jsf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'CG'

            Conjugate gradient method.

        `method` = 'SYMMLQ'

            Lanczos method (SYMMLQ).

    precon : str, length 1
        Specifies the type of preconditioning to be used.

        `precon` = 'N'

            No preconditioning.

        `precon` = 'J'

            Jacobi.

        `precon` = 'S'

            Symmetric successive-over-relaxation (SSOR).

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A.

    a : complex, shape (`nnz`)
        The nonzero elements of the lower triangular part of the matrix
        A, ordered by increasing row index, and by increasing column
        index within each row. Multiple entries for the same row and
        column indices are not permitted. The function
        ``complex_herm_sort`` may be used to order the elements in this
        way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    omega : float
        If `precon` = 'S', `omega` is the relaxation parameter omega to
        be used in the SSOR method. Otherwise `omega` need not be
        initialized.

    b : complex, shape (`n`)
        The right-hand side vector b.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : complex, shape (`n`), modified in place
        `On entry`: an initial approximation to the solution vector x.

        `On exit`: an improved approximation to the solution vector x.

    rdiag : float, shape (`n`), modified in place
        `On exit`: the elements of the diagonal matrix D^-1, where D is
        the diagonal part of A. Note that since A is Hermitian the
        elements of D^-1 are necessarily real.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||, where k is the
        output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `method` != 'CG' or 'SYMMLQ': `method` =
            *<value>*.

        (`errno` 1)
            On entry, `omega` = *<value>*.

            Constraint: 0.0 < `omega` < 2.0.

        (`errno` 1)
            On entry, `precon` != 'N', 'J' or 'S': `precon` = *<value>*.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            The matrix A has a zero diagonal entry in row *<value>*.

        (`errno` 3)
            The matrix A has no diagonal entry in row *<value>*.

        (`errno` 4)
            The required accuracy could not be obtained. However, a
            reasonable accuracy has been achieved and further iterations
            could not improve the result.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            The preconditioner appears not to be positive definite. The
            computation cannot continue.

        (`errno` 7)
            The matrix of the coefficients `a` appears not to be
            positive definite. The computation cannot continue.

        (`errno` 8)
            A serious error, code *<value>*, has occurred in an internal
            call. Check all function calls and array sizes. Seek expert
            help.

        (`errno` 9)
            The matrix A has a non-real diagonal entry in row *<value>*.

    Notes
    -----
    ``complex_herm_solve_jacssor`` solves a complex sparse Hermitian
    linear system of equations

        Ax = b,

    using a preconditioned conjugate gradient method (see Barrett `et
    al.` (1994)), or a preconditioned Lanczos method based on the
    algorithm SYMMLQ (see Paige and Saunders (1975)).
    The conjugate gradient method is more efficient if A is positive
    definite, but may fail to converge for indefinite matrices.
    In this case the Lanczos method should be used instead.
    For further details see Barrett `et al.` (1994).

    ``complex_herm_solve_jacssor`` allows the following choices for the
    preconditioner:

    -   no preconditioning;

    -   Jacobi preconditioning (see Young (1971));

    -   symmetric successive-over-relaxation (SSOR) preconditioning (see
        Young (1971)).

    For incomplete Cholesky (IC) preconditioning see
    ``complex_herm_solve_ilu``.

    The matrix A is represented in symmetric coordinate storage (SCS)
    format (see [the F11 Introduction]) in the arrays `a`, `irow` and
    `icol`.
    The array `a` holds the nonzero entries in the lower triangular part
    of the matrix, while `irow` and `icol` hold the corresponding row
    and column indices.

    References
    ----------
    Barrett, R, Berry, M, Chan, T F, Demmel, J, Donato, J, Dongarra, J,
    Eijkhout, V, Pozo, R, Romine, C and Van der Vorst, H, 1994,
    `Templates for the Solution of Linear Systems: Building Blocks for
    Iterative Methods`, SIAM, Philadelphia

    Paige, C C and Saunders, M A, 1975, `Solution of sparse indefinite
    systems of linear equations`, SIAM J. Numer. Anal. (12), 617--629

    Young, D, 1971, `Iterative Solution of Large Linear Systems`,
    Academic Press, New York
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_solve_jacssor'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._EngineComplex128ArrayType.c_type, # Input, b[n].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineFloat64ArrayType.c_type, # Output, rdiag[n].
        utils._EngineComplex128ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _rdiag_cls = utils._EngineFloat64ArrayType(
        rdiag, {'fun_name': fun_name, 'entity': 'rdiag'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _rdiag = _rdiag_cls.to_ctypes(can_reference_contents=False)
    _work = utils._EngineComplex128ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _omega,
        _b,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _rdiag,
        _work,
        _lwork,
        _iwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method != \'CG\' or \'SYMMLQ\': method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < omega < 2.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon != \'N\', \'J\' or \'S\': precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, icol[I-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has a zero diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy has been achieved and further iterations',
                    'could not improve the result.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The preconditioner appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients a appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99979:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error, code {:s}, has occurred in an internal'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'call. Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99980:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error, code {:s}, has occurred in an internal'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'call. Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99978:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has a non-real diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _rdiag_cls.refresh_py(_rdiag)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
    _err_class.manage()
    return return_data

def direct_real_gen_setup(spec, n, icolzp, irowix, iprm):
    u"""
    Real sparse nonsymmetric linear systems, setup for
    ``direct_real_gen_lu``.

    ``direct_real_gen_setup`` computes a column permutation suitable for
    LU factorization (by ``direct_real_gen_lu``) of a real sparse matrix
    in compressed column (Harwell--Boeing) format and applies it to the
    matrix.
    This function must be called prior to ``direct_real_gen_lu``.

    For full information please refer to the NAG Library document for
    f11md

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11mdf.html

    Parameters
    ----------
    spec : str, length 1
        Indicates the permutation to be applied.

        `spec` = 'N'

            The identity permutation is used (i.e., the columns are not
            permuted).

        `spec` = 'U'

            The permutation in the `iprm` array is used, as supplied by
            you.

        `spec` = 'M'

            The permutation computed by the COLAMD algorithm is used

    n : int
        n, the order of the matrix A.

    icolzp : int, shape (max(1, (`n`+1)))
        The new column index array of sparse matrix A. See [the F11
        Introduction].

    irowix : int, shape (max(1, (`icolzp`[max(1, (`n`+1))-1]-1)))
        `irowix`[i-1] contains the row index in A for element A(i). See
        [the F11 Introduction].

    iprm : int, shape (max(1, (7*`n`))), modified in place
        `On entry`: the first `n` entries contain the column permutation
        if supplied by the user. This will be used if `spec` = 'U', and
        ignored otherwise. If used, it must consist of a permutation of
        all the integers in the range [0,(`n`-1)], the leftmost column
        of the matrix A denoted by 0 and the rightmost by `n`-1.
        Labelling columns in this way, `iprm`[i] = j means that column
        i-1 of A is in position j in A*P_c, where P_r AP_c = LU
        expresses the factorization to be performed.

        `On exit`: The column permutation given or computed is returned
        in the second `n` entries. The rest of the array contains data
        structures that will be used by other functions in the suite.
        The function computes the column elimination tree for A and a
        post-order permutation on the tree. It then compounds the `iprm`
        permutation given or computed by the COLAMD algorthm with the
        post-order permutation and this permutation is returned in the
        first `n` entries. This whole array is needed by the LU
        factorization function ``direct_real_gen_lu`` and associated
        functions ``direct_real_gen_solve``, ``direct_real_gen_refine``
        and ``direct_real_gen_diag`` and should be passed to them
        unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 1)
            On entry, `spec` = *<value>*.

            Constraint: `spec` = 'N', 'U' or 'M'.

        (`errno` 2)
            Incorrect column permutations in array `iprm`.

        (`errno` 3)
            COLAMD algorithm failed.

        (`errno` 4)
            Incorrect specification of argument `icolzp`.

        (`errno` 5)
            Incorrect specification of argument `irowix`.

    Notes
    -----
    Given a sparse matrix in compressed column (Harwell--Boeing) format
    A and a choice of column permutation schemes, the function computes
    those data structures that will be needed by the LU factorization
    function ``direct_real_gen_lu`` and associated functions
    ``direct_real_gen_diag``, ``direct_real_gen_solve`` and
    ``direct_real_gen_refine``.
    The column permutation choices are:

        original order (that is, no permutation);

        user-supplied permutation;

        a permutation, computed by the function, designed to minimize
        fill-in during the LU factorization.

    The algorithm for this computed permutation is based on the
    approximate minimum degree column ordering algorithm COLAMD.
    The computed permutation is not sensitive to the magnitude of the
    nonzero values of A.

    References
    ----------
    Amestoy, P R, Davis, T A and Duff, I S, 1996, `An approximate
    minimum degree ordering algorithm`, SIAM J. Matrix Anal. Appl. (17),
    886--905

    Gilbert, J R and Larimore, S I, 2004, `A column approximate minimum
    degree ordering algorithm`, ACM Trans. Math. Software (30,3),
    353--376

    Gilbert, J R, Larimore, S I and Ng, E G, 2004, `Algorithm 836:
    COLAMD, an approximate minimum degree ordering algorithm`, ACM
    Trans. Math. Software (30, 3), 377--380
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_setup'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11mdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, spec(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, icolzp[max(1, (n+1))].
        utils._EngineIntArrayType.c_type, # Input, irowix[max(1, (icolzp[max(1, (n+1))-1]-1))].
        utils._EngineIntArrayType.c_type, # Input/Output, iprm[max(1, (7*n))].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of spec.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _spec = utils._EngineCharScalarType(
        spec, {'fun_name': fun_name, 'entity': 'spec'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _icolzp = utils._EngineIntArrayType(
        icolzp, {'fun_name': fun_name, 'entity': 'icolzp'},
        exp_shape=(max(1, (_n_py+1)),),
    ).to_ctypes()
    _irowix = utils._EngineIntArrayType(
        irowix, {'fun_name': fun_name, 'entity': 'irowix'},
        exp_shape=(max(1, (utils._EngineIntArrayType.get_element(icolzp, {'fun_name': fun_name, 'entity': 'icolzp'}, (max(1, (_n_py+1))-1,))-1)),),
    ).to_ctypes()
    _iprm_cls = utils._EngineIntArrayType(
        iprm, {'fun_name': fun_name, 'entity': 'iprm'},
        exp_shape=(max(1, (7*_n_py)),),
    )
    _iprm = _iprm_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _spec,
        _n,
        _icolzp,
        _irowix,
        _iprm,
        _errbuf,
        _ierr,
        _spec._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _spec_py = utils._EngineCharScalarType.to_py(_spec)
        if _ierr_py == 1 and _ifmt == -2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, spec = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_spec_py),
                    ),
                    'Constraint: spec = \'N\', \'U\' or \'M\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect column permutations in array iprm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'COLAMD algorithm failed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect specification of argument icolzp.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect specification of argument irowix.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iprm_cls.refresh_py(_iprm)
        return_data = None
    _err_class.manage()
    return return_data

def direct_real_gen_lu(n, irowix, a, iprm, thresh, nzlmx, nzlumx, nzumx, il, lval, iu, uval):
    u"""
    LU factorization of real sparse matrix.

    ``direct_real_gen_lu`` computes the LU factorization of a real
    sparse matrix in compressed column (Harwell--Boeing),
    column-permuted format.

    For full information please refer to the NAG Library document for
    f11me

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11mef.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    irowix : int, shape (:)
        The row index array of sparse matrix A. See [the F11
        Introduction].

    a : float, shape (:)
        The array of nonzero values in the sparse matrix A.

    iprm : int, shape (7*`n`), modified in place
        `On entry`: contains the column permutation which defines the
        permutation P_c and associated data structures as computed by
        function ``direct_real_gen_setup``.

        `On exit`: part of the array is modified to record the row
        permutation P_r determined by pivoting.

    thresh : float
        The diagonal pivoting threshold, t. At step j of the Gaussian
        elimination, if |A_{jj}| >= t*(max_{i >= j}|A_{ij}|), use A_{jj}
        as a pivot, otherwise use max_{i >= j}|A_{ij}|. A value of t = 1
        corresponds to partial pivoting, a value of t = 0 corresponds to
        always choosing the pivot on the diagonal (unless it is zero).

        Suggested value: `thresh` = 1.0. Smaller values may result in a
        faster factorization, but the benefits are likely to be small in
        most cases. It might be possible to use `thresh` = 0.0 if you
        are confident about the stability of the factorization, for
        example, if A is diagonally dominant.

    nzlmx : int
        Indicates the available size of array `il`. The dimension of
        `il` should be at least 7*`n`+`nzlmx`+4. A good range for
        `nzlmx` that works for many problems is nnz to 8*nnz, where nnz
        is the number of nonzeros in the sparse matrix A. If, on exit,
        `errno` = 2, the given `nzlmx` was too small and you should
        attempt to provide more storage and call the function again.

    nzlumx : int
        Indicates the available size of array `lval`. The dimension of
        `lval` should be at least `nzlumx`.

    nzumx : int
        Indicates the available sizes of arrays `iu` and `uval`. The
        dimension of `iu` should be at least 2*`n`+`nzumx`+1 and the
        dimension of `uval` should be at least `nzumx`. A good range for
        `nzumx` that works for many problems is nnz to 8*nnz, where nnz
        is the number of nonzeros in the sparse matrix A. If, on exit,
        `errno` = 3, the given `nzumx` was too small and you should
        attempt to provide more storage and call the function again.

    il : int, shape (7*`n`+`nzlmx`+4), modified in place
        `On exit`: encapsulates the sparsity pattern of matrix L.

    lval : float, shape (`nzlumx`), modified in place
        `On exit`: records the nonzero values of matrix L and some of
        the nonzero values of matrix U.

    iu : int, shape (2*`n`+`nzumx`+1), modified in place
        `On exit`: encapsulates the sparsity pattern of matrix U.

    uval : float, shape (`nzumx`), modified in place
        `On exit`: records some of the nonzero values of matrix U.

    Returns
    -------
    nzlumx : int
        If `errno` = 4, the given `nzlumx` was too small and is reset to
        a value that will be sufficient. You should then provide the
        indicated storage and call the function again.

    nnzl : int
        The number of nonzero values in the matrix L.

    nnzu : int
        The number of nonzero values in the matrix U.

    flop : float
        The number of floating-point operations performed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nzumx` = *<value>*.

            Constraint: `nzumx` >= 1.

        (`errno` 1)
            On entry, `nzlumx` = *<value>*.

            Constraint: `nzlumx` >= 1.

        (`errno` 1)
            On entry, `nzlmx` = *<value>*.

            Constraint: `nzlmx` >= 1.

        (`errno` 1)
            On entry, `thresh` = *<value>*.

            Constraint: 0.0 <= `thresh` <= 1.0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            Insufficient `nzlmx`.

        (`errno` 3)
            Insufficient `nzumx`.

        (`errno` 4)
            Insufficient `nzlumx`.

        (`errno` 5)
            The matrix is singular -- no factorization possible.

    Notes
    -----
    Given a real sparse matrix A, ``direct_real_gen_lu`` computes an LU
    factorization of A with partial pivoting, P_r AP_c = LU, where P_r
    is a row permutation matrix (computed by ``direct_real_gen_lu``),
    P_c is a (supplied) column permutation matrix, L is unit lower
    triangular and U is upper triangular.
    The column permutation matrix, P_c, must be computed by a prior call
    to ``direct_real_gen_setup``.
    The matrix A must be presented in the column permuted, compressed
    column (Harwell--Boeing) format.

    The LU factorization is output in the form of four one-dimensional
    arrays: integer arrays `il` and `iu` and real-valued arrays `lval`
    and `uval`.
    These describe the sparsity pattern and numerical values in the L
    and U matrices.
    The minimum required dimensions of these arrays cannot be given as a
    simple function of the size arguments (order and number of nonzero
    values) of the matrix A.
    This is due to unpredictable fill-in created by partial pivoting.
    ``direct_real_gen_lu`` will, on return, indicate which dimensions of
    these arrays were not adequate for the computation or (in the case
    of one of them) give a firm bound.
    You should then allocate more storage and try again.

    References
    ----------
    Demmel, J W, Eisenstat, S C, Gilbert, J R, Li, X S and Li, J W H,
    1999, `A supernodal approach to sparse partial pivoting`, SIAM J.
    Matrix Anal. Appl. (20), 720--755

    Demmel, J W, Gilbert, J R and Li, X S, 1999, `An asynchronous
    parallel supernodal algorithm for sparse gaussian elimination`, SIAM
    J. Matrix Anal. Appl. (20), 915--952
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_lu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11meft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, irowix[*].
        utils._EngineFloat64ArrayType.c_type, # Input, a[*].
        utils._EngineIntArrayType.c_type, # Input/Output, iprm[7*n].
        utils._BYREF_FLOAT64, # Input, thresh.
        utils._BYREF_INT, # Input, nzlmx.
        utils._BYREF_INT, # Input/Output, nzlumx.
        utils._BYREF_INT, # Input, nzumx.
        utils._EngineIntArrayType.c_type, # Output, il[7*n+nzlmx+4].
        utils._EngineFloat64ArrayType.c_type, # Output, lval[nzlumx].
        utils._EngineIntArrayType.c_type, # Output, iu[2*n+nzumx+1].
        utils._EngineFloat64ArrayType.c_type, # Output, uval[nzumx].
        utils._BYREF_INT, # Output, nnzl.
        utils._BYREF_INT, # Output, nnzu.
        utils._BYREF_FLOAT64, # Output, flop.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _irowix = utils._EngineIntArrayType(
        irowix, {'fun_name': fun_name, 'entity': 'irowix'},
        exp_shape=(None,),
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(None,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _iprm_cls = utils._EngineIntArrayType(
        iprm, {'fun_name': fun_name, 'entity': 'iprm'},
        exp_shape=(7*_n_py,),
    )
    _iprm = _iprm_cls.to_ctypes()
    _thresh = utils._EngineFloat64ScalarType(
        thresh, {'fun_name': fun_name, 'entity': 'thresh'},
    ).to_ctypes()
    _nzlmx = utils._EngineIntScalarType(
        nzlmx, {'fun_name': fun_name, 'entity': 'nzlmx'},
    ).to_ctypes()
    _nzlumx = utils._EngineIntScalarType(
        nzlumx, {'fun_name': fun_name, 'entity': 'nzlumx'},
    ).to_ctypes()
    _nzumx = utils._EngineIntScalarType(
        nzumx, {'fun_name': fun_name, 'entity': 'nzumx'},
    ).to_ctypes()
    _nzlmx_py = utils._EngineIntScalarType.to_py(_nzlmx)
    _il_cls = utils._EngineIntArrayType(
        il, {'fun_name': fun_name, 'entity': 'il'},
        exp_shape=(7*_n_py+_nzlmx_py+4,),
        check_contents=False,
    )
    _il = _il_cls.to_ctypes(can_reference_contents=False)
    _nzlumx_py = utils._EngineIntScalarType.to_py(_nzlumx)
    _lval_cls = utils._EngineFloat64ArrayType(
        lval, {'fun_name': fun_name, 'entity': 'lval'},
        exp_shape=(_nzlumx_py,),
        check_contents=False,
    )
    _lval = _lval_cls.to_ctypes(can_reference_contents=False)
    _nzumx_py = utils._EngineIntScalarType.to_py(_nzumx)
    _iu_cls = utils._EngineIntArrayType(
        iu, {'fun_name': fun_name, 'entity': 'iu'},
        exp_shape=(2*_n_py+_nzumx_py+1,),
        check_contents=False,
    )
    _iu = _iu_cls.to_ctypes(can_reference_contents=False)
    _uval_cls = utils._EngineFloat64ArrayType(
        uval, {'fun_name': fun_name, 'entity': 'uval'},
        exp_shape=(_nzumx_py,),
        check_contents=False,
    )
    _uval = _uval_cls.to_ctypes(can_reference_contents=False)
    _nnzl = utils._EngineIntScalarType.empty_ctypes()
    _nnzu = utils._EngineIntScalarType.empty_ctypes()
    _flop = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _irowix,
        _a,
        _iprm,
        _thresh,
        _nzlmx,
        _nzlumx,
        _nzumx,
        _il,
        _lval,
        _iu,
        _uval,
        _nnzl,
        _nnzu,
        _flop,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _nzlumx_py = utils._EngineIntScalarType.to_py(_nzlumx)
        _thresh_py = utils._EngineFloat64ScalarType.to_py(_thresh)
        if _ierr_py == 1 and _ifmt == -8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nzumx = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nzumx_py),
                    ),
                    'Constraint: nzumx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nzlumx = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nzlumx_py),
                    ),
                    'Constraint: nzlumx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nzlmx = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nzlmx_py),
                    ),
                    'Constraint: nzlmx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, thresh = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_thresh_py),
                    ),
                    'Constraint: 0.0 <= thresh <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Insufficient nzlmx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Insufficient nzumx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Insufficient nzlumx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix is singular -- no factorization possible.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iprm_cls.refresh_py(_iprm)
        _nzlumx_py = utils._EngineIntScalarType.to_py(_nzlumx)
        _il_cls.refresh_py(_il)
        _lval_cls.refresh_py(_lval)
        _iu_cls.refresh_py(_iu)
        _uval_cls.refresh_py(_uval)
        _nnzl_py = utils._EngineIntScalarType.to_py(_nnzl)
        _nnzu_py = utils._EngineIntScalarType.to_py(_nnzu)
        _flop_py = utils._EngineFloat64ScalarType.to_py(_flop)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nzlumx', 'nnzl', 'nnzu', 'flop'],
        )(
            _nzlumx_py,
            _nnzl_py,
            _nnzu_py,
            _flop_py,
        )
    _err_class.manage()
    return return_data

def direct_real_gen_solve(trans, n, iprm, il, lval, iu, uval, nrhs, b):
    u"""
    Solution of real sparse simultaneous linear equations (coefficient
    matrix already factorized).

    ``direct_real_gen_solve`` solves a real sparse system of linear
    equations with multiple right-hand sides given an LU factorization
    of the sparse matrix computed by ``direct_real_gen_lu``.

    For full information please refer to the NAG Library document for
    f11mf

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11mff.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether AX = B or A^TX = B is solved.

        `trans` = 'N'

            AX = B is solved.

        `trans` = 'T'

            A^TX = B is solved.

    n : int
        n, the order of the matrix A.

    iprm : int, shape (7*`n`)
        The column permutation which defines P_c, the row permutation
        which defines P_r, plus associated data structures as computed
        by ``direct_real_gen_lu``.

    il : int, shape (:)
        Records the sparsity pattern of matrix L as computed by
        ``direct_real_gen_lu``.

    lval : float, shape (:)
        Records the nonzero values of matrix L and some nonzero values
        of matrix U as computed by ``direct_real_gen_lu``.

    iu : int, shape (:)
        Records the sparsity pattern of matrix U as computed by
        ``direct_real_gen_lu``.

    uval : float, shape (:)
        Records some nonzero values of matrix U as computed by
        ``direct_real_gen_lu``.

    nrhs : int
        nrhs, the number of right-hand sides in B.

    b : float, shape (`n`, `nrhs`), modified in place
        `On entry`: the `n`*`nrhs` right-hand side matrix B.

        `On exit`: the `n`*`nrhs` solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 2)
            Incorrect row permutations in array `iprm`.

        (`errno` 3)
            Incorrect column permutations in array `iprm`.

    Notes
    -----
    ``direct_real_gen_solve`` solves a real system of linear equations
    with multiple right-hand sides AX = B or A^TX = B, according to the
    value of the argument `trans`, where the matrix factorization P_r
    AP_c = LU corresponds to an LU decomposition of a sparse matrix
    stored in compressed column (Harwell--Boeing) format, as computed by
    ``direct_real_gen_lu``.

    In the above decomposition L is a lower triangular sparse matrix
    with unit diagonal elements and U is an upper triangular sparse
    matrix; P_r and P_c are permutation matrices.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11mfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, iprm[7*n].
        utils._EngineIntArrayType.c_type, # Input, il[*].
        utils._EngineFloat64ArrayType.c_type, # Input, lval[*].
        utils._EngineIntArrayType.c_type, # Input, iu[*].
        utils._EngineFloat64ArrayType.c_type, # Input, uval[*].
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _iprm = utils._EngineIntArrayType(
        iprm, {'fun_name': fun_name, 'entity': 'iprm'},
        exp_shape=(7*_n_py,),
    ).to_ctypes()
    _il = utils._EngineIntArrayType(
        il, {'fun_name': fun_name, 'entity': 'il'},
        exp_shape=(None,),
    ).to_ctypes()
    _lval = utils._EngineFloat64ArrayType(
        lval, {'fun_name': fun_name, 'entity': 'lval'},
        exp_shape=(None,),
    ).to_ctypes()
    _iu = utils._EngineIntArrayType(
        iu, {'fun_name': fun_name, 'entity': 'iu'},
        exp_shape=(None,),
    ).to_ctypes()
    _uval = utils._EngineFloat64ArrayType(
        uval, {'fun_name': fun_name, 'entity': 'uval'},
        exp_shape=(None,),
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _iprm,
        _il,
        _lval,
        _iu,
        _uval,
        _nrhs,
        _b,
        _ldb,
        _errbuf,
        _ierr,
        _trans._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _trans_py = utils._EngineCharScalarType.to_py(_trans)
        if _ierr_py == 1 and _ifmt == -8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nrhs_py),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_trans_py),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect row permutations in array iprm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect column permutations in array iprm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _b_cls.refresh_py(_b)
        return_data = None
    _err_class.manage()
    return return_data

def direct_real_gen_cond(norm, n, il, lval, iu, uval, anorm):
    u"""
    Estimate condition number of real matrix, matrix already factorized
    by ``direct_real_gen_lu``.

    ``direct_real_gen_cond`` computes an estimate of the reciprocal of
    the condition number of a sparse matrix given an LU factorization of
    the matrix computed by ``direct_real_gen_lu``.

    For full information please refer to the NAG Library document for
    f11mg

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11mgf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is to be
        estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    n : int
        n, the order of the matrix A.

    il : int, shape (:)
        Records the sparsity pattern of matrix L as computed by
        ``direct_real_gen_lu``.

    lval : float, shape (:)
        Records the nonzero values of matrix L and some nonzero values
        of matrix U as computed by ``direct_real_gen_lu``.

    iu : int, shape (:)
        Records the sparsity pattern of matrix U as computed by
        ``direct_real_gen_lu``.

    uval : float, shape (:)
        Records some nonzero values of matrix U as computed by
        ``direct_real_gen_lu``.

    anorm : float
        If `norm` = '1' or 'O', the 1-norm of the matrix A.

        If `norm` = 'I', the infinity-norm of the matrix A.

        `anorm` may be computed by calling ``direct_real_gen_norm`` with
        the same value for the argument `norm`.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `anorm` = *<value>*.

            Constraint: `anorm` >= 0.0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 1)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = '1', 'O' or 'I'.

    Notes
    -----
    ``direct_real_gen_cond`` estimates the condition number of a real
    sparse matrix A, in either the 1-norm or the infinity-norm:

        kappa_1(A) = ||A||_1||A^-1||_1 or kappa_infinity (A) =
        ||A||_infinity||A^-1||_infinity.

    Note that kappa_infinity (A) = kappa_1(A^T).

    Because the condition number is infinite if A is singular, the
    function actually returns an estimate of the **reciprocal** of the
    condition number.

    The function should be preceded by a call to
    ``direct_real_gen_norm`` to compute ||A||_1 or ||A||_infinity, and a
    call to ``direct_real_gen_lu`` to compute the LU factorization of A.
    The function then estimates ||A^-1||_1 or ||A^-1||_infinity and
    computes the reciprocal of the condition number.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_cond'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11mgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, il[*].
        utils._EngineFloat64ArrayType.c_type, # Input, lval[*].
        utils._EngineIntArrayType.c_type, # Input, iu[*].
        utils._EngineFloat64ArrayType.c_type, # Input, uval[*].
        utils._BYREF_FLOAT64, # Input, anorm.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _il = utils._EngineIntArrayType(
        il, {'fun_name': fun_name, 'entity': 'il'},
        exp_shape=(None,),
    ).to_ctypes()
    _lval = utils._EngineFloat64ArrayType(
        lval, {'fun_name': fun_name, 'entity': 'lval'},
        exp_shape=(None,),
    ).to_ctypes()
    _iu = utils._EngineIntArrayType(
        iu, {'fun_name': fun_name, 'entity': 'iu'},
        exp_shape=(None,),
    ).to_ctypes()
    _uval = utils._EngineFloat64ArrayType(
        uval, {'fun_name': fun_name, 'entity': 'uval'},
        exp_shape=(None,),
    ).to_ctypes()
    _anorm = utils._EngineFloat64ScalarType(
        anorm, {'fun_name': fun_name, 'entity': 'anorm'},
    ).to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _norm,
        _n,
        _il,
        _lval,
        _iu,
        _uval,
        _anorm,
        _rcond,
        _errbuf,
        _ierr,
        _norm._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _n_py = utils._EngineIntScalarType.to_py(_n)
        _norm_py = utils._EngineCharScalarType.to_py(_norm)
        if _ierr_py == 1 and _ifmt == -7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, anorm = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_anorm_py),
                    ),
                    'Constraint: anorm >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: norm = \'1\', \'O\' or \'I\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        return_data = _rcond_py
    _err_class.manage()
    return return_data

def direct_real_gen_refine(trans, n, icolzp, irowix, a, iprm, il, lval, iu, uval, nrhs, b, x, ferr, berr):
    u"""
    Refined solution with error bounds of real system of linear
    equations, multiple right-hand sides.

    ``direct_real_gen_refine`` returns error bounds for the solution of
    a real sparse system of linear equations with multiple right-hand
    sides, AX = B or A^TX = B.
    It improves the solution by iterative refinement in standard
    precision, in order to reduce the backward error as much as
    possible.

    For full information please refer to the NAG Library document for
    f11mh

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11mhf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether AX = B or A^TX = B is solved.

        `trans` = 'N'

            AX = B is solved.

        `trans` = 'T'

            A^TX = B is solved.

    n : int
        n, the order of the matrix A.

    icolzp : int, shape (`n`+1)
        The new column index array of sparse matrix A. See [the F11
        Introduction].

    irowix : int, shape (asize)
        The row index array of sparse matrix A. See [the F11
        Introduction].

    a : float, shape (asize)
        The array of nonzero values in the sparse matrix A.

    iprm : int, shape (7*`n`)
        The column permutation which defines P_c, the row permutation
        which defines P_r, plus associated data structures as computed
        by ``direct_real_gen_lu``.

    il : int, shape (:)
        Records the sparsity pattern of matrix L as computed by
        ``direct_real_gen_lu``.

    lval : float, shape (:)
        Records the nonzero values of matrix L and some nonzero values
        of matrix U as computed by ``direct_real_gen_lu``.

    iu : int, shape (:)
        Records the sparsity pattern of matrix U as computed by
        ``direct_real_gen_lu``.

    uval : float, shape (:)
        Records some nonzero values of matrix U as computed by
        ``direct_real_gen_lu``.

    nrhs : int
        nrhs, the number of right-hand sides in B.

    b : float, shape (`n`, `nrhs`)
        The n*nrhs right-hand side matrix B.

    x : float, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*nrhs solution matrix X, as returned by
        ``direct_real_gen_solve``.

        `On exit`: the n*nrhs improved solution matrix X.

    ferr : float, shape (`nrhs`), modified in place
        `On exit`: `ferr`[j-1] contains an estimated error bound for the
        jth solution vector, that is, the jth column of X, for j =
        1,2,...,nrhs.

    berr : float, shape (`nrhs`), modified in place
        `On exit`: `berr`[j-1] contains the component-wise backward
        error bound beta for the jth solution vector, that is, the jth
        column of X, for j = 1,2,...,nrhs.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 2)
            Incorrect row permutations in array `iprm`.

        (`errno` 3)
            Incorrect column permutations in array `iprm`.

    Notes
    -----
    ``direct_real_gen_refine`` returns the backward errors and estimated
    bounds on the forward errors for the solution of a real system of
    linear equations with multiple right-hand sides AX = B or A^TX = B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``direct_real_gen_refine`` in terms of a single right-hand side b
    and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that if x is the exact solution of a
    perturbed system:

        (A+delta A) x = b + delta b; then |delta a_{ij}| <= beta
        |a_{ij}| and |delta b_i| <= beta |b_i| .

    Then the function estimates a bound for the `component-wise`
    `forward error` in the computed solution, defined by:

        max_i^|x_i-x^_i|/max_i^|x_i|

    where x^ is the true solution.

    The function uses the LU factorization P_r AP_c = LU computed by
    ``direct_real_gen_lu`` and the solution computed by
    ``direct_real_gen_solve``.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_refine'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11mhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, icolzp[n+1].
        utils._EngineIntArrayType.c_type, # Input, irowix[*].
        utils._EngineFloat64ArrayType.c_type, # Input, a[*].
        utils._EngineIntArrayType.c_type, # Input, iprm[7*n].
        utils._EngineIntArrayType.c_type, # Input, il[*].
        utils._EngineFloat64ArrayType.c_type, # Input, lval[*].
        utils._EngineIntArrayType.c_type, # Input, iu[*].
        utils._EngineFloat64ArrayType.c_type, # Input, uval[*].
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineFloat64ArrayType.c_type, # Input, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[ldx, nrhs].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineFloat64ArrayType.c_type, # Output, ferr[nrhs].
        utils._EngineFloat64ArrayType.c_type, # Output, berr[nrhs].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _icolzp = utils._EngineIntArrayType(
        icolzp, {'fun_name': fun_name, 'entity': 'icolzp'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    if _n_py >= 0:
        _vl_asize_py = utils._EngineIntArrayType.get_element(icolzp, {'fun_name': fun_name, 'entity': 'icolzp'}, ((_n_py+1)-1,))-1
    else:
        _vl_asize_py = 0
    _irowix = utils._EngineIntArrayType(
        irowix, {'fun_name': fun_name, 'entity': 'irowix'},
        exp_shape=(_vl_asize_py,),
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_vl_asize_py,),
    ).to_ctypes()
    _iprm = utils._EngineIntArrayType(
        iprm, {'fun_name': fun_name, 'entity': 'iprm'},
        exp_shape=(7*_n_py,),
    ).to_ctypes()
    _il = utils._EngineIntArrayType(
        il, {'fun_name': fun_name, 'entity': 'il'},
        exp_shape=(None,),
    ).to_ctypes()
    _lval = utils._EngineFloat64ArrayType(
        lval, {'fun_name': fun_name, 'entity': 'lval'},
        exp_shape=(None,),
    ).to_ctypes()
    _iu = utils._EngineIntArrayType(
        iu, {'fun_name': fun_name, 'entity': 'iu'},
        exp_shape=(None,),
    ).to_ctypes()
    _uval = utils._EngineFloat64ArrayType(
        uval, {'fun_name': fun_name, 'entity': 'uval'},
        exp_shape=(None,),
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldx = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _ferr_cls = utils._EngineFloat64ArrayType(
        ferr, {'fun_name': fun_name, 'entity': 'ferr'},
        exp_shape=(_nrhs_py,),
        check_contents=False,
    )
    _ferr = _ferr_cls.to_ctypes(can_reference_contents=False)
    _berr_cls = utils._EngineFloat64ArrayType(
        berr, {'fun_name': fun_name, 'entity': 'berr'},
        exp_shape=(_nrhs_py,),
        check_contents=False,
    )
    _berr = _berr_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _icolzp,
        _irowix,
        _a,
        _iprm,
        _il,
        _lval,
        _iu,
        _uval,
        _nrhs,
        _b,
        _ldb,
        _x,
        _ldx,
        _ferr,
        _berr,
        _errbuf,
        _ierr,
        _trans._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _trans_py = utils._EngineCharScalarType.to_py(_trans)
        if _ierr_py == 1 and _ifmt == -11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nrhs_py),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_trans_py),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect row permutations in array iprm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect column permutations in array iprm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _ferr_cls.refresh_py(_ferr)
        _berr_cls.refresh_py(_berr)
        return_data = None
    _err_class.manage()
    return return_data

def direct_real_gen_matmul(trans, n, m, alpha, icolzp, irowix, a, b, beta, c):
    u"""
    Real sparse nonsymmetric matrix-matrix multiply, compressed column
    storage.

    ``direct_real_gen_matmul`` computes a matrix-matrix or transposed
    matrix-matrix product involving a real, square, sparse nonsymmetric
    matrix stored in compressed column (Harwell--Boeing) format.

    For full information please refer to the NAG Library document for
    f11mk

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11mkf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether or not the matrix A is transposed.

        `trans` = 'N'

            alpha AB+beta C is computed.

        `trans` = 'T'

            alpha A^TB+beta C is computed.

    n : int
        n, the order of the matrix A.

    m : int
        m, the number of columns of matrices B and C.

    alpha : float
        alpha, the scalar factor in the matrix multiplication.

    icolzp : int, shape (`n`+1)
        The new column index array of sparse matrix A. See [the F11
        Introduction].

    irowix : int, shape (asize)
        The row index array of sparse matrix A. See [the F11
        Introduction].

    a : float, shape (asize)
        The array of nonzero values in the sparse matrix A.

    b : float, shape (`n`, `m`)
        The n*m matrix B.

    beta : float
        The scalar factor beta.

    c : float, shape (`n`, `m`), modified in place
        `On entry`: the n*m matrix C.

        `On exit`: C is overwritten by alpha AB+beta C or alpha
        A^TB+beta C depending on the value of `trans`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

    Notes
    -----
    ``direct_real_gen_matmul`` computes either the matrix-matrix product
    C←alpha AB+beta C, or the transposed matrix-matrix product C←alpha
    A^TB+beta C, according to the value of the argument `trans`, where A
    is a real n*n sparse nonsymmetric matrix, of arbitrary sparsity
    pattern with nnz nonzero elements, B and C are n*m real dense
    matrices.
    The matrix A is stored in compressed column (Harwell--Boeing)
    storage format.
    The array `a` stores all nonzero elements of A, while arrays
    `icolzp` and `irowix` store the compressed column indices and row
    indices of A respectively.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_matmul'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11mkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineIntArrayType.c_type, # Input, icolzp[n+1].
        utils._EngineIntArrayType.c_type, # Input, irowix[*].
        utils._EngineFloat64ArrayType.c_type, # Input, a[*].
        utils._EngineFloat64ArrayType.c_type, # Input, b[ldb, m].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, c[ldc, m].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'), (c, 'c')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _icolzp = utils._EngineIntArrayType(
        icolzp, {'fun_name': fun_name, 'entity': 'icolzp'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    if _n_py >= 0:
        _vl_asize_py = utils._EngineIntArrayType.get_element(icolzp, {'fun_name': fun_name, 'entity': 'icolzp'}, ((_n_py+1)-1,))-1
    else:
        _vl_asize_py = 0
    _irowix = utils._EngineIntArrayType(
        irowix, {'fun_name': fun_name, 'entity': 'irowix'},
        exp_shape=(_vl_asize_py,),
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_vl_asize_py,),
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _ldc = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _beta_py = utils._EngineFloat64ScalarType.to_py(_beta)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=_beta_py != 0.0,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=_beta_py != 0.0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _m,
        _alpha,
        _icolzp,
        _irowix,
        _a,
        _b,
        _ldb,
        _beta,
        _c,
        _ldc,
        _errbuf,
        _ierr,
        _trans._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _trans_py = utils._EngineCharScalarType.to_py(_trans)
        if _ierr_py == 1 and _ifmt == -3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT.format(_m_py),
                    ),
                    'Constraint: m >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_trans_py),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _c_cls.refresh_py(_c)
        return_data = None
    _err_class.manage()
    return return_data

def direct_real_gen_norm(norm, n, icolzp, irowix, a):
    u"""
    1-norm, infinity-norm, largest absolute element, real, square,
    sparse matrix.

    ``direct_real_gen_norm`` computes the 1-norm, the infinity-norm or
    the maximum absolute value of the elements of a real, square, sparse
    matrix which is held in compressed column (Harwell--Boeing) format.

    For full information please refer to the NAG Library document for
    f11ml

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11mlf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned in `anorm`.

        `norm` = '1' or 'O'

            The 1-norm ||A||_1 of the matrix is computed, that is max_{1
            <= j <= n}sum_i = 1^n|A_{ij}|.

        `norm` = 'I'

            The infinity-norm ||A||_infinity of the matrix is computed,
            that is max_{1 <= i <= n}sum_j = 1^n|A_{ij}|.

        `norm` = 'M'

            The value max_{1 <= i,j <= n}(|A_{ij}|) (not a norm).

    n : int
        n, the order of the matrix A.

    icolzp : int, shape (`n`+1)
        The new column index array of sparse matrix A. See [the F11
        Introduction].

    irowix : int, shape (asize)
        The row index array of sparse matrix A. See [the F11
        Introduction].

    a : float, shape (asize)
        The array of nonzero values in the sparse matrix A.

    Returns
    -------
    anorm : float
        The computed quantity relating the matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 1)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = '1', 'O', 'I' or 'M'.

    Notes
    -----
    ``direct_real_gen_norm`` computes various quantities relating to
    norms of a real, sparse n*n matrix A presented in compressed column
    (Harwell--Boeing) format.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_norm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11mlft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_FLOAT64, # Output, anorm.
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, icolzp[n+1].
        utils._EngineIntArrayType.c_type, # Input, irowix[*].
        utils._EngineFloat64ArrayType.c_type, # Input, a[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _anorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _icolzp = utils._EngineIntArrayType(
        icolzp, {'fun_name': fun_name, 'entity': 'icolzp'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    if _n_py >= 0:
        _vl_asize_py = utils._EngineIntArrayType.get_element(icolzp, {'fun_name': fun_name, 'entity': 'icolzp'}, ((_n_py+1)-1,))-1
    else:
        _vl_asize_py = 0
    _irowix = utils._EngineIntArrayType(
        irowix, {'fun_name': fun_name, 'entity': 'irowix'},
        exp_shape=(_vl_asize_py,),
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_vl_asize_py,),
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _norm,
        _anorm,
        _n,
        _icolzp,
        _irowix,
        _a,
        _errbuf,
        _ierr,
        _norm._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _norm_py = utils._EngineCharScalarType.to_py(_norm)
        if _ierr_py == 1 and _ifmt == -3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: norm = \'1\', \'O\', \'I\' or \'M\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        return_data = _anorm_py
    _err_class.manage()
    return return_data

def direct_real_gen_diag(n, icolzp, a, iprm, il, lval, iu, uval):
    u"""
    Real sparse nonsymmetric linear systems, diagnostic for
    ``direct_real_gen_lu``.

    ``direct_real_gen_diag`` computes the reciprocal pivot growth factor
    of an LU factorization of a real sparse matrix in compressed column
    (Harwell--Boeing) format.

    For full information please refer to the NAG Library document for
    f11mm

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11mmf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    icolzp : int, shape (`n`+1)
        The new column index array of sparse matrix A. See [the F11
        Introduction].

    a : float, shape (asize)
        The array of nonzero values in the sparse matrix A.

    iprm : int, shape (7*`n`)
        The column permutation which defines P_c, the row permutation
        which defines P_r, plus associated data structures as computed
        by ``direct_real_gen_lu``.

    il : int, shape (:)
        Records the sparsity pattern of matrix L as computed by
        ``direct_real_gen_lu``.

    lval : float, shape (:)
        Records the nonzero values of matrix L and some nonzero values
        of matrix U as computed by ``direct_real_gen_lu``.

    iu : int, shape (:)
        Records the sparsity pattern of matrix U as computed by
        ``direct_real_gen_lu``.

    uval : float, shape (:)
        Records some nonzero values of matrix U as computed by
        ``direct_real_gen_lu``.

    Returns
    -------
    rpg : float
        The reciprocal pivot growth factor
        max_j(||A_j||_infinity/||U_j||_infinity). If the reciprocal
        pivot growth factor is much less than 1, the stability of the LU
        factorization may be poor.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            Incorrect column permutations in array `iprm`.

    Notes
    -----
    ``direct_real_gen_diag`` computes the reciprocal pivot growth factor
    max_j(||A_j||_infinity/||U_j||_infinity) from the columns A_j and
    U_j of an LU factorization of the matrix A, P_r AP_c = LU where P_r
    is a row permutation matrix, P_c is a column permutation matrix, L
    is unit lower triangular and U is upper triangular as computed by
    ``direct_real_gen_lu``.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_diag'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11mmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, icolzp[n+1].
        utils._EngineFloat64ArrayType.c_type, # Input, a[*].
        utils._EngineIntArrayType.c_type, # Input, iprm[7*n].
        utils._EngineIntArrayType.c_type, # Input, il[*].
        utils._EngineFloat64ArrayType.c_type, # Input, lval[*].
        utils._EngineIntArrayType.c_type, # Input, iu[*].
        utils._EngineFloat64ArrayType.c_type, # Input, uval[*].
        utils._BYREF_FLOAT64, # Output, rpg.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _icolzp = utils._EngineIntArrayType(
        icolzp, {'fun_name': fun_name, 'entity': 'icolzp'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    if _n_py >= 0:
        _vl_asize_py = utils._EngineIntArrayType.get_element(icolzp, {'fun_name': fun_name, 'entity': 'icolzp'}, ((_n_py+1)-1,))-1
    else:
        _vl_asize_py = 0
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_vl_asize_py,),
    ).to_ctypes()
    _iprm = utils._EngineIntArrayType(
        iprm, {'fun_name': fun_name, 'entity': 'iprm'},
        exp_shape=(7*_n_py,),
    ).to_ctypes()
    _il = utils._EngineIntArrayType(
        il, {'fun_name': fun_name, 'entity': 'il'},
        exp_shape=(None,),
    ).to_ctypes()
    _lval = utils._EngineFloat64ArrayType(
        lval, {'fun_name': fun_name, 'entity': 'lval'},
        exp_shape=(None,),
    ).to_ctypes()
    _iu = utils._EngineIntArrayType(
        iu, {'fun_name': fun_name, 'entity': 'iu'},
        exp_shape=(None,),
    ).to_ctypes()
    _uval = utils._EngineFloat64ArrayType(
        uval, {'fun_name': fun_name, 'entity': 'uval'},
        exp_shape=(None,),
    ).to_ctypes()
    _rpg = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _icolzp,
        _a,
        _iprm,
        _il,
        _lval,
        _iu,
        _uval,
        _rpg,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect column permutations in array iprm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rpg_py = utils._EngineFloat64ScalarType.to_py(_rpg)
        return_data = _rpg_py
    _err_class.manage()
    return return_data

def real_gen_matvec(trans, n, nnz, a, irow, icol, check, x, y):
    u"""
    Real, sparse, nonsymmetric matrix-vector multiply.

    ``real_gen_matvec`` computes a matrix-vector or transposed
    matrix-vector product involving a real sparse nonsymmetric matrix
    stored in coordinate storage format.

    For full information please refer to the NAG Library document for
    f11xa

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11xaf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether or not the matrix A is transposed.

        `trans` = 'N'

            y = Ax is computed.

        `trans` = 'T'

            y = A^Tx is computed.

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : float, shape (`nnz`)
        The nonzero elements in the matrix A, ordered by increasing row
        index, and by increasing column index within each row. Multiple
        entries for the same row and column indices are not permitted.
        The function ``real_gen_sort`` may be used to order the elements
        in this way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    check : str, length 1
        Specifies whether or not the CS representation of the matrix A,
        values of `n`, `nnz`, `irow` and `icol` should be checked.

        `check` = 'C'

            Checks are carried on the values of `n`, `nnz`, `irow` and
            `icol`.

        `check` = 'N'

            None of these checks are carried out.

        See also [Further Comments].

    x : float, shape (`n`)
        The vector x.

    y : float, shape (`n`), modified in place
        `On exit`: the vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

    Notes
    -----
    ``real_gen_matvec`` computes either the matrix-vector product y =
    Ax, or the transposed matrix-vector product y = A^Tx, according to
    the value of the argument `trans`, where A is an n*n sparse
    nonsymmetric matrix, of arbitrary sparsity pattern.
    The matrix A is stored in coordinate storage (CS) format (see [the
    F11 Introduction]).
    The array `a` stores all nonzero elements of A, while arrays `irow`
    and `icol` store the corresponding row and column indices
    respectively.

    It is envisaged that a common use of ``real_gen_matvec`` will be to
    compute the matrix-vector product required in the application of
    ``real_gen_basic_solver`` to sparse linear systems.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_matvec'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11xaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, y[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _check,
        _x,
        _y,
        _errbuf,
        _ierr,
        _trans._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

def real_symm_matvec(n, nnz, a, irow, icol, check, x, y):
    u"""
    Real sparse symmetric matrix-vector multiply.

    ``real_symm_matvec`` computes a matrix-vector product involving a
    real sparse symmetric matrix stored in symmetric coordinate storage
    format.

    For full information please refer to the NAG Library document for
    f11xe

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11xef.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        A.

    a : float, shape (`nnz`)
        The nonzero elements in the lower triangular part of the matrix
        A, ordered by increasing row index, and by increasing column
        index within each row. Multiple entries for the same row and
        column indices are not permitted. The function
        ``real_symm_sort`` may be used to order the elements in this
        way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    check : str, length 1
        Specifies whether or not the SCS representation of the matrix A,
        values of `n`, `nnz`, `irow` and `icol` should be checked.

        `check` = 'C'

            Checks are carried out on the values of `n`, `nnz`, `irow`
            and `icol`.

        `check` = 'N'

            None of these checks are carried out.

        See also [Further Comments].

    x : float, shape (`n`)
        The vector x.

    y : float, shape (`n`), modified in place
        `On exit`: the vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

    Notes
    -----
    ``real_symm_matvec`` computes the matrix-vector product

        y = Ax

    where A is an n*n symmetric sparse matrix, of arbitrary sparsity
    pattern, stored in symmetric coordinate storage (SCS) format (see
    [the F11 Introduction]).
    The array `a` stores all nonzero elements in the lower triangular
    part of A, while arrays `irow` and `icol` store the corresponding
    row and column indices respectively.

    It is envisaged that a common use of ``real_symm_matvec`` will be to
    compute the matrix-vector product required in the application of
    ``real_symm_basic_solver`` to sparse symmetric linear systems.
    An illustration of this usage appears in
    ``real_symm_precon_ssor_solve``.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_matvec'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11xeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, y[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _check,
        _x,
        _y,
        _errbuf,
        _ierr,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

def complex_gen_matvec(trans, n, nnz, a, irow, icol, check, x, y):
    u"""
    Complex sparse non-Hermitian matrix-vector multiply.

    ``complex_gen_matvec`` computes a matrix-vector or conjugate
    transposed matrix-vector product involving a complex sparse
    non-Hermitian matrix stored in coordinate storage format.

    For full information please refer to the NAG Library document for
    f11xn

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11xnf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether or not the matrix A is conjugate transposed.

        `trans` = 'N'

            y = Ax is computed.

        `trans` = 'T'

            y = A^Hx is computed.

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : complex, shape (`nnz`)
        The nonzero elements in the matrix A, ordered by increasing row
        index, and by increasing column index within each row. Multiple
        entries for the same row and column indices are not permitted.
        The function ``complex_gen_sort`` may be used to order the
        elements in this way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    check : str, length 1
        Specifies whether or not the CS representation of the matrix A,
        values of `n`, `nnz`, `irow` and `icol` should be checked.

        `check` = 'C'

            Checks are carried on the values of `n`, `nnz`, `irow` and
            `icol`.

        `check` = 'N'

            None of these checks are carried out.

        See also [Further Comments].

    x : complex, shape (`n`)
        The vector x.

    y : complex, shape (`n`), modified in place
        `On exit`: the vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

    Notes
    -----
    ``complex_gen_matvec`` computes either the matrix-vector product y =
    Ax, or the conjugate transposed matrix-vector product y = A^Hx,
    according to the value of the argument `trans`, where A is a complex
    n*n sparse non-Hermitian matrix, of arbitrary sparsity pattern.
    The matrix A is stored in coordinate storage (CS) format (see [the
    F11 Introduction]).
    The array `a` stores all the nonzero elements of A, while arrays
    `irow` and `icol` store the corresponding row and column indices
    respectively.

    It is envisaged that a common use of ``complex_gen_matvec`` will be
    to compute the matrix-vector product required in the application of
    ``complex_gen_basic_solver`` to sparse complex linear systems.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_matvec'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11xnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineComplex128ArrayType.c_type, # Input, x[n].
        utils._EngineComplex128ArrayType.c_type, # Output, y[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _check,
        _x,
        _y,
        _errbuf,
        _ierr,
        _trans._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

def complex_herm_matvec(n, nnz, a, irow, icol, check, x, y):
    u"""
    Complex sparse Hermitian matrix-vector multiply.

    ``complex_herm_matvec`` computes a matrix-vector product involving a
    complex sparse Hermitian matrix stored in symmetric coordinate
    storage format.

    For full information please refer to the NAG Library document for
    f11xs

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11xsf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A.

    a : complex, shape (`nnz`)
        The nonzero elements in the lower triangular part of the matrix
        A, ordered by increasing row index, and by increasing column
        index within each row. Multiple entries for the same row and
        column indices are not permitted. The function
        ``complex_herm_sort`` may be used to order the elements in this
        way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    check : str, length 1
        Specifies whether or not the SCS representation of the matrix A,
        values of `n`, `nnz`, `irow` and `icol` should be checked.

        `check` = 'C'

            Checks are carried out on the values of `n`, `nnz`, `irow`
            and `icol`.

        `check` = 'N'

            None of these checks are carried out.

    x : complex, shape (`n`)
        The vector x.

    y : complex, shape (`n`), modified in place
        `On exit`: the vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

    Notes
    -----
    ``complex_herm_matvec`` computes the matrix-vector product

        y = Ax

    where A is an n*n complex Hermitian sparse matrix, of arbitrary
    sparsity pattern, stored in symmetric coordinate storage (SCS)
    format (see [the F11 Introduction]).
    The array `a` stores all the nonzero elements in the lower
    triangular part of A, while arrays `irow` and `icol` store the
    corresponding row and column indices respectively.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_matvec'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11xsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineComplex128ArrayType.c_type, # Input, x[n].
        utils._EngineComplex128ArrayType.c_type, # Output, y[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _check,
        _x,
        _y,
        _errbuf,
        _ierr,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

def sym_rcm(n, nnz, icolzp, irowix, lopts, mask, perm, info):
    u"""
    Reverse Cuthill--McKee reordering of a sparse symmetric matrix in
    CCS format.

    ``sym_rcm`` reduces the bandwidth of a sparse symmetric matrix
    stored in compressed column storage format using the Reverse
    Cuthill--McKee algorithm.

    For full information please refer to the NAG Library document for
    f11ye

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11yef.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    icolzp : int, shape (`n`+1)
        `icolzp` records the index into `irowix` which starts each new
        column.

    irowix : int, shape (`nnz`)
        The row indices corresponding to the nonzero elements in the
        matrix A.

    lopts : bool, shape (5)
        The options to be used by ``sym_rcm``.

        `lopts`[0] = True

            Row/column i of the matrix A will only be referenced if
            `mask`[i-1] != 0, otherwise `mask` will be ignored.

        `lopts`[1] = True

            The final permutation will not be reversed, that is, the
            Cuthill--McKee ordering will be returned. The bandwidth of
            the non-reversed matrix will be the same but the profile
            will be the same or larger (see Wai-Hung and Sherman
            (1976)).

        `lopts`[2] = True

            The matrix A will be checked for symmetrical sparsity
            pattern, otherwise not.

        `lopts`[3] = True

            The bandwidth and profile of the unpermuted matrix will be
            calculated, otherwise not.

        `lopts`[4] = True

            The bandwidth and profile of the permuted matrix will be
            calculated, otherwise not.

    mask : None or int, shape (:)
        Note: the required length for this argument is determined as
        follows: if `lopts`[0]= True: `n`; otherwise: 0.

        `mask` is only referenced if `lopts`[0] is True. A value of
        `mask`[i-1] = 0 indicates that the node corresponding to row or
        column i is not to be referenced. A value of `mask`[i-1] != 0
        indicates that the node corresponding to row or column i is to
        be referenced. In particular, rows and columns not referenced
        will not be permuted.

    perm : int, shape (`n`), modified in place
        `On exit`: this will contain the permutation vector that
        describes the permutation matrix P for the reordering of the
        matrix A. The elements of the permutation matrix P are zero
        except for the unit elements in row i and column `perm`[i-1], i
        = 1,2,...n.

    info : int, shape (4), modified in place
        `On exit`: statistics about the matrix A and the permuted
        matrix. The quantities below are calculated using any masking in
        effect otherwise the value zero is returned.

        `info`[0]

            The bandwidth of the matrix A, if `lopts`[3] = True.

        `info`[1]

            The profile of the matrix A, if `lopts`[3] = True.

        `info`[2]

            The bandwidth of the permuted matrix PAP^T, if `lopts`[4] =
            True.

        `info`[3]

            The profile of the permuted matrix PAP^T, if `lopts`[4] =
            True.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 3)
            On entry, `irowix`[*<value>*] = *<value>* and `n` =
            *<value>*.

            Constraint: 1 <= `irowix`[i-1] <= `n` for all i.

        (`errno` 4)
            On entry, `icolzp`[*<value>*] = *<value>* and `nnz` =
            *<value>*.

            Constraint: 1 <= `icolzp`[i-1] <= `nnz` for all i.

        (`errno` 5)
            On entry, `icolzp`[0] = *<value>*.

            Constraint: `icolzp`[0] = 1.

        (`errno` 5)
            On entry, `icolzp`[`n`] = *<value>* and `nnz` = *<value>*.

            Constraint: `icolzp`[`n`] = `nnz`+1.

        (`errno` 6)
            On entry, the matrix A is not symmetric.

            Element (*<value>*,*<value>*) has no symmetric element.

    Notes
    -----
    ``sym_rcm`` takes the compressed column storage (CCS) representation
    (see [the F11 Introduction]) of an n*n symmetric matrix A and
    applies the Reverse Cuthill--McKee (RCM) algorithm which aims to
    minimize the bandwidth of the matrix A by reordering the rows and
    columns symmetrically.
    This also results in a lower profile of the matrix (see [Further
    Comments]).

    ``sym_rcm`` can be useful for solving systems of equations Ax = b,
    as the permuted system PAP^T(Px) = Pb (where P is the permutation
    matrix described by the vector `perm` returned by ``sym_rcm``) may
    require less storage space and/or less computational steps when
    solving (see Wai-Hung and Sherman (1976)).

    ``sym_rcm`` may be used prior to ``real_symm_precon_ichol`` and
    ``real_symm_precon_ichol_solve``.

    References
    ----------
    Pissanetsky, S, 1984, `Sparse Matrix Technology`, Academic Press

    Wai-Hung, L and Sherman, A H, 1976, `Comparative analysis of the
    Cuthill--McKee and the reverse Cuthill--McKee ordering algorithms
    for sparse matrices`, SIAM J. Numer. Anal. (13(2)), 198--213
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.sym_rcm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11yeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineIntArrayType.c_type, # Input, icolzp[n+1].
        utils._EngineIntArrayType.c_type, # Input, irowix[nnz].
        utils._EngineLogicalArrayType.c_type, # Input, lopts[5].
        utils._EngineIntArrayType.c_type, # Input, mask[*].
        utils._EngineIntArrayType.c_type, # Output, perm[n].
        utils._EngineIntArrayType.c_type, # Output, info[4].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _icolzp = utils._EngineIntArrayType(
        icolzp, {'fun_name': fun_name, 'entity': 'icolzp'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _irowix = utils._EngineIntArrayType(
        irowix, {'fun_name': fun_name, 'entity': 'irowix'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _lopts = utils._EngineLogicalArrayType(
        lopts, {'fun_name': fun_name, 'entity': 'lopts'},
        exp_shape=(5,),
    ).to_ctypes()
    if utils._EngineLogicalArrayType.get_element(lopts, {'fun_name': fun_name, 'entity': 'lopts'}, (0,)):
        if mask is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'mask'},
                'must not be None when lopts[0]',
            )
    if utils._EngineLogicalArrayType.get_element(lopts, {'fun_name': fun_name, 'entity': 'lopts'}, (0,)):
        _mask_dim1 = _n_py
    else:
        _mask_dim1 = 0
    exp_shape_mask = (_mask_dim1,)
    _mask = utils._EngineIntArrayType(
        mask, {'fun_name': fun_name, 'entity': 'mask'},
        exp_shape=exp_shape_mask,
        check_contents=utils._EngineLogicalArrayType.get_element(lopts, {'fun_name': fun_name, 'entity': 'lopts'}, (0,)),
        allow_none=True,
    ).to_ctypes(can_reference_contents=utils._EngineLogicalArrayType.get_element(lopts, {'fun_name': fun_name, 'entity': 'lopts'}, (0,)))
    _perm_cls = utils._EngineIntArrayType(
        perm, {'fun_name': fun_name, 'entity': 'perm'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _perm = _perm_cls.to_ctypes(can_reference_contents=False)
    _info_cls = utils._EngineIntArrayType(
        info, {'fun_name': fun_name, 'entity': 'info'},
        exp_shape=(4,),
        check_contents=False,
    )
    _info = _info_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _icolzp,
        _irowix,
        _lopts,
        _mask,
        _perm,
        _info,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1020:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1030:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irowix[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irowix[i-1] <= n for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1040:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icolzp[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= icolzp[i-1] <= nnz for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1050:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icolzp[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: icolzp[0] = 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1055:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icolzp[n] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: icolzp[n] = nnz+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 1060:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the matrix A is not symmetric.',
                    'Element ({:s},{:s}) has no symmetric'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'element.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _perm_cls.refresh_py(_perm)
        _info_cls.refresh_py(_info)
        return_data = None
    _err_class.manage()
    return return_data

def real_gen_sort(n, nnz, a, irow, icol, dup, zer, istr):
    u"""
    Real sparse nonsymmetric matrix reorder function.

    ``real_gen_sort`` sorts the nonzero elements of a real sparse
    nonsymmetric matrix, represented in coordinate storage format.

    For full information please refer to the NAG Library document for
    f11za

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11zaf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of elements supplied in the array `a`.

    a : float, shape (`nnz`), modified in place
        `On entry`: the nonzero elements of the matrix A. These may be
        in any order and there may be multiple nonzero elements with the
        same row and column indices.

        `On exit`: the nonzero elements ordered by increasing row index,
        and by increasing column index within each row. Each nonzero
        element has a unique row and column index.

    irow : int, shape (`nnz`), modified in place
        `On entry`: the row indices corresponding to the elements
        supplied in the array `a`.

        `On exit`: the first `nnz` elements contain the row indices
        corresponding to the elements returned in the array `a`.

    icol : int, shape (`nnz`), modified in place
        `On entry`: the column indices corresponding to the elements
        supplied in the array `a`.

        `On exit`: the first `nnz` elements contain the column indices
        corresponding to the elements returned in the array `a`.

    dup : str, length 1
        Indicates how elements in `a` with duplicate row and column
        indices are to be treated.

        `dup` = 'R'

            Duplicate entries are removed, only the first entry is kept.

        `dup` = 'S'

            The relevant values in `a` are summed.

        `dup` = 'F'

            The function fails with `errno` = 3 on detecting a
            duplicate.

    zer : str, length 1
        Indicates how elements in `a` with zero values are to be
        treated.

        `zer` = 'R'

            The entries are removed.

        `zer` = 'K'

            The entries are kept.

        `zer` = 'F'

            The function fails with `errno` = 4 on detecting a zero.

    istr : int, shape (`n`+1), modified in place
        `On exit`: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix A. `istr`[`n`]-1 is the address of the last element in
        `a` plus one.

    Returns
    -------
    nnz : int
        The number of elements with unique row and column indices.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `zer` = *<value>*.

            Constraint: `zer` = 'R', 'K' or 'F'.

        (`errno` 1)
            On entry, `dup` = *<value>*.

            Constraint: `dup` = 'R', 'S' or 'F'.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, a duplicate entry has been found in row i and
            column j: i = *<value>*, j = *<value>*.

        (`errno` 4)
            On entry, a zero entry has been found in row i and column j:
            i = *<value>*, j = *<value>*.

    Notes
    -----
    ``real_gen_sort`` takes a coordinate storage (CS) representation
    (see [the F11 Introduction]) of a real n*n sparse nonsymmetric
    matrix A, and reorders the nonzero elements by increasing row index
    and increasing column index within each row.
    Entries with duplicate row and column indices may be removed.
    Alternatively, duplicate entries may be summed, which facilitates
    spare matrix addition (see [Further Comments]).
    Any entries with zero values may optionally be removed.

    ``real_gen_sort`` also returns a pointer array `istr` to the
    starting address of each row in A.
    This can be used to construct a compressed column storage (CCS)
    representation of the matrix (see [Further Comments]).
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_sort'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11zaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input/Output, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[nnz].
        utils._BYREF_STRING, # Input, dup(len=1).
        utils._BYREF_STRING, # Input, zer(len=1).
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dup.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of zer.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _dup = utils._EngineCharScalarType(
        dup, {'fun_name': fun_name, 'entity': 'dup'}, exp_chrlen=1,
    ).to_ctypes()
    _zer = utils._EngineCharScalarType(
        zer, {'fun_name': fun_name, 'entity': 'zer'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _dup,
        _zer,
        _istr,
        _iwork,
        _errbuf,
        _ierr,
        _dup._length_,
        _zer._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 6 and _ifmt == 99994:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zer = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zer = \'R\', \'K\' or \'F\'.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dup = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dup = \'R\', \'S\' or \'F\'.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99992:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99993:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 99988:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a duplicate entry has been found in row i and column j:',
                    'i = {:s}, j = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 99987:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a zero entry has been found in row i and column j:',
                    'i = {:s}, j = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _istr_cls.refresh_py(_istr)
        return_data = _nnz_py
    _err_class.manage()
    return return_data

def real_symm_sort(n, nnz, a, irow, icol, dup, zer, istr):
    u"""
    Real sparse symmetric matrix reorder function.

    ``real_symm_sort`` sorts the nonzero elements of a real sparse
    symmetric matrix, represented in symmetric coordinate storage
    format.

    For full information please refer to the NAG Library document for
    f11zb

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11zbf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of elements supplied in the array `a`.

    a : float, shape (`nnz`), modified in place
        `On entry`: the nonzero elements of the lower triangular part of
        the real matrix A. These may be in any order and there may be
        multiple nonzero elements with the same row and column indices.

        `On exit`: the lower triangular nonzero elements ordered by
        increasing row index, and by increasing column index within each
        row. Each nonzero element has a unique row and column index.

    irow : int, shape (`nnz`), modified in place
        `On entry`: the row indices corresponding to the elements
        supplied in the array `a`.

        `On exit`: the first `nnz` elements contain the row indices
        corresponding to the elements returned in the array `a`.

    icol : int, shape (`nnz`), modified in place
        `On entry`: the column indices corresponding to the elements
        supplied in the array `a`.

        `On exit`: the first `nnz` elements contain the column indices
        corresponding to the elements returned in the array `a`.

    dup : str, length 1
        Indicates how elements in `a` with duplicate row and column
        indices are to be treated.

        `dup` = 'R'

            Duplicate entries are removed, only the first entry is kept.

        `dup` = 'S'

            The relevant values in `a` are summed.

        `dup` = 'F'

            The function fails with `errno` = 3 on detecting a
            duplicate.

    zer : str, length 1
        Indicates how elements in `a` with zero values are to be
        treated.

        `zer` = 'R'

            The entries are removed.

        `zer` = 'K'

            The entries are kept.

        `zer` = 'F'

            The function fails with `errno` = 4 on detecting a zero.

    istr : int, shape (`n`+1), modified in place
        `On exit`: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix A. `istr`[`n`]-1 is the address of the last element in
        `a` plus one.

    Returns
    -------
    nnz : int
        The number of elements with unique row and column indices.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `zer` = *<value>*.

            Constraint: `zer` = 'R', 'K' or 'F'.

        (`errno` 1)
            On entry, `dup` = *<value>*.

            Constraint: `dup` = 'R', 'S' or 'F'.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and
            `irow`[i-1] = *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `irow`[i-1].

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, a duplicate entry has been found in row I and
            column J: I = *<value>*, J = *<value>*.

        (`errno` 4)
            On entry, a zero entry has been found in row I and column J:
            I = *<value>*, J = *<value>*.

    Notes
    -----
    ``real_symm_sort`` takes a symmetric coordinate storage (SCS)
    representation (see [the F11 Introduction]) of a real n*n sparse
    symmetric matrix A, and reorders the nonzero elements by increasing
    row index and increasing column index within each row.
    Entries with duplicate row and column indices may be removed.
    Alternatively, duplicate entries may be summed, which facilitates
    spare matrix addition (see [Further Comments]).
    Any entries with zero values may optionally be removed.

    ``real_symm_sort`` also returns a pointer array `istr` to the
    starting address of each row in A.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_sort'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11zbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input/Output, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[nnz].
        utils._BYREF_STRING, # Input, dup(len=1).
        utils._BYREF_STRING, # Input, zer(len=1).
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dup.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of zer.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _dup = utils._EngineCharScalarType(
        dup, {'fun_name': fun_name, 'entity': 'dup'}, exp_chrlen=1,
    ).to_ctypes()
    _zer = utils._EngineCharScalarType(
        zer, {'fun_name': fun_name, 'entity': 'zer'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _dup,
        _zer,
        _istr,
        _iwork,
        _errbuf,
        _ierr,
        _dup._length_,
        _zer._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zer = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zer = \'R\', \'K\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dup = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dup = \'R\', \'S\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and irow[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= irow[i-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a duplicate entry has been found in row I and column J:',
                    'I = {:s}, J = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a zero entry has been found in row I and column J:',
                    'I = {:s}, J = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _istr_cls.refresh_py(_istr)
        return_data = _nnz_py
    _err_class.manage()
    return return_data

def real_rect_sort(m, n, nnz, a, irow, icol, istc, store, dup, zer):
    u"""
    Sorts and merges elements of a real sparse rectangular matrix,
    represented in coordinate storage format, and provides the
    compressed column storage format of the result.

    ``real_rect_sort`` sorts the nonzero elements of a real sparse
    rectangular matrix, represented in coordinate storage or compressed
    column storage format.

    For full information please refer to the NAG Library document for
    f11zc

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11zcf.html

    Parameters
    ----------
    m : int
        m, the number of rows in the matrix A.

    n : int
        n, the number of columns in the matrix A.

    nnz : int
        The number of elements supplied in the array `a`.

    a : float, shape (`nnz`), modified in place
        `On entry`: the nonzero elements of the matrix A. If `store` =
        1, the elements may be in any order. If `store` = 2, the
        elements must be ordered by increasing column index. There may
        be multiple nonzero elements with the same row and column
        indices.

        `On exit`: the nonzero elements ordered by increasing column
        index, and by increasing row index within each column. Each
        nonzero element has a unique row and column index.

    irow : int, shape (`nnz`), modified in place
        `On entry`: the row indices corresponding to the elements
        supplied in the array `a`.

        `On exit`: the first `nnz` elements contain the row indices
        corresponding to the elements returned in the array `a`.

    icol : int, shape (`nnz`), modified in place
        `On entry`: if `store` = 1, the column indices corresponding to
        the elements supplied in the array `a`.

        If `store` = 2, `icol` need not be set.

        `On exit`: the first `nnz` elements contain the column indices
        corresponding to the elements returned in the array `a`.

    istc : int, shape (`n`+1), modified in place
        `On entry`: if `store` = 2, the starting address of each column,
        as supplied in the array `a`.

        If `store` = 1, `istc` need not be set.

        `On exit`: the starting address of each column, as returned in
        the array `a`. `istc`[`n`]-1 is the address of the last element
        in `a` plus one.

    store : int
        Indicates which storage format the matrix A is represented with
        on entry.

        `store` = 1

            A is represented in coordinate storage (CS) format using
            `a`, `irow` and `icol`.

        `store` = 2

            A is represented in compressed column storage (CCS) format
            using `a`, `irow` and `istc`.

    dup : str, length 1
        Indicates how elements in `a` with duplicate row and column
        indices are to be treated.

        `dup` = 'R'

            Duplicate entries are removed, only the first entry is kept.

        `dup` = 'S'

            The relevant values in `a` are summed.

        `dup` = 'F'

            The function fails with `errno` = 12 on detecting a
            duplicate.

    zer : str, length 1
        Indicates how elements in `a` with zero values are to be
        treated.

        `zer` = 'R'

            The entries are removed.

        `zer` = 'K'

            The entries are kept.

        `zer` = 'F'

            The function fails with `errno` = 13 on detecting a zero.

    Returns
    -------
    nnz : int
        The number of elements with unique row and column indices.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 0.

        (`errno` 4)
            On entry, `store` = *<value>*.

            Constraint: `store` = 1 or 2.

        (`errno` 5)
            On entry, `dup` = *<value>*.

            Constraint: `dup` = 'R', 'S' or 'F'.

        (`errno` 6)
            On entry, `zer` = *<value>*.

            Constraint: `zer` = 'R', 'K' or 'F'.

        (`errno` 7)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `m` =
            *<value>*.

            Constraint: 1 <= `irow`[i-1] <= `m`.

        (`errno` 8)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: if `store` = 1, then 1 <= `icol`[i-1] <= `n`.

        (`errno` 9)
            On entry, `istc`[0] = *<value>*.

            Constraint: `istc`[0] = 1.

        (`errno` 10)
            On entry, i = *<value>*, `istc`[i-1] = *<value>* and
            `istc`[i] = *<value>*.

            Constraint: `istc`[i-1] <= `istc`[i], for i = 1,2,...,`n`

        (`errno` 11)
            On entry, `n` = *<value>*, `istc`[`n`] = *<value>* and `nnz`
            = *<value>*.

            Constraint: `istc`[`n`] = `nnz`+1.

        (`errno` 12)
            On entry, a duplicate entry has been found in row *<value>*
            and column *<value>*.

        (`errno` 13)
            On entry, a zero entry has been found in row *<value>* and
            column *<value>*.

    Notes
    -----
    ``real_rect_sort`` takes a coordinate storage (CS) representation
    (see [the F11 Introduction]), or compressed column storage (CCS)
    representation (see [the F11 Introduction]) of a real m*n sparse
    rectangular matrix A, and reorders the nonzero elements by
    increasing column index and increasing row index within each column.

    Entries with duplicate row and column indices may be removed.
    Alternatively, duplicate entries may be summed, which facilitates
    sparse matrix addition (see [Further Comments]).
    Any entries with zero values may optionally be removed.

    Both CS and CCS representations of the resulting matrix are output,
    which allows ``real_rect_sort`` to be used to convert between the
    two formats (see [Further Comments]).
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_rect_sort'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11zcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input/Output, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, istc[n+1].
        utils._BYREF_INT, # Input, store.
        utils._BYREF_STRING, # Input, dup(len=1).
        utils._BYREF_STRING, # Input, zer(len=1).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dup.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of zer.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _istc_cls = utils._EngineIntArrayType(
        istc, {'fun_name': fun_name, 'entity': 'istc'},
        exp_shape=(_n_py+1,),
    )
    _istc = _istc_cls.to_ctypes()
    _store = utils._EngineIntScalarType(
        store, {'fun_name': fun_name, 'entity': 'store'},
    ).to_ctypes()
    _dup = utils._EngineCharScalarType(
        dup, {'fun_name': fun_name, 'entity': 'dup'}, exp_chrlen=1,
    ).to_ctypes()
    _zer = utils._EngineCharScalarType(
        zer, {'fun_name': fun_name, 'entity': 'zer'}, exp_chrlen=1,
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _istc,
        _store,
        _dup,
        _zer,
        _errbuf,
        _ierr,
        _dup._length_,
        _zer._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, store = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: store = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dup = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dup = \'R\', \'S\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zer = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zer = \'R\', \'K\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irow[i-1] <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if store = 1, then 1 <= icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istc[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: istc[0] = 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, istc[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and istc[i] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: istc[i-1] <= istc[i], for i = 1,2,...,n',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}, istc[n] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: istc[n] = nnz+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a duplicate entry has been found in row {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'and column {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a zero entry has been found in row {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'and column {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _istc_cls.refresh_py(_istc)
        return_data = _nnz_py
    _err_class.manage()
    return return_data

def complex_gen_sort(n, nnz, a, irow, icol, dup, zer, istr):
    u"""
    Complex sparse non-Hermitian matrix reorder function.

    ``complex_gen_sort`` sorts the nonzero elements of a complex sparse
    non-Hermitian matrix, represented in coordinate storage format.

    For full information please refer to the NAG Library document for
    f11zn

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11znf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of elements supplied in the array `a`.

    a : complex, shape (`nnz`), modified in place
        `On entry`: the nonzero elements of the matrix A. These may be
        in any order and there may be multiple nonzero elements with the
        same row and column indices.

        `On exit`: the nonzero elements ordered by increasing row index,
        and by increasing column index within each row. Each nonzero
        element has a unique row and column index.

    irow : int, shape (`nnz`), modified in place
        `On entry`: the row indices corresponding to the elements
        supplied in the array `a`.

        `On exit`: the first `nnz` elements contain the row indices
        corresponding to the elements returned in the array `a`.

    icol : int, shape (`nnz`), modified in place
        `On entry`: the column indices corresponding to the elements
        supplied in the array `a`.

        `On exit`: the first `nnz` elements contain the column indices
        corresponding to the elements returned in the array `a`.

    dup : str, length 1
        Indicates how elements in `a` with duplicate row and column
        indices are to be treated.

        `dup` = 'R'

            Duplicate entries are removed, only the first entry is kept.

        `dup` = 'S'

            The relevant values in `a` are summed.

        `dup` = 'F'

            The function fails with `errno` = 3 on detecting a
            duplicate.

    zer : str, length 1
        Indicates how elements in `a` with zero values are to be
        treated.

        `zer` = 'R'

            The entries are removed.

        `zer` = 'K'

            The entries are kept.

        `zer` = 'F'

            The function fails with `errno` = 4 on detecting a zero.

    istr : int, shape (`n`+1), modified in place
        `On exit`: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix A. `istr`[`n`]-1 is the address of the last element in
        `a` plus one.

    Returns
    -------
    nnz : int
        The number of elements with unique row and column indices.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `zer` = *<value>*.

            Constraint: `zer` = 'R', 'K' or 'F'.

        (`errno` 1)
            On entry, `dup` = *<value>*.

            Constraint: `dup` = 'R', 'S' or 'F'.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, a duplicate entry has been found in row I and
            column J: I = *<value>*, J = *<value>*.

        (`errno` 4)
            On entry, a zero entry has been found in row I and column J:
            I = *<value>*, J = *<value>*.

    Notes
    -----
    ``complex_gen_sort`` takes a coordinate storage (CS) representation
    (see [the F11 Introduction]) of a sparse n*n complex non-Hermitian
    matrix A, and reorders the nonzero elements by increasing row index
    and increasing column index within each row.
    Entries with duplicate row and column indices may be removed.
    Alternatively, duplicate entries may be summed, which facilitates
    spare matrix addition (see [Further Comments]).
    Any entries with zero values may optionally be removed.

    ``complex_gen_sort`` also returns a pointer array `istr` to the
    starting address of each row in A.
    This can be used to construct a compressed column storage (CCS)
    representation of the matrix (see [Further Comments]).
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_sort'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11znft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input/Output, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[nnz].
        utils._BYREF_STRING, # Input, dup(len=1).
        utils._BYREF_STRING, # Input, zer(len=1).
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dup.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of zer.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _dup = utils._EngineCharScalarType(
        dup, {'fun_name': fun_name, 'entity': 'dup'}, exp_chrlen=1,
    ).to_ctypes()
    _zer = utils._EngineCharScalarType(
        zer, {'fun_name': fun_name, 'entity': 'zer'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _dup,
        _zer,
        _istr,
        _iwork,
        _errbuf,
        _ierr,
        _dup._length_,
        _zer._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zer = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zer = \'R\', \'K\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dup = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dup = \'R\', \'S\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a duplicate entry has been found in row I and column J:',
                    'I = {:s}, J = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a zero entry has been found in row I and column J:',
                    'I = {:s}, J = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _istr_cls.refresh_py(_istr)
        return_data = _nnz_py
    _err_class.manage()
    return return_data

def complex_herm_sort(n, nnz, a, irow, icol, dup, zer, istr):
    u"""
    Complex sparse Hermitian matrix reorder function.

    ``complex_herm_sort`` sorts the nonzero elements of a sparse complex
    Hermitian matrix, represented in symmetric coordinate storage
    format.

    For full information please refer to the NAG Library document for
    f11zp

    https://www.nag.com/numeric/nl/nagdoc_27.1/flhtml/f11/f11zpf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of elements supplied in the array `a`.

    a : complex, shape (`nnz`), modified in place
        `On entry`: the nonzero elements of the lower triangular part of
        the complex matrix A. These may be in any order and there may be
        multiple nonzero elements with the same row and column indices.

        `On exit`: the lower triangular nonzero elements ordered by
        increasing row index, and by increasing column index within each
        row. Each nonzero element has a unique row and column index.

    irow : int, shape (`nnz`), modified in place
        `On entry`: the row indices corresponding to the elements
        supplied in the array `a`.

        `On exit`: the first `nnz` elements contain the row indices
        corresponding to the elements returned in the array `a`.

    icol : int, shape (`nnz`), modified in place
        `On entry`: the column indices corresponding to the elements
        supplied in the array `a`.

        `On exit`: the first `nnz` elements contain the column indices
        corresponding to the elements returned in the array `a`.

    dup : str, length 1
        Indicates how elements in `a` with duplicate row and column
        indices are to be treated.

        `dup` = 'R'

            Duplicate entries are removed, only the first entry is kept.

        `dup` = 'S'

            The relevant values in `a` are summed.

        `dup` = 'F'

            The function fails with `errno` = 3 on detecting a
            duplicate.

    zer : str, length 1
        Indicates how elements in `a` with zero values are to be
        treated.

        `zer` = 'R'

            The entries are removed.

        `zer` = 'K'

            The entries are kept.

        `zer` = 'F'

            The function fails with `errno` = 4 on detecting a zero.

    istr : int, shape (`n`+1), modified in place
        `On exit`: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix A. `istr`[`n`]-1 is the address of the last element in
        `a` plus one.

    Returns
    -------
    nnz : int
        The number of elements with unique row and column indices.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `zer` = *<value>*.

            Constraint: `zer` = 'R', 'K' or 'F'.

        (`errno` 1)
            On entry, `dup` = *<value>*.

            Constraint: `dup` = 'R', 'S' or 'F'.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, a duplicate entry has been found in row I and
            column J: I = *<value>*, J = *<value>*.

        (`errno` 4)
            On entry, a zero entry has been found in row I and column J:
            I = *<value>*, J = *<value>*.

    Notes
    -----
    ``complex_herm_sort`` takes a symmetric coordinate storage (SCS)
    representation (see [the F11 Introduction]) of a sparse n*n complex
    Hermitian matrix A, and reorders the nonzero elements by increasing
    row index and increasing column index within each row.
    Entries with duplicate row and column indices may be removed.
    Alternatively, duplicate entries may be summed, which facilitates
    spare matrix addition (see [Further Comments]).
    Any entries with zero values may optionally be removed.

    ``complex_herm_sort`` also returns a pointer array `istr` to the
    starting address of each row in A.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_sort'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11zpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input/Output, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[nnz].
        utils._BYREF_STRING, # Input, dup(len=1).
        utils._BYREF_STRING, # Input, zer(len=1).
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dup.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of zer.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _dup = utils._EngineCharScalarType(
        dup, {'fun_name': fun_name, 'entity': 'dup'}, exp_chrlen=1,
    ).to_ctypes()
    _zer = utils._EngineCharScalarType(
        zer, {'fun_name': fun_name, 'entity': 'zer'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _dup,
        _zer,
        _istr,
        _iwork,
        _errbuf,
        _ierr,
        _dup._length_,
        _zer._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zer = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zer = \'R\', \'K\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dup = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dup = \'R\', \'S\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, icol[I-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a duplicate entry has been found in row I and column J:',
                    'I = {:s}, J = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a zero entry has been found in row I and column J:',
                    'I = {:s}, J = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _istr_cls.refresh_py(_istr)
        return_data = _nnz_py
    _err_class.manage()
    return return_data
